// <header>
/**
 *  ThemaMap : The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
package fr.unicaen.openmap.dataStorage;

import com.bbn.openmap.proj.coords.LatLonPoint;
import com.bbn.openmap.omGraphics.OMPoint;
import com.bbn.openmap.util.PropUtils;
import com.bbn.openmap.util.Debug;

/**
 * Stores a bounding Box using Lon/Lat TopLeft and LowerRight points in double.
 * <p>
 * Does NOT use Rectangle2D.Double as internal representation for the bounding box.
 * <p>
 * Intersects relies on the box never crosses the dateline (lon 180 or -180 meridian).
 */
public class BoundingBox extends Object {
    
    private double upperLeftLon = 0d;
    private double upperLeftLat = 0d;
    private double lowerRightLon = 0d;
    private double lowerRightLat = 0d;
    
    /**
     * Create a bounding box with 4 coordinates.
     *
     * The call must be done using correct order for coordinates. Names are explicit here.
     */
    public BoundingBox (double upperLeftLon, double upperLeftLat, double lowerRightLon, double lowerRightLat){
	this.upperLeftLon = upperLeftLon;
	this.upperLeftLat = upperLeftLat;
	this.lowerRightLon = lowerRightLon;
	this.lowerRightLat = lowerRightLat;
    }
    
    /**
     * Create a bounding box with its upperLeft and lowerRight points
     * as <code>LatLonPoint</code> objects.
     *
     * Please remember that internal representation here is in Lon/Lat (as X/Y).
     */
    public BoundingBox (LatLonPoint upperLeft, LatLonPoint lowerRight) {
	this.upperLeftLon = upperLeft.getLongitude();
	this.upperLeftLat = upperLeft.getLatitude();
	this.lowerRightLon = lowerRight.getLongitude();
	this.lowerRightLat = lowerRight.getLatitude();
    }
    
    /**
     * Intersection of tested BB in current BB.
     * Returns true if tested BB and this object have a non null common rectangle
     * area, but does not actually compute the resulting object (we don't need it).
     *
     * Computed using 3 cases :
     * - tested UpperLeft point is in BB (no matter where is LowerRight)
     * - tested UpperLeft is above BB (LowerRight must be in a 1/4 plan (that is 2 tests))
     * - tested UpperLeft is left of BB (same as above with another 1/4 plan)
     *
     * Total is only 3*4 tests. Little optimisation is possible avoiding 1 duplicate test. Better ?
     *
     <code>
            (tested.upperLeftLon &lt;= lowerRightLon &amp;&amp; tested.upperLeftLon &gt;= upperLeftLon &amp;&amp;
	     tested.upperLeftLat &lt;= upperLeftLat &amp;&amp; tested.upperLeftLat &gt;= lowerRightLat)||
	    
	    (tested.upperLeftLon &lt;= lowerRightLon &amp;&amp; tested.upperLeftLat &gt;= upperLeftLat &amp;&amp;
	     tested.lowerRightLat &lt;= upperLeftLat &amp;&amp; tested.lowerRightLon &gt;= upperLeftLon)||
	    
	    (tested.upperLeftLon &gt;= upperLeftLon &amp;&amp; tested.upperLeftLat &gt;= lowerRightLat
	     &amp;&amp; tested.lowerRightLon &gt;= upperLeftLon &amp;&amp; tested.lowerRightLat &lt;= upperLeftLat);
     </code>
     */
    public boolean intersects (BoundingBox tested) {
	boolean res = 	   
	    (tested.upperLeftLon <= lowerRightLon && tested.upperLeftLon >= upperLeftLon &&
	     tested.upperLeftLat <= upperLeftLat && tested.upperLeftLat >= lowerRightLat)
	    ||
	    (tested.upperLeftLon <= lowerRightLon && tested.upperLeftLat >= upperLeftLat &&
	     tested.lowerRightLat <= upperLeftLat && tested.lowerRightLon >= upperLeftLon)
	    ||
	    (tested.upperLeftLon <= upperLeftLon && tested.upperLeftLat >= lowerRightLat
	     && tested.lowerRightLon >= upperLeftLon && tested.lowerRightLat <= upperLeftLat);

	Debug.message("BoundingBox", this + " inter " + tested + " = " + res);
	return res;
    }
    
    public boolean contains(OMPoint point){
        double lat = point.getLat();
        double lon = point.getLon();
        
//        System.out.println("Test de lat : "+lat+", lon : "+lon +"with UpperLeftLat "+upperLeftLat+" , lowerRightLat "+lowerRightLat+" , upperLeftLon "+upperLeftLon+" , lowerRightLon " + lowerRightLon);
        
        if (lat<=upperLeftLat && lat >=lowerRightLat && lon>=upperLeftLon && lon<=lowerRightLon ) return true;
        else return false;
        
    }

    /**
     * Verbose printing.
     */
    public String toString () {
	return "BBox[UL,BR](lon,lat):[(" + this.upperLeftLon + ", " + this.upperLeftLat + "), ("
	    + this.lowerRightLon + ", " + this.lowerRightLat + ")]";
	
    }

    /**
     * adapt the bounding box to include the point given with lon/lat
     */
    public void extendTo(double lon, double lat) {
	if (lon > lowerRightLon)
	    lowerRightLon = lon;
	else if (lon < upperLeftLon)
	    upperLeftLon = lon;

	if (lat > upperLeftLat)
	    upperLeftLat = lat;
	else if (lat < lowerRightLat)
	    lowerRightLat = lat;
    }

    public double getUpperLeftLon() { return upperLeftLon; }
    public double getUpperLeftLat() { return upperLeftLat; }
    public double getLowerRightLon() { return lowerRightLon; }
    public double getLowerRightLat() { return lowerRightLat; }

    public void setUpperLeftLon(double l) { upperLeftLon = l; }
    public void setUpperLeftLat(double l) { upperLeftLat = l; }
    public void setLowerRightLon(double l) { lowerRightLon =l; }
    public void setLowerRightLat(double l) { lowerRightLat =l; }

    static final public String ullonPropertyString = "ullon";
    static final public String ullatPropertyString = "ullat";
    static final public String lrlonPropertyString = "lrlon";
    static final public String lrlatPropertyString = "lrlat";
    public void setProperties(String prefix, java.util.Properties properties) {
        prefix = PropUtils.getScopedPropertyPrefix(prefix);
	upperLeftLon = PropUtils.doubleFromProperties(properties, prefix + ullonPropertyString, 0D);
	upperLeftLat = PropUtils.doubleFromProperties(properties, prefix + ullatPropertyString, 0D);
	lowerRightLon = PropUtils.doubleFromProperties(properties, prefix + lrlonPropertyString, 0D);
	lowerRightLat = PropUtils.doubleFromProperties(properties, prefix + lrlatPropertyString, 0D);
    }

    public java.util.Properties getProperties(String prefix, java.util.Properties properties) {
        prefix = PropUtils.getScopedPropertyPrefix(prefix);

	properties.put(prefix + ullonPropertyString, String.valueOf(upperLeftLon));
	properties.put(prefix + ullatPropertyString, String.valueOf(upperLeftLat));
	properties.put(prefix + lrlonPropertyString, String.valueOf(lowerRightLon));
	properties.put(prefix + lrlatPropertyString, String.valueOf(lowerRightLat));

        return properties;
    }

}
