// <header>
/**
 *  ThemaMap : The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
package fr.unicaen.openmap.dataStorage;
import java.util.regex.*;

/**
 * Tokenizer for character separated values files, as generated
 * by excel of most localizations. The delimiter can be one char 
 * of the given string constructor parameter. 
 * Decimal numbers can have their decimal fraction separated 
 * with a dot or a comma unless comma is in the field separator list.
 * <p>
 * Specifications of this tokenizer is the same as
 * @see com.bnn.openmap.util.CSVTokenizer, from which this class derived 
 * and from which several package protected method are simply copied ! 
 * <p>
 * token() returns the next token, which can be either:
 * <ul>
 * <li>null, indicating an empty field.
 * <li>a Double, indicating a numeric field, but as a String
 * <li>a String, indicating an alphanumeric field.
 * <li>the NEWLINE object, indicating the end of a record.
 * <li>the EOF object, test with isEOF(), indicating the end of file.
 * </ul>
 * I really must rewrite this tokenizer from scratch.
 * Numbers are returned as String objects because there is no real schema, 
 * and that the data type of a column is deduced from the first row element !!!
 */
public class CSVTokenizer extends com.bbn.openmap.util.CSVTokenizer {

    private static Pattern doublePattern = Pattern.compile("[-+]?([0-9 ]+([.,]?[0-9 ]*))|([.,][0-9 ]+)");
    // private static Pattern integerPattern = Pattern.compile("[-+]?[0-9 ]+");
    private static Pattern spacePattern = Pattern.compile(" ");

    public CSVTokenizer(java.io.Reader in) {
        this(in, ",");
    }

    public CSVTokenizer(java.io.Reader in, String delimiters) {
        super(in);
        this.delimiters = delimiters;
	decimalSeparators = ".";
	if (! isTrueDelimiter(','))
	    decimalSeparators += ",";
    }
    
    private String decimalSeparators;
    protected String delimiters;

    public String toString() {
	return "CSVTokenizer with: ["+delimiters+"] as fields delimiters and: ["+decimalSeparators+"] as decimal separators";
    }

    public boolean isTrueDelimiter(int c) {
	for(int i = 0; i < delimiters.length(); i++)
	    if (delimiters.codePointAt(i) == c)
		return true;
	return false;
    }

    public boolean isDecimalSeparator(int c) {
	for(int i = 0; i < decimalSeparators.length(); i++)
	    if (decimalSeparators.codePointAt(i) == c)
		return true;
	return false;
    }

    /** as NEWLINE is package protected in com.bbn.openmap.util.Tokenizer, 
     * I rebuild this object here ! **/
    protected static Object NEWLINE = new Object() {
	    public String toString() {
		return "<newline>";
	    }
	};

    /** as EOF is package protected in com.bbn.openmap.util.Tokenizer, 
     * I rebuild this object here ! **/
    protected static Object EOF = new Object() {
	    public String toString() {
		return "<EOF>";
	    }
	};

    public boolean isNewline(Object o) {
        return o == NEWLINE;
    }

    public boolean isEOF(Object o) {
        return o == EOF;
    }

    protected StringBuffer b = new StringBuffer(80);

    public void bpush(int c) {
        this.b.append((char) c);
    }

    public String bclear() {
        String result = this.b.toString().trim();
        this.b.setLength(0);
	return result;
    }

    /**
     * Return the next object read from the stream.
     * <p>Changes with com.bnn.openmap.util.CSVTokenizer.token() 
     * is that we check if c is one of the delimiters instead of just a comma.
     */
    public Object token() {
        int c = next();
	while (c == ' ') // trim leading spaces
	    c = next();
        if (isTrueDelimiter(c)) 
            return null;
	if (c == '\n')
            return NEWLINE;
	if (c == '"')
            return tokenString(next());
	if (c == '\\')
            return tokenString(c);
	if (c == '-'  || isDigit(c) || isDecimalSeparator(c) || c == '+')
            return tokenNumber(c);
	if (c == -1)
            return EOF;
	return tokenAny(c);
    }

    /**
     * Merely checks for the end of the record: either a true delimiter or a '\n'.
     * The '\n' is out back in the stream.
     * <p>
     * isDelimiter.set(alt(isTrueDelimiter(), is(-1), seq(is('\n'), putback)));
     */
    protected boolean isDelimiter(int c) {
        if (isTrueDelimiter(c) || c == -1) {
            return true;
        } else if (c == '\n') {
            putback(c); // Wait for next token().
            return true;
        } else {
            return false;
        }
    }

    /**
     * Return a number or a string. Comma may be used as a decimal separator.
     */
    protected Object tokenNumber(int c) {
        String result = tokenAny(c);
	Matcher m = doublePattern.matcher(result);
	if (m.matches()) { // it can be a Double or an integer
	    if (isDecimalSeparator(',')) // always . as decimal separator 
		result = result.replace(',', '.');
	    // strip spaces
	    result = spacePattern.matcher(result).replaceAll("");
	}
	return result;
	
        // try {
	//     Double d = isDecimalSeparator(',') 
	// 	? new Double((String) result.replace(',', '.'))
	// 	: new Double((String) result);
        //     return d;
        // } catch (NumberFormatException e) {
        //     return result;
        // }
    }

    /** Just a copy of com.bnn.openmap.util.CSVTokenizer as it is package protected !-[ 
     *  and correct a mistake in case of a string of the form "9F"
    **/
    protected Object tokenString(int c) {
	while(true) {
            // Enable escapes to force characters into string.
            if (c == '\\') { // interpret \n and \t
		switch(c = next()) {
		case 'n': bpush('\n'); break;
		case 't': bpush('\t'); break;

		}
                c = next();
            } else if (c == '"') {
                c = next();
		if (c == '"') { // it was "", push a single "
		    bpush(c); 
		    continue;
		}
                if (isDelimiter(c)) {
		    String result = bclear();
		    if (isDecimalSeparator(',')) {
			Matcher m;
			m = doublePattern.matcher(result);
			if (m.matches()) { // it can be a Double or an integer
			    result = result.replace(',', '.');
			}
		    }
		    return result;
		    // Matcher m;
		    // m = doublePattern.matcher(result);
		    // if (m.matches()) { // it can be a Double or an integer
		    // 	m = integerPattern.matcher(result);
		    // 	if (m.matches()) { // it is an integer
		    // 	    try {
		    // 		Integer i = new Integer(result);
		    // 		return i;
		    // 	    } catch (NumberFormatException e) { 
		    // 		System.err.println("!!! Should never be written !!!");
		    // 	    } 
		    // 	} else { // it's a double
		    // 	    try {
		    //          if (isDecimalSeparator(','))
		    //	            result = result.replace(',', '.');
		    // 		Double d = new Double(result);
		    // 		return d;
		    // 	    } catch (NumberFormatException e) { 
		    // 		System.err.println("!!! Should never be written !!!");
		    // 	    }
		    // 	} 
		    // }
		    //// It's a String
		    // return result;
		}
            } else if (isAny(c)) {
                bpush(c);
                c = next();
            } else {
                return bclear();
            }
        }
    }

    /** Just a copy of com.bnn.openmap.util.CSVTokenizer as it is package protected !-[ **/
    protected String tokenAny(int c) {
        while (true) {
            if (isDelimiter(c)) {
                return bclear().trim();
            } else {
                bpush(c);
                c = next();
            }
        }
    }

    public static void main(String[] args) {
        try {
            CSVTokenizer csv = 
		new CSVTokenizer(new java.io.BufferedReader(new java.io.FileReader(args[0])),
				 args[0].matches(".*_fr\\..*") ? ";" : ",");

            while (true) {
                Object token = csv.token();
                if (csv.isEOF(token))
                    return;
		if (csv.isNewline(token))
		    System.out.println();
		else
		    System.out.print(fr.unicaen.openmap.util.Util.className(token) + ":[" + token + "] ");
            }
        } catch (Exception e) {
            System.out.println(e);
        }
    }
}

