// <header>
/**
 *  ThemaMap : The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
/*
 * DelaunayTriangulation.java
 *
 * Created on 15 december 2009
 *
 * @author Jacques.Madelaine@info.unicaen.fr
 *
 */
package fr.unicaen.openmap.dataStorage;

import com.bbn.openmap.util.Debug;

import com.bbn.openmap.util.PropUtils;
import com.bbn.openmap.omGraphics.OMLine;
import com.bbn.openmap.omGraphics.OMPoint;
import com.bbn.openmap.omGraphics.OMPoly;
import com.bbn.openmap.omGraphics.OMGraphic;
import com.bbn.openmap.omGraphics.OMGraphicList;

import fr.unicaen.openmap.dataStorage.InformationDataStorage;
import fr.unicaen.openmap.dataStorage.gui.DataTableModel;
import fr.unicaen.openmap.dataStorage.gui.DataTableStoragePanel;
import fr.unicaen.openmap.dataStorage.DataTable;
import fr.unicaen.openmap.dataStorage.triangulation.paulchew.DataPoint;
import fr.unicaen.openmap.dataStorage.triangulation.paulchew.DelaunayTriangulation;
import fr.unicaen.openmap.dataStorage.triangulation.paulchew.Pnt;
import fr.unicaen.openmap.dataStorage.triangulation.paulchew.Simplex;
import fr.unicaen.openmap.gui.component.CheckListItem;
import fr.unicaen.openmap.gui.component.CheckListRenderer;
import fr.unicaen.openmap.util.Util;

import java.util.*;
import javax.swing.*;
import java.awt.event.*;
import java.awt.Color;
import java.awt.Component;
import java.util.Collection;

/**
 * DelaunayTriangulation is a DataStorage Adaptor that build a triangulation
 * with a data points cloud defined in the DataStorage
 */
public class DelaunayTriangulationDataStorage extends RemovableInformationDataStorage implements DataStorageListener {

    public static final String delaunayDSPropertyString = "delaunayDS";
    public static final String dataStorageClassPropertyString = "dataStorageClass";
    public static final String geometryColumnNamePropertyString = "geometryColumnName";
    public static final String dataColumnNamesPropertyString = "dataColumnNames";
    public static final String keepVertexesPropertyString = "keepVertexes";
    public static final String drawVoronoiPropertyString = "drawVoronoi";
    public static final String voronoiNamePropertyString = "voronoiName";
    public static final String vertexesNamePropertyString = "vertexesName";

    public static final String convexHullEdgeAttribute = "convexHullEdge";
    public static  String omNameTriangle = "triangle";
    public static  String omNameVertex = "vertex_";
    public static  String omNameVoronoiLine = "voronoï_line";

    protected String geometryColumnName;

    protected ArrayList<String> dataColumnNames = new ArrayList<String>();

    protected boolean keepVertexes = false;
    protected boolean drawVoronoi = false;

    //we should use a DelaunayGUIPanel
    private JTextField fieldVertexesName;
    private JTextField fieldVoronoiName;

    protected void setKeepVertexes(boolean b) { keepVertexes = b; }
    protected void setDrawVoronoi(boolean b) { drawVoronoi = b; }

    private boolean initialized = false;
    protected boolean getInitialized() { return initialized; }
    protected void setInitialized(boolean b) { 
	Debug.message("DataStorage", classNameForDebug + ".initialized = "+b);
	initialized = b; 
    }
        
    /**
     * Tells if the dataStorage has its data.
     */
    public boolean isReady() {
	return getDataStorage() != null &&  getDataStorage().isReady() && initialized;
    }
       
    public DelaunayTriangulationDataStorage(DataStorageListener dsl) {
	this(dsl, null);
    }

    public DelaunayTriangulationDataStorage(DataStorageListener dsl, DataStorage dataStorage) {
	super(dsl, dataStorage);
    }    

    /*
      public String getDataColumnName() { return dataColumnName; }
      public void setDataColumnName(String name) { dataColumnName = name; }
    */
    public String getGeometryColumnName() { return geometryColumnName; }
    public void setGeometryColumnName(String name) { geometryColumnName = name; }

    // vertices of the global triangle containing all the other points  
    private static final DataPoint dp0 = new DataPoint(-181, 91, null);
    private static final DataPoint dp1 = new DataPoint(-181, -271, null);
    private static final DataPoint dp2 = new DataPoint(541, 91, null);

    /** List of triangles having one of dp0, dp1 or dp2 as vertex */
    private List<Simplex<DataPoint>> artefactTriangles;
    private boolean isArtefactTriangles(Simplex<DataPoint> s) {
	return artefactTriangles.contains(s);
    }
    /** List of triangles having one edge beeing on the convex hull of the triangualtion.
     * These triangles are all neighbours one of the artefact triangles
     */
    private Set<Simplex<DataPoint>> boundaryTriangles;
    private boolean isBoundaryTriangles(Simplex<DataPoint> s) {
	return boundaryTriangles.contains(s);
    }
    /**
     */
    private List<ArrayList<DataPoint>> convexHullEdges;
    ArrayList<DataPoint> getConvexHullEdge(Simplex<DataPoint> s) {
	for (ArrayList<DataPoint> e: convexHullEdges)
	    if (s.contains(e.get(0)) && s.contains(e.get(1))) {
		return e;
	    }
	return null;
    }
    private void initBoundaryTriangles(DelaunayTriangulation dt) {
	artefactTriangles = new ArrayList<Simplex<DataPoint>>();
	boundaryTriangles = new HashSet<Simplex<DataPoint>>();
	convexHullEdges = new ArrayList<ArrayList<DataPoint>>();
	for (Simplex<DataPoint> triangle: dt)

	    // if the triangle contains any of the initial triangle vertices
	    if (triangle.contains(dp0) || triangle.contains(dp1) || triangle.contains(dp2)) {
		// it is an artefact
		artefactTriangles.add(triangle);
		boolean convexHullEdgeAdded = false;
		for (Simplex<DataPoint> neighbor: dt.neighbors(triangle)) {
		    if (neighbor.contains(dp0) || neighbor.contains(dp1) || neighbor.contains(dp2))
			continue; // we skip artefact triangle
		    boundaryTriangles.add(neighbor);
		    if (convexHullEdgeAdded)
			continue;
		    // Which is the edge on the convex hull?
		    // Answer: the edge of triangle that doesn't contain dp0 nor dp1 nor dp2
		    ArrayList<DataPoint> convexHullEdge = new ArrayList<DataPoint>(2);
		    for(DataPoint dp: triangle) {
			if (dp == dp0 || dp == dp1 || dp == dp2)
			    continue;
			convexHullEdge.add(dp);
		    }
		    convexHullEdges.add(convexHullEdge);
		    convexHullEdgeAdded = true;
		}
	    }
    }

    public void initialize() {
	if (getDataStorage() == null || ! getDataStorage().isReady()) {
	    Debug.error("[DelaunayTriangulationDataStorage] can't initialize because dataStorage is not ready !!!!");
	    return;
	}

	DataTableModel dtm = getDataStorage().getDefaultDataTableModel();
	if (geometryColumnName == null || "".equals(geometryColumnName)) {
	    if (dtm.getOMColumnNames().length < 1) {
		Debug.error("[DelaunayTriangulationDataStorage] can't initialize because there is no geometry !!!!");
		return;
	    }
	    geometryColumnName = dtm.getOMColumnName(0);
	    Debug.error("[DelaunayTriangulationDataStorage] set geometry to "+ geometryColumnName);
	}

	// create the triangulation
	DelaunayTriangulation dt = new DelaunayTriangulation(new Simplex<DataPoint>(dp0, dp1, dp2));

	// add data points from the underlying dataStorage
	for (int i = 0; i < getDataStorage().getCount(); i++) {
	    DataObject data = getDataStorage().getElementAt(i);
	    if (data == null)
		continue;
	    OMDataLocation omdl = data.getOMDataLocation(geometryColumnName);
	    // Debug.message("DelaunayTriangulationDataStorage", "Delaunay Triangulation Place " + geometryColumnName+ " " + omdl + " "+ data);
	    if (omdl == null)
		continue;
	    
	    Centroid centroid = omdl.getCentroid();
	    if (centroid == null)
		continue;
	    // Debug.message("DelaunayTriangulationDataStorage", "Delaunay Triangulation Place " + centroid + " "+ data);
	    dt.delaunayPlace(new DataPoint(centroid.getLon(), centroid.getLat(), data));
	}

	Debug.message("DelaunayTriangulationDataStorage", "Delaunay Triangulation done: " + dt);
	Debug.message("DelaunayTriangulationDataStorage", "Convex Hull: " + convexHullEdges);

	ArrayList<DataObject> dataObjects = new ArrayList<DataObject>();
	ArrayList<String> omColumnNames = new ArrayList<String>();
	ArrayList<String> dataColumnNames = new ArrayList<String>();
	ArrayList<String> dataColumnAliases = new ArrayList<String>();
	
	omColumnNames.add(omNameTriangle);
	for (int i = 0; i < 3; i++)
	    for (String dataColumnName : this.dataColumnNames)
		dataColumnNames.add(dataColumnName+"_"+i);
	
	if (drawVoronoi)
	    omColumnNames.add(omNameVoronoiLine);

	if (keepVertexes)
	    for (int i = 0; i < 3; i++)
		omColumnNames.add(omNameVertex+i);

	initBoundaryTriangles(dt);
	
	for (Simplex<DataPoint> triangle: dt) {

	    if (isArtefactTriangles(triangle))
		continue; // skeep the triangle as being degenerated
	    
	    DataObject data = new DataObject();
	    BoundingBox bb = null;
	    // 0- Voronoï -- only lines at the moment...
	    if (drawVoronoi) {
		OMGraphicList omgl = new OMGraphicList();
		Pnt p = Pnt.circumcenter(triangle.toArray(new Pnt[3]));
		for (Simplex<DataPoint> other: dt.neighbors(triangle)) {
		    if (isArtefactTriangles(other))
			continue; // skeep the triangle as being degenerated
	    
		    Pnt q = Pnt.circumcenter(other.toArray(new Pnt[3]));
		    omgl.add(new OMLine(p.coord(1), p.coord(0),
					q.coord(1), q.coord(0),
					OMGraphic.LINETYPE_STRAIGHT));
		    if (bb == null) 
			bb = new BoundingBox(p.coord(0), p.coord(1), q.coord(0), q.coord(1));
		    else {
			bb.extendTo(p.coord(0), p.coord(1));
			bb.extendTo(q.coord(0), q.coord(1));
		    }
		}
		data.addOMDataLocation(omNameVoronoiLine, 	
				       new OMDataLocation(omgl, 
							  new Centroid(p.coord(0), p.coord(1)),
							  bb));
	    }

	    {
		// 1-  add the vertexes associated data
		int i = 0;
		for (DataPoint pt : triangle) {
		    if (keepVertexes) // add the 3 vertices
			data.addOMDataLocation(omNameVertex+i, 	
					       new OMDataLocation(new OMPoint(pt.coord(1), pt.coord(0)),
								  new Centroid(pt.coord(0), pt.coord(1)),
								  new BoundingBox(pt.coord(0), pt.coord(1),
										  pt.coord(0), pt.coord(1))));
		    
		    for (String dataColumnName : this.dataColumnNames)
			data.addData(dataColumnName+"_"+i, ((DataObject)pt.getData()).getData(dataColumnName));
		    ++i;
		}
	    }
	    
	    {
		// 2- convert the simplex triangle into an OMPoly
		double [] ll = new double[2*(3+1)];
		int t = 0;
		double clon = 0, clat = 0;
		bb = null;
		for (DataPoint pt : triangle) {
		    double lon = pt.coord(0);
		    double lat = pt.coord(1);
		    if (bb == null) 
			bb = new BoundingBox(lon, lat, lon, lat);
		    else 
			bb.extendTo(lon, lat);
		    clon += lon;
		    clat += lat;
		    ll[t++] = lat;
		    ll[t++] = lon;
		}
		// close the triangle
		ll[t++] = ll[0];
		ll[t++] = ll[1];

		OMGraphic omTriangle = new OMPoly(ll, OMGraphic.DECIMAL_DEGREES, 
						OMGraphic.LINETYPE_STRAIGHT);
		if (isBoundaryTriangles(triangle)) { // add convex hull info
		    omTriangle.putAttribute(convexHullEdgeAttribute, getConvexHullEdge(triangle));
		}

		data.addOMDataLocation(omNameTriangle, 	
				       new OMDataLocation(omTriangle,
							  new Centroid(clon/3, clat/3), 
							  bb)); 
	    }
	    dataObjects.add(data);
	}

	dataTable.setTable(dataObjects,
			   dataColumnNames,
			   omColumnNames,
			   dataColumnAliases);

	// data are stored, we can create dataTableModel and panel to view them.
	dataTableModel = new DataTableModel(dataTable);
	resetDataTableStoragePanel(dataTableModel);
	initialized = true;

	// tell the listener we're ready
	fireDataStorageEvent();
    }
    
    public  JComponent getGUI() {
	
	JPanel pan = new JPanel();
	Box delaunayPanel = Box.createVerticalBox();

	//delaunayPanel.add(new JLabel("<html>Delaunay<p>We shall add two lists<br>in order to select<br> dataColumnNames</html>"));
	int index = 0;
        final String[] datas = getDataStorage().getDefaultDataTableModel().getDataColumnNames();
        CheckListItem[] checkListItem = new CheckListItem[datas.length];
        for (String v : datas){ 
            checkListItem[index] = new CheckListItem(v);
            checkListItem[index].setSelected(dataColumnNames.contains(v));
            index++;
        }
        final JList list = new JList(checkListItem);
        list.setCellRenderer(new CheckListRenderer());
        list.setSelectionMode(
                ListSelectionModel.SINGLE_SELECTION);
        list.addMouseListener(new MouseAdapter() {
            @Override
            public void mouseClicked(MouseEvent event) {
                dataColumnNames.clear();
                //filterColumnValues.clear();
                JList list = (JList) event.getSource();
                // Get index of item clicked
                int index = list.locationToIndex(event.getPoint());
                CheckListItem item = (CheckListItem) list.getModel().getElementAt(index);
                // Toggle selected state
                item.setSelected(!item.isSelected());
                // Repaint cell
                list.repaint(list.getCellBounds(index, index));

                for (int i = 0; i < list.getModel().getSize(); i++) {
                    CheckListItem cItem = (CheckListItem) list.getModel().getElementAt(i);
                    if (cItem.isSelected()) {
                        dataColumnNames.add(datas[i]);
                    }
                }

                //setFilterColumnValues(filterColumnValues);
            }
        });
        final JCheckBox checkAll = new JCheckBox(i18n.get(this, "selectall", "Select all"));
        checkAll.addActionListener(new ActionListener() {

            public void actionPerformed(ActionEvent arg0) {
                dataColumnNames.clear();
                for (int i = 0; i < list.getModel().getSize(); i++) {
                    CheckListItem cItem = (CheckListItem) list.getModel().getElementAt(i);
                    cItem.setSelected(checkAll.isSelected());
                    if (checkAll.isSelected()) {
                        dataColumnNames.add(datas[i]);
                    }
                }
                list.repaint();
            }
        });


        JComboBox geometryList = null;
	try {
	    geometryList = new JComboBox(getDataStorage().getDefaultDataTableModel().getOMColumnNames());
	} catch (Exception ex) {
            Debug.error("Geometry Column Name List is unavailable because of "+ ex);
            geometryList = new JComboBox();
	}

	geometryList.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                try {
                    int index = ((JComboBox) e.getSource()).getSelectedIndex();
                    setGeometryColumnName(getDataStorage().getDefaultDataTableModel().getRealOMColumnName(index));
                }
		catch (ArrayIndexOutOfBoundsException ex) {
                }
            }
        });

        JPanel pGeom = new JPanel();
        pGeom.add(new JLabel(i18n.get(this, "centroid", "Centroid")));
        pGeom.add(geometryList);
	delaunayPanel.add(pGeom);
        JPanel pTop = new JPanel();
        pTop.add(new JLabel(i18n.get(this, "data", "Data")));
        pTop.add(checkAll);
        delaunayPanel.add(pTop);
        delaunayPanel.add(new JScrollPane(list));
	final JCheckBox checkDrawVoronoi = new JCheckBox(i18n.get(this, "drawVoronoi", "Draw Voronoï"));
	checkDrawVoronoi.setSelected(drawVoronoi);
	checkDrawVoronoi.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent arg0) {
		    setDrawVoronoi(checkDrawVoronoi.isSelected());
            }
        });
        fieldVoronoiName = new JTextField(omNameVoronoiLine, 15);
        JPanel pVoronoi = new JPanel();
        pVoronoi.add(checkDrawVoronoi);
        pVoronoi.add(new JLabel(i18n.get(this, "attrName", " using name:")));
        pVoronoi.add(fieldVoronoiName);

	delaunayPanel.add(pVoronoi);

	final JCheckBox checkKeepVertexes = new JCheckBox(i18n.get(this, "keepVertexes", "Add Vertexes"));
	checkKeepVertexes.setSelected(keepVertexes);
	checkKeepVertexes.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent arg0) {
		    setKeepVertexes(checkKeepVertexes.isSelected());
            }
        });
        fieldVertexesName = new JTextField(omNameVertex, 15);
        JPanel pVertexes = new JPanel();
        pVertexes.add(checkKeepVertexes);
        pVertexes.add(new JLabel(i18n.get(this, "attrName", " using name:")));
        pVertexes.add(fieldVertexesName);
	delaunayPanel.add(pVertexes);

	JButton validateButton = new JButton(i18n.get(this, "validate", "Validate"));
	validateButton.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
                    omNameVertex = fieldVertexesName.getText();
                    omNameVoronoiLine = fieldVoronoiName.getText();
		    initialize();
		}
	    });

	delaunayPanel.add(validateButton);

	pan.add(delaunayPanel);

	Component dsGUI = getDataStorage().getGUI();
        
        pan.add(dsGUI);

        // wrap the panel in a flow layout panel in order to shrink the
        // components
        JPanel wrapPanel = new JPanel();
        wrapPanel.add(pan);

        //wrapPanel.add(testLabel);

        // set opacity
        wrapPanel.setBackground(Color.GRAY);
        wrapPanel.setOpaque(false);
       
	// pan.setBackground(Color.GRAY);
        // pan.setOpaque(false);
        // panssionPanel.setBackground(Color.GRAY);
        // panssionPanel.setOpaque(false);
        // pan.setVisible(false);
        // pan.setVisible(true);
        wrapPanel.setBorder(BorderFactory.createTitledBorder(i18n.get(this, "name", "Triangulation")));
        return wrapPanel;
    }
    
    public void setProperties(String prefix, java.util.Properties properties) { 
	prefix = PropUtils.getScopedPropertyPrefix(prefix);
	super.setProperties(prefix, properties);

	prefix = PropUtils.getScopedPropertyPrefix(prefix) + delaunayDSPropertyString + ".";
	String dsClassName = properties.getProperty(prefix + dataStorageClassPropertyString, "");
	if ("".equals(dsClassName))
	    // empty Delaunay Data Storage ????
	    return;
	setDataStorage(Util.loadDataStorage(dsClassName, prefix, properties, this));

        geometryColumnName = properties.getProperty(prefix + geometryColumnNamePropertyString, "shape");
        String[] dcn = properties.getProperty(prefix + dataColumnNamesPropertyString, "").split("[ \t]+");
        for(String s : dcn)
         dataColumnNames.add(s);
	
	drawVoronoi = Util.isTrue(properties.getProperty(prefix + drawVoronoiPropertyString, "false"));
	keepVertexes = Util.isTrue(properties.getProperty(prefix + keepVertexesPropertyString, "false"));

        omNameVertex = properties.getProperty(prefix + vertexesNamePropertyString, omNameVertex);
        omNameVoronoiLine = properties.getProperty(prefix + voronoiNamePropertyString, omNameVoronoiLine);
	initialize();
    }

    public java.util.Properties getProperties(java.util.Properties properties) {
	return getProperties("", properties);
    }

    public java.util.Properties getProperties(String prefix, java.util.Properties properties) {
	properties = super.getProperties(prefix, properties);
	prefix = PropUtils.getScopedPropertyPrefix(prefix) + delaunayDSPropertyString + "."; 
	if (getDataStorage() != null) {
	    properties.put(prefix + dataStorageClassPropertyString, getDataStorage().getClass().getCanonicalName());
	    // Debug.message("DataStorage", "PivotDataStorage.getProperties() - DataStorage member is a: "+getDataStorage().getClass().getCanonicalName());
	    properties = getDataStorage().getProperties(prefix, properties);
	}
	
	StringBuffer dataColumnNamesString = new StringBuffer();
        for (int i = 0; i < dataColumnNames.size(); i++) {
            dataColumnNamesString.append(dataColumnNames.get(i));
            dataColumnNamesString.append(" ");
        }
	properties.put(prefix + dataColumnNamesPropertyString, dataColumnNamesString.toString());
        properties.put(prefix + geometryColumnNamePropertyString, geometryColumnName);

	properties.put(prefix + drawVoronoiPropertyString, new Boolean(drawVoronoi).toString());
	properties.put(prefix + keepVertexesPropertyString, new Boolean(keepVertexes).toString());
        properties.put(prefix + vertexesNamePropertyString, omNameVertex);
        properties.put(prefix + voronoiNamePropertyString, omNameVoronoiLine);
	return properties;
    }

}
