// <header>
/**
 *  ThemaMap : The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
/*
 * GridDataStorage.java
 *
 * Created on 8 january 2009
 *
 * @author Jacques.Madelaine@info.unicaen.fr
 *
 */
package fr.unicaen.openmap.dataStorage;

import com.bbn.openmap.util.Debug;

import com.bbn.openmap.util.PropUtils;

import javax.swing.*;




import java.util.*;

import fr.unicaen.openmap.util.*;
import fr.unicaen.openmap.dataStorage.gui.DataTableModel;
import fr.unicaen.openmap.dataStorage.gui.DataTableStoragePanel;

import com.bbn.openmap.omGraphics.OMGraphic;
import com.bbn.openmap.omGraphics.OMRect;
import fr.unicaen.openmap.dataStorage.gui.GridDataStoragePanel;
import fr.unicaen.openmap.exception.UnavailableColumnNameException;

/**
 * Creates a regular grid given : a size and an origin.
 * populate the grid by aggregating values in a cell with a given aggregate function
 * as soon as the centroid of its DataLocation is within that cell
 */
public class GridDataStorage extends PivotDataStorage {
    /* --------------------------  Begin   Properties  ---------------------*/

    public static final String gridDSAdaptorPropertyString = "gridDSA";
    public static final String gridGeometryColumnNamePropertyString = "gridGeometryColumnName";
    public static final String latitudeResolutionPropertyString = "latitudeResolution";
    public static final String longitudeResolutionPropertyString = "longitudeResolution";
    public static final String geometryColumnNamePropertyString = "geometryColumnName";
    /* ---------------------------  End     Properties  ---------------------*/
    /** Geometry used for setting grid cell */
    protected String geometryColumnName = "";
    /** Geometry column name of grid cells in the new GridDataStorage */
    private String gridGeometryColumnName = "grid";
    // protected float LatOrigin = 0F, LonOrigin = 0F; // always (0,0) !!!
    /** The vertical/latitude interval between rows in decimal degrees */
    protected float latitudeResolution = 0.1F;
    /** The horizontal/longitude interval between columns in decimal degrees */
    protected float longitudeResolution = 0.1F;
    private LonLatNames lonLatNames;    // suffixes of column names starting with longitude or latitude
    private int nbStat = 0;

    public GridDataStorage(DataStorageListener dsl) {
        super(dsl);
    }

    public GridDataStorage(DataStorageListener dsl, DataStorage dataStorage) {
        super(dsl, dataStorage);
    }

    public void setProperties(String prefix, java.util.Properties properties) {
        setInitialized(false);
        prefix = PropUtils.getScopedPropertyPrefix(prefix);
        try {
	    // do it before super.setProperties in order not to have attribute name conflict
            setGridGeometryColumnName(properties.getProperty(prefix + gridGeometryColumnNamePropertyString, getGridGeometryColumnName()));
        } catch (UnavailableColumnNameException e) {
            JOptionPane.showMessageDialog(null, e.getMessage(), "", JOptionPane.ERROR_MESSAGE);
        }
        super.setProperties(prefix, properties);
        geometryColumnName =
                properties.getProperty(prefix + pivotDSPropertyString + "." +
                geometryColumnNamePropertyString, "");
        latitudeResolution = PropUtils.floatFromProperties(properties, prefix + latitudeResolutionPropertyString, 0.1F);
        longitudeResolution = PropUtils.floatFromProperties(properties, prefix + longitudeResolutionPropertyString, 0.1F);
        initialize(); // in case we received a DataStorageEvent meanwhile...
    }

    public java.util.Properties getProperties(String prefix, java.util.Properties properties) {
        properties = super.getProperties(prefix, properties);
        properties.put(prefix + pivotDSPropertyString + "." +
                geometryColumnNamePropertyString,
                geometryColumnName);
        properties.put(prefix + gridGeometryColumnNamePropertyString, getGridGeometryColumnName());
        properties.put(prefix + latitudeResolutionPropertyString, String.valueOf(latitudeResolution));
        properties.put(prefix + longitudeResolutionPropertyString, String.valueOf(longitudeResolution));
        return properties;
    }
    private int minLatIndex, maxLatIndex, minLonIndex, maxLonIndex;

    protected void initMinMaxIndices() {
        minLatIndex = (int) (-90 / latitudeResolution) + 1;
        maxLatIndex = (int) (90 / latitudeResolution) - 1;
        minLonIndex = (int) (-180 / longitudeResolution) + 1;
        maxLonIndex = (int) (180 / longitudeResolution) - 1;
        Debug.message("DataStorage", "[GridDataStorage] initMinMaxIndices " + minLatIndex + ", " + minLonIndex + " x  " + maxLatIndex + ", " + maxLonIndex + " res " + latitudeResolution + " x " + longitudeResolution + " " + getInitialized());
    }

    /**
     * @return the gridGeometryColumnName
     */
    public String getGridGeometryColumnName() {
        return gridGeometryColumnName;
    }

    /**
     * @param gridGeometryColumnName the gridGeometryColumnName to set
     */
    public void setGridGeometryColumnName(String gridGeometryColumnName) throws UnavailableColumnNameException {
	if (this.gridGeometryColumnName.equals(gridGeometryColumnName)) {
	    // in case of reload
	    return;
	}

        if (isAnExistingColumnName(gridGeometryColumnName))
            throw new UnavailableColumnNameException(UnavailableColumnNameException.DS_COLUMN_MESS);
	this.gridGeometryColumnName = gridGeometryColumnName;
    }

    /** 
     * Utility class used to identify a grid cell with a lat, lon pair.
     * May compute an <code>OMDataLocation</code> representing that cell with 
     * the method <code>newOMDataLocation()</code>.
     */
    public class GridCellID {

	/**  beware these are only indices and not true lat/lon */
        public int lat, lon; 

        public GridCellID(int lat, int lon) {
            this.lat = lat;
            this.lon = lon;
        }

        public GridCellID(double lat, double lon) {
            int _lat = (int) Math.floor(lat / latitudeResolution);
            int _lon = (int) Math.floor(lon / longitudeResolution);
            if (_lat > maxLatIndex) {
                _lat = maxLatIndex;
            } else if (_lat < minLatIndex) {
                _lat = minLatIndex;
            }
            if (_lon > maxLonIndex) {
                _lon = maxLonIndex;
            } else if (_lon < minLonIndex) {
                _lon = minLonIndex;
            }
            this.lat = _lat;
            this.lon = _lon;
            // Debug.message("DataStorage", "GridCellId("+lat+", "+lon+") => "+ this);
        }

        /** computes an OMRect representing that grid cell */
        public OMDataLocation newOMDataLocation() {
            float lt1 = (lat + 1) * latitudeResolution; // latitude of north edge, decimal degrees.
            float ln1 = lon * longitudeResolution;  // longitude of west edge, decimal degrees.
            float lt2 = lat * latitudeResolution;   // latitude of south edge, decimal degrees.
            float ln2 = (lon + 1) * longitudeResolution;// longitude of east edge, decimal degrees.

            OMDataLocation omdl =
                    new OMDataLocation(new OMRect(lt1, ln1, lt2, ln2, OMGraphic.LINETYPE_STRAIGHT),
                    new Centroid((lon + 0.5) * longitudeResolution, (lat + 0.5) * latitudeResolution),
                    new BoundingBox(ln1, lt1, ln2, lt2));
            // System.out.println("Grid new omdl for ("+lat+", "+lon+") => "+omdl);
            return omdl;
        }

        public int hashCode() {
            return lat * 100000 + lon;
        }

        public boolean equals(Object o) {
            return lat == ((GridCellID) o).lat && lon == ((GridCellID) o).lon;
        }

        public String toString() {
            return "(" + lat + ", " + lon + ")";
        }
    } // end class GridCellID

    public void setGeometryColumnName(String name) {
        geometryColumnName = name;
    }

    public String getGeometryColumnName() {
        return geometryColumnName;
    }

    public void setLatitudeResolution(float res) {
        latitudeResolution = res;
    }

    public void setLongitudeResolution(float res) {
        longitudeResolution = res;
    }

    public float getLatitudeResolution() {
        return latitudeResolution;
    }

    public float getLongitudeResolution() {
        return longitudeResolution;
    }

    public void initialize() {
        if (getDataStorage() == null || !getDataStorage().isReady()) {
            Debug.error("[GridDataStorage] can't initialize because dataStorage is not ready !!!!");
            return;
        }
        Debug.message("GridDataStorage", "GridDataStorage.initialize() called.");

        DataTableModel dtm = getDataStorage().getDefaultDataTableModel();
        if ("".equals(geometryColumnName)) {
            if (dtm.getOMColumnNames().length < 1) {
                Debug.error("[GridDataStorage] can't initialize because there is no geometry !!!!");
                return;
            }
            geometryColumnName = dtm.getOMColumnName(0);
            Debug.error("[GridDataStorage] set geometry to " + geometryColumnName);
        }

        // test if gridGeometryColumnName already exists
        for (String name : dtm.getOMColumnNames()) {
            if (getGridGeometryColumnName().equals(name)) {
                try {
                    setGridGeometryColumnName(getGridGeometryColumnName() + "_a");

                } catch (UnavailableColumnNameException e) {
                    JOptionPane.showMessageDialog(null, e.getMessage(), "", JOptionPane.ERROR_MESSAGE);
                }
            }
        }

        Debug.message("DataStorage", "[GridDataStorage] start initialize");

        lonLatNames = new LonLatNames();

	// new data column names
        // merge stat and criteria
        ArrayList<String> dataColumnNames = new ArrayList<String>(getStatisticsAttributesList());
        for (String a : getCriteriaAttributesList()) {
            dataColumnNames.add(a);
        }
        dataColumnNames.add("lon_centroid_" + getGridGeometryColumnName());
        dataColumnNames.add("lat_centroid_" + getGridGeometryColumnName());

        ArrayList<String> dataColumnAliases = null; // should be updated --TBD

        // we need a temporary map in order to store
        // for each pair < GridCellID, DataObject with criteria values > a DataObject with stats
        HashMap<GridCellID, HashMap<DataObject, DataObject>> grid =
                new HashMap<GridCellID, HashMap<DataObject, DataObject>>();

        initMinMaxIndices(); // indices used by GridCellId

        for (int i = 0; i < getDataStorage().getCount(); i++) {
            DataObject currentDataObject = getDataStorage().getElementAt(i);

            // get the centroid
            OMDataLocation omdl = currentDataObject.getOMDataLocation(geometryColumnName);
            if (omdl == null) {
                continue;
            }
            Centroid centroid = omdl.getCentroid();
            if (centroid == null) {
                continue;
            }

            GridCellID gcid = new GridCellID(centroid.getLat(), centroid.getLon());

            // collect criteria attributes values
            DataObject criteriaDataObject = new DataObject();
            for (String criterion : getCriteriaAttributesList()) {
                criteriaDataObject.addData(criterion, currentDataObject.getData(criterion));
            }

            // gridData is a map that gives for each criteria attributes value
            // a map of statistics aggregate value for each statistic attribute
            // (should be called cellData ?)
            HashMap<DataObject, DataObject> gridData = grid.get(gcid);
            if (gridData == null) {
                gridData = new HashMap<DataObject, DataObject>();
                // System.out.println("new grid cell " + gcid);
            }
            //
            DataObject gridDataForCriteria = gridData.get(criteriaDataObject);
            if (gridDataForCriteria == null) {
                gridDataForCriteria = new DataObject();
                // System.out.println("new grid data " + gcid + " for " + criteriaDataObject);
            }
            for (String name : getStatisticsAttributesList()) {
                // get and add data according to the value of pivot columns
                Object previous = gridDataForCriteria.getData(name);
                Object current = currentDataObject.getData(name);
                gridDataForCriteria.addData(name, operate(previous, current, getAggregateFunction(name)));
            }
            gridData.put(criteriaDataObject, gridDataForCriteria);
            grid.put(gcid, gridData);
        }

        for (String attr : getStatisticsAttributesList()) {
            lonLatNames.getLonLatType(attr);
        }

        for (String attr : getCriteriaAttributesList()) // Is it useful ???
        {
            lonLatNames.getLonLatType(attr);
        }

        boolean buildCentroid = lonLatNames.getLonLatNamesSuffixes().size() > 0;

        if (buildCentroid) {
            Debug.message("DataStorage", "[GridDataStorage] build centroid " + lonLatNames);
        }

        ArrayList<DataObject> gridDataObjects = new ArrayList<DataObject>();
        // gather all data in gridDataObjects
        for (Map.Entry<GridCellID, HashMap<DataObject, DataObject>> gridEntry : grid.entrySet()) {
            OMDataLocation gridOMDataLocation = gridEntry.getKey().newOMDataLocation();
            // for each criteria value, we apply the aggregate functions to the statistics values
            for (Map.Entry<DataObject, DataObject> e : gridEntry.getValue().entrySet()) {
                DataObject data = e.getValue();
                // 1- complete the computation of aggregate values
                replaceAggregateResultByRealResult(data, dataColumnNames);
                // 2- add the grid cell as a OMGraphic
                data.addOMDataLocation(getGridGeometryColumnName(), gridOMDataLocation);
                // 2b- add the centroid lat/lon to the data
                data.addData("lon_centroid_" + getGridGeometryColumnName(), gridOMDataLocation.getCentroid().getLon());
                data.addData("lat_centroid_" + getGridGeometryColumnName(), gridOMDataLocation.getCentroid().getLat());
                // 3- merge the criteria values
                data.putAll(e.getKey());
                // 4- built lon/lat centroïd
                if (buildCentroid) {
                    for (String suffix : lonLatNames.getLonLatNamesSuffixes()) {
                        try {
                            lonLatNames.putLon(suffix,
                                    data.getData(lonLatNames.getLonName(suffix)));
                            lonLatNames.putLat(suffix,
                                    data.getData(lonLatNames.getLatName(suffix)));
                        } catch (java.lang.NullPointerException exception) {
                            // one lat or lon does not exist...
                        } catch (java.lang.Exception exception) {
                            System.err.println("*** GridDataStorage bad cast to Double building centroid for \"" + data.getData(lonLatNames.getLonName(suffix)) + "\", \"" + data.getData(lonLatNames.getLatName(suffix)) + "\"");
                        }
                    }
                    lonLatNames.putCentroids(data);
                }
                // 5- add the data row to the table
                gridDataObjects.add(data);
                // 6- if there exists lat/lon pair in the attributes build a centroid
            }
        }

	// add omColumnNames
        ArrayList<String> omColumnNames = lonLatNames.getOMGraphicNames();
        omColumnNames.add(0, getGridGeometryColumnName());

	// store the data
        dataTable.setTable(gridDataObjects, dataColumnNames, omColumnNames, dataColumnAliases);

        // data are stored, we can create dataTableModel and panel to view them.
        dataTableModel = new DataTableModel(dataTable);
        resetDataTableStoragePanel(dataTableModel);
        setInitialized(true);
        Debug.message("DataStorage", "[GridDataStorage] (re)initialized");

        // tell the listener we're ready
        fireDataStorageEvent();
    }

    public JComponent getGUI() {
        return new GridDataStoragePanel(this);
    }
}
