// <header>
/**
 *  ThemaMap : The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
package fr.unicaen.openmap.dataStorage;

import com.bbn.openmap.I18n;
import com.bbn.openmap.util.Debug;
import com.bbn.openmap.util.PropUtils;

import java.io.*;
import java.sql.*;
import java.util.*;
import java.awt.Color;
import java.awt.event.*;
import java.awt.Dimension;
import java.awt.BorderLayout;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

import javax.swing.*;

import fr.unicaen.openmap.util.Util;
import fr.unicaen.openmap.util.LonLatNames;
import fr.unicaen.openmap.util.propertyEditor.Inspector;
import fr.unicaen.openmap.dataStorage.gui.DataTableModel;
import fr.unicaen.openmap.dataStorage.gui.DataTableStoragePanel;
import fr.unicaen.openmap.util.propertyEditor.DataStoragePropertyEditor;
import fr.unicaen.openmap.dataStorage.gui.SQLEditor;
import fr.unicaen.openmap.dataStorage.postgis.db.Factory;

public class JDBCDataStorage extends InformationDataStorage
{
	// ---- DATA

	// Data
    private ArrayList<DataObject> dataObjects;
    private ArrayList<String> columnNames;
    private ArrayList<String> typeNames;      // string, lon or lat
    private LonLatNames lonLatNames;    // suffixes of column names starting with longitude or latitude

	// Properties
	private final static String baseNamePropertyString = "baseName";
	private final static String driverPropertyString = "driver";
	private final static String queryPropertyString = "query";
	private String baseName = "";
	private String driver = "";
	private String query = "";

    protected boolean initialized = false;

	// ---- METHODS

    public JDBCDataStorage(DataStorageListener dsl)
	{
        super(dsl);
        Debug.message("DataStorage", this + " JDBCDataStorage(DataStorageListener dsl)");
    }

    public String getName()
	{
        return baseName;
    }

    public boolean isReady()
	{
        return initialized;
    }

    public void setProperties(String prefix, java.util.Properties properties)
	// file -> class
	{
        prefix = PropUtils.getScopedPropertyPrefix(prefix);
        super.setProperties(prefix, properties);

		baseName = Util.unEscapePropertyString( properties.getProperty(prefix + baseNamePropertyString) );
		driver = properties.getProperty(prefix + driverPropertyString);
		query = properties.getProperty(prefix + queryPropertyString);

        initialize();
    }

    public Properties getProperties(String prefix, Properties properties)
	// class -> file
	{
        properties = super.getProperties(prefix, properties);
        prefix = PropUtils.getScopedPropertyPrefix(prefix);

		properties.put(prefix + baseNamePropertyString, baseName);
		properties.put(prefix + driverPropertyString, driver);
		properties.put(prefix + queryPropertyString, query == null ? "" : query );

        return properties;
    }

    public Properties getPropertyInfo(Properties list)
	{
		// Build configuration screen

        list = super.getPropertyInfo(list);
		String propertyList = baseNamePropertyString + " " + driverPropertyString;

        PropUtils.setI18NPropertyInfo(i18n,
                list,
                JDBCDataStorage.class,
                baseNamePropertyString,
                baseNamePropertyString, //label
                "Database URL", //tooltip
                null);

        PropUtils.setI18NPropertyInfo(i18n,
                list,
                JDBCDataStorage.class,
                driverPropertyString,
                driverPropertyString, //label
                "Java class to use as JDBC driver", //tooltip
                "fr.unicaen.openmap.util.propertyEditor.JDBCDriverPropertyEditor");

		// Add SQL property editor only in option panel
		if( isReady() )
		{
			PropUtils.setI18NPropertyInfo(i18n,
                list,
                JDBCDataStorage.class,
                queryPropertyString,
                queryPropertyString, //label
                "SQL query to execute", //tooltip
                "fr.unicaen.openmap.util.propertyEditor.TextAreaPropertyEditor");	

			propertyList = propertyList + " " + queryPropertyString;
		}

        list.put(initPropertiesProperty, propertyList); 
        return list;
    }

	private String getTypeName(int i)
	{
		if (i >= typeNames.size())
		{
            typeNames.add("string");
        }
        return typeNames.get(i);
    }

    public void initialize()
	{
        initialized = false;
        dataObjects = new ArrayList<DataObject>();
        columnNames = new ArrayList<String>();
        typeNames = new ArrayList<String>();
        lonLatNames = new LonLatNames();

		Connection conn = null;
		Statement stmt = null;
		ResultSet rs = null;

        Inspector inspector = new Inspector();

		// First configuration screen ( Driver + DB URL )
        if ( baseName == null || driver == null || baseName.equals("") || driver.equals("") )
		{
			do
			{
				JPanel p = new JPanel(new BorderLayout());
				String f = "";
				if( baseName == null )
				{
					f = new String("Base");
					baseName = new String("");
				}
				if( driver == null )
				{
					f = new String("Driver");
					driver = new String("");
				}
				JLabel lab = new JLabel("<html><b>Field '" + f + "' is empty</b></html>");
				lab.setForeground(Color.RED);
				p.add(lab, BorderLayout.NORTH);
				p.add(DataStoragePropertyEditor.provideSimpleCustomPanel(this, inspector), BorderLayout.CENTER);

				int choice = JOptionPane.showConfirmDialog(null, p, "Please", JOptionPane.OK_CANCEL_OPTION, JOptionPane.INFORMATION_MESSAGE);
				if (choice == JOptionPane.CANCEL_OPTION)
				{
					return;
				}
				setProperties(inspector.collectProperties());
			}
			while( baseName == null || driver == null || baseName.equals("") || driver.equals("") );

            return;
        }
		else // Check parameters
		{
			try
			{
				// Driver ?
				Object o = Class.forName(driver).newInstance();
				Debug.message("DataStorage", this + " JDBC driver loaded ("+driver+")");

				if( ! Driver.class.isAssignableFrom( o.getClass() ) )
				{
					throw new Exception();
				}
			}
			catch( Exception e )
			{
				// Alert
				JOptionPane.showMessageDialog(null, "Class \""+driver+"\" is not a valid JDBC driver", "Error", JOptionPane.INFORMATION_MESSAGE);
				driver = null;
				initialize();
			}

			try
			{
				// Base Valid ?
				DriverManager.getConnection( baseName );
				Debug.message("DataStorage", this + " Connection OK ("+baseName+")");
			}
			catch( Exception e )
			{
				// Alert
				e.printStackTrace();
				JOptionPane.showMessageDialog(null, "Database URL \""+baseName+"\" is invalid", "Error", JOptionPane.INFORMATION_MESSAGE);
				baseName = null;
				initialize();
			}
		}

		// Connection OK
		// Get query
		if( query == null || query.equals("") )
		{
			do
			{
				SQLEditor sql = null;
				try
				{
					conn = DriverManager.getConnection( baseName );
					sql = new SQLEditor( conn );
				}
				catch( Exception e )
				{
					e.printStackTrace();
				}

				int choice = JOptionPane.showConfirmDialog(null, sql.getPanel(), "Please", JOptionPane.OK_CANCEL_OPTION, JOptionPane.INFORMATION_MESSAGE);
				if (choice == JOptionPane.CANCEL_OPTION)
				{
					return;
				}
				query = sql.getText();
				initialize();
			}
			while( query == null || query.equals("") );

            return;
        }

		ArrayList<String> omColumnNames = new ArrayList<String>();

		try
		{
			conn = DriverManager.getConnection( baseName );
            stmt = conn.createStatement();
            rs = stmt.executeQuery(query);

			Debug.message("DataStorage", this + " Statement OK ("+query+")");

			ResultSetMetaData rsmd = rs.getMetaData();
			int columnCount = rsmd.getColumnCount();

			for( int i = 1 ; i <= columnCount ; i++ )
			{
				String colName = rsmd.getColumnName(i);
				String type = lonLatNames.getLonLatType(colName);

				columnNames.add( colName );

                if (type != null) // we have a lon or lat column name
                {
                    typeNames.add(type);
                }
				else
				{
                    typeNames.add("string");
                }

				// WKT Handling
				if( colName.startsWith("geometry_" ) || colName.startsWith("centroid_" ) || colName.startsWith("bounding_" ) )
				{
					omColumnNames.add(colName);
				}
			}

			int lineCount = 1;
			while (rs.next())
            {
				DataObject dataObject = new DataObject();
                for ( int i = 1 ; i <= columnCount ; i++ )
                {
					String type = getTypeName(i-1);
					String name = rsmd.getColumnName(i);
					String data = ""+rs.getObject(i);

					// WKT Handling
					if( name.startsWith("geometry_" ) )
					{
						dataObject.addOMDataLocation( name,  new OMDataLocation( Factory.createOMGraphic(data) , null, null));
					}
					else
					if( name.startsWith("centroid_" ) )
					{
						dataObject.addOMDataLocation( name,  new OMDataLocation( null , Factory.createCentroid(data) , null));
					}
					else
					if( name.startsWith("bounding_" ) )
					{
						dataObject.addOMDataLocation( name,  new OMDataLocation( null , null,  Factory.createBoundingBox(data)));
					}

					dataObject.addData(rsmd.getColumnName(i), data );

					try
					{
						if (LonLatNames.LON.equals(type))
						{
							lonLatNames.putLon(i-1, data);
						}
						else
						if (LonLatNames.LAT.equals(type))
						{
                        	lonLatNames.putLat(i-1, data);
						}
					}
					catch (java.lang.ClassCastException e)
					{
                    	System.err.println("*** CSVFileDataStorage bad cast to Double line " + lineCount + " for " + type + " \"" + data + "\"");
					}
					catch (java.lang.NumberFormatException e)
					{
						System.err.println("*** CSVFileDataStorage number format exception line " + lineCount + " for " + type + " \"" + data + "\"");
					}
                }
				lonLatNames.putCentroids(dataObject);
				dataObjects.add(dataObject);
				lineCount++;
            }
			conn.close();
		}
		catch (Exception e)
        {
            System.err.println("Exception: "+e.getMessage());
            e.printStackTrace();
        }

		omColumnNames.addAll(0, lonLatNames.getOMGraphicNames());
        dataTable.setTable(dataObjects, getDataAttributesNamesToDisplay(columnNames), omColumnNames, getDataAttributesNamesAliases());

        //// data are stored, we can create dataTableModel and panel to view them.
        dataTableModel = new DataTableModel(dataTable);
        resetDataTableStoragePanel(dataTableModel);
        initialized = true;
        getDataStorageListener().dataStorageEvent(this);
    }

    public JComponent getGUI()
	{
        JPanel panel = new JPanel();
        panel.add(DataStoragePropertyEditor.provideCustomPanel(this));
        return panel;
    }
}
