// <header>
/**
 *  ThemaMap : The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
package fr.unicaen.openmap.dataStorage;

import com.bbn.openmap.util.Debug;
import com.bbn.openmap.util.FileUtils;
import com.bbn.openmap.I18n;

import com.bbn.openmap.util.PropUtils;

import java.util.*;
import java.io.*;
import javax.swing.*;
import java.awt.event.*;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.ProtocolException;

import fr.unicaen.openmap.util.Util;
import fr.unicaen.openmap.util.LonLatNames;
import fr.unicaen.openmap.util.propertyEditor.DataStoragePropertyEditor;
import fr.unicaen.openmap.util.propertyEditor.FileTypeDescription;
import fr.unicaen.openmap.dataStorage.gui.DataTableModel;
import fr.unicaen.openmap.dataStorage.gui.DataTableStoragePanel;
import fr.unicaen.openmap.util.propertyEditor.Inspector;
import java.awt.BorderLayout;
import java.awt.Color;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.net.URL;

import org.json.*;

/**
 * JSONDataStorage imports its content using a REST server
 *
 * @author Jacques.Madelaine@info.unicaen.fr
 */
public class JSONDataStorage extends InformationDataStorage {

    public static final String urlPropertyString = "url";
    public static final String queryPropertyString = "query";
    ///** used to pack all the properties in one property for the SharedDataStoragePropertyEditor */
    //public final static String GLOBAL = "JSONDS.global"; 

    private ArrayList<DataObject> dataObjects;
    private ArrayList<String> columnNames;
    private Map<String, String> typeNames;      // string, lon or lat
    private LonLatNames lonLatNames;    // suffixes of column names starting with longitude or latitude

    private String url = "http://localhost:8080";
    private String query =  "";

    public JSONDataStorage(DataStorageListener dsl) {
        super(dsl);
        Debug.message("DataStorage", this + " JSONDataStorage(DataStorageListener dsl)");
    }
    protected boolean initialized = false;

    public String getName() {
        return "".equals(query) ? name : ("JSONDS "+query);
    }

    public boolean isReady() {
        return initialized;
    }

    public void setProperties(String prefix, java.util.Properties properties) {
        Debug.message("DataStorage" , "[JSONDataStorage] setProperties " + properties + " prefix: '"+prefix+"'" );
        prefix = PropUtils.getScopedPropertyPrefix(prefix);
        super.setProperties(prefix, properties);

        query = Util.unEscapePropertyString(properties.getProperty(prefix + queryPropertyString,
								   ""));
	url = Util.unEscapePropertyString(properties.getProperty(prefix + urlPropertyString, "http://localhost:8080"));

        Debug.message("DataStorage", "[JSONDataStorage] using url: " + url + " query: " + query);

        initialize();
    }

    public Properties getProperties(String prefix, Properties properties) {
        properties = super.getProperties(prefix, properties);
        prefix = PropUtils.getScopedPropertyPrefix(prefix);

        properties.put(prefix + queryPropertyString, query);
        properties.put(prefix + urlPropertyString, url);

        return properties;
    }

    public Properties getPropertyInfo(Properties list) {
        list = super.getPropertyInfo(list);
        PropUtils.setI18NPropertyInfo(i18n,
				      list,
				      JSONDataStorage.class,
				      urlPropertyString,
				      "URL", //label
				      "JSON server URL", //tooltip
				      null);
        PropUtils.setI18NPropertyInfo(i18n,
				      list,
				      JSONDataStorage.class,
				      queryPropertyString,
				      "Query", //label
				      "Query", //tooltip
				      null);

	list.put(initPropertiesProperty, getInitPropertiesPropertyString());

        return list;
    }

    /** which properties to use by the inspector */
    public String getInitPropertiesPropertyString() {
        return urlPropertyString /*+ " " + queryPropertyString */;
    }

    /** get ith column name. Automagically set it if lacking. 
     * Warning column names must be retrieved in the right order for the first time, 
     * as they are set in that order! 
     */
    private String getColumnName(int i) {
        if (i >= columnNames.size()) {
            columnNames.add("Attr " + i);
        }
        return columnNames.get(i);
    }

    private String getTypeName(String colName) {
        return typeNames.get(colName);
    }

    protected void addColumnName(String colName) {
	if (columnNames.contains(colName))
	    return;
	columnNames.add(colName);
	String type = lonLatNames.getLonLatType(colName);
	typeNames.put(colName, type != null ? type : "string");
    }

    public void initialize() {
        initialized = false;
	Debug.message("DataStorage", "[JSONDataStorage] initialize using url: " + url + " query: " + query);

	try {
	    HttpURLConnection connection = (HttpURLConnection)(new URL(url)).openConnection();
	    connection.setRequestMethod("GET");
	    connection.setReadTimeout(10000);
	    
	    connection.connect();
	    
	    BufferedReader rd  = new BufferedReader(new InputStreamReader(connection.getInputStream()));
	    StringBuilder sb = new StringBuilder();
	    String line;
	    while ((line = rd.readLine()) != null)
		sb.append(line + '\n');
	                        
	    connection.disconnect();
	    
	    JSONArray jsonArray = new JSONArray(sb.toString());
	    Debug.message("DataStorage", this + " found " + jsonArray.length() + " entries");

	    dataObjects = new ArrayList<DataObject>();
	    columnNames = new ArrayList<String>();
	    typeNames = new HashMap<String, String>();
	    lonLatNames = new LonLatNames();
	    int nbErrors = 0;

	    for(int lineCount = 0; lineCount < jsonArray.length(); lineCount++) {
		JSONObject jsonObject = jsonArray.getJSONObject(lineCount);
		DataObject dataObject = new DataObject();
		Iterator iterator = jsonObject.keys();
		while (iterator.hasNext()) {
		    String key = (String)iterator.next();
		    addColumnName(key);
		    String type = getTypeName(key);
		    Object o = jsonObject.get(key);
		    if (o == null)
			continue;
		    dataObject.addData(key,  
				       o instanceof Number ? ((Number)o).doubleValue()
				       : o.toString());
		    try {
			if (LonLatNames.LON.equals(type)) {
			    lonLatNames.putLon(lonLatNames.getLonLatName(key), o);
			} else if (LonLatNames.LAT.equals(type)) {
			    lonLatNames.putLat(lonLatNames.getLonLatName(key), o);
			}
		    } catch (java.lang.ClassCastException e) {
			nbErrors++;
			if (nbErrors < 10)
			    System.err.println("*** JSONDataStorage bad cast to Double line " + lineCount + " for " + type + " \"" + o + "\"");
		    } catch (java.lang.NumberFormatException e) {
			nbErrors++;
			if (nbErrors < 10)
			    System.err.println("*** JSONDataStorage number format exception line " + lineCount + " for " + type + " \"" + o + "\"");
		    }
 
		}        
		// gather the centroids
		lonLatNames.putCentroids(dataObject);
		// System.out.println("# "+dataObject);
		dataObjects.add(dataObject);
	    }

	    // System.out.println(columnNames + " -> " + getDataAttributesNamesToDisplay(columnNames) + " " +   lonLatNames.getOMGraphicNames());
	    
	    dataTable.setTable(dataObjects,
			       getDataAttributesNamesToDisplay(columnNames),
			       lonLatNames.getOMGraphicNames(),
			       getDataAttributesNamesAliases());
	    
	    // data are stored, we can create dataTableModel and panel to view them.
	    dataTableModel = new DataTableModel(dataTable);
	    resetDataTableStoragePanel(dataTableModel);
	    initialized = true;
	    fireDataStorageEvent();
	    
	} catch (MalformedURLException e) {
	    e.printStackTrace();
	} catch (ProtocolException e) {
	    e.printStackTrace();
	} catch (IOException e) {
	    e.printStackTrace();
	} catch (JSONException e) {
	    e.printStackTrace();
	}
    }

    public JComponent getGUI() {
	java.awt.Component gui = DataStoragePropertyEditor.provideCustomPanel(this);
	fr.unicaen.openmap.util.gui.ComponentSizer.fixComponentSize(gui, new java.awt.Dimension(500, 100));
         JPanel panel = new JPanel();
         panel.add(gui);
         return panel;
    }
}

