// <header>
/**
 *  ThemaMap : The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
/*
 * KernelGridDataStorage.java
 *
 * Created on 10 June 2009
 *
 * @author Adrien.Andre@etu.info.unicaen.fr
 *
 */
package fr.unicaen.openmap.dataStorage;

import com.bbn.openmap.util.Debug;

import com.bbn.openmap.util.PropUtils;


import java.awt.geom.Point2D;



import java.util.*;

import fr.unicaen.openmap.util.*;
import fr.unicaen.openmap.dataStorage.gui.DataTableModel;
import fr.unicaen.openmap.dataStorage.kernel.BandwidthKernel;
import fr.unicaen.openmap.dataStorage.gui.KernelDataStoragePanel;
import fr.unicaen.openmap.dataStorage.kernel.Cosine;
import fr.unicaen.openmap.dataStorage.kernel.Epanechnikov;
import fr.unicaen.openmap.dataStorage.kernel.Gaussian;
import fr.unicaen.openmap.dataStorage.kernel.Kernel;
import fr.unicaen.openmap.dataStorage.kernel.Quartic;
import fr.unicaen.openmap.dataStorage.kernel.Triangle;
import fr.unicaen.openmap.dataStorage.kernel.Triweight;
import fr.unicaen.openmap.dataStorage.kernel.Uniform;
import fr.unicaen.openmap.dataStorage.BoundingBox;
import java.lang.reflect.Constructor;
import javax.swing.JComponent;

/**
 * Warning: 
 * -- we should avoid computing when data is equal to 0.
 * -- we should buffered the results of the computation.
 */
public class KernelGridDataStorage extends GridDataStorage {

    static final public String kernelDSPropertyString = "kernelDS";
    static final public String kernelTypePropertyString = "kernel";
    static final public String kernelRadiusPropertyString = "kernelRadius";
    static final public String kernelBoundingBoxPropertyString = "kernelBoundingBox";

    private double kernelRadius = 0.3;
    private BoundingBox kernelBoundingBox = new BoundingBox(-180, 90, 180, -90);
    private KernelType kType = KernelType.UNIFORM;
    public enum KernelType {

        UNIFORM(Uniform.class.getName()),
        COSINE(Cosine.class.getName()),
        EPANECHNIKOV(Epanechnikov.class.getName()),
        GAUSSIAN(Gaussian.class.getName()),
        QUARTIC(Quartic.class.getName()),
        TRIANGLE(Triangle.class.getName()),
	TRIWEIGHT(Triweight.class.getName());

        private String className;

        KernelType(String s) {
            className = s;
        }

        public String getClassName() {
            return className;
        }
    }

    public KernelGridDataStorage(DataStorageListener dsl) {
        super(dsl);
        Debug.message("KernelGridDataStorage", "KernelGridDataStorage(DataStorageListener) called.");
    }

    public KernelGridDataStorage(DataStorageListener dsl, DataStorage dataStorage) {
        super(dsl, dataStorage);
        Debug.message("KernelGridDataStorage", "KernelGridDataStorage(DataStorageListener, DataStorage) called.");
    }

    public void setProperties(String prefix, java.util.Properties properties) {
        prefix = PropUtils.getScopedPropertyPrefix(prefix);
	setKernelType(KernelType.valueOf(properties.getProperty(prefix + 
								kernelDSPropertyString + "." +
								kernelTypePropertyString, "UNIFORM")));
	setRadius(PropUtils.doubleFromProperties(properties, 
						 prefix + kernelDSPropertyString + "." + 
						 kernelRadiusPropertyString, 5D));
	kernelBoundingBox.setProperties(prefix + kernelDSPropertyString 
					+ "." + kernelBoundingBoxPropertyString, properties);
      super.setProperties(prefix, properties);
    }

    public java.util.Properties getProperties(String prefix, java.util.Properties properties) {
        properties = super.getProperties(prefix, properties);
        properties.put(prefix + kernelDSPropertyString + "." + kernelTypePropertyString,
		       getKernelType().toString());
	properties.put(prefix + kernelDSPropertyString + "." + kernelRadiusPropertyString, 
		       String.valueOf(getRadius()));
	properties = kernelBoundingBox.getProperties(prefix + kernelDSPropertyString 
						     + "." + kernelBoundingBoxPropertyString, properties);

        return properties;
    }

    public double getRadius() {
        return kernelRadius;
    }
    public void setRadius(double r) {
        kernelRadius = r;
    }
    public BoundingBox getKernelBoundingBox() {
	return kernelBoundingBox;
    }
    protected Kernel getKernel() {
        Kernel k;
        try {
            Class classe = Class.forName(kType.getClassName());
            Constructor constructor = null;
            for (Constructor c : classe.getConstructors()) {
                java.lang.reflect.Type[] params = c.getGenericParameterTypes();
                if(params.length == 1 /*&& params[0] == Class.forName("double")*/)
                    constructor =c;
            }
            k = (Kernel) constructor.newInstance(new Object[]{getRadius()});
        } catch (Exception e) {
            e.printStackTrace();
            k = new Uniform(getRadius());
        }
	k.prepare(longitudeResolution, latitudeResolution, getKernelBoundingBox());
	 Debug.message("KernelGridDataStorage", "[KernelGridDataStorage] getKernel -> " + kType + " - " + k);
	return k;
    }

    public void initialize() {

        Debug.message("KernelGridDataStorage", "KernelGridDataStorage.initialize() called.");
        if (getDataStorage() == null || !getDataStorage().isReady()) {
            Debug.error("[KernelGridDataStorage] can't initialize because dataStorage is null or not ready ! end.");
            return;
        }

        if ("".equals(geometryColumnName)) {
            DataTableModel dtm = getDataStorage().getDefaultDataTableModel();
            if (dtm.getOMColumnNames().length < 1) {
                Debug.error("[KernelGridDataStorage] can't initialize because there is no geometry ! end.");
                return;
            }
            geometryColumnName = dtm.getOMColumnName(0);
            Debug.error("[KernelGridDataStorage] set geometry to " + geometryColumnName);
        }

        Debug.message("DataStorage", "[KernelGridDataStorage] start initialize");

        // we need a temporary map in order to store
        // for each pair <KernelGridCellID, DataObject with CRITERIA values> a DataObject with STATS
        HashMap<KernelGridCellID, HashMap<DataObject, DataObject>> grid = 
	    new HashMap<KernelGridCellID, HashMap<DataObject, DataObject>>();

        initMinMaxIndices(); // indices used by GridCellId
	Kernel kernel = getKernel();
	Debug.message("DataStorage", "[KernelGridDataStorage] getKernel <-" + kernel);
	ArrayList<KernelGridCellID> surroundingCells = null; // used by the kernel for optimization

	// to hold the data statistics sums foreach criteria value in order to normalize afterwards
	HashMap<DataObject, DataObject> sumDataObject = new HashMap<DataObject, DataObject>();

	// for each DataObject of the dataStorage
        for (int i = 0; i < getDataStorage().getCount(); i++) {
            DataObject currentDataObject = getDataStorage().getElementAt(i);

            // get the dataObject centroid
            OMDataLocation omdl = currentDataObject.getOMDataLocation(geometryColumnName);
            if (omdl == null)
                continue;
            
            Centroid centroid = omdl.getCentroid();
            if (centroid == null)
                continue;
            
            double lon = centroid.getLon();
            double lat = centroid.getLat();

            // select CRITERIA attributes values
            DataObject criteriaDataObject = new DataObject();
            for (String criterion : getCriteriaAttributesList()) {
                Debug.message("DataStorage2", "[KernelGridDataStorage] criterion = \"" + criterion + "\"");
                criteriaDataObject.addData(criterion, currentDataObject.getData(criterion)); // fill with "criterion name - value"
            }

	    // Built the surronding cells ids we are interested in for the kernel to operate
	    // We don't rebuilt the list in case it exists and we use a Gaussian kernel
	    // because it is independant of current lon, lat.
	    if (surroundingCells == null || kernel instanceof BandwidthKernel) {
		surroundingCells = new ArrayList<KernelGridCellID>();
		KernelGridCellID kgcID = new KernelGridCellID(lat, lon); // used for rounding (lat, lon) according to the grid
		ArrayList<Integer> cellIndices = kernel.getSurroundingCellIndices(kgcID.lon, kgcID.lat);
		for (int j = 0; j < cellIndices.size(); j += 2) 
		    surroundingCells.add(new KernelGridCellID(cellIndices.get(j + 1).intValue(), cellIndices.get(j).intValue()));
	    }

	    DataObject sum =  sumDataObject.get(criteriaDataObject);
	    if(sum == null) {
		sumDataObject.put(criteriaDataObject, sum = new DataObject());
	    }
            // for each STATISTIC attribute
            for (String name : getStatisticsAttributesList()) {
                // Debug.message("DataStorage", "[KernelGridDataStorage] Statistics attribute name = \"" + name + "\"");
                double current = Util.parseDouble(currentDataObject.getData(name));   // new datum
		// compute the sum in order to normalize the result
		sum.addData(name, (Double)(Util.parseDouble(sum.getData(name)) + current)); 

                // for each surronding cell 
                for (KernelGridCellID cID : surroundingCells) {
                    HashMap<DataObject, DataObject> cellData = grid.get(cID);
                    if (cellData == null) {
                        // Debug.message("DataStorage", "new empty grid cell " + cID);
                        cellData = new HashMap<DataObject, DataObject>();
                    }

                    // STATS for the given CRITERIA
                    DataObject gridDataForCriteria = cellData.get(criteriaDataObject);
                    if (gridDataForCriteria == null) {
                        // Debug.message("DataStorage", "new grid data " + cID + " for " + criteriaDataObject);
                        gridDataForCriteria = new DataObject();
                    }

		    gridDataForCriteria.addData(name,
						kernelOperate(gridDataForCriteria.getData(name), // previous datum
							      current,
							      kernel,
							      cID.distance(lon, lat)
							      ));
                    cellData.put(criteriaDataObject, gridDataForCriteria);
                    grid.put(cID, cellData);
                } // end for each cell
            } // end for each STATISTIC attribute
        } // end for each DataObject of the dataStorage


	////////////////////////////////////
        // gather all the data //
	////////////////////////////////////

	// the rows of the DS
        ArrayList<DataObject> gridDataObjects = new ArrayList<DataObject>();

	// in order to afterwards normalize the value, we must compute 
	// the sums of the data statistics contained in the grid foreach criteria values  
	// HashMap<DataObject with criteria values, DataObect with stats sums>
	HashMap<DataObject, DataObject> sumGridDataObject = new HashMap<DataObject, DataObject>();

	// foreach grid cell
        for (Map.Entry<KernelGridCellID, HashMap<DataObject, DataObject>> gridEntry : grid.entrySet()) {
            OMDataLocation gridOMDataLocation = gridEntry.getKey().newOMDataLocation();

            // Build a dataObject (a row in the DataStorage) foreach criteria (given the gridcell)
            for (Map.Entry<DataObject, DataObject> e : gridEntry.getValue().entrySet()) {
                DataObject data = e.getValue();
                // 1- compute the sum
		DataObject sum = sumGridDataObject.get(e.getKey());
		if (sum == null) sum = new DataObject();
		// foreach stats value
		for (Map.Entry<String, Object> dataEntry : data.getDatas().entrySet()) {
		    Object o = sum.getData(dataEntry.getKey());
		    sum.addData(dataEntry.getKey(),
				((o != null && o instanceof Double)? (Double)o : 0)+Util.parseDouble(dataEntry.getValue()));
		}
		sumGridDataObject.put(e.getKey(), sum);
                // 2- add the grid cell as a OMGraphic
                data.addOMDataLocation(getGridGeometryColumnName(),gridOMDataLocation);
                // 3- merge the criteria values
                data.putAll(e.getKey());
                // 4- add the data row to the table
                gridDataObjects.add(data);
            }
        }
	Debug.message("DataStorage", "Kernel normalization factor: "+sumDataObject+"/"+sumGridDataObject);

	// normalize the data...
	for(DataObject data : gridDataObjects) {
            DataObject criteriaDataObject = new DataObject();
            for (String criterion : getCriteriaAttributesList())
                criteriaDataObject.addData(criterion, data.getData(criterion)); // fill with "criterion name - value"
            // Debug.message("KernelDataStorage", "Kernel normalize: "+data);
	    for (String name : getStatisticsAttributesList()) 
		data.addData(name, 
			     (Double)data.getData(name) * (Double)sumDataObject.get(criteriaDataObject).getData(name)
			     / (Double)sumGridDataObject.get(criteriaDataObject).getData(name));  
	}

	//////////////////////////////////////////////////////////////
	// Set dataTable and dataTableModel //
	//////////////////////////////////////////////////////////////
 
        // merge stat and criteria names
        ArrayList<String> dataColumnNames = new ArrayList<String>(getStatisticsAttributesList());
        for (String a : getCriteriaAttributesList()) {
            dataColumnNames.add(a);
        }
        ArrayList<String> dataColumnAliases = null; // should be updated --TBD

        ArrayList<String> omColumnNames = new ArrayList<String>(1);
        omColumnNames.add(getGridGeometryColumnName());
        dataTable.setTable(gridDataObjects,
                dataColumnNames,
                omColumnNames,
                dataColumnAliases);

        // data are stored, we can create dataTableModel and panel to view them.
        dataTableModel = new DataTableModel(dataTable);
        resetDataTableStoragePanel(dataTableModel);
        setInitialized(true);
        Debug.message("DataStorage", "[KernelGridDataStorage] (re)initialized");

        // tell the listener we're ready
        fireDataStorageEvent();
    }

    /** operates the aggregateFunction over value and data. Value turns to be in fact a Double. */
    protected Double kernelOperate(Object value, double dataVal, Kernel kernel, double distance) {
        //	Debug.message("KernelGridDataStorage", "kernelOperate() called.");
	double kernelVal = kernel.k(distance)*dataVal;	
        if (value == null) {
	    return new Double(kernelVal);
        }

        if (value instanceof Double) {
	    return (Double)value+kernelVal;
        }

        Debug.error("[KernelGridDataStorage] *** Error in aggregate for (" + value + ", " + dataVal + ")");
        return null;
    }

    public void setKernelType(KernelType kt) {
        this.kType = kt;
    }

    public KernelType getKernelType() {
	return kType;
    }

    @Override
    public JComponent getGUI() {
        return new KernelDataStoragePanel(this);
    }

    public class KernelGridCellID extends GridCellID {

        public KernelGridCellID(int lat, int lon) {
            super(lat, lon);
        }

        public KernelGridCellID(double lat, double lon) {
            super(lat, lon); // this constructor should be modified to manage lat out of [-90, 90] and lon out of [-180, 180]
        }

        /** Returns the distance from this grid cell center to a specified point.
         * Should use Geo ?
         * @param lon the point longitude
         * @param lat the point latitude
         * @return the distandce in degrees
         */
        public double distance(double lon, double lat) {
            return Point2D.distance(((double) this.lon + 0.5) * longitudeResolution, ((double) this.lat + 0.5) * latitudeResolution, lon, lat);
        }
    }
}
