// <header>
/**
 *  ThemaMap : The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
package fr.unicaen.openmap.dataStorage;

import java.lang.Math;

import java.util.*;
import java.io.*;
import javax.swing.*;
import java.awt.event.*;
import java.awt.BorderLayout;
import java.awt.Color;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

import fr.unicaen.openmap.util.Util;
import fr.unicaen.openmap.util.LonLatNames;
import fr.unicaen.openmap.util.propertyEditor.DataStoragePropertyEditor;
import fr.unicaen.openmap.util.propertyEditor.FileTypeDescription;
import fr.unicaen.openmap.dataStorage.gui.DataTableModelNetCDF;
import fr.unicaen.openmap.dataStorage.gui.DataTableStoragePanel;
import fr.unicaen.openmap.util.propertyEditor.Inspector;
import fr.unicaen.openmap.dataStorage.Centroid;
import fr.unicaen.openmap.dataStorage.OMDataLocation;
import fr.unicaen.openmap.dataStorage.ProjectionSettable;

import com.bbn.openmap.util.Debug;
import com.bbn.openmap.util.FileUtils;
import com.bbn.openmap.I18n;
import com.bbn.openmap.layer.location.LocationLayer;
import com.bbn.openmap.util.PropUtils;
import com.bbn.openmap.proj.Projection;
import com.bbn.openmap.proj.coords.LatLonPoint;

import ucar.unidata.geoloc.LatLonRect;
import ucar.nc2.NetcdfFile;
import ucar.nc2.VariableSimpleIF;
import ucar.nc2.dataset.NetcdfDataset;
import ucar.nc2.dt.grid.GridDataset;
import ucar.nc2.dt.grid.GeoGrid;
import ucar.nc2.dt.GridCoordSystem;
import ucar.nc2.dataset.CoordinateAxis;
import ucar.nc2.dataset.CoordinateAxis1D;
import ucar.nc2.dataset.CoordinateAxis1DTime;

import thredds.catalog.DataType;

public class NetCDFDataStorage extends DataStorage implements ProjectionSettable
{

	private static final String netCDFFilePropertyString = "netCDFFile";
	private static final String NETCDF_BY_INFO = "netCDF.info";
	private static final String clippingPropertyString = "clipping";
	private static final String samplingPropertyString = "sampling";
	private static final String replaceNaNPropertyString = "replaceNaN";

	private String netCDFFile = "";
	private boolean clipping = false;
	private int sampling = 0;
	private double replaceNaN = Double.NaN;

	private ArrayList<String> columnNames;
	private ArrayList<String> omColumnNames;
	private List<GeoGrid> geoGrids;
	private GridDataset gridDS;
	private GridCoordSystem gcs;
	private CoordinateAxis xAxis;
	private CoordinateAxis yAxis;
	private double[] xValues = null;
	private double[] yValues = null;
	private int timeSteps;
	protected boolean initialized = false;
	private Projection projection;

	private ArrayList<Integer> iList = null;
	private double hdx = 1.0;
	private double hdy = 1.0;


	public NetCDFDataStorage(DataStorageListener dsl)
	{
		super(dsl);
		Debug.message("DataStorage", this + " NetCDFDataStorage(DataStorageListener dsl)");
		columnNames = new ArrayList<String>();
		omColumnNames = new ArrayList<String>();
	}

	public String getName()
	{
		return "".equals(netCDFFile) ? name : stripFileNameWithClassPathDirectories(netCDFFile);
	}

	// Abstract in DataStorage
	public boolean isReady()
	{
		return initialized;
	}

	// Abstract in DataStorage
	public int getCount()
	{
		if( ! initialized )
		{
			return 0;
		}

		if( clipping || sampling != 0 )
		{
			return iList == null ? 0 : iList.size();
		}

		return ( xValues.length > 0 ? xValues.length : 1 ) *
			( yValues.length > 0 ? yValues.length : 1 ) *
			timeSteps;
	}

	public int getColumnCount()
	{
		if(initialized)
		{
			return columnNames.size();
		}
		return 0;
	}

	public String getColumnName(int col)
	{
		return columnNames.get(col);
	}

	public int getOMColumnCount()
	{
		if(initialized)
		{
			return omColumnNames.size();
		}
		return 0;
	}

	public String getOMColumnName(int col)
	{
		return omColumnNames.get(col);
	}

	// Abstract in DataStorage
	public DataObject getElementAt(int i)
	{
		DataObject ret = new DataObject();

		if( clipping || sampling != 0 )
		{
			i = iList.get(i);
		}

		int tIdx = i % timeSteps;
		i = (int) Math.floor(i/timeSteps);
		int xIdx = i % xValues.length;
		int yIdx = (int) Math.floor(i/xValues.length);

		try
		{
			ret.addData("latitude", ""+yValues[yIdx]);
			ret.addData("longitude", ""+xValues[xIdx]);
			ret.addData("time", ""+tIdx);
		}
		catch( Exception e )
		{
			e.printStackTrace();
			return null;
		}

		for (GeoGrid v : geoGrids)
		{
			double data[] = null;
			try
			{
				//								   T    Z   X     Y   	
				data = (double[]) v.readDataSlice(tIdx, 0, yIdx, xIdx).get1DJavaArray(Double.TYPE);
			}
			catch(IOException ex)
			{
				Debug.error(" Error reading NetCDF cell ( T : " + tIdx + ", X : " + xIdx + ", Y : " + yIdx + ")");
				ex.printStackTrace();
			}

			if( data != null )
			{
				ret.addData(v.getName(), ""+ ( Double.isNaN(data[0]) ? replaceNaN : data[0] ) );
			}
			else
			{
				ret.addData(v.getName(), "");
			}
		}

		//ret.addOMDataLocation("latitude/longitude", new OMDataLocation(new com.bbn.openmap.omGraphics.OMPoint(yValues[yIdx],xValues[xIdx]), new Centroid(xValues[xIdx],yValues[yIdx]), null));

		if( clipping == false && sampling == 0 )
		{
			if(xValues.length > 1 )
			{
				if( xIdx > 0 )
				{
					/* distance between curent point and it's predecessor */
					hdx = Math.abs( xValues[xIdx]-xValues[xIdx-1] );
				}
				else
				if( xIdx < ( xValues.length - 1 ) )
				{
					/* distance between curent point and it's successor */
					hdx = Math.abs( xValues[xIdx]-xValues[xIdx+1] );
				}
			}

			if(yValues.length > 1 )
			{
				if( yIdx > 0 )
				{
					/* distance between curent point and it's predecessor */
					hdy = Math.abs( yValues[yIdx]-yValues[yIdx-1] );
				}
				else
				if( yIdx < ( yValues.length - 1 ) )
				{
					/* distance between curent point and it's successor */
					hdy = Math.abs( yValues[yIdx]-yValues[yIdx+1] );
				}
			}
		}

		ret.addOMDataLocation("latitude/longitude",
													new OMDataLocation(new com.bbn.openmap.omGraphics.OMRect(yValues[yIdx],
																																									 xValues[xIdx],
																																									 yValues[yIdx]-hdy,
																																									 xValues[xIdx]-hdy,
																																									 com.bbn.openmap.omGraphics.OMGraphicConstants.LINETYPE_STRAIGHT),
													new Centroid(xValues[xIdx],yValues[yIdx]), null));
		return ret;
	}

	public void setProperties(String prefix, java.util.Properties properties)
		// file -> class
	{
		super.setProperties(prefix, properties);
		prefix = PropUtils.getScopedPropertyPrefix(prefix);

		netCDFFile = properties.getProperty(prefix + netCDFFilePropertyString, properties.getProperty(prefix + NETCDF_BY_INFO));
		Debug.message("NetCDFDataStorage", "[NetCDFDataStorage] netCDFFile : " + netCDFFile + ".");

		clipping = Util.isTrue(properties.getProperty(prefix + clippingPropertyString, "false"));
		Debug.message("NetCDFDataStorage", "[NetCDFDataStorage] clipping : " + clipping + ".");

		sampling = Integer.parseInt(properties.getProperty(prefix + samplingPropertyString, "0"));
		Debug.message("NetCDFDataStorage", "[NetCDFDataStorage] sampling : " + sampling + ".");

		replaceNaN = Double.parseDouble(properties.getProperty(prefix + replaceNaNPropertyString, ""+Double.NaN));
		Debug.message("NetCDFDataStorage", "[NetCDFDataStorage] replaceNaN : " + replaceNaN + ".");

		initialize();
	}

	public Properties getProperties(String prefix, Properties properties)
		// class -> file
	{
		properties = super.getProperties(prefix, properties);
		prefix = PropUtils.getScopedPropertyPrefix(prefix);

		properties.put(prefix + netCDFFilePropertyString, stripFileNameWithClassPathDirectories(netCDFFile));
		properties.put(prefix + NETCDF_BY_INFO, new FileTypeDescription(FileTypeDescription.NC_DESCRIPTION, netCDFFile, FileTypeDescription.NC_PATTERN));

		properties.put(prefix + clippingPropertyString, String.valueOf(clipping));
		properties.put(prefix + samplingPropertyString, String.valueOf(sampling));
		properties.put(prefix + replaceNaNPropertyString, String.valueOf(replaceNaN));

		return properties;
	}

	public Properties getPropertyInfo(Properties properties)
	{
		properties = super.getPropertyInfo(properties);

		PropUtils.setI18NPropertyInfo(i18n,
				properties,
				NetCDFDataStorage.class,
				NETCDF_BY_INFO,
				"NetCDF file", //label
				"NetCDF file to load", //tooltip
				"fr.unicaen.openmap.util.propertyEditor.FileTypePropertyEditor");

		PropUtils.setI18NPropertyInfo(i18n,
				properties,
				NetCDFDataStorage.class,
				clippingPropertyString,
				"Enable clipping", //label
				"Enable clipping", //tooltip
				"fr.unicaen.openmap.util.propertyEditor.CheckBoxPropertyEditor");

		PropUtils.setI18NPropertyInfo(i18n,
				properties,
				NetCDFDataStorage.class,
				samplingPropertyString,
				"Sampling", //label
				"Maximum amount of points to be drawn at once (0 = sampling disabled)", //tooltip
				null);

		PropUtils.setI18NPropertyInfo(i18n,
				properties,
				NetCDFDataStorage.class,
				replaceNaNPropertyString,
				"Replace NaN by", //label
				"Value to replace NaN (empty = no replacing)", //tooltip
				null);

		properties.put(initPropertiesProperty, NETCDF_BY_INFO + " " + clippingPropertyString + " " + samplingPropertyString + " " + replaceNaNPropertyString );

		return properties;
	}

	@SuppressWarnings("unchecked")
		public void initialize()
		{
			try
			{
				gridDS = new GridDataset( new NetcdfDataset(NetcdfFile.open(netCDFFile)) );
			}
			catch(IOException ex)
			{
				Inspector inspector = new Inspector();
				do
				{
					JPanel p = new JPanel(new BorderLayout());
					String f = i18n.get(this, "file", "The file") + " ";
					String nf = " " + i18n.get(this, "not_found", "was not found - please browse to select a new one");
					JLabel lab = new JLabel("<html><b>" + f + "'" + netCDFFile + "'" + nf + "</b></html>");
					lab.setForeground(Color.RED);
					p.add(lab, BorderLayout.NORTH);
					netCDFFile = "#";
					p.add(DataStoragePropertyEditor.provideSimpleCustomPanel(this, inspector), BorderLayout.CENTER);
					int choice = JOptionPane.showConfirmDialog(null, p, "Please Select a NetCDF File", JOptionPane.OK_CANCEL_OPTION, JOptionPane.INFORMATION_MESSAGE);
					if (choice == JOptionPane.CANCEL_OPTION)
					{
						return;
					}
					setProperties(inspector.collectProperties());

				}
				while (!isReady());

				return;
			}

			geoGrids = (List<GeoGrid>) gridDS.getGrids();

			if( ! initialized ) /* Avoid column duplication */
			{
				columnNames.add("latitude");
				columnNames.add("longitude");
				columnNames.add("time");

				omColumnNames.add("latitude/longitude");
			}

			if (geoGrids.size() != 0)
			{   
					int gIdx = 0;
					GeoGrid g;
					do  
					{   
						g = (GeoGrid) geoGrids.get(gIdx++);

						if( ! initialized ) /* Avoid column duplication */
						{
							columnNames.add(g.getName());
						}
					}   
					while ((gIdx != geoGrids.size()) && (g.getDimensions().size() != gridDS.getNetcdfDataset().getDimensions().size()));

				// Time Axis
				gcs = g.getCoordinateSystem();
				if (gcs.hasTimeAxis1D())
				{   
					timeSteps = gcs.getTimeAxis1D().getTimeDates().length;
				}
				else
				{
					timeSteps = 1;
				}

				// X Axis
				xAxis = gcs.getXHorizAxis();
				if (xAxis != null)
				{
					try
					{
						xValues = ((double[])xAxis.read().get1DJavaArray(Double.TYPE));
					}
					catch (IOException ex)
					{
						ex.printStackTrace();
						return;
					}
				}

				// Y Axis
				yAxis = gcs.getYHorizAxis();
				if (yAxis != null)
				{
					try
					{
						yValues = ((double[])yAxis.read().get1DJavaArray(Double.TYPE));
					}
					catch (IOException ex)
					{
						ex.printStackTrace();
						return;
					}
				}

				initialized = true;
				dataTableModel = new DataTableModelNetCDF(this);

				if (dtsp == null)
				{
					dtsp = new DataTableStoragePanel(this, dataTableModel);
				}
				else
				{
					((DataTableStoragePanel)dtsp).setModel(dataTableModel);
				}

				getDataStorageListener().dataStorageEvent(this);
			}
		}

	public JComponent getGUI()
	{
		JPanel panel = new JPanel();
		panel.add(DataStoragePropertyEditor.provideCustomPanel(this));
		return panel;
	}

	public void setProjection(Projection proj)
	{
		// adapter pour découpler le sampling et le clipping
		if( projection == proj )
		{
			return;
		}

		projection = proj;

		if( clipping == false || sampling == 0 )
		{
			return;
		}

		if( isReady() )
		{
			LatLonPoint.Double ul = proj.getUpperLeft();
			LatLonPoint.Double lr = proj.getLowerRight();

			double lonMin = ul.getLongitude();
			double lonMax = lr.getLongitude();
			double latMin = lr.getLatitude();
			double latMax = ul.getLatitude();

			LinkedList<Integer> xList = new LinkedList<Integer>();
			LinkedList<Integer> yList = new LinkedList<Integer>();

			for( int lon=0 ; lon < xValues.length ; ++lon )
			{
				if( xValues[lon] <= lonMax && xValues[lon] >= lonMin )
				{
					xList.add(lon);
				}
			}

			for( int lat=0 ; lat < yValues.length ; ++lat )
			{
				if( yValues[lat] <= latMax && yValues[lat] >= latMin )
				{
					yList.add(lat);
				}
			}

			double sample = ( ((double)sampling) / ( ((double)yList.size()) * ((double)xList.size()) ) ) / ((double)timeSteps);
			int take = (int) Math.ceil(1.0/sample);
			int sampleCntX = 0;
			int sampleCntY = 0;

			int nbX = 0;
			int nbY = 0;
			double minLat = 1000;
			double maxLat = -1000;
			double minLon = 1000;
			double maxLon = -1000;

			iList = new ArrayList<Integer>( xList.size()*yList.size()*timeSteps );

			for( int y : yList )
			{
				sampleCntY = ++sampleCntY % take;
				if( sampleCntY != 0 ) { continue; }

				++nbY;
				minLat = Math.min( yValues[y], minLat ) ;
				maxLat = Math.max( yValues[y], maxLat ) ;

				sampleCntX = 0;
				for( int x : xList )
				{
					sampleCntX = ++sampleCntX % take;
					if( sampleCntX != 0 ) { continue; }

					if(nbY == 1) /* only once */
					{
						++nbX;
						minLon = Math.min ( xValues[x], minLon ) ;
						maxLon = Math.max ( xValues[x], maxLon ) ;
					}

					for( int t = 0 ; t< timeSteps ; ++t )
					{
						int i = (y * timeSteps * xValues.length) + ( x * timeSteps ) + t;
						iList.add(i);
					}
				}
			}

			hdx = Math.abs( Math.abs(minLon)-Math.abs(maxLon)) / ((double)nbX);
			hdy = Math.abs( Math.abs(minLat)-Math.abs(maxLat)) / ((double)nbY);
		}


		getDataStorageListener().dataStorageEvent(this);
	}
}
