// <header>
/**
 *  ThemaMap : The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
package fr.unicaen.openmap.dataStorage;

import com.bbn.openmap.omGraphics.*;
import com.bbn.openmap.proj.ProjMath;

import java.util.Iterator;

/**
 * Stores a data location made up with a OMGraphic and optionally its centroid 
 * and bounding box. centroid and bounding box may be null. 
 */
public class OMDataLocation {

    private OMGraphic omGraphic = null;
    private Centroid centroid = null;
    private BoundingBox boundingBox = null;

    /**
     * Constructor
     * @param omGraphic an OMGraphic (generally a shape or a line)
     * @param centroid its associated centroid defined in lat,lon
     * @param boundingBox the boundingBox box of the OMGraphic. 
     */
    public OMDataLocation(OMGraphic omGraphic, Centroid centroid, BoundingBox boundingBox){
	this.omGraphic = omGraphic;
	this.centroid  = centroid;
	this.boundingBox  = boundingBox;
    }
    
    /**
     * Simple constructor.
     * centroid and boundingBox will be null
     * @param omGraphic an OMGraphic (generally a shape or a line)
     */
    public OMDataLocation(OMGraphic omGraphic) { this(omGraphic, null, null); }
    
    /**
     * We need this constructor to set a completely empty OMDataLocation.
     *
     */
    public OMDataLocation () { }

    /**
     * Get the Centroid object associated to OMGraphic.
     * @return a Centroid or null
     */
    public Centroid getCentroid() {
	if (centroid == null) {
	    // try to build it
	    if (omGraphic == null)
		return null;
	    
	    Centroid c = fr.unicaen.openmap.dataStorage.expr.SpatialFunction.centroid(omGraphic);
	    if (c != null)
		centroid = c;
	    else {
		// fall back use the center of the bounding box
		if (boundingBox != null)
		    centroid =  new Centroid((boundingBox.getUpperLeftLon() + boundingBox.getLowerRightLon())/2.0,
					    (boundingBox.getUpperLeftLat() + boundingBox.getLowerRightLat())/2.0);
	    }
	}
	return centroid;
    }

    /**
     * Get the BoundingBox object associated to OMGraphic.
     * @return a BoundingBox or null.
     */
    public BoundingBox getBoundingBox() {
	return boundingBox;
    }

    /**
     * Set the Centroid object associated to OMGraphic.
     */
    public void setCentroid(Centroid c){
	centroid = c;
    }

    /**
     * Set the BoundingBox object associated to OMGraphic.
     */
    public void setBoundingBox(BoundingBox b){
	boundingBox = b;
    }

    /**
     * Get the OMGraphic
     * @return OMGraphic
     */
     public OMGraphic getOMGraphic(){
	 if (omGraphic != null)
	     return omGraphic;
	 if (centroid == null) 
	     return null;
	 // Assume it's a point
	 return omGraphic = new OMPoint(centroid.getLat(), centroid.getLon());
    }
    /**
     *
     */
    public void setOMGraphic(OMGraphic omGraphic) {
	this.omGraphic = omGraphic;
    }

    public String toString() {
	return "{"+omGraphic+", "+centroid+", "+boundingBox+"}";
    }

}
