// <header>
/**
 *  ThemaMap : The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
/*
 * PivotDataStorage.java
 *
 * Created on 23 may 2008
 *
 * @author Jacques.Madelaine@info.unicaen.fr
 *
 */
package fr.unicaen.openmap.dataStorage;

import com.bbn.openmap.util.Debug;

import com.bbn.openmap.util.PropUtils;

import javax.swing.*;
import javax.swing.event.TableModelEvent;
import javax.swing.event.TableModelListener;
import javax.swing.table.DefaultTableCellRenderer;
import javax.swing.table.TableCellRenderer;
import javax.swing.table.TableColumn;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.GridLayout;
import java.awt.event.*;
import java.util.*;

import fr.unicaen.openmap.util.*;
import fr.unicaen.openmap.dataStorage.gui.DataTableModel;
import fr.unicaen.openmap.dataStorage.gui.DataTableStoragePanel;
import fr.unicaen.openmap.dataStorage.DataTable;
import fr.unicaen.openmap.gui.component.JComponentCellEditor;
import fr.unicaen.openmap.gui.component.JComponentCellRenderer;
import fr.unicaen.openmap.gui.component.JComponentTable;
import fr.unicaen.openmap.gui.component.JComponentTableModel;
import fr.unicaen.openmap.util.classification.BoxPlot;
import fr.unicaen.openmap.util.classification.Quantile;

import com.bbn.openmap.I18n;
import com.bbn.openmap.BasicI18n;

/**
 * Pivot a DataStorage given 3 lists of attributes:
 * <ol><li>the criteria (atrributes used as a group by)
 *     <li>the attributes used to pivot
 *     <li>the statistics leading to a new column attributes per value 
 *         aggregate for each value of the pivot attributes
 * </ol>
 * and an aggregate function used in case several tuples exist for a characteristics value.
 * <p>Danger: the OMDataLocations associated with the criteria follow the criteria in the result.
 * <p>Bug: don't really manage attributes aliases.
 */
public class PivotDataStorage extends RemovableInformationDataStorage {

    /* --------------------------  Begin   Properties  ---------------------*/
    public static final String pivotDSPropertyString = "pivotDS";
    public static final String dataStorageClassPropertyString = "dataStorageClass";
    public static final String criteriaAttributesListPropertyString = "criteriaAttributes";
    public static final String pivotAttributesListPropertyString = "pivotAttributes";
    public static final String statisticsAttributesListPropertyString = "statisticsAttributes";
    public static final String aggregateFunctionPropertyString = "aggregateFunction";
    public static final String alwaysPrefixPivotWithStatisticsNamePropertyString = "alwaysPrefixPivotWithStatisticsName";
    /* ---------------------------  End     Properties  ---------------------*/
    private ArrayList<String> criteriaAttributesList = null;

    protected ArrayList<String> getCriteriaAttributesList() {
        return criteriaAttributesList;
    }
    private ArrayList<String> pivotAttributesList = null;

    protected ArrayList<String> getPivotAttributesList() {
        return pivotAttributesList;
    }
    private ArrayList<String> statisticsAttributesList = null;

    protected ArrayList<String> getStatisticsAttributesList() {
        return statisticsAttributesList;
    }
    protected JButton validateButton;

    public enum AttributeType {

        CRITERIA, PIVOT, STATISTIC, DISCARD
    };
    /** attributesHashMap is used by the GUI to edit the lists 
     *     criteriaAttributesList, pivotAttributesList, statisticsAttributesList 
     */
    protected HashMap<String, AttributeType> attributesHashMap = null;

    public enum AggregateFunction {

        COUNT(i18n.get(PivotDataStorage.class, "COUNT", "Count")),
	    SUM(i18n.get(PivotDataStorage.class, "SUM", "Sum")),
	    AVG(i18n.get(PivotDataStorage.class, "AVG", "Avg")),
	    STDEV(i18n.get(PivotDataStorage.class, "STDEV", "StDev")),
	    MAX(i18n.get(PivotDataStorage.class, "MAX", "Max")),
	    MIN(i18n.get(PivotDataStorage.class, "MIN", "Min")),
	    MEDIAN(i18n.get(PivotDataStorage.class, "MEDIAN", "Median")),
	    BOXPLOT(i18n.get(PivotDataStorage.class, "BOXPOT", "BoxPlot")),
	    FIRST(i18n.get(PivotDataStorage.class, "FIRST", "First")),
	    LAST(i18n.get(PivotDataStorage.class, "LAST", "Last")),
	    CONCAT(i18n.get(PivotDataStorage.class, "CONCAT", "Concat"));

	private final String name;
	public String toString() { return name; }
	AggregateFunction(String name) {
	    this.name = name;
	}
    }

    private HashMap<String, AggregateFunction> aggregateFunctionMap = new HashMap<String, AggregateFunction>();

    protected AggregateFunction getAggregateFunction(String attribute) {
        AggregateFunction aggregateFunction = aggregateFunctionMap.get(attribute);
        return aggregateFunction != null ? aggregateFunction : AggregateFunction.SUM;
    }

    public void setAggregateFunction(String attribute, AggregateFunction a) {
        aggregateFunctionMap.put(attribute, a);
    }

    /** operates the aggregateFunction over value and data. Value turns to be in fact a AggregateResult. */
    @SuppressWarnings("unchecked")
	protected Object operate(Object value, Object data, AggregateFunction aggregateFunction) {
	// skip null values
	if (data == null || (data instanceof String && "".equals((String)data))) 
	    return value;
	if (data instanceof Double) {
	    Double d = (Double)data;
	    if (d.isNaN() || d == Double.NEGATIVE_INFINITY || d == Double.POSITIVE_INFINITY)
		return value;
	}
	
        if (value == null) {
            switch (aggregateFunction) {
	    case COUNT:
		return new AggregateResult("");
	    case SUM:
	    case AVG:
	    case STDEV:
	    case MAX:
	    case MIN:
		{
		    Double d = Util.parseDouble(data);		    
		    AggregateResult ret = new AggregateResult(d);
		    if (aggregateFunction == AggregateFunction.STDEV)
			ret.setRes2(d*d);
		    return ret;
		}
	    case MEDIAN:
	    case BOXPLOT:
		{
		    // just collect the data
		    ArrayList<Double> datalist = new ArrayList<Double>();
		    datalist.add(Util.parseDouble(data));
		    return new AggregateResult(datalist);
		}
	    case FIRST:
	    case LAST:
		return new AggregateResult(data);
	    case CONCAT:
		return new AggregateResult(new StringBuffer(data.toString()));
            }
        }
        if (value instanceof AggregateResult) {
            switch (aggregateFunction) {
	    case COUNT:
		((AggregateResult) value).incrNb();
		return value;
	    case AVG:
	    case STDEV:
		((AggregateResult) value).incrNb();
                // no break, we must also add the value
	    case SUM:
		{
		    Double d = Util.parseDouble(data);
		    ((AggregateResult) value).setRes(((Double) ((AggregateResult) value).getRes())+d);
		    if (aggregateFunction == AggregateFunction.STDEV)
			((AggregateResult) value).setRes2(((Double) ((AggregateResult) value).getRes2())+d*d);
		}
		return value;
	    case MAX:
	    case MIN:
		{
		    Double d = Util.parseDouble(data);
		    if (d > ((Double) ((AggregateResult) value).getRes())) {
			if (aggregateFunction == AggregateFunction.MAX) 
			    ((AggregateResult) value).setRes(d);
		    } else {
			if (aggregateFunction == AggregateFunction.MIN)
			    ((AggregateResult) value).setRes(d);
		    }
		}
		return value;
	    case MEDIAN:
	    case BOXPLOT:
		{
		    Double d = Util.parseDouble(data);
		    ((ArrayList<Double>)(((AggregateResult) value).getRes())).add(d);
		    return value;
		}
	    case FIRST:
		return value;
	    case LAST:
		((AggregateResult) value).setRes(data);
		return value;
	    case CONCAT:
		((StringBuffer) ((AggregateResult) value).getRes()).append(data.toString());
		return value;
            }
        }
        Debug.error("[PivotDataStorage] *** Error in aggregate for (" + value + ", " + data + ")");
        return null;
    }

    private boolean alwaysPrefixPivotWithStatisticsName = false;

    protected boolean getAlwaysPrefixPivotWithStatisticsName() {
        return alwaysPrefixPivotWithStatisticsName;
    }

    protected void setAlwaysPrefixPivotWithStatisticsName(boolean b) {
        if (alwaysPrefixPivotWithStatisticsName == b) {
            return;
        }
        alwaysPrefixPivotWithStatisticsName = b;
        // we should change the data table model...
    }

    public PivotDataStorage(DataStorageListener dsl) {
        this(dsl, null);
    }

    public PivotDataStorage(DataStorageListener dsl, DataStorage dataStorage) {
	super(dsl, dataStorage);
    }

    public void setProperties(String prefix, java.util.Properties properties) {
        prefix = PropUtils.getScopedPropertyPrefix(prefix);
        super.setProperties(prefix, properties);

        prefix = PropUtils.getScopedPropertyPrefix(prefix) + pivotDSPropertyString + ".";
        String dsClassName = properties.getProperty(prefix + dataStorageClassPropertyString, "");
        if ("".equals(dsClassName)) // empty Pivot Data Storage ????
        {
            return;
        }
        setDataStorage(Util.loadDataStorage(dsClassName, prefix, properties, this));

        criteriaAttributesList = UnicaenPropUtils.arrayListFromProperties(properties, prefix + criteriaAttributesListPropertyString, listDelim);
        pivotAttributesList = UnicaenPropUtils.arrayListFromProperties(properties, prefix + pivotAttributesListPropertyString, listDelim);
        statisticsAttributesList = UnicaenPropUtils.arrayListFromProperties(properties, prefix + statisticsAttributesListPropertyString, listDelim);
        for (String attribute : statisticsAttributesList) {
            setAggregateFunction(attribute,
                    Enum.valueOf(AggregateFunction.class,
                    properties.getProperty(prefix + aggregateFunctionPropertyString + "." + attribute, "SUM")));
        }
        alwaysPrefixPivotWithStatisticsName = Util.isTrue(properties.getProperty(prefix + alwaysPrefixPivotWithStatisticsNamePropertyString, "false"));
        ;

        initialize();
    }

    public java.util.Properties getProperties(java.util.Properties properties) {
        return getProperties("", properties);
    }

    public java.util.Properties getProperties(String prefix, java.util.Properties properties) {
        properties = super.getProperties(prefix, properties);
        prefix = PropUtils.getScopedPropertyPrefix(prefix) + pivotDSPropertyString + ".";
        if (getDataStorage() != null) {
            properties.put(prefix + dataStorageClassPropertyString, getDataStorage().getClass().getCanonicalName());
            // Debug.message("DataStorage", "PivotDataStorage.getProperties() - DataStorage member is a: "+getDataStorage().getClass().getCanonicalName());
            properties = getDataStorage().getProperties(prefix, properties);
        }

        UnicaenPropUtils.arrayListToPropertyString(properties, prefix + criteriaAttributesListPropertyString, criteriaAttributesList, listDelim);
        UnicaenPropUtils.arrayListToPropertyString(properties, prefix + pivotAttributesListPropertyString, pivotAttributesList, listDelim);
        UnicaenPropUtils.arrayListToPropertyString(properties, prefix + statisticsAttributesListPropertyString, statisticsAttributesList, listDelim);

        for (String attribute : statisticsAttributesList) {
            properties.put(prefix + aggregateFunctionPropertyString + "." + attribute, getAggregateFunction(attribute).name());
        }

        properties.put(prefix + alwaysPrefixPivotWithStatisticsNamePropertyString, new Boolean(alwaysPrefixPivotWithStatisticsName).toString());

        return properties;
    }
    private boolean initialized = false;

    protected boolean getInitialized() {
        return initialized;
    }

    protected void setInitialized(boolean b) {
        Debug.message("DataStorage", classNameForDebug + ".initialized = " + b);
        initialized = b;
    }

    /**
     * Tells if the dataStorage has its data.
     */
    public boolean isReady() {
        return getDataStorage() != null && getDataStorage().isReady() && initialized;
    }

    /** 
     * Replace all AggregateResult by its result in the DataObject.
     * Update dataColumnNames with new names (see BOXPLOT).
     */
   @SuppressWarnings("unchecked")
       protected void replaceAggregateResultByRealResult(DataObject data, ArrayList<String> dataColumnNames) {
        Map<String, Object> datas = data.getDatas();
        Map<String, Object> addedDatas = null;
        for (Map.Entry<String, Object> entry : datas.entrySet()) {
            Object value = entry.getValue();
            if (value instanceof AggregateResult) {
                AggregateResult v = (AggregateResult) value;
		// System.out.println(entry.getKey() + " uses " +getAggregateFunction(entry.getKey()) + " " + v);
                switch (getAggregateFunction(entry.getKey())) {
		case AVG:
		    {
			// beware to add 1 to nb.
			entry.setValue(((Double) v.getRes()) / (v.getNb() + 1));
			break;
		    }
		case STDEV:
		    {  // uses a naive formula for unbaised
			Integer n = new Integer(v.getNb() + 1);
			Double r = ((Double) v.getRes());;
			Double r2 = ((Double) v.getRes2());
			entry.setValue(Math.sqrt( (r2 - (r*r)/n) / (n-1) ));
			break;
		    }
		case COUNT:
		    {
			Integer n = new Integer(v.getNb() + 1);
			entry.setValue(n);
			break;
		    }
		case MEDIAN:
		    {
			final int nbClasses = 100;
			Quantile classif = new Quantile(((ArrayList<Double>)(v.getRes())), nbClasses);		       
			entry.setValue(classif.getNthBreak(nbClasses/2));
			break;
		    }
		case BOXPLOT:
		    {
			final int nbClasses = 6;
			BoxPlot classif = new BoxPlot(((ArrayList<Double>)(v.getRes())), nbClasses);
			if (addedDatas == null)
			    addedDatas = new Hashtable<String, Object>(); 
			for (int c = 1; c < nbClasses; c++) {
			    int i = c - 1;
			    String name_Q_i = entry.getKey()+"_Q_"+i;
			    if (dataColumnNames.indexOf(name_Q_i) == -1) {
				dataColumnNames.add(name_Q_i);
				// dataColumnAliases.add(name_Q_i); // should take aliases into account !!
			    }
			    dataColumnNames.remove(entry.getKey());
			    addedDatas.put(name_Q_i, classif.getNthBreak(c)); 
			}
			entry.setValue(""); // should remove the entry! TBD
			break;
		    }
		default:
		    entry.setValue(v.getRes());
                }
            }
        }
	if (addedDatas != null)
	    datas.putAll(addedDatas);
    }

    public void initialize() {
        if (getDataStorage() == null || !getDataStorage().isReady()) {
            Debug.error("[PivotDataStorage] can't initialize because dataStorage is not ready !!!!");
            return;
        }

        ArrayList<DataObject> dataObjects = new ArrayList<DataObject>(getDataStorage().getCount());
        // copy data from the DataStorage, we need to sort, and may be the dataStorage has no dataTable !
        for (int i = 0; i < getDataStorage().getCount(); i++) {
            dataObjects.add(getDataStorage().getElementAt(i));
        }

        // sort data on criteria
        Collections.sort(dataObjects, new Comparator<DataObject>() {

            public int compare(DataObject o1, DataObject o2) {
                for (String name : getCriteriaAttributesList()) {
                    Object oo1 = o1.getData(name);
                    Object oo2 = o2.getData(name);
                    if (oo1 == null && oo2 == null) {
                        continue;
                    }
                    if (oo1 == null) {
                        return -1;
                    }
                    if (oo2 == null) {
                        return +1;
                    }
                    int compare = Util.compare(oo1, oo2);
                    if (compare != 0) {
                        return compare;
                    }
                }
                return 0;
            }
        });

        DataTableModel dtm = getDataStorage().getDefaultDataTableModel();
        ArrayList<String> dataColumnNames = new ArrayList<String>(criteriaAttributesList);
	
        ArrayList<String> dataColumnAliases = null; // dataStorage.getDataAttributesNamesAliases();

        ArrayList<DataObject> pivotDataObjects = new ArrayList<DataObject>();

        ArrayList<Object> previousCriteria = null;
        DataObject currentDataObject = null;
        for (DataObject data : dataObjects) {
            // get the current criteria
            ArrayList<Object> currentCriteria = new ArrayList<Object>(criteriaAttributesList.size());
            boolean equalPrevious = true;
	    if (criteriaAttributesList.size() > 0) {
		int n = 0;
		for (String name : criteriaAttributesList) {
		    Object o = data.getData(name);
		    if (previousCriteria == null) {
			equalPrevious = false;
		    } else if (o == null) {
			equalPrevious &= null == previousCriteria.get(n);
		    } else {
			equalPrevious &= o.equals(previousCriteria.get(n));
		    }
		    currentCriteria.add(o);
		    n++;
		}
	    } else { // no criteria
		equalPrevious = true;
		if (currentDataObject == null)
		    currentDataObject = new DataObject();
	    }

            if (!equalPrevious) {
                // store the currentDataObject
                if (currentDataObject != null) {
                    replaceAggregateResultByRealResult(currentDataObject, dataColumnNames);
                    pivotDataObjects.add(currentDataObject);
                }
                // create a new DataObject ...
                Map<String, Object> datas = new HashMap<String, Object>();
                // ... with the criteria attributes ...
                for (String name : criteriaAttributesList) {
                    if (datas == null || data == null || name == null || data.getData(name) == null) {
                        continue;
                    }
                    datas.put(name, data.getData(name));
                }
                // ... and the omdls
                currentDataObject = new DataObject(datas, data.getOmdls());
            }

            // compute the pivot attributes value in order to set statPivotColumnNames
            // (always starts with an underscore)
            String pivotAttributesValue = "";
            for (String name : pivotAttributesList) {
                pivotAttributesValue += "_" + data.getData(name);
            }

            // add the statistics corresponding to the pivot attributes values
            for (String name : statisticsAttributesList) {
                String statPivotName =
                        (getAlwaysPrefixPivotWithStatisticsName() || statisticsAttributesList.size() > 1 || "".equals(pivotAttributesValue))
                        ? name + pivotAttributesValue
                        : pivotAttributesValue.substring(1);
                if (dataColumnNames.indexOf(statPivotName) == -1) {
                    dataColumnNames.add(statPivotName);
                    // dataColumnAliases.add(statPivotName); // should take aliases into account !!
                }
                Object current = data.getData(name);
                Object previous = currentDataObject.getData(statPivotName);
		AggregateFunction aggregateFunction = getAggregateFunction(name);
		// set the aggregateFunction of statPivotName as the aggregateFunction of name
		// as afterwards in replaceAggregateResultByRealResult we cannot guess name from statPivotName
		setAggregateFunction(statPivotName, aggregateFunction); 
		currentDataObject.addData(statPivotName, operate(previous, current, aggregateFunction));
            }
            previousCriteria = currentCriteria;
        }
        // store the last currentDataObject
        if (currentDataObject != null) {
            replaceAggregateResultByRealResult(currentDataObject, dataColumnNames);
            pivotDataObjects.add(currentDataObject);
        }

        ArrayList<String> omColumnNames = new ArrayList<String>(dtm.getOMColumnCount());
        for (int i = 0; i < dtm.getOMColumnCount(); i++) {
            omColumnNames.add(dtm.getOMColumnName(i));
        }

        dataTable.setTable(pivotDataObjects,
                dataColumnNames,
                omColumnNames,
                dataColumnAliases);

        // data are stored, we can create dataTableModel and panel to view them.
        dataTableModel = new DataTableModel(dataTable);
        resetDataTableStoragePanel(dataTableModel);
        initialized = true;

        // tell the listener we're ready
        fireDataStorageEvent();
    }

    public class AggregateFunctionComboBox extends JComboBox implements ItemListener {

        private String attribute;

        public AggregateFunctionComboBox(String attribute) {
            super(AggregateFunction.values());
            this.attribute = attribute;
            setSelectedIndex(getAggregateFunction(attribute).ordinal());
            addItemListener(this);
        }

        public void itemStateChanged(ItemEvent e) {
            int index = ((JComboBox) e.getSource()).getSelectedIndex();
            setAggregateFunction(attribute, AggregateFunction.values()[index]);
        }
    }

    public JComponent getGUI() {

        Box pivotPanel = Box.createVerticalBox();

        JCheckBox alwaysPrefixPivotWithStatisticsNameCB = new JCheckBox(i18n.get(this, "allwaysEtc", "Always prefix pivot column names with statistitics column name"));
        alwaysPrefixPivotWithStatisticsNameCB.setSelected(alwaysPrefixPivotWithStatisticsName);
        alwaysPrefixPivotWithStatisticsNameCB.addItemListener(new ItemListener() {

            public void itemStateChanged(ItemEvent e) {
                setAlwaysPrefixPivotWithStatisticsName(((AbstractButton) (e.getItemSelectable())).isSelected());
            }
        });
        pivotPanel.add(alwaysPrefixPivotWithStatisticsNameCB);

        // add here specific controls for Pivot
        JPanel attributePanel = new JPanel();

        String[] attributesNames = getDataStorage().getDefaultDataTableModel().getDataColumnNames();

        initAttributesHashMapFromLists(attributesNames);


        String criteriaToolTipText = i18n.get(this, "criteria", I18n.TOOLTIP, "Lead to one row per criteria value");
        String pivotToolTipText = i18n.get(this, "pivot", I18n.TOOLTIP, "Lead to one statistic attribute per pivot value");
        String statisticsToolTipText = i18n.get(this, "statistics", I18n.TOOLTIP, "The pivoted values");
        String discardToolTipText = i18n.get(this, "discard", I18n.TOOLTIP, "The discarded attributes");
        String aggregateFunctionToolTipText = i18n.get(this, "aggregateFunction", I18n.TOOLTIP, "The aggregate function applied to the statistic attribute");

        String[] columnNames = {
            i18n.get(this, "attributes", "Attributes"),
            i18n.get(this, "discard", "Discard"),
            i18n.get(this, "criteria", "Criteria"),
            i18n.get(this, "pivot", "Pivot"),
            i18n.get(this, "statistics", "Statistics"),
            i18n.get(this, "aggregateFunction", "Aggregate Function")};
        Object[][] data = new Object[attributesNames.length][columnNames.length];
        JComponentTableModel model = new JComponentTableModel(data, columnNames);
        JComponentTable table = new JComponentTable(model);
        table.setRowHeight(25);
        table.setBackground(new Color(238, 238, 238));
        table.setDefaultRenderer(JComponent.class, new JComponentCellRenderer());
        DefaultTableCellRenderer renderer = new DefaultTableCellRenderer();
        renderer.setHorizontalAlignment(SwingConstants.CENTER);
        TableColumn col1 = table.getColumnModel().getColumn(0);
        col1.setCellRenderer(renderer);
        table.setDefaultEditor(JComponent.class, new JComponentCellEditor());

        int index = 0;
        for (String name : attributesNames) {
            table.setValueAt(name, index, 0);

            final String attribute = name;
            JComboBox cb = new AggregateFunctionComboBox(attribute);

            cb.setToolTipText(aggregateFunctionToolTipText);

            table.setValueAt(cb, index, 5);
            cb.setEnabled(attributesHashMap.get(name) == AttributeType.STATISTIC);

            ButtonGroup attributeRBGroup = new ButtonGroup();
            JRadioButton rb1, rb2, rb3, rb4;

            // 4 times the same lines!

            rb1 = new JRadioButton(new RBAction(name, AttributeType.DISCARD, cb, table));
            rb1.setToolTipText(discardToolTipText);
            rb1.setVerticalAlignment(JRadioButton.CENTER);
            rb1.setSelected(hasAttributeType(name, AttributeType.DISCARD));
            attributeRBGroup.add(rb1);

            table.setValueAt(rb1, index, 1);


            rb2 = new JRadioButton(new RBAction(name, AttributeType.CRITERIA, cb, table));
            rb2.setToolTipText(criteriaToolTipText);
            rb2.setSelected(hasAttributeType(name, AttributeType.CRITERIA));
            attributeRBGroup.add(rb2);
            table.setValueAt(rb2, index, 2);


            rb3 = new JRadioButton(new RBAction(name, AttributeType.PIVOT, cb, table));
            rb3.setToolTipText(pivotToolTipText);
            rb3.setSelected(hasAttributeType(name, AttributeType.PIVOT));
            attributeRBGroup.add(rb3);
            table.setValueAt(rb3, index, 3);


            rb4 = new JRadioButton(new RBAction(name, AttributeType.STATISTIC, cb, table));
            rb4.setToolTipText(statisticsToolTipText);
            rb4.setSelected(hasAttributeType(name, AttributeType.STATISTIC));
            attributeRBGroup.add(rb4);
            table.setValueAt(rb4, index, 4);


            index++;
        }



        validateButton = new JButton(i18n.get(this, "validate", "Validate"));
        // should be active only if at least a criteria is chosen -- TBD.
        validateButton.addActionListener(new ActionListener() {

            public void actionPerformed(ActionEvent e) {
                initAttributesListsFromHashMap();
                initialize();
            }
        });
        /*attributeAndValidatePanel*/
        JPanel validWrap = new JPanel();
        validWrap.add(validateButton);

        pivotPanel.add(attributePanel);
        pivotPanel.add(new JScrollPane(table));
        pivotPanel.add(validWrap);

        Box panel = Box.createVerticalBox();
        panel.add(pivotPanel);
        // add the datastorage controls
        panel.add(getDataStorage().getGUI());

        table.setFillsViewportHeight(true);
        table.updateUI();
        pivotPanel.setBorder(BorderFactory.createTitledBorder(i18n.get(this, "name", "Pivot")));

        return panel;
    }

    public boolean hasAttributeType(String n, AttributeType att) {
        return attributesHashMap.get(n) == att;
    }

    public void initAttributesHashMapFromLists(String[] attributesNames) {
        attributesHashMap = new HashMap<String, AttributeType>(attributesNames.length);
        for (String name : attributesNames) {
            attributesHashMap.put(name, AttributeType.DISCARD);
        }
        // update map with lists
        if (criteriaAttributesList != null) {
            for (String name : criteriaAttributesList) {
                attributesHashMap.put(name, AttributeType.CRITERIA);
            }
        }
        if (statisticsAttributesList != null) {
            for (String name : statisticsAttributesList) {
                attributesHashMap.put(name, AttributeType.STATISTIC);
            }
        }
        if (pivotAttributesList != null) {
            for (String name : pivotAttributesList) {
                attributesHashMap.put(name, AttributeType.PIVOT);
            }
        }
    }

    /** AttributesLists are made of RealColumnNames */
    public void initAttributesListsFromHashMap() {
        criteriaAttributesList = new ArrayList<String>();
        statisticsAttributesList = new ArrayList<String>();
        pivotAttributesList = new ArrayList<String>();

        for (Map.Entry<String, AttributeType> e : attributesHashMap.entrySet()) {
            switch (e.getValue()) {
                case DISCARD:
                    break;
                case CRITERIA:
                    criteriaAttributesList.add(e.getKey());
                    break;
                case STATISTIC:
                    statisticsAttributesList.add(e.getKey());
                    break;
                case PIVOT:
                    pivotAttributesList.add(e.getKey());
                    break;
            }
        }
        Debug.message("DataStorage", "[PivotDataStorage] Criteria: " + criteriaAttributesList + " Stats: " + statisticsAttributesList + " Pivot: " + pivotAttributesList);
    }

    public class RBAction extends AbstractAction {

        protected String name;
        protected AttributeType type;
        protected JComboBox aggregateFunctionComboBox;
        protected JTable table;

        // public RBAction(String name, AttributeType type) {
        //     this(name, type, null, null, null);
        // }

        public RBAction(String name, AttributeType type, JComboBox aggregateFunctionComboBox, JTable table) {
            this.name = name;
            this.type = type;
            this.aggregateFunctionComboBox = aggregateFunctionComboBox;
            this.table = table;
        }

        public void actionPerformed(ActionEvent e) {
            attributesHashMap.put(name, type);
            if (aggregateFunctionComboBox != null) // enable the aggregateFunctionComboBox
            {
                aggregateFunctionComboBox.setEnabled(type == AttributeType.STATISTIC);
            }
            table.updateUI();
        }
    }

    /** Used as a place holder to store intermediate result of aggregates :
     * the sum for SUM and AVG aggregates, the concat for CONCAT, ...
     */
    class AggregateResult {
	
        private Object res;
        private Object res2;
        private int nb = 0;

        public AggregateResult(Object o) {
            res = o;
        }

        public Object getRes() {
            return res;
        }

        public void setRes(Object r) {
            res = r;
        }

        public Object getRes2() {
            return res2;
        }

        public void setRes2(Object r) {
            res2 = r;
        }

        public int getNb() {
            return nb;
        }

        public void incrNb() {
            nb++;
        }
    
	public String toString() {
	    return "AggregateResult " + nb + " " + res + " " + res2;
	}
    }

}
