// <header>
/**
 *  ThemaMap : The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
package fr.unicaen.openmap.dataStorage;

import fr.unicaen.openmap.dataStorage.InformationDataStorage;
import fr.unicaen.openmap.dataStorage.QualitativeAnalysisTools;
import fr.unicaen.openmap.dataStorage.gui.DataTableModel;
import fr.unicaen.openmap.dataStorage.DataStorageIteratorAdapter;
import fr.unicaen.openmap.layer.thematicDataDisplayLayer.visu.Thematic;
import fr.unicaen.openmap.layer.thematicDataDisplayLayer.ThematicDataDisplayLayer;
import fr.unicaen.openmap.util.propertyEditor.DataStoragePropertyEditor;
import fr.unicaen.openmap.util.Util;

import com.bbn.openmap.util.PropUtils;
import com.bbn.openmap.util.Debug;
import com.bbn.openmap.omGraphics.*;
import com.bbn.openmap.omGraphics.geom.BasicGeometry;

import java.util.*;
import javax.swing.JPanel;
import javax.swing.JComponent;
import java.awt.Color;
import java.awt.geom.Rectangle2D;
import java.awt.geom.Point2D;

public class QualitativeAnalysisDataStorage extends InformationDataStorage
{
	private boolean ready;

	// Constants
	private final int ETATS = 2;

	// Source
	private Thematic thematic=null;
	private String layerAndThematicName = new String();
	private String layerName = new String();
	private String thematicName = new String();

	// Neural network parameters
	private double distance=170.0;
	private double difference=1.0;
	private int maxIterations=10000;
	private boolean autoStop=true;
	private boolean printSteps=false;
	private int modeDifference = QualitativeAnalysisTools.Criterion.DIFFERENCE_AREA;

	// PropertyStrings
	private final String distancePropertyString = "distance";
	private final String layerAndThematicNamePropertyString = "layerAndThematicName";
	private final String differencePropertyString = "difference";
	private final String printStepsPropertyString = "printSteps";
	private final String maxIterationsPropertyString = "maxIterations";
	private final String autoStopPropertyString = "autoStop";
	private final String modeDifferencePropertyString = "modeDifference";

	// ------------- Methods

	public QualitativeAnalysisDataStorage(DataStorageListener dsl)
	{
		super(dsl);
		Debug.message("DataStorage", this + " QualitativeAnalysisDataStorage(DataStorageListener dsl)");
	}

	public boolean isReady()
	{
		return ready;
	}

	public void setProperties(String prefix, java.util.Properties properties)
	{
		prefix = PropUtils.getScopedPropertyPrefix(prefix);
		super.setProperties(prefix, properties);

		layerAndThematicName = properties.getProperty(prefix + layerAndThematicNamePropertyString , "none");
		printSteps = Util.isTrue(properties.getProperty(prefix + printStepsPropertyString, ""+printSteps));
		autoStop = Util.isTrue(properties.getProperty(prefix + autoStopPropertyString, ""+autoStop));

		try
		{
			distance = Double.parseDouble( properties.getProperty(prefix + distancePropertyString , ""+distance ));
		}
		catch( NumberFormatException e )
		{}

		try
		{
			difference = Double.parseDouble(properties.getProperty(prefix + differencePropertyString , ""+difference));
		}
		catch( NumberFormatException e )
		{}

		try
		{
			maxIterations = Integer.parseInt(properties.getProperty(prefix + maxIterationsPropertyString , ""+maxIterations));
		}
		catch( NumberFormatException e )
		{}

		try
		{
			modeDifference = Integer.parseInt(properties.getProperty(prefix + modeDifferencePropertyString , ""+modeDifference));
		}
		catch( NumberFormatException e )
		{}

		String lAndt[] = layerAndThematicName.split(",");
		if( lAndt.length == 2 )
		{
			layerName = lAndt[0];
			thematicName = lAndt[1];
		}

		Debug.message("DataStorage", "[QualitativeAnalysisDataStorage] : layer("+layerName+", "+thematicName+"), "+
																																		"distance("+distance+"), "+
																																		"modeDifference("+modeDifference+"), "+
																																		"difference("+difference+"), "+
																																		"maxIterations("+maxIterations+"), "+
																																		"autoStop("+autoStop+"), "+
																																		"printSteps("+printSteps+")");

		try
		{
			initialize();
		}
		catch( Exception e )
		{
			e.printStackTrace();
		}
	}

	public Properties getProperties(String prefix, Properties properties)
	{
		properties = super.getProperties(prefix, properties);
		prefix = PropUtils.getScopedPropertyPrefix(prefix);

		properties.put(prefix + layerAndThematicNamePropertyString, layerAndThematicName);
		properties.put(prefix + distancePropertyString, ""+distance);
		properties.put(prefix + modeDifferencePropertyString, ""+modeDifference);
		properties.put(prefix + differencePropertyString, ""+difference);
		properties.put(prefix + maxIterationsPropertyString, ""+maxIterations);
		properties.put(prefix + autoStopPropertyString, ""+autoStop);
		properties.put(prefix + printStepsPropertyString, ""+printSteps);

		return properties;
	}

	public Properties getPropertyInfo(Properties list)
	{
		list = super.getPropertyInfo(list);

		PropUtils.setI18NPropertyInfo(i18n,
				list,
				QualitativeAnalysisDataStorage.class,
				layerAndThematicNamePropertyString,
				"Data source", //label
				"Select layer and thematic source", //tooltip
				"fr.unicaen.openmap.util.propertyEditor.LayerAndThematicNamePropertyEditor");

		PropUtils.setI18NPropertyInfo(i18n,
				list,
				QualitativeAnalysisDataStorage.class,
				distancePropertyString,
				"\"Far\" threshold", //label
				"The threshold for two graphic elements to be considered as \"far\"", //tooltip
				null);

		PropUtils.setI18NPropertyInfo(i18n,
				list,
				QualitativeAnalysisDataStorage.class,
				modeDifferencePropertyString,
				"\"Difference\" criterion", //label
				"The criterion used to compute \"Difference\"", //tooltip
				"fr.unicaen.openmap.util.propertyEditor.QADifferenceModePropertyEditor");

		PropUtils.setI18NPropertyInfo(i18n,
				list,
				QualitativeAnalysisDataStorage.class,
				differencePropertyString,
				"\"Difference\" threshold", //label
				"The threshold for two graphic elements to be considered as \"different\"", //tooltip
				null);

		PropUtils.setI18NPropertyInfo(i18n,
				list,
				QualitativeAnalysisDataStorage.class,
				maxIterationsPropertyString,
				"Max iterations", //label
				"Max iterations", //tooltip
				null);

		PropUtils.setI18NPropertyInfo(i18n,
				list,
				QualitativeAnalysisDataStorage.class,
				autoStopPropertyString,
				"Auto stop", //label
				"Automatically end iterations when the network has not been updated during last turn", //tooltip
				"fr.unicaen.openmap.util.propertyEditor.CheckBoxPropertyEditor");

		PropUtils.setI18NPropertyInfo(i18n,
				list,
				QualitativeAnalysisDataStorage.class,
				printStepsPropertyString,
				"With steps", //label
				"With steps", //tooltip
				"fr.unicaen.openmap.util.propertyEditor.CheckBoxPropertyEditor");


		list.put(initPropertiesProperty, layerAndThematicNamePropertyString+" "+
																		 distancePropertyString+" "+
																		 modeDifferencePropertyString+" "+
																		 differencePropertyString+" "+
																		 maxIterationsPropertyString+" "+
																		 autoStopPropertyString+" "+
																		 printStepsPropertyString);

		return list;
	}

	void initialize()
	{
		ArrayList<String> columnNames = new ArrayList<String>();
		ArrayList<String> omColumnNames = new ArrayList<String>();
		ArrayList<DataObject> dataObjects = null;

		// Get the target thematic from it's layer name and thematic name
		if( thematic == null )
		{
			for( ThematicDataDisplayLayer tddl : SharedDataStorage.getThematicDataDisplayLayers() )
			{ 
				if(tddl.getName().replaceAll("[ \\t]", "_").equals(layerName))
				{
					for( Thematic t : tddl.getThematicArray() )
					{
						if(t.getName().replaceAll("[ \\t]", "_").equals(thematicName))
						{
							thematic = t;
							break;
						}
					}
					break;
				}
			}
		}

		// Get a dataStorage iterator on the thematic's datastorage
		DataStorageIterator dsi = new DataStorageIteratorAdapter( thematic.getLayer().getDataStorageIterator().getDataStorage() );

		// Filling dataTable
		OMGraphicList omgl = getFlatOMGraphicList( thematic.generateGraphics(dsi) );
		//omgl = (omgl == null ) ? new OMGraphicList() : omgl;

		// All GraphicElements
		ArrayList<QualitativeAnalysisTools.GraphicElement> graphicElements = new ArrayList<QualitativeAnalysisTools.GraphicElement>( omgl.size() );

		// All Cells
		ArrayList<QualitativeAnalysisTools.Cell> cells = new ArrayList<QualitativeAnalysisTools.Cell>(graphicElements.size()*ETATS);

		// Fill Graphic and Cells ArrayList
		for( OMGraphic omg : omgl )
		{
			try
			{
				QualitativeAnalysisTools.Criterion crit = null;
				BasicGeometry geom = ((BasicGeometry)omg);
				geom.regenerate( thematic.getProjection() );
				Rectangle2D bBox = geom.getShape().getBounds2D();

				if( modeDifference == QualitativeAnalysisTools.Criterion.DIFFERENCE_AREA )
				{
					//System.out.println("--------------- Area : "+bBox.getWidth() * bBox.getHeight());
					crit = new QualitativeAnalysisTools.Criterion( modeDifference, bBox.getWidth() * bBox.getHeight() ); // Need to be improved : exact area based on shape
				}
				else
				if( modeDifference == QualitativeAnalysisTools.Criterion.DIFFERENCE_COLOR )
				{
					Color color = ((OMGraphicAdapter) omg).getFillColor();
					int red   = color.getRed();
					int green = color.getGreen();
					int blue  = color.getBlue();
					//System.out.println("--------------- RGB("+red+", "+green+", "+blue+")");
					crit = new QualitativeAnalysisTools.Criterion( modeDifference, red+green*1000+blue*1000000 );
				}

				QualitativeAnalysisTools.GraphicElement g = new QualitativeAnalysisTools.GraphicElement( crit, bBox, ETATS );
				graphicElements.add(g);

				for( QualitativeAnalysisTools.Cell c : g.getCells() )
				{
					cells.add(c);
				}
			}
			catch( Exception e )
			{
				Debug.error("--------- Cast error : "+omg);
				e.printStackTrace();
				continue;
			}
		}

		// Link weight matrix
		QualitativeAnalysisTools.Link linkMatrix[][] = new QualitativeAnalysisTools.Link[cells.size()][cells.size()];

		// Initialization rules
		QualitativeAnalysisTools.RuleTable rules = new QualitativeAnalysisTools.RuleTable( distance, difference, thematic.getProjection());

		// Initializing link values from rules
		for( int i = 0 ; i<cells.size() ; ++i )
		{
			for( int j = 0 ; j<cells.size() ; ++j )
			{
				linkMatrix[i][j] = rules.getRule( cells.get(i),  cells.get(j) );
			}
		}

		// Adding column names to dataStorage
		columnNames.add("lat");
		columnNames.add("lon");
		columnNames.add("step");
		for( int i = 0 ; i < ETATS ; ++i )
		{
			columnNames.add("Cell_"+i);
		}
		columnNames.add("result");

		// Adding geometry column names to dataStorage
		omColumnNames.add("lat/lon");

		// Initialize dataObjects ArrayList and reserve space to avoid re-allocations during iterations
		if( printSteps )
		{
			dataObjects = new ArrayList<DataObject>(cells.size()*maxIterations);
		}
		else
		{
			dataObjects = new ArrayList<DataObject>(cells.size());
		}

		// Computing
		double globalVariation;
		for( int iteration = 0 ; iteration < maxIterations ; ++iteration )
		{
			globalVariation = 0;
			// For each cell
			for( int idxC1 = 0 ; idxC1<cells.size() ; ++idxC1 )
			{
				QualitativeAnalysisTools.Cell current = cells.get(idxC1);
				double sum = current.getValue();

				// For each cell (again)
				for( int idxC2 = 0 ; idxC2<cells.size() ; ++idxC2 )
				{
					sum += linkMatrix[idxC1][idxC2].getScore(current);
				}

				sum = QualitativeAnalysisTools.sigmoid(sum);

				globalVariation += Math.abs( current.getValue()-sum );
				current.setValue(sum);
			}

			// Add velues to dataTable if printSteps is true or if it is the last iteration
			if( printSteps || globalVariation == 0.0f || iteration+1 == maxIterations )
			{
				// Filling table
				for( QualitativeAnalysisTools.GraphicElement ge : graphicElements )
				{
					DataObject dataObj = new DataObject();
					Rectangle2D bBox = ge.getBbox();
					Point2D latLonPoint = thematic.getProjection().inverse( bBox.getCenterX(), bBox.getCenterY());
					double lat = latLonPoint.getY();
					double lon = latLonPoint.getX();
					dataObj.addData("lat", ""+lat);
					dataObj.addData("lon", ""+lon);
					dataObj.addData("step", ""+iteration);
					dataObj.addOMDataLocation("lat/lon", new OMDataLocation(new com.bbn.openmap.omGraphics.OMPoint(lat,lon), new Centroid(lon,lat), null));

					Double max = Double.NEGATIVE_INFINITY;
					int type = -1;
					for( QualitativeAnalysisTools.Cell c : ge.getCells() )
					{
						if( c.getValue() > max )
						{
							max = c.getValue();
							type = c.getType();
						}

						dataObj.addData("Cell_"+c.getType(), ""+c.getValue());
					}

					dataObj.addData("result", ""+type);
					dataObjects.add(dataObj);
				}
			}

			// Autodetect end
			if(globalVariation == 0.0f)
			{
				break;
			}
		}

		// Set dataTable
		dataTable.setTable(dataObjects, columnNames, omColumnNames, new ArrayList<String>() );

		// set dataTableModel
		if( ! ready )
		{
			dataTableModel = new DataTableModel(dataTable);
			resetDataTableStoragePanel(dataTableModel);
		}

		// "ready" event
		ready = true;
		fireDataStorageEvent();
	}

	public JComponent getGUI()
	{
		JPanel panel = new JPanel();
		panel.add(DataStoragePropertyEditor.provideCustomPanel(this));
		return panel;
	}  

	private OMGraphicList getFlatOMGraphicList( OMGraphicList omgl )
	{
		OMGraphicList out = new OMGraphicList();

		if( omgl != null )
		{
			for( OMGraphic omg : omgl )
			{
				if( omg instanceof OMGraphicList )
				{
					out.addAll( getFlatOMGraphicList( (OMGraphicList) omg ) );
				}
				else
				{
					out.add(omg);
				}
			}
		}

		return out;
	}

}
