// <header>
/**
 *  ThemaMap : The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
package fr.unicaen.openmap.dataStorage;

import java.util.*;
import java.lang.*;
import java.awt.geom.Rectangle2D;
import fr.unicaen.openmap.layer.thematicDataDisplayLayer.omGraphics.*;
import com.bbn.openmap.proj.Projection;
import com.bbn.openmap.omGraphics.geom.BasicGeometry;
import java.awt.geom.Point2D;

public class QualitativeAnalysisTools
{

	public static class Criterion
	{
		public static final int DIFFERENCE_COLOR=1;
		public static final int DIFFERENCE_AREA=2;

		private final int _mode;
		private final double _value;

		public Criterion( int mode, double value )
		{
			_mode = mode;
			_value = value;
		}

		public double difference( Criterion other )
		{
			if( other._mode != _mode )
			{
				return Double.NaN;
			}

			switch( _mode )
			{
				case 1:
					// color Mode
					int self_color[] = new int[3];
					int other_color[] = new int[3];

					// values
					double self_value  =  this._value;
					double other_value = other._value;

					// Dirty trick used to get the 3 bytes in one double
					self_color[2]  = (int) (self_value / 1000000.0);
					other_color[2] = (int) (other_value / 1000000.0);
					self_value  -= 1000000.0 *  self_color[2];
					other_value -= 1000000.0 * other_color[2];
					self_color[1]  = (int) (self_value / 1000.0);
					other_color[1] = (int) (other_value / 1000.0);
					self_value  -= 1000.0 *  self_color[1];
					other_value -= 1000.0 * other_color[1];
					self_color[0] = (int) self_value;
					other_color[0] = (int) other_value;

					double sum = 0.0;
					for( int i = 0 ; i<3 ; ++i )
					{
						sum += Math.abs( self_color[i] - other_color[i] );
					}

					return sum;
				case 2:
					return Math.abs(this._value - other._value);
				default:
					return Double.NaN;
			}
		}
	}

	public static class Cell
	{
		protected Object _parent;	// parent GraphicElement
		protected int _type;			// cell type number
		protected double _value;	// cell score

		public Cell( Object parent, int type, double value )
		{
			_parent = parent;
			_type = type;
			_value = value;
		}

		public double getValue()
		{
			return _value;
		}

		public void setValue(double value)
		{
			_value = value;
		}

		public Object getParent()
		{
			return _parent;
		}

		public int getType()
		{
			return _type;
		}
	}

	public static class GraphicElement
	{
		protected Cell _cells[];		// Cells representing element
		protected Criterion _crit;		// the "difference" parameter (based on size, color, shape, ...)
		protected Rectangle2D _bBox;

		public GraphicElement( Criterion crit, Rectangle2D bBox, int stateNb )
		{
			_cells = new Cell[stateNb];
			_crit = crit;
			_bBox = bBox;

			for( int i = 0 ; i<stateNb ; ++i )
			{
				_cells[i] = new Cell( this, i, 0.5 );
			}
		}

		public Cell[] getCells()
		{
			return _cells;
		}

		public Rectangle2D getBbox()
		{
			return _bBox;
		}

		public Criterion getCriterion()
		{
			return _crit;
		}
	}

	public static class Link
	{
		private double _weight;		// link weight
		private Cell _a, _b;			// the two cells linked

		public Link( Cell a, Cell b, double weight )
		{
			_a = a;
			_b = b;
			_weight = weight;
		}

		public Cell getOther( Cell self )
		{
			return self.equals(_a) ? _b : _a ;
		}

		public double getScore( Cell self )
		{
			return ( self.equals(_a) ? _b.getValue() : _a.getValue() ) * _weight ;
		}
	}

	public static class RuleTable
	{
		protected double _table[][][][];			// Rule table (could also be a tree)
		protected double _farDistThreshold;		// Far threshold
		protected double _differentThreshold;	// Difference threshold
		protected Projection _proj;

		public RuleTable( double farDistThreshold, double differentThreshold, Projection proj )
		{
			_farDistThreshold = farDistThreshold;
			_differentThreshold = differentThreshold;
			_proj = proj;

			_table = new double[2][2][2][2];

			/*
				 First idx :
				 - 0 = near
				 - 1 = far

				 Second idx :
				 - 0 = similar
				 - 1 = different

				 Third idx :
				 Type of first cell

				 Fourth idx :
				 Type of second cell
			 */

			// ----------- Sample rule table detecting divisions ------------

			// UU : "uniform" cell to "uniform" cell
			// UD : "uniform" cell to "division" cell
			// DU : "division" cell to "uniform" cell
			// DD : "division" cell to "division" cell

			// near and similar
			_table[0][0][0][0] = +1.0;	// UU
			_table[0][0][0][1] =  0.0;  // UD
			_table[0][0][1][0] =  0.0;  // DU
			_table[0][0][1][1] = -1.0;  // DD

			// near and different
			_table[0][1][0][0] = -1.0; 	// UU
			_table[0][1][0][1] =  0.0;  // UD
			_table[0][1][1][0] =  0.0;  // DU
			_table[0][1][1][1] = +1.0;  // DD

			// far and similar
			_table[1][0][0][0] =  0.0;	// UU
			_table[1][0][0][1] =  0.0;  // UD
			_table[1][0][1][0] =  0.0;  // DU
			_table[1][0][1][1] =  0.0;  // DD

			// far and different
			_table[1][1][0][0] =  0.0;	// UU
			_table[1][1][0][1] =  0.0;  // UD
			_table[1][1][1][0] =  0.0;  // DU
			_table[1][1][1][1] =  0.0;  // DD
		}

		public Link getRule( Cell a, Cell b )
		{
			// Cells from same parent have no influence
			if( a.getParent() == b.getParent() )
			{
				return new Link( a, b, 0);
			}

			// Get centroids
			Rectangle2D bBox_a = ((GraphicElement) a.getParent() ).getBbox();
			Rectangle2D bBox_b = ((GraphicElement) b.getParent() ).getBbox();

			//System.out.println( "--------------------- Center : "+bBox_a.getCenterX()+"\t"+bBox_a.getCenterY() );

			// Compute distance and difference between graphic objects
			//double distance = _proj.forward( omg_a.getLat(), omg_a.getLon() ).distance( _proj.forward( omg_b.getLat(), omg_b.getLon() ) );
			double distance = Point2D.distance( bBox_a.getCenterX(), bBox_a.getCenterY(), bBox_b.getCenterX(), bBox_b.getCenterY() );
			//double difference = Math.abs( ((GraphicElement) a.getParent() ).getValue() - ((GraphicElement) b.getParent() ).getValue() );
			double difference = ((GraphicElement) a.getParent() ).getCriterion().difference( ((GraphicElement) b.getParent() ).getCriterion() );
			int idxDist = distance > _farDistThreshold ? 1 : 0 ;
			int idxDiff = difference > _differentThreshold ? 1 : 0 ;
			int idxTypeA= a.getType();
			int idxTypeB= b.getType();

			// Weight
			double weight = _table[idxDist][idxDiff][idxTypeA][idxTypeB];

			return new Link( a, b, weight);
		}
	}

	// Utility : a sigmoid function between 0 an 1
	public static double sigmoid( double x )
	{
		return 1.0 / ( 1.0 + Math.exp( -x ) );
	}
}
