// <header>
/**
 *  ThemaMap : The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
/*
 * SelectionDataStorage.java
 *
 * Created on 9 novembre 2005, 23:31
 *
 * @author Christophe.Turbout@info.unicaen.fr
 *
 */

package fr.unicaen.openmap.dataStorage;
import com.bbn.openmap.util.ColorFactory;
import com.bbn.openmap.util.Debug;
import com.bbn.openmap.util.PropUtils;
import fr.unicaen.openmap.dataStorage.gui.DataAndSelectionModel;
import fr.unicaen.openmap.dataStorage.gui.DataTableModel;
import fr.unicaen.openmap.dataStorage.gui.SelectionStoragePanel;
import fr.unicaen.openmap.layer.isochroneLayer.IsochroneDataObject;
import fr.unicaen.openmap.layer.preLayer.UnicaenLayerListener;
import fr.unicaen.openmap.util.Util;
import java.awt.Color;
import java.util.ArrayList;
import java.util.Properties;
import javax.swing.JComponent;

/**
 *
 * @author Christophe.Turbout@info.unicaen.fr
 *
 * <p>Sample implementation of an Selection dataStorage.
 *
 * Example for openmap.properties :
 * <pre>
 *   selection.class             = fr.unicaen.openmap.layer.dataDisplayLayer.InteractiveDataDisplayLayer
 *   selection.dataStorageClass  = fr.unicaen.openmap.dataStorage.SelectionDataStorage
 *   selection.storageComponents   = xmltable
 *   selection.defaultSelectionName = Zone de chalandise
 *   selection.defaultSelectionName = Sous zone
 *   selection.defaultColor1 = C9FFFF99
 *   selection.defaultColor2 = C9FEF2F8
 *   selection.defaultColor3 = C9D9F0D3
 *   selection.defaultColor4 = C9E6E9F3
 *   selection.defaultColor5 = C9FFB300
 *   selection.defaultColor6 = C9EC93C1
 *   selection.defaultColor7 = C974C476
 *   selection.defaultColor8 = C980A6CE
 *   selection.multipleSelectionModel = no
 *   selection.quickLocationMode = yes
 *   selection.quickLocationColumn = nom
 *   selection.xmltable.dataStorageClass  = fr.unicaen.openmap.dataStorage.xmltable.XmlTableDataStorage
 *   selection.xmltable.prettyName        = communes
 *   selection.xmltable.xmlFile  = communes.xml
 *   selection.geometryColumnName = commune
 *   selection.lineColor         = FF000000
 *   selection.fillColor         = FFFFFFFF
 *   selection.selectedLineColor = FF990099
 *   selection.selectedFillColor = C900FF55
 *   selection.fill              = yes
 *   selection.drawToolTips      = yes
 *   selection.selectionDatas=<?xml version="1.0" encoding="utf-8" ?><storage><selection><name>Zone de chalandise 1</name><nbColumnCreated>3</nbColumnCreated><subselection><name>Sous zone  1</name><color>c9ffb300</color><value>5282</value><subselection><subselection><name>Sous zone  2</name><color>c9ec93c1</color><value>11225,18838</value></selection></storage>
 *   selection.storageComponents=xmltable
 *   selection.toolTipsColumnName= nom
 * </pre>
 */

public class SelectionDataStorage  extends DataStorage implements DataStorageListener {
    
    /* --------------------------  Begin   Properties  ---------------------*/
    public static final String storageComponentsString = "storageComponents";
    public static final String selectionFileString = "fileStorage";
    public static final String defaultSelectionString = "defaultSelectionName";
    public static final String defaultSubSelectionString = "defaultSubSelectionName";
    public static final String defaultColor1String = "defaultColor1";
    public static final String defaultColor2String = "defaultColor2";
    public static final String defaultColor3String = "defaultColor3";
    public static final String defaultColor4String = "defaultColor4";
    public static final String defaultColor5String = "defaultColor5";
    public static final String defaultColor6String = "defaultColor6";
    public static final String defaultColor7String = "defaultColor7";
    public static final String defaultColor8String = "defaultColor8";
    public static final String multipleSelectionModelString = "multipleSelectionModel";
    public static final String datasSelectionPropertyString = "selectionDatas";
    /**
     * Property string for the DataStorage class name.
     */
    static public String dataStorageClassPropertyString = "dataStorageClass";
    /**
     * Class name for DataStorage.
     */
    protected String className = "";
    
    DataStorage dataStorage = null;
    private String storageName = null;
    private String defaultSelectionName = "selection";
    private String defaultSubSelectionName = "sub selection";
    private boolean multipleSelectionModel = false;
    private String storageFileName = "unnamedFilename";
    private String name = "unnamedSelectionDataStorage";
    private String initialDatas = "";
    
    /* --------------------------  End   Properties    ---------------------*/
    
    
    /* --------------------------  Begin   Variables  ---------------------*/
    
    private DataStorage datas = null;
    private ArrayList<Color> defaultColors = null;
    
    private boolean readyDataStorage = false;
    private boolean readyDatas = false;
    private SelectionManager selections = null;
    private DataStorageIterator datasDataStorageIterator = null; // iterator on initial data
    private DataAndSelectionModel dataAndSelectionModel =  null;
    private UnicaenLayerListener unicaenLayerListener = null;
    /* --------------------------  End Variables      ---------------------*/
    
    
    
    /* default constructor. Nothing is set. This is done by setProperties method. */
    public SelectionDataStorage(DataStorageListener dsl) {
        super(dsl);
        Debug.message(Util.className(this), this + " SelectionDataStorage(DataStorageListener dsl)");
    }
    
    /** Set the DataStorage used in the selectionDataStorage */
    public void setDataStorage(DataStorage ds){
        
        if (ds != null) {
            datas = ds;
            readyDataStorage = true;
            datasDataStorageIterator = new DataStorageIteratorAdapter(ds);
            dataTableModel = (DataTableModel) datas.getDefaultDataTableModel();
            
            try {
                // First remove the unicaen listener
                if (dtsp!=null) {
                    removeUnicaenListener();
                    ((SelectionStoragePanel)dtsp).removeUnicaenListener();
                }
                selections = new SelectionManager(dataTableModel,defaultSelectionName,defaultSubSelectionName,defaultColors,multipleSelectionModel);
                // the dataAndSelectionModel to view datas and the first selection
                dataAndSelectionModel= new DataAndSelectionModel(dataTableModel,selections.getCurrentSelectionMatrix(),selections.getCurrentName());
                // the panel used to view all those informations.
                dtsp = new SelectionStoragePanel(this);
                ((SelectionStoragePanel)dtsp).addUnicaenListener(unicaenLayerListener);
            } catch (Exception e){
                System.err.println("===> [ "+Util.className(this)+" ] :  Error "+e.toString());
                e.printStackTrace();
            }
        } else {
            datas = null;
            readyDataStorage = false;
            datasDataStorageIterator = null;
            dataTableModel = null;
            dtsp = null;
            dataAndSelectionModel = null;
        }
    }
    
    
    /**
     * Let's find our properties.
     * @param prefix string prefix used in the properties file for the layer
     * @param properties the properties set in the properties file
     */
    public void setProperties(String prefix, Properties properties) {
        super.setProperties(prefix, properties);
        String col = null;
        Color color =null;
        
        Debug.message(Util.className(this) , this + " setProperties() : "+prefix);
        
        storageName = properties.getProperty(prefix + storageComponentsString, "dataStorage");
        defaultSelectionName=properties.getProperty(prefix +  defaultSelectionString, "defaultSelectionName");
        defaultSubSelectionName=properties.getProperty(prefix +  defaultSubSelectionString, "defaultSubSelectionName");
        multipleSelectionModel = Util.isTrue(properties.getProperty(prefix + multipleSelectionModelString, "false"));
        // colors ... stupid implementation  !
        // need to be optimized ...
        defaultColors = new ArrayList<Color>();
        
        col = properties.getProperty(prefix +  defaultColor1String, "C9FFFF99");
        color = ColorFactory.parseColor(col);
        if (color != null) defaultColors.add(color);
        col = properties.getProperty(prefix +  defaultColor2String, "C9FEF2F8");
        color = ColorFactory.parseColor(col);
        if (color != null) defaultColors.add(color);
        col = properties.getProperty(prefix +  defaultColor3String, "C9D9F0D3");
        color = ColorFactory.parseColor(col);
        if (color != null) defaultColors.add(color);
        col = properties.getProperty(prefix +  defaultColor4String, "C9E6E9F3");
        color = ColorFactory.parseColor(col);
        if (color != null) defaultColors.add(color);
        col = properties.getProperty(prefix +  defaultColor5String, "C9FFB300");
        color = ColorFactory.parseColor(col);
        if (color != null) defaultColors.add(color);
        col = properties.getProperty(prefix +  defaultColor6String, "C9EC93C1");
        color = ColorFactory.parseColor(col);
        if (color != null) defaultColors.add(color);
        col = properties.getProperty(prefix +  defaultColor7String, "C974C476");
        color = ColorFactory.parseColor(col);
        if (color != null) defaultColors.add(color);
        col = properties.getProperty(prefix +  defaultColor8String, "C980A6CE");
        color = ColorFactory.parseColor(col);
        if (color != null) defaultColors.add(color);
        
        String prefix2 = PropUtils.getScopedPropertyPrefix(prefix+storageName);
        className = properties.getProperty(prefix2 + dataStorageClassPropertyString);
        dataStorage = Util.loadDataStorage(className,prefix2,properties,this);
        setDataStorage(dataStorage);
        
        initialDatas=properties.getProperty(prefix +  datasSelectionPropertyString,"");
        // we need to load Selections stored in selectionFile
        if (initialDatas!=null && initialDatas.length()>0){
            loadSelections(initialDatas);
        } else Debug.message(Util.className(this) , this + "Erreur, pas de données :"+initialDatas);
    }
    
    //---------------------------     implements DataStorage -----------------------------------------------
    /**
     * The properties and prefix are managed and decoded here, for the standard uses of SelectionDataStorage
     * Notes this method allows to define what is present in file "save Map as properties file".
     * @return Properties.
     */
    public Properties getProperties(String prefix,Properties properties) {
        Debug.message(Util.className(this) , this + " SelectionDataStorage getProperties() : ");
        
        properties.put(prefix +  storageComponentsString, storageName);
        properties.put(prefix +  defaultSelectionString, defaultSelectionName);
        properties.put(prefix +  defaultSubSelectionString, defaultSubSelectionName);
        properties.put(prefix +  multipleSelectionModelString, new Boolean(multipleSelectionModel).toString());
        if (defaultColors.size()==8) {
            properties.put(prefix+defaultColor1String, ColorFactory.getHexColorString(defaultColors.get(0)));
            properties.put(prefix+defaultColor2String, ColorFactory.getHexColorString(defaultColors.get(1)));
            properties.put(prefix+defaultColor3String, ColorFactory.getHexColorString(defaultColors.get(2)));
            properties.put(prefix+defaultColor4String, ColorFactory.getHexColorString(defaultColors.get(3)));
            properties.put(prefix+defaultColor5String, ColorFactory.getHexColorString(defaultColors.get(4)));
            properties.put(prefix+defaultColor6String, ColorFactory.getHexColorString(defaultColors.get(5)));
            properties.put(prefix+defaultColor7String, ColorFactory.getHexColorString(defaultColors.get(6)));
            properties.put(prefix+defaultColor8String, ColorFactory.getHexColorString(defaultColors.get(7)));
        }
        
        String xmlversion = saveSelections();
        properties.put(prefix+datasSelectionPropertyString,xmlversion);
        
        // ajout nono test
        String prefix2 = PropUtils.getScopedPropertyPrefix(prefix + storageName);
        properties.put(prefix2 + dataStorageClassPropertyString , className);
        
        // save DataStorage Properties
        if (dataStorage!=null) dataStorage.getProperties(prefix2,properties);
        
        return properties;
    }
    
    public DataObject getElementAt(int i){
        if (datas!=null && i<datas.getCount()) return datas.getElementAt(i);
        return null;
    }
    
    public int getCount(){
        if (datas!=null) return datas.getCount();
        return 0;
    }
    
    
    public boolean isReady() { return readyDataStorage && readyDatas; }
    
    /**
     *  @return JComponent built with components GUI.
     **/
    public JComponent getGUI() {
        // nothing to add at the moment.
        // if we want to add an expert mode we could add a file chooser to find
        // a DataStorage ... a little difficult to but not impossible.
        return datas.getGUI();
    }
    /**
     * Set the active selection in SelectionDataStorage and notifies the change to all components
     *
     * @see SelectionMatrix
     * @see DataAndSelectionModel
     */
    public void setCurrentSelection(int selection){
        Debug.message(Util.className(this) , this + "Sélection "+selection+" active.");
        try {
            selections.setCurrentSelection(selection);
            dataAndSelectionModel.setMatrix(selections.getCurrentSelectionMatrix(),selections.getCurrentName());
        } catch (Exception e){
            System.err.println("===> [ "+Util.className(this)+" ] :  Error "+e.toString());
            e.printStackTrace();
        }
        if (dtsp!=null) ((SelectionStoragePanel)dtsp).fireStructureChanged();
    }
    
    
    public Object getValueAt(int row, int col){ return dataAndSelectionModel.getValueAt(row,col);}
    
    //---------------------------  end implements DataStorage -----------------------------------------------
    
    //---------------------------     implements DataStorageListener -----------------------------------------------
    
     /*
      * if the component of this datastorage is modified we can know it !
      **/
    public void dataStorageEvent(DataStorage source){
        Debug.message(Util.className(this) , this + " SelectionDataStorage, DataStorageEvent from "+source);
    }
    
    //---------------------------  end implements DataStorage -----------------------------------------------
      
    /**
     * Add an Empty Matrix with a default name.
     **/
    public void createEmptySelection(){
        selections.createEmptySelection(dataTableModel,multipleSelectionModel);
    }
     
    /**
     * Add a new Empty Matrix.
     **/
    public void createNewEmptySelection(String name){
        int selection = selections.createNewEmptySelection(name,dataTableModel,multipleSelectionModel);
        setCurrentSelection(selection);
    }

    
    //---------------------------  other  methods -----------------------------------------------
    
   /*
    * Load and save information in XML Format.
    * DataStorage is linked with some informations. We can store a copy of the data selected or index.
    * We store index. DataStorage must be the same otherwise the stored informations doesn't mean anything.
    *
    * <storage>
    *   <nbZoneCreated>1</nbZoneCreated>
    *   <isochroneDataObject>
    *   <autoroute>115</autoroute> ...
    *   </isochroneDataObject>
    *   <selection>
    *     <name>Zone de Chalandise 1</name>
    *     <nbColumnCreated>3</nbColumnCreated>
    *     <subselection>
    *        <name>sous zone 1</name>
    *        <color>FF00FF00</color>
    *        <value>1,3,5,7,8</value>
    *     <subselection>
    *     <subselection>
    *        <name>sous zone 3</name>
    *        <color>FF0000FF</color>
    *        <value>2,4,6,7,9</value>
    *     </subselection>
    *   </selection>
    * </storage>
    **/
    private void loadSelections(String datas){
        // reset all variables to load what is stored.
        selections.loadSelections(datas,dataTableModel,multipleSelectionModel);
        // the panel used to view all those informations.
        dtsp = new SelectionStoragePanel(this);
        
        if (selections.size()>0) setActiveSelection(0);
        readyDatas = true;
        fireDataStorageEvent();
    }
    
    /*
     * Need to use a thread to read information. This is not finished at all !
     **/
    private String saveSelections(){
        Debug.message(Util.className(this) , this + "Début de trace des matrix");
        StringBuffer buffer = new StringBuffer();
        buffer.append("<?xml version=\"1.0\" encoding=\"utf-8\" ?><storage>");
        buffer.append(selections.saveSelections());
        buffer.append("</storage>");
        Debug.message(Util.className(this) , this + buffer.toString());
        return buffer.toString();
    }
    
    /**
     *
     */
    public boolean isTitleExists(String newZoneName) {
        for (int i = 0; i<selections.size(); i++) {
            String name = getSelectionName(i);
            if (name!=null && newZoneName.trim().toUpperCase().equals(name.toUpperCase())) return true;
        }
        return false;
    }
    
//---------------------------  end Private methods -----------------------------------------------
    
    
    
//---------------------------  begin accessor  methods -----------------------------------------------
    /**
     * Return the current color in the current selection.
     *
     * @return Color
     **/
    public Color getCurrentColorSelection() {
        return selections.getCurrentColorSelection();
    }
    
    /**
     * return the currentSelection
     *
     * @see SelectionMatrix
     **/
    public int getCurrentSelection(){
        return selections.getCurrentSelection();
    }
    
    /**
     * return the current SubSelection in current Selection
     *
     * @see SelectionMatrix
     **/
    public int getCurrentSubSelection(){
        return selections.getCurrentSubSelection();
    }
    
    /**
     * return the name of the current Selection
     *
     **/
    public String getCurrentSelectionName(){
        return selections.getSelectionName(selections.getCurrentSelection());
    }
    
    /**
     * Return the model used to view data and current selection.
     **/
    public DataAndSelectionModel getDataAndSelectionModel(){
        return dataAndSelectionModel;
    }
    
    
    /**
     * get the color of the selected omgraphics if it is selected.
     * Do nothing otherwise.
     **/
    public Color getOMGraphicColor(int index){
        return selections.getOMGraphicColor(index);
    }
    
    /**
     * return the Color vector used by the given selection, null otherwise
     **/
    public ArrayList<Color> getSelectionColors(int selection){
        return selections.getSelectionColors(selection);
    }
    
    /**
     * return the SelectionMatrix of the selection
     **/
    public SelectionMatrix getSelectionMatrix(int selection){
        return selections.getSelectionMatrix(selection);
    }
    
    /**
     * Retrun the name of the selection.
     **/
    public String getSelectionName(int selection){
        return selections.getSelectionName(selection);
    }
    
    /**
     * Return the color of the subselection in the selection
     **/
    public Color getSubSelectionColor(int selection, int subselection){
        return selections.getSubSelectionColor(selection,subselection);
    }
    
    /**
     * Retrun the name of the subselection in the selection.
     **/
    public String getSubSelectionName(int selection, int subselection){
        return selections.getSubSelectionName(selection,subselection);
    }
    
    /**
     * Retrun the names of the selection.
     **/
    public ArrayList<String> getSubSelectionNames(int selection){
        return selections.getSubSelectionNames(selection);
    }
    
    /**
     * Retrun the number of selections of the selection.
     **/
    public int getSelectionsSize(){
        int nbsel = -1;
        try {
            nbsel = selections.size();
        } catch (Exception e){
            System.out.println("===> [ "+Util.className(this)+" ] :  Error "+e.toString());
            e.printStackTrace();
        }
        return nbsel;
    }
    
    /**
     * return number of selection in the current selection
     */
    public int getCurrentSelectionSize() {        
        if (selections!=null && selections.size()>selections.getCurrentSelection()) {
            return selections.getCurrentSelectionMatrix().getSubSelectionsSize();
        }
        else return -1;
    }
    
    /**
     * Set the active selection in SelectionDataStorage and notifies the change only to DataAndSelectionModel.
     *
     * @see SelectionMatrix
     * @see DataAndSelectionModel
     */
    public void setActiveSelection(int selection){
        if (selection<selections.size() && selection>=0) {
            selections.setCurrentSelection(selection);
            try {
                dataAndSelectionModel.setMatrix(selections.getCurrentSelectionMatrix() , selections.getCurrentName());
                if (unicaenLayerListener!=null) unicaenLayerListener.refreshGraphics();
            } catch (Exception e){
                System.err.println("===> [ "+Util.className(this)+" ] :  Error "+e.toString());
                e.printStackTrace();
            }
        }
    }
    
    /**
     * Set the active selection in CustomTabbedPane.
     *
     * @see SelectionMatrix
     * @see DataAndSelectionModel
     */
    public void setTabSelectedIndex(int selection){
        if (selection<selections.size() && selection>=0) {
            selections.setCurrentSelection(selection);
            try {
                setCurrentSelection(selection);
                // selectIndexInModel(selection);
                if (dtsp!=null) ((SelectionStoragePanel)dtsp).setActiveSelection(selection);
                if (unicaenLayerListener!=null) unicaenLayerListener.refreshGraphics();
            } catch (Exception e){
                System.err.println("===> [ "+Util.className(this)+" ] :  Error "+e.toString());
                e.printStackTrace();
            }
        }
    }
    
    
    
    public void setCurrentSelectionName(String newName){
        selections.setCurrentSelectionName(newName);
        ((SelectionStoragePanel)dtsp).fireSelectionNameChanged(getCurrentSelection(),newName);
    }
    
    /*
     * Rename the subselection index in the current selection.
     */
    public void setCurrentSelectionSubName(int index, String newName){
        selections.setSubSelectionName(getCurrentSelection(), index,newName);
        ((SelectionStoragePanel)dtsp).fireSelectionSubNameChanged(getCurrentSelection());
    }
    
    /**
     * Set the color of the subselection in the currentSelection.
     **/
    public void setSubSelectionColor(int subselection, Color color){
        selections.setSubSelectionColor(subselection,color);
    }
    
    /**
     * Set the active subselection in SelectionDataStorage. We store this information and change the
     * data in selectionMatrixModel and notifies the change to DataAndSelectionModel listeners.
     *
     * @see SelectionMatrix
     * @see fr.unicaen.openmap.dataStorage.gui.SelectionMatrixModel
     * @see DataAndSelectionModel
     */
    public void setCurrentSubSelection(int subselection){
        selections.setCurrentSubSelection(subselection);
    }
    
    public String getNextSubSelectionName(){
        String subname = new String(defaultSubSelectionName+" "+(selections.getCurrentSelectionMatrix().getNbColumnCreated()+1));
        return subname;
    }
    
    /**
     * Return the default name of the next selection using DefaultSelectionName.
     **/
    public String getNextSelectionName(){
        String subname = new String(defaultSelectionName+" "+(selections.getNbZoneCreated()+1));
        return subname;
    }
    
//---------------------------  end accessor  method -----------------------------------------------
    
//---------------------------  Other methods -----------------------------------------------
    
    /**
     * return the value of the column at specified index row.
     **/
    
    public String getRowValue(int rowindex, String columnName){
        return (String)((DataObject)getElementAt(rowindex)).getData(columnName);
    }
    
    /**
     * Select the index row in current SelectionMatrix and currentSubSelection.
     * return true if the index is selected, false otherwise.
     * This depends on multipleSelectionModel.
     **/
    public boolean selectIndexInModel(int index) {
        SelectionMatrix matrix = selections.getCurrentSelectionMatrix();
        if (matrix.getComposedValue(index) && !multipleSelectionModel) {
            // already selected and mode tell us it can be selected only one time.
            return false;
        }
        // we select it
        matrix.selectRow(index);
        // need to fire other view ...
        ((SelectionStoragePanel)dtsp).fireDataChanged(selections.getCurrentSelection(),matrix.getCurrentSubSelection(),index,true);
        return true;
    }
    
    /**
     * Deselect the index row in current SelectionMatrix and currentSubSelection.
     **/
    public boolean deselectIndexInModel(int index) {
        SelectionMatrix matrix = selections.getCurrentSelectionMatrix();
        // we need to know if index is selected in the current subselection.
        if (!matrix.isSelectedCurrentSubSelection(index)) return false;
        matrix.deselectRow(index);
        ((SelectionStoragePanel)dtsp).fireDataChanged(selections.getCurrentSelection(),matrix.getCurrentSubSelection(),index,false);
        return true;
    }
    
    
    /**
     * Deselect the index row in current SelectionMatrix and currentSubSelection.
     * and select this row in subSelection
     **/
    public boolean moveIndexInModel(int index, int subSelection) {
        SelectionMatrix matrix = selections.getCurrentSelectionMatrix();
        // we need to know if index is selected in the current subselection.
        if (!matrix.isSelectedCurrentSubSelection(index)) return false;
        matrix.selectRowInSubSelection(index,subSelection);
        ((SelectionStoragePanel)dtsp).fireDataChanged(selections.getCurrentSelection(),matrix.getCurrentSubSelection(),index,false);
        ((SelectionStoragePanel)dtsp).fireDataChanged(selections.getCurrentSelection(),subSelection,index,true);
        return true;
    }
    
    
    /**
     * Select all rows in the current Selection.
     **/
    public void selectAllDefaultSelection(){
        SelectionMatrix currentMatrix = selections.getCurrentSelectionMatrix();
        currentMatrix.setColumnState(true);
        // need to fire other view ...
        ((SelectionStoragePanel)dtsp).fireDataChanged(selections.getCurrentSelection(),currentMatrix.getCurrentSubSelection(),-1,true);
    }
    
    /**
     * Deselect all rows in the current Selection.
     **/
    public void deselectAllDefaultSelection(){
        SelectionMatrix currentMatrix = selections.getCurrentSelectionMatrix();
        currentMatrix.setColumnState(false);
        // need to fire other view ...
        ((SelectionStoragePanel)dtsp).fireDataChanged(selections.getCurrentSelection(),currentMatrix.getCurrentSubSelection(),-1,false);
    }
    
    /**
     * Add a new Selection Matrix which is the copy of an exiting selectionMatrix.
     **/
    public void createSelectionCopy(int selection, String name){
        int newselection = selections.createSelectionCopy(selection,name);
        setCurrentSelection(newselection);
    }
    
    
    
    /**
     * Create a new subselection.
     */
    public void createSubSelection(){
        int currentSubSelection = selections.createSubSelection(null);
        // need to fire the panel (gui) : this colum become the active currentSubSelection Column
        ((SelectionStoragePanel)dtsp).fireStructureColumnAdd(selections.getCurrentSelection(), currentSubSelection);
    }
    
    /**
     * Create a new subselection having the given subname
     */
    public void createSubSelection(String subSelectionName){
        Debug.message(Util.className(this),this+"Création de la sous-sélection : "+subSelectionName);
        int currentSubSelection = selections.createSubSelection(subSelectionName);
        // need to fire the panel (gui) : this colum become the active currentSubSelection Column
        ((SelectionStoragePanel)dtsp).fireStructureColumnAdd(selections.getCurrentSelection(), currentSubSelection);
    }
    
    /**
     * remove current subselection.
     */
    public void deleteSubSelection(){
        int oldSelection = selections.deleteCurrentSubSelection();
        ((SelectionStoragePanel)dtsp).fireStructureColumnRemoved(selections.getCurrentSelection(),oldSelection);
        if (unicaenLayerListener!=null) unicaenLayerListener.refreshGraphics();
    }

    /**
     * remove current selection.
     */
    public void deleteCurrentSelection(){
        selections.deleteCurrentSelection();
        ((SelectionStoragePanel)dtsp).fireStructureChanged();
        if (unicaenLayerListener!=null) unicaenLayerListener.refreshGraphics();
    }
    
    /**
     *  Add UnicaenLayerListener to refresh layer();
     */
    public void addUnicaenListener(UnicaenLayerListener ull){
        unicaenLayerListener=ull;
        ((SelectionStoragePanel)dtsp).addUnicaenListener(ull);
    }
    
    /**
     *  Remove the UnicaenLayerListener;
     */
    public void removeUnicaenListener(){
        unicaenLayerListener=null;
    }
    
    public void fireUpdateColors(){
        ((SelectionStoragePanel)dtsp).fireUpdateColors();
    }
    
    public void setCurrentSelectionInformation(IsochroneDataObject isochroneData){
       selections.getCurrentSelectionMatrix().setIsochroneDataObject(isochroneData);
    }
    
    public IsochroneDataObject getCurrentSelectionInformation(){
       return selections.getCurrentSelectionMatrix().getIsochroneDataObject();
    }
    
    public IsochroneDataObject getLastSelectionInformation() {
        for (int i = this.getSelectionsSize(); i>0;i--) {
            selections.setCurrentSelection(i-1);
            IsochroneDataObject info = this.getCurrentSelectionInformation();
            if (info!=null) return info;
        }
        return null;
    }
    
}
