// <header>
/**
 *  ThemaMap : The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
/*
 * CustomTabbedPane.java
 *
 * Created on 9 décembre 2005, 15:10
 *
 */

package fr.unicaen.openmap.dataStorage.gui;
import fr.unicaen.openmap.dataStorage.SelectionDataStorage;
import fr.unicaen.openmap.layer.isochroneLayer.IsochroneDataObject;
import fr.unicaen.openmap.util.Util;
import java.awt.Component;
import java.awt.Frame;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import javax.swing.AbstractButton;
import javax.swing.ImageIcon;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPopupMenu;
import javax.swing.JTabbedPane;

/**
 * Custom JTabbedPane
 * Allows mouse right click popup menu with add sub tab, rename, duplicate and close actions
 * @author Arnaud Daret, CERTIC, Université de Caen, daret@info.unicaen.fr
 */
class CustomTabbedPane extends JTabbedPane implements ActionListener{
    
    /** Options to display to the user. */
    protected Object[] options = {"Valider", "Annuler"};
    /** String action command for add a new sub tab. */
    private final static String NEW_SUB_TAB = "0";
    /** String action command for delete a sub tab. */
    private final static String DELETE_SUB_TAB = "1";
    /** String action command for rename a tab. */
    private final static String RENAME = "2";
    /** String action command for duplicate a tab. */
    private final static String DUPLICATE = "3";
    /** String action command for a new selection. */
    private final static String NEW_SELECTION = "4";
    /** String action command for rename a subtab. */
    private final static String RENAME_SUB_TAB = "5";
    /** String action command for delete a tab. */
    private final static String DELETE_TAB = "6";
    /** String action command for display informations. */
    private final static String DISPLAY = "7";
    
    private SelectionDataStorage selectionDataStorage = null;
    
    /** Value for PopupMenu options. */
    protected int popupMenuMode = 0;
    /** PopupMenu allows close operation only */
    static int CLOSE_ONLY = 0;
    /** PopupMenu allows all options */
    static int ALL_OPTIONS = 1;
    
    /* imageIcon for delete menu */
    private ImageIcon deleteButtonIcon = Util.loadIcon("delete.png");
    
    /**
     * Creates an empty TabbedPane with a default tab placement of JTabbedPane.TOP
     * and a default layout JTabbedPane.SCROLL_TAB_LAYOUT
     * popupMenuMode is set to default (CustomTabbedPane.CLOSE_ONLY).
     * @see JTabbedPane
     */
    public CustomTabbedPane(SelectionDataStorage selectionDataStorage) {
        super(JTabbedPane.TOP,JTabbedPane.SCROLL_TAB_LAYOUT);
        popupMenuMode = CLOSE_ONLY;
        this.selectionDataStorage=selectionDataStorage;
    }
    
    /**
     * Creates an empty TabbedPane with a default tab placement, a default layout and a specified
     * tab edit mode value which may be either CustomTabbedPane.CLOSE_ONLY or CustomTabbedPane.ALL_OPTIONS.
     * @param mode Tab edit mode value
     * @see #CustomTabbedPane()
     */
    public CustomTabbedPane(int mode, SelectionDataStorage selectionDataStorage) {
        this(selectionDataStorage);
        setMode(mode);
    }
    
    /**
     * Creates an empty TabbedPane with the specified tab placement and tab layout policy.
     * Tab placement may be either: JTabbedPane.TOP, JTabbedPane.BOTTOM, JTabbedPane.LEFT,
     * or JTabbedPane.RIGHT. Tab layout policy may be either: JTabbedPane.WRAP_TAB_LAYOUT
     * or JTabbedPane.SCROLL_TAB_LAYOUT.
     * @param tabPlacement the placement for the tabs relative to the content
     * @param layoutPolicy the policy for laying out tabs when all tabs will not fit on one run
     * @see JTabbedPane
     */
    public CustomTabbedPane(int tabPlacement, int layoutPolicy, int mode,SelectionDataStorage selectionDataStorage) {
        super(tabPlacement, layoutPolicy);
        setMode(mode);
        this.selectionDataStorage=selectionDataStorage;
    }
    
    public void setMode(int mode) {
        if (isValidMode(mode)) popupMenuMode = mode;
        else throw new IllegalArgumentException("Invalid mode : "+ mode +"\n mode may be either CustomTabbedPane.CLOSE_ONLY or CustomTabbedPane.ALL_OPTIONS.");
    }
    
    public boolean isValidMode(int mode) {
        if (mode==CLOSE_ONLY || mode==ALL_OPTIONS) return true;
        else return false;
    }
    
    /**
     * Processes mouse events occurring on this component by dispatching
     * them to any registered {@link MouseListener} objects.
     * @param evt The mouse event
     */
    protected void processMouseEvent(MouseEvent evt) {
        if (evt.getID() == MouseEvent.MOUSE_PRESSED && evt.getButton() == MouseEvent.BUTTON3) {
            int index = indexAtLocation(evt.getX(), evt.getY());
            if (getTabPopup(index)!=null)
                getTabPopup(index).show(this, evt.getX(), evt.getY());
        } else {
            super.processMouseEvent(evt);
        }
    }
    
    /**
     * Returns a new popup menu for the tab corresponds to the specified index.
     * @param index - the index of the item being queried
     * @return a new {@link JPopupMenu} object
     */
    private JPopupMenu getTabPopup(int index) {
        if (index >= 0 && index <= getTabCount()) // prevent bad click with an index out of range (-1)
        {
            super.setSelectedIndex(index); // select the clicked tab
            String title = getTitleAt(index); // get the tab title
            JPopupMenu pm = new JPopupMenu(); // construct a new JPopupMenu
            
            JMenuItem addNewSubTab = new JMenuItem("Ajouter une sous-zone");
            addNewSubTab.setHorizontalTextPosition(AbstractButton.LEADING);
            addNewSubTab.setActionCommand(NEW_SUB_TAB);
            addNewSubTab.addActionListener(this);
            
            JMenuItem rename = new JMenuItem("Renommer " + title);
            rename.setActionCommand(RENAME);
            rename.setHorizontalTextPosition(AbstractButton.LEADING);
            rename.addActionListener(this);
            
            JMenuItem renametab = new JMenuItem("Renommer " + title);
            renametab.setActionCommand(RENAME_SUB_TAB);
            renametab.setHorizontalTextPosition(AbstractButton.LEADING);
            renametab.addActionListener(this);
            
            JMenuItem duplicate = new JMenuItem("Dupliquer " + title);
            duplicate.setActionCommand(DUPLICATE);
            duplicate.setHorizontalTextPosition(AbstractButton.LEADING);
            duplicate.addActionListener(this);
            
            JMenuItem new_selection = new JMenuItem("Créer une nouvelle zone de chalandise");
            new_selection.setActionCommand(NEW_SELECTION);
            new_selection.addActionListener(this);
            
            JMenuItem affichage_informations = new JMenuItem("Afficher les informations de création de la zone de chalandise");
            affichage_informations.setActionCommand(DISPLAY);
            affichage_informations.addActionListener(this);
            
            
            JMenuItem delete = new JMenuItem("Supprimer " + title );
            delete.setHorizontalTextPosition(AbstractButton.LEADING);
            delete.setActionCommand(DELETE_SUB_TAB);
            delete.addActionListener(this);
            
            JMenuItem delete_tab = new JMenuItem("Supprimer " + title );
            delete_tab.setHorizontalTextPosition(AbstractButton.LEADING);
            delete_tab.setActionCommand(DELETE_TAB);
            delete_tab.addActionListener(this);
            
            if (popupMenuMode==ALL_OPTIONS) {
                
                pm.add(rename);
                pm.add(duplicate);
                pm.add(affichage_informations);
                // s'il n'y a pas d'objets d'informations, on grise le menu.
                if (selectionDataStorage.getCurrentSelectionInformation()==null) {
                    affichage_informations.setEnabled(false);
                }
                pm.add(new JPopupMenu.Separator());
                pm.add(addNewSubTab);
                pm.add(new JPopupMenu.Separator());
                pm.add(new_selection);
                pm.add(delete_tab);
                
            } else {
                pm.add(renametab);
                pm.add(new JPopupMenu.Separator());
                pm.add(delete);
            }
            return pm;
        } else return null;
    }
    
    /**
     * Invoked when an action occurs.
     * @param e the action event
     */
    public void actionPerformed(ActionEvent e) {
        if (e.getActionCommand().equals(DELETE_SUB_TAB))
            removeSubTab(getSelectedIndex(),getTitleAt(getSelectedIndex()));
        
        if (e.getActionCommand().equals(DELETE_TAB))
            removeTab(getSelectedIndex(),getTitleAt(getSelectedIndex()));
        
        if (e.getActionCommand().equals(NEW_SUB_TAB))
            newSubTab(getSelectedIndex());
        
        if (e.getActionCommand().equals(RENAME))
            renameTab(getSelectedIndex(),getTitleAt(getSelectedIndex()));
        
        if (e.getActionCommand().equals(DUPLICATE))
            duplicateTab(getSelectedIndex(),getTitleAt(getSelectedIndex()));
        
        if (e.getActionCommand().equals(NEW_SELECTION))
            newTab(getSelectedIndex());
        
        if (e.getActionCommand().equals(RENAME_SUB_TAB))
            renameSubTab(getSelectedIndex(),getTitleAt(getSelectedIndex()));
        
        if (e.getActionCommand().equals(DISPLAY))
            displayTab(getSelectedIndex(),getTitleAt(getSelectedIndex()));
    }
    
    /**
     * Removes the tab and component which corresponds to the specified index.
     * Prompts a warning dialog to confirm
     * @param index the index of the item being queried
     * @param title the title for the tab
     */
    private void removeSubTab(int index, String title) {
        int initsel = selectionDataStorage.getCurrentSelection();
        String msg = "Êtes-vous sûr de vouloir supprimer l'onglet "+title+" ?";
        if (JOptionPane.showConfirmDialog(new Frame(),msg,"Suppression...", JOptionPane.WARNING_MESSAGE,
                JOptionPane.OK_CANCEL_OPTION) == JOptionPane.OK_OPTION){
            selectionDataStorage.deleteSubSelection();
            setSelectedIndex(getTabCount()-1);
            ((CustomTabbedPane)getParent().getParent()).setSelectedIndex(initsel);
        }
    }
    
    /**
     * Removes the tab and component which corresponds to the specified index.
     * Prompts a warning dialog to confirm
     * @param index the index of the item being queried
     * @param title the title for the tab
     */
    private void removeTab(int index, String title) {
        //
        //int initsel = selectionDataStorage.getCurrentSelection();
        String msg = "Êtes-vous sûr de vouloir supprimer l'onglet "+title+" ?";
        if (JOptionPane.showConfirmDialog(new Frame(),msg,"Suppression...", JOptionPane.WARNING_MESSAGE,
                JOptionPane.OK_CANCEL_OPTION) == JOptionPane.OK_OPTION){
            selectionDataStorage.deleteCurrentSelection();
            setSelectedIndex(getTabCount()-1);
            //((CustomTabbedPane)getParent()).setSelectedIndex(getTabCount()-1);
        }
    }
    
    /**
     * Inserts a new sub tab for the tab corresponds to the specified index.
     * Prompts an input dialog to enter the name of the sub tab
     * @param index the index of the item being queried
     */
    private void newSubTab(int index) {
        String s = JOptionPane.showInputDialog("Entrer le nom de la nouvelle sous-zone puis valider", selectionDataStorage.getNextSubSelectionName());
        if (s!=null) selectionDataStorage.createSubSelection(s);
    }
    
    /**
     * Sets the title for the tab the tab corresponds to the specified index.
     * Prompts an input dialog to enter the new title.
     * @param index the index of the item being queried
     * @param title the current title for the tab
     */
    private void renameTab(int index, String title) {
        String s = JOptionPane.showInputDialog("Entrer le nouveau nom puis valider", title);
        if (s==null) { return; }
        if (s.trim().equals("")) {
            JOptionPane.showMessageDialog(this,"Nom invalide pour une zone de chalandise.","Erreur",JOptionPane.ERROR_MESSAGE);
            renameTab(index,"");
            return;
        }
        if (s!=null && !isTitleExists(s,index)) {
            setTitleAt(index,s.trim());
            selectionDataStorage.setCurrentSelectionName(s.trim());
            setSelectedIndex(index);
        } else {
            JOptionPane.showMessageDialog(this,"La zone de chalandise \""+s.trim()+"\" existe déjà !","Erreur",JOptionPane.ERROR_MESSAGE);
            renameTab(index,title);
        }
    }
    
    /**
     * Sets the title for the sub tab that matches the specified index.
     * Prompts an input dialog to enter the new title.
     * @param index the index of the item being queried
     * @param title the current title for the tab
     */
    private void renameSubTab(int index, String title) {
        String s = JOptionPane.showInputDialog("Entrer le nouveau nom puis valider", title);
        if (s!=null) {
            setTitleAt(index,s);
            selectionDataStorage.setCurrentSelectionSubName(index,s);
        }
    }
    
    /**
     * Display informations which have been used to build a selection
     * Prompts a dialog to display it.
     * @param index the index of the item being queried
     * @param title the current title for the tab
     */
    private void displayTab(int index, String title) {
        IsochroneDataObject datas = selectionDataStorage.getCurrentSelectionInformation();
        StringBuffer buffer = new StringBuffer("Latitude : ");
        buffer.append(datas.getOrigineLatitude());
        buffer.append("\n");
        buffer.append("Longitude : ");
        buffer.append(datas.getOrigineLongitude());
        buffer.append("\n");
        buffer.append("Autoroute : ");
        buffer.append(datas.getAutorouteSpeedCoef());
        buffer.append(" km/h \n");
        buffer.append("Nationale : ");
        buffer.append(datas.getNationaleSpeedCoef());
        buffer.append(" km/h \n");
        buffer.append("Départementale : ");
        buffer.append(datas.getDepartementaleSpeedCoef());
        buffer.append(" km/h \n");
        buffer.append("Route de liaison : ");
        buffer.append(datas.getLiaisonSpeedCoef());
        buffer.append(" km/h \n");
        
        int[] zones = datas.getZones();
        if (zones!=null) {
            for (int i=0;i<zones.length;i++){
                buffer.append("Zone ");
                buffer.append(i+1);
                buffer.append(" : ");
                buffer.append(zones[i]);
                buffer.append(" minutes \n");            
            }
        }
        JOptionPane.showMessageDialog(this,buffer.toString(),"Informations sur la génération de l'isochrone",JOptionPane.INFORMATION_MESSAGE);
    }
    
    /**
     * Duplicates the tab corresponds to the specified index.
     * @param index the index of the item being queried
     * @param title the current title for the tab
     */
    private void duplicateTab(int index, String title) {
        Component c = getComponentAt(index);
        String s = JOptionPane.showInputDialog("Entrer le nom de la nouvelle sélection puis valider", selectionDataStorage.getNextSelectionName());
        if (s==null) { setSelectedIndex(index); return; }
        if (s.trim().equals("")) {
            JOptionPane.showMessageDialog(this,"Nom invalide pour une zone de chalandise.","Erreur",JOptionPane.ERROR_MESSAGE);
            duplicateTab(index,"");
            return;
        }
        if (!isTitleExists(s,-1)) {
            selectionDataStorage.createSelectionCopy(index,s.trim());
            setSelectedIndex(getTabCount()-1);
        } else {
            JOptionPane.showMessageDialog(this,"La zone de chalandise \""+s.trim()+"\" existe déjà !","Erreur",JOptionPane.ERROR_MESSAGE);
            duplicateTab(index,title);
        }
    }
    
    /**
     * Add a new Selection
     * @param index the index of the item being queried
     * @param title the current title for the tab
     */
    private void newTab(int index) {
        String s = JOptionPane.showInputDialog("Entrer le nom de la nouvelle sélection puis valider", selectionDataStorage.getNextSelectionName());
        if (s==null) { return; }
        if (s.trim().equals("")) {
            JOptionPane.showMessageDialog(this,"Nom invalide pour une zone de chalandise.","Erreur",JOptionPane.ERROR_MESSAGE);
            newTab(index);
            return;
        }
        if (s!=null && !isTitleExists(s,-1)) {
            selectionDataStorage.createNewEmptySelection(s.trim());
            setSelectedIndex(getTabCount()-1);
        } else {
            JOptionPane.showMessageDialog(this,"La zone de chalandise \""+s.trim()+"\" existe déjà !","Erreur",JOptionPane.ERROR_MESSAGE);
            newTab(index);
        }
    }
    
    /**
     * Tests if a title allready exists.
     * @param title the title to test
     * @param ignoredIndex the index of the renamed tab or -1 for a new created tab
     * @return true if title match an existing tab title
     */
    public boolean isTitleExists(String title, int ignoredIndex) {
        for(int i = 0; i<this.getTabCount(); i++) {
            if (i!=ignoredIndex && getTitleAt(i)!=null && title.trim().toUpperCase().equals(getTitleAt(i).toUpperCase())) return true;
        }
        return false;
    }
}

