// <header>
/**
 *  ThemaMap : The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
package fr.unicaen.openmap.dataStorage.postgis.db;

import fr.unicaen.openmap.dataStorage.*;

import com.bbn.openmap.omGraphics.*;
import com.bbn.openmap.util.Debug;
import org.postgis.*;


/**
 * This class has capability to transform postgis geometry into
 * OpenMap objects.
 *
 * [postgis] => [openmap]
 * Point   => OMPoint
 * Line    => OMLine or OMPoly as a polyline with no fillColor.
 * MultiLine => OMGraphicList composed of OMLines (or OMPolys as a polyline with no fillColor).
 * Polygon => OMGraphicList (composed of OMPolys) because of the structure with LinearRings
 * MultiPolygon => OMGraphicList (composed of OMGraphicLists)
 *
 * Currently, i've got one bug : OMPoly's that shouldn't be closed
 * 'are' closed (rivers by example).
 * Workaround : for the moment, this kind of geometry will be replaced
 * by OMGraphicList of OMLines.
 */
public class Factory {
    
    /** Since all members are static, we won't construct an object. */
    private Factory() { }
    
    /**
     * Transform a Point to an OMPoint.
     * - 2007-04-25 : swap getX and getY as in lineStringToOMGraphicList.
     */
    public static OMPoint pointToOMPoint (Point point) {
	// Debug.message("factory", "[Factory] " + point.getX() + ", " + point.getY());
	return new OMPoint ((float) point.getY(), (float) point.getX());
    }
    
    /**
     *
     */
    public static OMGraphicList multiPointToOMGraphicList(MultiPoint multiPoint) {
        OMGraphicList oml = new OMGraphicList();
        for (int i = 0 ; i < multiPoint.numPoints() ; i++)
            oml.add(pointToOMPoint(multiPoint.getPoint(i)));
        return oml;
    }
    
    /**
     *
     */
    public static OMLine lineStringToOMLine(LineString lineString) {
        return new OMLine((float) lineString.getPoint(0).getX(), (float) lineString.getPoint(0).getY(),
			  (float) lineString.getPoint(1).getX(), (float) lineString.getPoint(1).getY(), OMLine.LINETYPE_STRAIGHT);
    }
    
    // Not used because OMPoly are always closed...
    //    /**
    //     * The bug (for rivers by example) may be located here but i don't have
    //     * much time to see what's going wrong, so take a look at
    //     */
    //    public static OMPoly lineStringToOMPoly (LineString lineString) {
    //	float [] points = new float[(2 * lineString.numPoints())];
    //	for (int i = 0 ; i < lineString.numPoints(); i++) {
    //	    points[2 * i] = (float) lineString.getPoint(i).getX();
    //	    points[2 * i + 1] = (float) lineString.getPoint(i).getY();
    //	}
    //	// OMPoly om = new OMPoly (points, OMPoly.DECIMAL_DEGREES, OMPoly.LINETYPE_UNKNOWN);
    //	OMPoly om = new OMPoly (points, OMPoly.DECIMAL_DEGREES, OMPoly.LINETYPE_STRAIGHT);
    //	om.setIsPolygon(false);
    //	//System.err.println(om.isPolygon());
    //	om.setRenderType(OMGraphic.RENDERTYPE_LATLON);
    //	return om;
    //    }
    
    /**
     * Don't call this method directly, use lineStringToOMGraphic instead !!!
     * Don't call this method when numPoints() < 1 !!!
     */
    public static OMGraphicList lineStringToOMGraphicList(LineString lineString) {
        if (lineString.numPoints() < 2) return null;
        float lat_begin, lon_begin, lat_end, lon_end;
        OMGraphicList om = new OMGraphicList();
        
        // add the first line.
        lat_begin = (float) lineString.getPoint(0).getY();
        lon_begin = (float) lineString.getPoint(0).getX();
        lat_end = (float) lineString.getPoint(1).getY();
        lon_end = (float) lineString.getPoint(1).getX();
        om.add(new OMLine(lat_begin, lon_begin, lat_end, lon_end, OMLine.LINETYPE_STRAIGHT));
        
        // for all the points, end becomes begin, get next end ... till the last point.
        for (int i = 2 ; i < lineString.numPoints(); i++) {
            lat_begin = lat_end;
            lon_begin = lon_end;
            lat_end = (float) lineString.getPoint(i).getY(); // should be getX();
            lon_end = (float) lineString.getPoint(i).getX(); // should be getY();
            om.add(new OMLine(lat_begin, lon_begin, lat_end, lon_end, OMLine.LINETYPE_STRAIGHT));
        }
        om.setRenderType(OMGraphic.RENDERTYPE_LATLON);
        return om;
    }
    
    /**
     *
     */
    public static OMGraphic lineStringToOMGraphic(LineString lineString) {
        return lineStringToOMGraphicList(lineString);
        //	// OMLine is not a good object !
        //	if (lineString.numPoints() == 2)
        //	    return lineStringToOMLine(lineString);
        //	else {
        //	    // more than 2 points in the line :
        //	    // BUG: we won't use this method for the moment ...
        //	    // return lineStringToOMPoly(lineString);
        //	    return lineStringToOMGraphicList(lineString);
        //	}
    }
    
    /**
     *
     */
    public static OMGraphicList multiLineStringToOMGraphicList(MultiLineString multiLineString) {
        OMGraphicList oml = new OMGraphicList();
        for (int i = 0 ; i < multiLineString.numLines() ; i++)
            oml.add(lineStringToOMGraphic(multiLineString.getLine(i)));
        return oml;
    }
    
    /**
     *
     */
    public static OMPoly linearRingToOMPoly(LinearRing ring) {
        double [] points = new double[(2 * ring.numPoints())];
        for (int i = 0 ; i < ring.numPoints() ; i++) {
            points[2 * i] = ring.getPoint(i).getY();
            points[2 * i + 1] =  ring.getPoint(i).getX();
        }
        OMPoly om = new OMPoly(points, OMPoly.DECIMAL_DEGREES, OMPoly.LINETYPE_STRAIGHT);
        //	OMPoly om = new OMPoly (points, OMPoly.DECIMAL_DEGREES, OMPoly.LINETYPE_UNKNOWN);
        om.setIsPolygon(true);
        return om;
    }
    
    /**
     *
     */
    public static OMGraphicList polygonToOMGraphicList(Polygon polygon) {
        OMGraphicList oml = new OMGraphicList();
        for (int i = 0 ; i < polygon.numRings() ; i++)
            oml.add(linearRingToOMPoly(polygon.getRing(i)));
        return oml;
    }
    
    /**
     *
     */
    //   public static OMGraphic multiPolygonToOMGraphicList(MultiPolygon multiPolygon) {
    public static OMGraphicList multiPolygonToOMGraphicList(MultiPolygon multiPolygon) {
	OMGraphicList oml = new OMGraphicList();
        for (int i = 0 ; i < multiPolygon.numPolygons() ; i++)
            oml.add(polygonToOMGraphicList(multiPolygon.getPolygon(i)));
        return oml;
 
        
        
	//      return linearRingToOMPoly(multiPolygon.getPolygon(0).getRing(0));
    }
    
    /**
     *
     */
    public static OMGraphic geometryToOMGraphic(Geometry g) {
        switch (g.getType()) {
	case Geometry.POINT:
	    return pointToOMPoint((Point) g);
	case Geometry.MULTIPOINT:
	    return multiPointToOMGraphicList((MultiPoint) g);
	case Geometry.LINESTRING:
	    return lineStringToOMGraphic((LineString) g);
	case Geometry.MULTILINESTRING:
	    return multiLineStringToOMGraphicList((MultiLineString) g);
	case Geometry.POLYGON:
	    return polygonToOMGraphicList((Polygon)g);
	case Geometry.MULTIPOLYGON:
	    return multiPolygonToOMGraphicList((MultiPolygon) g);
	case Geometry.GEOMETRYCOLLECTION:
	    // Not supported ... for the moment ...
	    return null;
	default:
	    // is there a problem ?
	    return null;
        }
    }
    
    /**
     * This method creates and returns an OMGraphic corresponding
     * to a string. The string is passed to a PGgeometry object
     * for construction and parsing purpose, then data and type are
     * deducted from the underlying Geometry
     * If an srid information is located at the begining of the
     * string, it is simply ignored.
     * @param pgString the string to get OMGraphic from.
     * @return An OMGraphic or null if an error occurs.
     */
    public static OMGraphic createOMGraphic (String pgString) throws java.sql.SQLException {
	// what about srid ?
	int index = pgString.indexOf(";");
	if (index != -1) {
	    pgString = pgString.substring(index + 1, pgString.length());
	}
	PGgeometry pgGeom = new PGgeometry (pgString);
	//Debug.message("factory", "[Factory] : createOMGraphic() : \n" + pgString.trim());
	
	return geometryToOMGraphic(pgGeom.getGeometry());
    }
    
    /**
     * This method creates and returns a Centroid corresponding to a
     * string. The string is passed to a PGgeometry object for
     * construction and parsing purpose, then data and type are
     * deducted from the underlying Geometry If an srid information is
     * located at the begining of the string, it is simply ignored.
     * @param pgString the string to get Centroid from.
     * @return A Centroid or null if an error occurs.  */
    public static Centroid createCentroid (String pgString) throws java.sql.SQLException {
	// what about srid ?
	int index = pgString.indexOf(";");
	if (index != -1) {
	    pgString = pgString.substring(index + 1, pgString.length());
	}

	//Debug.message("factory", "[Factory] : createCentroid() : \n" + pgString);
	
	PGgeometry pgGeom = new PGgeometry (pgString);
	return new Centroid(((Point) pgGeom.getGeometry()).getX(), ((Point) pgGeom.getGeometry()).getY());
    }
    
    /**
     * This method creates and returns a BoundingBox corresponding to a
     * string. The string is passed to a PGgeometry object for
     * construction and parsing purpose, then data and type are
     * deducted from the underlying Geometry If an srid information is
     * located at the begining of the string, it is simply ignored.
     * @param pg the string to get BoundingBox from.
     * @return A BoundingBox or null if an error occurs.  */
    public static BoundingBox createBoundingBox (String pg) /* throws java.sql.SQLException */ {
	// what about srid ?
	int index = pg.indexOf(";");
	if (index != -1) {
	    pg = pg.substring(index + 1, pg.length());
	}

	//	Debug.message("factorybb", "[Factory0] : createBoundingBox() : " + pg);
	
	pg = pg.trim(); // trim the string as our parser in not very robust...
	double tlon= 0d;
	double tlat= 0d;
	double blon= 0d;
	double blat= 0d;
	String[] tokens = null;
	try {
	    tokens = pg.split("[, \t\r\n()]+");
	    if (! tokens[0].equals("BOX3D")) // equalsIgnoreCase?
		throw(new Exception("createBoundingBox: not BOX3D found"));
	    tlon = Double.parseDouble(tokens[1]);
	    tlat = Double.parseDouble(tokens[2]);
	    blon = Double.parseDouble(tokens[4]);
	    blat = Double.parseDouble(tokens[5]);
	} catch (Exception e) {
	    StringBuilder buf = new StringBuilder("[Factory] Exception, ");
	    buf.append(e.getMessage() + " '" + pg + "'");
	    for (int i = 0; i < tokens.length; i++)
		buf.append("["+i+"]"+"'"+tokens[i]+"' ");
	    Debug.error(buf.toString());
	}
	// Debug.message("factory", "[Factory] : createBoundingBox() : " + pg + " -> " + tlon+" "+tlat+" "+blon+" "+blat);

	//return new BoundingBox(tlon, tlat, blon, blat);
        return new BoundingBox(tlon, blat, blon, tlat);
    }
}


