// <header>
/**
 *  ThemaMap : The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
package fr.unicaen.openmap.dataStorage.postgis.gui;

import java.awt.*;
import java.awt.event.*;
import java.lang.reflect.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.text.*;

/**
 * Take care of the string parsing behaviour : it's expensive and
 * probably not optimized : all data get parsed for strings when :
 *   - a single '\' is entered or removed before a '
 *   - a single ' is entered or removed but no '\' before it.
 *   - a group of chars is added or removed with one or more '
 *     or '\' inside.
 *
 * Todo : i need to verify code for exceptions to \ and ' tests ...
 * i think some code is used but shouldn't => slows the program down.
 *
 * Found : bug : paste inside a string.
 * optimize : if ' or \ is removed, only reparse next chars for strings.
 * optimize : in cuted strings, if there is an odd number of ' (execpt from \')
 * then we do not need to reparse anything execpt keywords.
 */
public class HighlightedDocument extends DefaultStyledDocument {
    
    private JTextPane textPane = null;
    private boolean inString = false;
    private boolean lastIsDQ = false;
    private boolean lastIsBS = false;

    private static String WHITE_SPACES = " \t\r\n";
    private static String SEPARATORS   = "\"'{([|`)]}=+-*/,;.:!%$";
    protected static String ALL        = WHITE_SPACES + SEPARATORS;

    protected static char charDelim = '\'';
    protected static String stringDelim = "'";
    
    protected static Hashtable<String,Style> highlihtedItems;
    
    protected static Style STYLE_NORMAL;
    protected static Style STYLE_KEYWORD;
    protected static Style STYLE_FUNCTION;
    protected static Style STYLE_STRING;
    protected static Style STYLE_OPERATOR;
    protected static Style STYLE_SEPARATOR;
    
    public HighlightedDocument (JTextPane textPane) {
	this.textPane = textPane;
	initStyles();
    }
    
    private void initStyles() {
	STYLE_NORMAL = addStyle("STYLE_NORMAL", StyleContext.getDefaultStyleContext().getStyle(StyleContext.DEFAULT_STYLE));
	StyleConstants.setFontFamily(STYLE_NORMAL, "SansSerif");
	StyleConstants.setBold(STYLE_KEYWORD = addStyle("STYLE_KEYWORD", STYLE_NORMAL), true);
	StyleConstants.setForeground(STYLE_KEYWORD, new Color(65, 105, 225));
	StyleConstants.setBold(STYLE_FUNCTION = addStyle("STYLE_FUNCTION", STYLE_NORMAL), true);
	StyleConstants.setForeground(STYLE_FUNCTION, new Color(0, 0, 255));
	StyleConstants.setForeground(STYLE_STRING = addStyle("STYLE_STRING", STYLE_NORMAL), new Color(160, 82, 45));
	// StyleConstants.setItalic(STYLE_STRING, true);
	StyleConstants.setBold(STYLE_OPERATOR = addStyle("STYLE_OPERATOR", STYLE_NORMAL), true);
	StyleConstants.setForeground(STYLE_OPERATOR, new Color(85, 107, 47));
	StyleConstants.setBold(STYLE_SEPARATOR = addStyle("STYLE_SEPARATOR", STYLE_NORMAL), true);
	StyleConstants.setForeground(STYLE_SEPARATOR, new Color(95, 158, 160));
    }
    
    private AttributeSet getHighlightedStyle(String s) {
	return (AttributeSet) highlihtedItems.get(s.toLowerCase());
    }
    
    private String begin = "";
    private String end = "";
    
    public void insertString(int offs, String str, AttributeSet a) throws BadLocationException {
	updateData(offs);
	if (str.length() == 1) insertChar(offs, str.charAt(0));
	// not a single char => one or more words.
	else insertWords(offs, str);
    }
    
    /**
     * When single char is entered.
     */
    private void insertChar(int offs, char c) throws BadLocationException {
	AttributeSet s;
	if (c == charDelim) {
	    super.insertString(offs, "" + c, STYLE_STRING);
	    // System.err.println("inString ? " + inString);
	    inString = (!inString) || lastIsBS;
	    // System.err.println("inString ? " + inString);
	    lastIsDQ = true;
	    if (!lastIsBS) reparseStrings(offs);
	    lastIsBS = false;
	    return;
	}
	if (c == '\\') {
	    super.insertString(offs, "" + c, (inString ? STYLE_STRING : STYLE_NORMAL));
	    lastIsBS = true;
	    if(isDQ(offs + 1)) reparseStrings(offs);
	    return;
	}
	if (inString) {
	    super.insertString(offs, "" + c, STYLE_STRING);
	} else {
	    super.insertString(offs, "" + c, null);
	    reparse(offs - begin.length(), offs + end.length() + 1);
	}
	lastIsDQ = false;
	lastIsBS = false;
    }
    
    /**
     * Insert one or more words. This method will only be
     * called internally, with string.length() > 1 as in other
     * case the code for a single char applies => count >= 1.
     */
    private void insertWords(int offs, String str) throws BadLocationException {
	super.insertString(offs, str, null);
	// reparse for strings if necessary.
	if (str.indexOf(stringDelim) != - 1 || str.indexOf("\\") != - 1)
	    reparseStrings(offs);
	else {
	    reparse(offs - begin.length(), offs + str.length() + end.length());
	}
    }
    
    public void remove(int offs, int len) throws BadLocationException {
	updateData(offs);
	AttributeSet s;
	boolean stringParsing = isDQ(offs) || (isBS(offs) && isDQ(offs + 1))
	    || textPane.getText().substring(offs, offs + len).indexOf(stringDelim) != - 1
	    || textPane.getText().substring(offs, offs + len).indexOf("\\") != - 1;
	
	super.remove(offs, len);
	
	synchronized(this) {
	    begin = getLastWordAt(offs);
	    end = getNextWordAt(offs);
	}
	
	if (stringParsing) {
	    reparseStrings(offs);
	    return;
	}
	
	if (inString) {
	    super.setCharacterAttributes(offs - begin.length(), begin.length(),STYLE_STRING, true);
	    super.setCharacterAttributes(offs, end.length(), STYLE_STRING, true);
	} else if ((s = getHighlightedStyle(begin + end)) != null) {
	    super.setCharacterAttributes(offs - begin.length(), begin.length(), s, true);
	    super.setCharacterAttributes(offs, end.length(), s, true);
	} else if ((s = getHighlightedStyle(begin)) != null && Character.isWhitespace(end.charAt(0))) {
	    super.setCharacterAttributes(offs - begin.length(), begin.length(), s, true);
	} else if ((s = getHighlightedStyle(end)) != null && Character.isWhitespace(begin.charAt(0))) {
	    super.setCharacterAttributes(offs, end.length(), s, true);
	} else {
	    super.setCharacterAttributes(offs - begin.length(), begin.length(),STYLE_NORMAL, true);
	    super.setCharacterAttributes(offs, end.length(), STYLE_NORMAL, true);
	}
    }
    
    private void reparseStrings(int offset) {
	// find all single ' (without \).
	ArrayList<Integer> indexes = new ArrayList<Integer>(15);
	int offs = -1;
	String text = textPane.getText();
	text = text.substring(offset);
	boolean hasBS = false;
	while ((offs = text.indexOf(charDelim, offs+1)) != -1) {
	    try { hasBS = (text.charAt(offs - 1) == '\\');
	    } catch (Exception e) { hasBS = false; }
	    if (!hasBS) {
		indexes.add(new Integer(offs));
	    }
	}
	// then hilight string data and reparse non string data.
	boolean string = false;
	int begin = offset;
	int end = offset;
	Iterator iter = indexes.iterator();
	while (iter.hasNext()) {
	    end = ((Integer) iter.next()).intValue();
	    if (string) super.setCharacterAttributes(begin, end - begin, STYLE_STRING, true);
	    else reparse(begin, end);
	    string = !string;
	    begin = end;
	}
	end = text.length();
	if (string) super.setCharacterAttributes(begin, end - begin, STYLE_STRING, true);
	else reparse(begin, end);
    }

    private void reparse(int begin, int end) {
	StringTokenizer st;
	try { st = new StringTokenizer(textPane.getText().substring(begin, end), ALL, true); }
	catch (Exception e) { return; }
	String token = "";
	while (st.hasMoreTokens()) {
	    token = st.nextToken();
	    AttributeSet s = getHighlightedStyle(token);
	    super.setCharacterAttributes(begin, token.length(), (s != null) ? s : STYLE_NORMAL, true);
	    begin += token.length();
	}
    }
    
    private String getNextWordAt(int index) {
	String text = textPane.getText().substring(index);
	StringTokenizer st = new StringTokenizer(text, HighlightedDocument.ALL, true);
	if (st.hasMoreElements())
	    return st.nextToken();
	else return text;
    }
    
    private String getLastWordAt(int index) {
	String text = textPane.getText().substring(0, index);
	StringTokenizer st = new StringTokenizer(text, HighlightedDocument.ALL, true);
	String result = text;
	while (st.hasMoreElements()) result = st.nextToken();
	return result;
    }
    
    private boolean isBS(int index) {
	try {
	    return (textPane.getText().charAt(index) == '\\');
	} catch (Exception e) { return false; }
    }
    
    private boolean isDQ(int index) {
	try {
	    return (textPane.getText().charAt(index) == charDelim);
	} catch (Exception e) { return false; }
    }
    
    private int stringCount(String s, String p) {
	int index = s.indexOf(p);
	if (index == -1) return 0;
	return 1 + stringCount(s.substring(index + p.length()), p);
    }
    
    private boolean inString(int index) {
	String text = textPane.getText().substring(0, index);
	return !(((stringCount(text, stringDelim) - stringCount(text, "\\" + stringDelim)) & 0x000001) == 0);
    }
    
    public void updateData(int offs) {
	synchronized(this) {
	    begin = getLastWordAt(offs);
	    end = getNextWordAt(offs);
	    inString = inString(offs);
	    lastIsDQ = isDQ(offs);
	    lastIsBS = isBS(offs);
	}
    }
}
