// <header>
/**
 *  ThemaMap : The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
/*
 * ShapeFileDataStorage.java
 *
 * Created on march 2006
 *
 * @author Jacques.Madelaine@info.unicaen.fr
 *
 */
package fr.unicaen.openmap.dataStorage.shapefile;

import fr.unicaen.openmap.dataStorage.*;

import com.bbn.openmap.util.PropUtils;
import com.bbn.openmap.util.Debug;
import com.bbn.openmap.util.propertyEditor.Inspector;
import com.bbn.openmap.omGraphics.DrawingAttributes;
import com.bbn.openmap.layer.shape.ESRIBoundingBox;
import com.bbn.openmap.layer.shape.ESRIRecord;
import com.bbn.openmap.layer.shape.ShapeFile;
import com.bbn.openmap.proj.coords.LatLonPoint;
import com.bbn.openmap.omGraphics.OMGraphicList;
import com.bbn.openmap.dataAccess.shape.input.DbfInputStream;
import com.bbn.openmap.io.FormatException;
import com.bbn.openmap.layer.shape.SpatialIndex;
import com.bbn.openmap.util.PropertyStringFormatException;
import fr.unicaen.openmap.dataStorage.gui.DataTableModel;
import fr.unicaen.openmap.dataStorage.gui.DataTableStoragePanel;
import fr.unicaen.openmap.event.GUIChangeEvent;

import fr.unicaen.openmap.util.Util;
import fr.unicaen.openmap.util.propertyEditor.DataStoragePropertyEditor;
import fr.unicaen.openmap.util.propertyEditor.FileTypePropertyEditor;

import fr.unicaen.openmap.event.GUIChangeListener;
import fr.unicaen.openmap.gui.FileNamePanel;


import javax.swing.*;
import java.awt.event.*;
import java.awt.*;
import java.net.MalformedURLException;
import java.util.Iterator;
import java.util.ArrayList;
import java.util.Set;
import java.util.Properties;
import java.net.URL;

/**
 * DataStorage getting its data from a shape file and a dbf companion file
 * <p>
 * Exemple of properties :
 * <pre>
 * exemple.class=fr.unicaen.openmap.layer.dataDisplayLayer.GeometryDataDisplayLayer
 * exemple.prettyName= Exemple
 * exemple.lineColor=FF000000
 * exemple.fillColor=FFFFFFFF
 * exemple.selectedLineColor=FF990099
 * exemple.selectedFillColor=C900FF55
 * exemple.fill=no
 * exemple.geometryColumnName=shape
 * exemple.projectionChangePolicy=pcp
 * exemple.removable=true
 * exemple.renderPolicy=rp
 * exemple.rp.class=com.bbn.openmap.layer.policy.BufferedImageRenderPolicy
 * # properties specific to the DataStorage
 * exemple.dataStorageClass=fr.unicaen.openmap.dataStorage.shapefile.ShapeFileDataStorage
 * exemple.shp=/home/jacques/shape/frdptsdd.shp
 * exemple.dbf=/home/jacques/shape/frdptsdd.dbf
 * </pre>
 *
 * The default name for the geometry in the DataStorage is geometryColumnName.
 * This name followed with a '_' is used to name the geometry id column (shape number). 
 *
 * @author Jacques.Madelaine@info.unicaen.fr
 *
 **/
public class ShapeFileDataStorage extends InformationDataStorage {// implements GUIChangeListener {

    protected DrawingAttributes drawingAttributes = new DrawingAttributes();
    public static final java.lang.String PARAM_SHP = "shp";
    public static final java.lang.String PARAM_DBF = "dbf";
    public static final java.lang.String PARAM_GEOMETRYCOLUMNNAME = "geometryColumnName";
    public static final java.lang.String PARAM_MAXNUMBEROFRECORDS = "maxNumberOfRecords";
    public static final java.lang.String PARAM_USEDBFWITHSAMENAME = "useDbfWithSameName";
    /** used to pack all the properties in one property for the SharedDataStoragePropertyEditor */
    public final static String GLOBAL = "shapefileDS.global";
    /**
     * Name of the shape file
     */
    protected String shpFile = "";
    /**
     * Name of the dbf file
     */
    protected String dbfFile = "";
    protected String geometryColumnName = "shape";
    protected int maxNumberOfRecords = -1;
    protected boolean useDbfWithSameName = false;

    public ShapeFileDataStorage(DataStorageListener dsl) {
        super(dsl);
    }
    protected boolean initialized = false;

    public String getName() {
        return "".equals(shpFile) ? name : stripFileNameWithClassPathDirectories(shpFile) + "/" + stripFileNameWithClassPathDirectories(dbfFile);
    }

    /**
     * Tells if the dataStorage has it's data.
     */
    public boolean isReady() {
        return initialized;
    }

    //Now, we can use PropertyEditor from getPropertyInfo() as GUI.
    public JComponent getGUI() {

        JPanel panel = new JPanel();
        panel.add(DataStoragePropertyEditor.provideCustomPanel(this));

        return panel;
        //return new ClosablePanel("ShapeFile: "+shpFile, panel);
    }

    /** hack pour CommercialDataDisplayLayer!!!*/
    public JTable getJTable() {
        return ((DataTableStoragePanel) dtsp).getJTable();
    }

    /**
     * most of this code is borrowed from com.bbn.openmap.plugin.esri.EsriLayer.setProperties
     */
    public void setProperties(String prefix, Properties properties) {
        super.setProperties(prefix, properties);
        prefix = PropUtils.getScopedPropertyPrefix(prefix);

        // From PropertyInfo
        Debug.message("DataStorage", "ShapeFileDataStorage setProperties from >" + properties.getProperty(prefix + GLOBAL) + "< " + fr.unicaen.openmap.util.propertyEditor.ShapeFileDataStoragePropertyEditor.propertySeparator + " " + fr.unicaen.openmap.util.propertyEditor.ShapeFileDataStoragePropertyEditor.keyValueSeparator);
        try { // if called via PropertyInfo, remove "global" from the keys
            Properties global = PropUtils.parsePropertyList(properties.getProperty(prefix + GLOBAL), fr.unicaen.openmap.util.propertyEditor.ShapeFileDataStoragePropertyEditor.propertySeparator, fr.unicaen.openmap.util.propertyEditor.ShapeFileDataStoragePropertyEditor.keyValueSeparator);
            Set<String> keys = global.stringPropertyNames();
            String key = "";
            for (Iterator<String> iterator = keys.iterator(); iterator.hasNext();) {
                key = iterator.next();
                properties.put(prefix + key, global.getProperty(key));
                properties.remove(key); //useless, but clean
            }
        } catch (PropertyStringFormatException exc) {
            System.err.println("Warning Parse error" + properties.getProperty(prefix + GLOBAL));
        }

        drawingAttributes.setProperties(prefix, properties);

        shpFile = properties.getProperty(prefix + PARAM_SHP, "cntry02.shp");

       // System.out.println("SHAPE_FILE = " + shpFile + ", PREFIX = " + prefix);
        //System.out.println("PRoperties = " + properties);
        
        dbfFile = properties.getProperty(prefix + PARAM_DBF, "cntry02.dbf");

        //	if ((dbfFile == null || dbfFile.equals("")) && shpFile != null)
        //	    dbfFile = shpFile.substring(0, shpFile.lastIndexOf('.') + 1) + PARAM_DBF;

        geometryColumnName = properties.getProperty(prefix + PARAM_GEOMETRYCOLUMNNAME, geometryColumnName);
        maxNumberOfRecords = com.bbn.openmap.util.PropUtils.intFromProperties(properties, prefix + PARAM_MAXNUMBEROFRECORDS, -1);

        initialize();
    }

    public String getShapeFile() {
        return shpFile;
    }

    public String getDbfFile() {
        return dbfFile;
    }

    public java.util.Properties getProperties(String prefix, Properties properties) {
        properties = super.getProperties(prefix, properties);
        prefix = PropUtils.getScopedPropertyPrefix(prefix);

        properties.put(prefix + PARAM_SHP, stripFileNameWithClassPathDirectories(shpFile));
        if (dbfFile != null && !"".equals(dbfFile)) {
            properties.put(prefix + PARAM_DBF, stripFileNameWithClassPathDirectories(dbfFile));
        }
        properties.put(prefix + PARAM_GEOMETRYCOLUMNNAME, geometryColumnName);

        //Management for getPropertyInfo()
        //Note : these properties can't be saved in a properties file but can be used during getPropertyInfo to load properties from a PropertyEditor.
        //Obsolete if we use default FilePropertyEditor but necessary with our FileNamePanel
        properties.put(
                prefix + GLOBAL,
                this);

        return properties;
    }

    /**
     * Provides the essential information in order to initialize this Data Storage.
     */
    public java.util.Properties getPropertyInfo(Properties info) {

        info = super.getPropertyInfo(info);

        /*
        //Using default FilePropertyEditor
        PropUtils.setI18NPropertyInfo(i18n,
        info,
        ShapeFileDataStorage.class,
        PARAM_SHP, //property
        PARAM_SHP, //ShapeFileDataStorage.shp in I18N*.properties *or* this default label
        PARAM_SHP, //ShapeFileDataStorage.shp.tooltip in I18N*.properties *or* this default tooltip
        "com.bbn.openmap.util.propertyEditor.FilePropertyEditor"
        );

        PropUtils.setI18NPropertyInfo(i18n,
        info,
        ShapeFileDataStorage.class,
        PARAM_DBF,
        PARAM_DBF,
        PARAM_DBF,
        "com.bbn.openmap.util.propertyEditor.FilePropertyEditor"
        //				      "fr.unicaen.openmap.util.propertyEditor.DBFFileTypePropertyEditor"
        );
         */

        //Group shpFile & dbfFile properties
        PropUtils.setI18NPropertyInfo(i18n,
                info,
                ShapeFileDataStorage.class,
                GLOBAL,
                "Data files", //GLOBAL,
                "Data files to used.", //GLOBAL,
                "fr.unicaen.openmap.util.propertyEditor.ShapeFileDataStoragePropertyEditor");

        //TO DO : custom editor with ComboBox & dsi data column names ?
        PropUtils.setI18NPropertyInfo(i18n,
                info,
                ShapeFileDataStorage.class,
                PARAM_GEOMETRYCOLUMNNAME,
                PARAM_GEOMETRYCOLUMNNAME,
                PARAM_GEOMETRYCOLUMNNAME,
                null //default editor : textfield
                );

        info.put(initPropertiesProperty,
                //		 getInitPropertiesPropertyString()
                GLOBAL + " " + getInitPropertiesPropertyString());

        return info;
    }

    //notes : init *just* for this class
    public String getInitPropertiesPropertyString() {
        return //	    PARAM_SHP + " " +
                //	    PARAM_DBF + " " +
                PARAM_GEOMETRYCOLUMNNAME;
    }

    protected void initialize() {
        Debug.message("DataStorage", "[ShapeFileDataStorage] using shapefile \"" + this.shpFile + "\"" + " and DBF file \"" + dbfFile + "\" => \"" + geometryColumnName + "\" geometry column name" + (maxNumberOfRecords > 0 ? (" limited to " + maxNumberOfRecords + " records") : ""));
	// openWithFailOver rewrite properties with a valid shape file name
	Util.openWithFailOver(this.shpFile, true, this);

        // @see XmlTableDataStorage and XmlTableParser...
        // There should be a way to share code between DbfDataStorage, but it is simply pasted here !

        ArrayList<String> columnNames = new ArrayList<String>(); // column names from the dbf file
        Iterator dbfIterator = null;
        if (dbfFile != null) {
            try {
                URL dbf = PropUtils.getResourceOrFileOrURL(dbfFile);
                DbfInputStream dis = new DbfInputStream(dbf.openStream());

                dbfIterator = dis.getRecords().iterator();

                String[] dbfNames = dis.getColumnNames();
                for (int i = 0; i < dbfNames.length; i++) {
                    columnNames.add(dbfNames[i]);
                }
                Debug.message("DataStorage", "[ShapeFileDataStorage] dbf column names: " + columnNames);
            } catch (java.io.IOException e) {
                Debug.error("ShapeFileDataStorage: IOException for " + dbfFile + " " + e.getMessage());
            } catch (java.lang.Exception e) {
                Debug.error("ShapeFileDataStorage: Exception for " + dbfFile + " " + e.getMessage());
            }
        }

        try {
            ArrayList<DataObject> dataObjects = new ArrayList<DataObject>();
            ArrayList<String> omGraphicNames = new ArrayList<String>();

            omGraphicNames.add(geometryColumnName);
            String geometryColumnIdName = geometryColumnName + "_";
            int numbrecord = 0;
            try{
		SpatialIndex spatialIndex = SpatialIndex.locateAndSetShapeData(shpFile);
		ESRIRecord [] esriRecords = spatialIndex.locateRecords(
								       Integer.MIN_VALUE,
								       Integer.MIN_VALUE,
								       Integer.MAX_VALUE,
								       Integer.MAX_VALUE);
		//// while ((esriRecord = shapeFile.getNextRecord()) != null) {
		for(ESRIRecord esriRecord : esriRecords){
		    //<limit to maxrecords>
		    numbrecord++;
		    if (maxNumberOfRecords > 0 && numbrecord > maxNumberOfRecords) {
			break;
		    }
		    //</limit to maxrecords>
		    OMDataLocation omdl = newOMDataLocation(esriRecord, dataObjects.size());

		    DataObject tempDataObject = new DataObject();
		    tempDataObject.addOMDataLocation(geometryColumnName, omdl);
		    // add the esri record number to tempDataObject
		    tempDataObject.addData(geometryColumnIdName, esriRecord.getRecordNumber());

		    // add data from the next dbf file record
		    // #### we should check for the equality of the esri record number and the dbfShapeNumberColumn value
		    if (dbfIterator != null && dbfIterator.hasNext()) {
			ArrayList record = (ArrayList) dbfIterator.next();
			Iterator<String> c = columnNames.iterator();
			for (Iterator i = record.iterator(); i.hasNext();) {
			    // add the attribute only if it is member of the attribute list
			    String cname = c.next();
			    Object obj = i.next();
			    if (isDataAttributeNameToLoad(cname)) {
				tempDataObject.addData(cname, obj);
			    }
			}
		    }
		    dataObjects.add(tempDataObject);
		    // Debug.message("DataStorage", "[ShapeFileDataStorage] adding " + tempDataObject);
		}
            }
            catch(/*MalformedURLException|FormatException|NullPointerException*/ Exception exc){
                System.err.println("No SSX file");
                ESRIRecord esriRecord;
                String newFileName = Util.locateFile(shpFile);
                ShapeFileReadOnly shapeFile = new ShapeFileReadOnly(newFileName);
                while ((esriRecord = shapeFile.getNextRecord()) != null){
		    if (maxNumberOfRecords > 0 && numbrecord > maxNumberOfRecords) {
			break;
		    }
		    //</limit to maxrecords>
		    OMDataLocation omdl = newOMDataLocation(esriRecord, dataObjects.size());

		    DataObject tempDataObject = new DataObject();
		    tempDataObject.addOMDataLocation(geometryColumnName, omdl);
		    // add the esri record number to tempDataObject
		    tempDataObject.addData(geometryColumnIdName, esriRecord.getRecordNumber());

		    // add data from the next dbf file record
		    // #### we should check for the equality of the esri record number and the dbfShapeNumberColumn value
		    if (dbfIterator != null && dbfIterator.hasNext()) {
			ArrayList record = (ArrayList) dbfIterator.next();
			Iterator<String> c = columnNames.iterator();
			for (Iterator i = record.iterator(); i.hasNext();) {
			    // add the attribute only if it is member of the attribute list
			    String cname = c.next();
			    Object obj = i.next();
			    if (isDataAttributeNameToLoad(cname)) {
				tempDataObject.addData(cname, obj);
			    }
			}
		    }
		    dataObjects.add(tempDataObject);

                }

            }
            dataTable.setTable(dataObjects, getDataAttributesNamesToDisplay(columnNames, geometryColumnIdName), omGraphicNames, getDataAttributesNamesAliases());
            // data are stored, we can create a dataTableModel and the panel to view them.
            // if they don't already exist
            if (dataTableModel == null) {
                dataTableModel = new DataTableModel(dataTable);
            }
            resetDataTableStoragePanel(dataTableModel);

            initialized = true;

        } catch (java.io.IOException e) {
            Debug.error("ShapeFileDataStorage: IOException for " + shpFile + " " + e.getMessage());
	    // e.printStackTrace();
            initialized = false;
            return;
        }
	
	// tell we are ready
        fireDataStorageEvent();
    }

    protected OMDataLocation newOMDataLocation(ESRIRecord esriRecord, int dataStorageIndex) {

        OMDataLocation omdl = new OMDataLocation();

        //bounding box
        ESRIBoundingBox esriBB = esriRecord.getBoundingBox();
        omdl.setBoundingBox(new BoundingBox(new LatLonPoint.Double(esriBB.max.y, esriBB.min.x),
                new LatLonPoint.Double(esriBB.min.y, esriBB.max.x)));

        //omgraphics
        OMGraphicList omgl = new OMGraphicList();

        esriRecord.addOMGraphics(omgl, drawingAttributes);

        // we put the index relative to the DataStorage inside the OMGraphic
        omgl.putAttribute(Util.OMGraphicIndex, dataStorageIndex);

        omdl.setOMGraphic(omgl);

        return omdl;
    }
}
