// <header>
/**
 *  ThemaMap : The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
/*
 * ShapeFile.java
 *
 * Created on june 2009
 *
 * @author Jacques.Madelaine@info.unicaen.fr
 *
 */
package fr.unicaen.openmap.dataStorage.shapefile;


import com.bbn.openmap.dataAccess.shape.ShapeUtils;

import com.bbn.openmap.layer.shape.ESRIBoundingBox;
import com.bbn.openmap.layer.shape.ESRIRecord;
import com.bbn.openmap.layer.shape.ESRIPointRecord;
import com.bbn.openmap.layer.shape.ESRIPolygonRecord;

import com.bbn.openmap.util.Debug;
import com.bbn.openmap.util.PropUtils;

import java.io.File;
import java.io.InputStream;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.net.URL;


/** Stupid copy/paste of from com.bbn.openmap.layer.shape.ShapeFile
 *  that reads only the file.
 *  It can also accept an URL for the file name.
 */

public class ShapeFileReadOnly {

    /** A Shape File's magic number. */
    public static final int SHAPE_FILE_CODE = 9994;

    /** The currently handled version of Shape Files. */
    public static final int SHAPE_FILE_VERSION = 1000;

    /** A default record size. Automatically increased on demand. */
    public static final int DEFAULT_RECORD_BUFFER_SIZE = 50000;

    /** The read class for shape files. */
    //protected InputStream raf;
    protected RandomAccessFile raf;

    /** The buffer that holds the 100 byte header. */
    protected byte header[];

    /** Holds the length of the file, in bytes. */
    protected long fileLength;

    /** Holds the version of the file, as an int. */
    protected int fileVersion;

    /** Holds the shape type of the file. */
    protected int fileShapeType;

    /** Holds the bounds of the file (four doubles). */
    protected ESRIBoundingBox fileBounds;

    /** A buffer for current record's header. */
    protected byte recHdr[];

    /** A buffer for the current record's data. */
    protected byte recBuf[];

    /**
     * Construct a <code>ShapeFileReadOnly</code> from a file name.
     * 
     * @exception IOException if something goes wrong opening or
     *            reading the file.
     */
    public ShapeFileReadOnly(String name) throws IOException {
	URL shp = PropUtils.getResourceOrFileOrURL(name);
//        if(shp == null){
//            Debug.error("Error loading " + name);
//        }
        //raf = shp.openStream();
        raf = new RandomAccessFile(name, "rw");
        recHdr = new byte[ShapeUtils.SHAPE_FILE_RECORD_HEADER_LENGTH];
        recBuf = new byte[DEFAULT_RECORD_BUFFER_SIZE];
        initHeader();
    }

    /**
     * Construct a <code>ShapeFileReadOnly</code> from the given
     * <code>File</code>.
     * 
     * @param file A file object representing an ESRI Shape File
     * 
     * @exception IOException if something goes wrong opening or
     *            reading the file.
     */
    public ShapeFileReadOnly(File file) throws IOException {
        this(file.getPath());
    }

    /**
     * Reads the header of a Shape file. 
     * <p>
     * After this function runs, the file pointer is set to byte 100,
     * the first byte of the first record in the file.
     * 
     * @exception IOException if something goes wrong reading
     *            the shape file
     */
    protected void initHeader() throws IOException {
        readHeader();
    }

    /**
     * Reads and parses the header of the file. Values from the header
     * are stored in the fields of this class.
     * 
     * @exception IOException if something goes wrong reading the file
     * @see #header
     * @see #fileVersion
     * @see #fileLength
     * @see #fileShapeType
     * @see #fileBounds
     */
    protected void readHeader() throws IOException {
        header = new byte[ShapeUtils.SHAPE_FILE_HEADER_LENGTH];
        raf.seek(0); // Make sure we're at the beginning of
        // the file
        raf.read(header, 0, ShapeUtils.SHAPE_FILE_HEADER_LENGTH);
        int fileCode = ShapeUtils.readBEInt(header, 0);
        if (fileCode != SHAPE_FILE_CODE) {
            throw new IOException("Invalid file code, "
                    + "probably not a shape file");
        }

        fileVersion = ShapeUtils.readLEInt(header, 28);
        if (fileVersion != SHAPE_FILE_VERSION) {
            throw new IOException("Unable to read shape files with version "
                    + fileVersion);
        }

        fileLength = ShapeUtils.readBEInt(header, 24);
        fileLength *= 2; // convert from 16-bit words to 8-bit bytes
        fileShapeType = ShapeUtils.readLEInt(header, 32);
        fileBounds = ShapeUtils.readBox(header, 36);
    }

    /**
     * Returns the length of the file in bytes.
     * 
     * @return the file length
     */
    public long getFileLength() {
        return fileLength;
    }

    /**
     * Returns the version of the file. The only currently supported
     * version is 1000 (which represents version 1).
     * 
     * @return the file version
     */
    public int getFileVersion() {
        return fileVersion;
    }

    /**
     * Returns the shape type of the file. Shape files do not mix
     * shape types; all the shapes are of the same type.
     * 
     * @return the file's shape type
     */
    public int getShapeType() {
        return fileShapeType;
    }

    /**
     * Returns the bounding box of this shape file. The bounding box
     * is the smallest rectangle that encloses all the shapes in the
     * file.
     * 
     * @return the bounding box
     */
    public ESRIBoundingBox getBoundingBox() {
        return fileBounds;
    }

    /**
     * Returns the next record from the shape file as an
     * <code>ESRIRecord</code>. Each successive call gets the next
     * record. There is no way to go back a record. When there are no
     * more records, <code>null</code> is returned.
     * 
     * @return a record, or null if there are no more records
     * @exception IOException if something goes wrong reading the file
     */
    public ESRIRecord getNextRecord() throws IOException {
        // Debug.output("getNextRecord: ptr = "+raf.getFilePointer());
        int result = raf.read(recHdr,
                0,
                ShapeUtils.SHAPE_FILE_RECORD_HEADER_LENGTH);
        if (result == -1) { // EOF
            // Debug.output("getNextRecord: EOF");
            return null;
        }

        int contentLength = ShapeUtils.readBEInt(recHdr, 4);
        int bytesToRead = contentLength * 2;
        int fullRecordSize = bytesToRead + 8;
        if (recBuf.length < fullRecordSize) {
            if (Debug.debugging("shape")) {
                Debug.output("record size: " + fullRecordSize);
            }
            recBuf = new byte[fullRecordSize];
        }
        System.arraycopy(recHdr,
                0,
                recBuf,
                0,
                ShapeUtils.SHAPE_FILE_RECORD_HEADER_LENGTH);
        raf.read(recBuf,
                ShapeUtils.SHAPE_FILE_RECORD_HEADER_LENGTH,
                bytesToRead);

        switch (fileShapeType) {

        case ShapeUtils.SHAPE_TYPE_NULL:
            throw new IOException("Can't parse NULL shape type");

        case ShapeUtils.SHAPE_TYPE_POINT:
            return new ESRIPointRecord(recBuf, 0);

        case ShapeUtils.SHAPE_TYPE_ARC:
            // case ShapeUtils.SHAPE_TYPE_POLYLINE:
            return new ESRIPolygonRecord(recBuf, 0);

        case ShapeUtils.SHAPE_TYPE_POLYGON:
            return new ESRIPolygonRecord(recBuf, 0);

        case ShapeUtils.SHAPE_TYPE_MULTIPOINT:
            throw new IOException("Multipoint shape not yet implemented");

        default:
            throw new IOException("Unknown shape type: " + fileShapeType);
        }

    }

    /**
     * Closes the shape file and disposes of resources.
     * 
     * @exception IOException if something goes wrong closing the file
     */
    public void close() throws IOException {
        raf.close();
        raf = null;
    }

} 
