// <header>
/**
 *  ThemaMap : The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
package fr.unicaen.openmap.jts;

import java.util.List;
import fr.unicaen.openmap.dataStorage.DataObject;
import fr.unicaen.openmap.dataStorage.OMDataLocation;
import fr.unicaen.openmap.dataStorage.BoundingBox;
import fr.unicaen.openmap.dataStorage.Centroid;
import fr.unicaen.openmap.dataStorage.expr.SpatialFunction;
import com.bbn.openmap.omGraphics.OMCircle;
import com.bbn.openmap.omGraphics.OMGraphicList;
import com.bbn.openmap.omGraphics.OMGraphic;
import com.bbn.openmap.omGraphics.OMPoint;
import com.bbn.openmap.omGraphics.OMLine;
import com.bbn.openmap.omGraphics.OMPoly;
import com.bbn.openmap.omGraphics.OMRect;
import com.bbn.openmap.proj.GreatCircle;
import com.bbn.openmap.proj.ProjMath;
import com.bbn.openmap.geo.Geo;


import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.GeometryCollection;
import com.vividsolutions.jts.geom.GeometryFactory;
import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.PrecisionModel;
import com.vividsolutions.jts.geom.LinearRing;
import com.vividsolutions.jts.geom.LineString;
import com.vividsolutions.jts.geom.Polygon;
import com.vividsolutions.jts.geom.Point;
import com.vividsolutions.jts.geom.Envelope;

import fr.unicaen.openmap.dataStorage.OMDataLocation;

/** 
 * JTS Geometry to OMGraphic
 *
 * Coordinates are suppose to be always in decimal degrees
 */
public class JTSFactory {
    static public GeometryFactory defaultFactory = 
	new GeometryFactory(new PrecisionModel(), 4326);

    static public Geometry LonLatToJTSGeometry(double lon, double lat) {
	return defaultFactory.createPoint(new Coordinate(lon, lat));
    }

    static public Geometry LonLatToJTSGeometry(double lon, double lat, GeometryFactory factory) {
	return factory.createPoint(new Coordinate(lon, lat));
    }
    
    /** From OpenMap structures to JTS structures */
    static public Geometry OMGraphicToJTSGeometry(OMGraphic omg) {	
	return OMGraphicToJTSGeometry(omg, defaultFactory);
    }

    static public Geometry OMGraphicToJTSGeometry(OMGraphic omg, GeometryFactory factory) {
	if (omg instanceof OMGraphicList) 
	    return OMGraphicToJTSGeometry((OMGraphicList)omg, factory);

	// OMGraphicList is of RENDERTYPE_UNKNOWN so we test for it only now.
	if (omg.getRenderType() != OMGraphic.RENDERTYPE_LATLON) {
	    String errmsg = "fr.unicaen.openmap.jts.JTSFactory.OMGraphicToJTSGeometry: support only for LAT/LON Render Type "+omg.getRenderType() + " " + omg;
	    assert false : errmsg;
	    System.err.println(errmsg);
	    return null;
	}

	if (omg instanceof OMPoint)
	    return OMGraphicToJTSGeometry((OMPoint)omg, factory);
	if (omg instanceof OMLine)
	    return OMGraphicToJTSGeometry((OMLine)omg, factory);
	if (omg instanceof OMPoly)
	    return OMGraphicToJTSGeometry((OMPoly)omg, factory);
	if (omg instanceof OMRect)
	    return OMGraphicToJTSGeometry((OMRect)omg, factory);
	if (omg instanceof OMCircle)
	    return OMGraphicToJTSGeometry((OMCircle)omg, factory);
	    
	String errmsg = "fr.unicaen.openmap.jts.JTSFactory.OMGraphicToJTSGeometry: unsupported class for "+omg;
	assert false : errmsg;
	System.err.println(errmsg);
	return null;
    }

	static public Geometry OMGraphicToJTSGeometry(OMGraphicList omglist, GeometryFactory factory) {
	int size = omglist.size();
	Geometry geometries[] = new Geometry[size];
	for(int i = 0; i < size; i++)
	    geometries[i] = OMGraphicToJTSGeometry(omglist.get(i));
	return factory.createGeometryCollection(geometries);
    }

    static public Geometry OMGraphicToJTSGeometry(OMPoint omp, GeometryFactory factory) {
	return factory.createPoint(new Coordinate(omp.getLon(), omp.getLat()));
    }

    static public Geometry OMGraphicToJTSGeometry(OMLine oml, GeometryFactory factory) {
	return factory.createLineString(extractCoordinates(oml));
    }

    static public Geometry OMGraphicToJTSGeometry(OMPoly omp, GeometryFactory factory) {
	Coordinate[] coords = extractCoordinates(omp);
	if (coords[0].equals(coords[coords.length-1]))
	    return factory.createPolygon(factory.createLinearRing(coords), null);
	return factory.createLineString(coords);
    }

    static public Geometry OMGraphicToJTSGeometry(OMRect omp, GeometryFactory factory) {
	return factory.createPolygon(factory.createLinearRing(extractCoordinates(omp)), null);
    }

    static public Geometry OMGraphicToJTSGeometry(OMCircle omc, GeometryFactory factory) {
	return factory.createPolygon(factory.createLinearRing(extractCoordinates(omc)), null);
    }

    /** From raw llpoints in radians to Coordinate in degrees */
    static public Coordinate[] coordRadiansToDegrees(double[] rawllpts) {

 	int nbPoints = rawllpts.length/2;
	Coordinate coords[] = new Coordinate[nbPoints];
	for(int i = 0; i < nbPoints; i++)
	    coords[i] = new Coordinate(Geo.degrees(rawllpts[2*i+1]), 
						   Geo.degrees(rawllpts[2*i]));
	
	return coords;
    }

    /** extract Coordinates from OMGraphic */
    static public Coordinate[] extractCoordinates(OMGraphic omg) {
	if (omg instanceof OMGraphicList) 
	    return extractCoordinates((OMGraphicList)omg);
	if (omg.getRenderType() != OMGraphic.RENDERTYPE_LATLON) {
	    String errmsg = "fr.unicaen.openmap.jts.JTSFactory.extractCoordinates: support only for LAT/LON Render Type";
	    assert false : errmsg;
	    System.err.println(errmsg);
	    return null;
	}
	if (omg instanceof OMPoint)
	    return extractCoordinates((OMPoint)omg);
	if (omg instanceof OMLine)
	    return extractCoordinates((OMLine)omg);
	if (omg instanceof OMPoly)
	    return extractCoordinates((OMPoly)omg);
	if (omg instanceof OMRect)
	    return extractCoordinates((OMRect)omg);
	if (omg instanceof OMCircle)
	    return extractCoordinates((OMCircle)omg);
	    
	String errmsg = "fr.unicaen.openmap.jts.JTSFactory.extractCoordinates: unsupported class for "+omg;
	assert false : errmsg;
	System.err.println(errmsg);
	return null;
    }

    static public Coordinate[] extractCoordinates(OMGraphicList omglist) {
	int size = omglist.size();
	Coordinate[][] coords = new Coordinate[size][];
	int nbCoordinates = 0;
	for (int i = 0; i < size; i++) {
	    coords[i] = extractCoordinates(omglist.get(i));
	    nbCoordinates += coords[i] == null ? 0 : coords[i].length;
	}
	// merge the coords elements into the result
	Coordinate[] coordinates = new Coordinate[nbCoordinates];
	int j = 0;
	for (int i = 0; i < size; i++)
	    if (coords[i] != null)
		for (Coordinate c : coords[i])
		    coordinates[j++] = c;
	return coordinates;
    }

    static public Coordinate[] extractCoordinates(OMPoint omp) {
	Coordinate[] coords = new Coordinate[1];
	coords[0] = new Coordinate(omp.getLon(), omp.getLat());
	return coords;
    }

    static public Coordinate[] extractCoordinates(OMLine oml) {
	double lls[] = oml.getLL();
	int nbPoints = lls.length/2;
	Coordinate coords[] = new Coordinate[nbPoints];
	for(int i = 0; i < nbPoints; i++)
	    coords[i] = new Coordinate(lls[2*i+1], lls[2*i]);
	return coords;
    }

    static public Coordinate[] extractCoordinates(OMPoly omp) {
	return coordRadiansToDegrees(omp.getRawllpts());
    }

    static public Coordinate[] extractCoordinates(OMRect omp) {
	Coordinate coords[] = new Coordinate[5];
	
	coords[0] = new Coordinate(omp.getEastLon(), omp.getNorthLat());
	coords[1] = new Coordinate(omp.getWestLon(), omp.getNorthLat());
	coords[2] = new Coordinate(omp.getWestLon(), omp.getSouthLat());
	coords[3] = new Coordinate(omp.getEastLon(), omp.getSouthLat());
	coords[4] = coords[0]; // close the LinearRing

	return coords;
    }

    static public Coordinate[] extractCoordinates(OMCircle omp) {
	int nverts = omp.getNumVerts();
	if (nverts < 3)
	    nverts = 3;
	// 
	double[] rawllpts = new double[2*nverts + 2];
	
	GreatCircle.earthCircle(omp.getCenter().getRadLat(), 
				omp.getCenter().getRadLon(),
				ProjMath.degToRad(omp.getRadius()), nverts,
				rawllpts);	
	// connect the vertices.
	rawllpts[rawllpts.length - 2] = rawllpts[0];
	rawllpts[rawllpts.length - 1] = rawllpts[1];

	return coordRadiansToDegrees(rawllpts);
    }

    /** From JTS structures to OpenMap structures */
    static public OMGraphic JTSGeometryToOMGraphic(Geometry geom) {
	if (geom instanceof GeometryCollection)
	    return JTSGeometryToOMGraphic((GeometryCollection)geom);
	if (geom instanceof LinearRing)
	    return JTSGeometryToOMGraphic((LinearRing)geom);
	if (geom instanceof LineString)
	    return JTSGeometryToOMGraphic((LineString)geom);
	if (geom instanceof Point)
	    return JTSGeometryToOMGraphic((Point)geom);
	if (geom instanceof Polygon)
	    return JTSGeometryToOMGraphic((Polygon)geom);
	String errmsg = "fr.unicaen.openmap.jts.JTSFactory.JTSGeometryToOMGraphic: unsupported class for "+geom;
	// assert false : errmsg;
	System.err.println(errmsg);
	return null;
    } 

    static public OMGraphic JTSGeometryToOMGraphic(GeometryCollection geom) {
	OMGraphicList omgl = new OMGraphicList();
	for (int i = 0; i < geom.getNumGeometries(); i++) {
	    omgl.add(JTSGeometryToOMGraphic(geom.getGeometryN(i)));
	}
	return omgl;
    }

    static public OMGraphic JTSGeometryToOMGraphic(LinearRing geom) {
 	Coordinate coords[] = geom.getCoordinates();
	double rawllpts[] = new double[coords.length *2];
	for (int i = 0; i < coords.length; i++) {
	    rawllpts[i*2] = coords[i].y;
	    rawllpts[i*2+1] = coords[i].x;
	}
	return new OMPoly(rawllpts, OMGraphic.DECIMAL_DEGREES, OMGraphic.LINETYPE_STRAIGHT);   
    }

    static public OMGraphic JTSGeometryToOMGraphic(LineString geom) {
	Coordinate coords[] = geom.getCoordinates();
	if (coords.length == 2) 
	    return new OMLine(coords[0].y, coords[0].x, coords[1].y, coords[1].x, OMGraphic.LINETYPE_STRAIGHT);
	double rawllpts[] = new double[coords.length *2];
	for (int i = 0; i < coords.length; i++) {
	    rawllpts[i*2] = coords[i].y;
	    rawllpts[i*2+1] = coords[i].x;
	}
	return new OMPoly(rawllpts, OMGraphic.DECIMAL_DEGREES, OMGraphic.LINETYPE_STRAIGHT);
    }

    static public OMGraphic JTSGeometryToOMGraphic(Point geom) {
	return new OMPoint(geom.getY(), geom.getX());
    }

    static public OMGraphic JTSGeometryToOMGraphic(Polygon geom) {
	assert geom.getNumInteriorRing() == 0 : "Interior ring discarded";
	return JTSGeometryToOMGraphic(geom.getExteriorRing());
    }

    static public OMDataLocation JTSGeometryToOMDataLocation(Geometry geom) {
	if (geom == null)
	    return null;
	Envelope env = geom.getEnvelopeInternal();
	Centroid centroid = null;
	BoundingBox bb = null;
	OMGraphic omg = JTSFactory.JTSGeometryToOMGraphic(geom);
	Point geomCentroid = geom.getCentroid();
	if (geomCentroid != null) {
	    Coordinate c = geom.getCentroid().getCoordinate();
	    centroid = new Centroid(c.x, c.y);
	    bb = new BoundingBox(env.getMinX(), env.getMaxY(), env.getMaxX(), env.getMinY());
	} 
	
	return new OMDataLocation(JTSGeometryToOMGraphic(geom),
				  centroid,
				  bb);
    }

    //   static public OMDataLocation testJTS(OMDataLocation omdl) {
    // 	return JTSGeometryToOMDataLocation(OMGraphicToJTSGeometry(omdl.getOMGraphic()));
    //   }
}
