// <header>
/**
 *  ThemaMap : The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
package fr.unicaen.openmap.layer.dataDisplayLayer;

import fr.unicaen.openmap.layer.dataDisplayLayer.gui.*;
import fr.unicaen.openmap.dataStorage.*;
import java.util.Properties;

import com.bbn.openmap.*;
import com.bbn.openmap.omGraphics.*;
import com.bbn.openmap.proj.*;
import com.bbn.openmap.proj.coords.LatLonPoint;
import com.bbn.openmap.util.*;
import fr.unicaen.openmap.util.Colors;
import fr.unicaen.openmap.util.Util;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Graphics;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import javax.swing.BorderFactory;
import javax.swing.ButtonGroup;
import javax.swing.JComponent;
import javax.swing.JPanel;
import javax.swing.JRadioButton;



/**
 * This class is the DataDisplayLayer for Geometry type objects.
 *
 * openmap.properties which used GeometryDataDisplay layer contains :
 *
 * test.class               = fr.unicaen.openmap.layer.dataDisplayLayer.GeometryDataDisplayLayer
 * test.dataStorageClass    = fr.unicaen.openmap.dataStorage.xmltable.XmlTableDataStorage
 * test.xmlFile             = departement96.xml.gz
 * test.geometryColumnName  = dep
 * test.prettyName          = Départements
 * test.lineColor           = FFFFFF33
 * test.fillColor           = 00FFFFFF
 * test.selectedLineColor   = FF990099
 * test.selectedFillColor   = C900FF55
 * test.fill                = no
 * test.drawGraphicNameMode = 1
 * test.drawColumnName      = nom
 * test.drawColumnNameColor = FFFF0000
 */
public class GeometryDataDisplayLayer extends SimpleDataDisplayLayer implements ActionListener, ComponentListener
        
{
    public static final String geometryNamePropertyString = "geometryColumnName";
    public static final String lineColorPropertyString = "lineColor";
    public static final String fillColorPropertyString = "fillColor";
    public static final String selectedFillColorPropertyString = "selectedFillColor";
    public static final String selectedLineColorPropertyString = "selectedLineColor";
    public static final String fillPropertyString = "fill";
    public static final String drawGraphicNameModePropertyString = "drawGraphicNameMode";
    public static final String drawColumnNamePropertyString = "drawColumnName";
    public static final String drawColumnNameColorPropertyString = "drawColumnNameColor";
    
    
    protected String geometryColumnName = "g";
    protected String drawColumnName = "column";
    protected Integer drawGraphicNameMode = 0;
    protected Color drawColumnNameColor = null;
    protected Component fatherGui = null;
    protected ButtonGroup drawingGroup = null;
    
    protected JRadioButton noDrawingButton = null;
    protected JRadioButton drawingAllButton = null;
    protected JRadioButton drawingSelectedButton = null;
    protected JPanel radioPanel = null;
    
    /**
     * Graphical properties for the layer.
     */
    protected Colors colors = null;
    
    protected GraphicsPanel graphicsPanel = null;
    
    protected transient JPanel box = null;
    /**
     * We need to know which DatStorage we will use, so the
     * constructor does nothing with it.
     */
    public GeometryDataDisplayLayer()  {
        Debug.message(Util.className(this), this + " new GeometryDataDisplayLayer()");
        colors = new Colors();
    }
    
    /**
     * Let's find our properties.
     * @param prefix string prefix used in the properties file for this layer
     * @param properties the properties set in the properties file
     */
    public void setProperties(String prefix, Properties properties) {
        super.setProperties(prefix, properties);
        Debug.message(Util.className(this), this + " setProperties : GeometryDataDisplayLayer");
        
        prefix = PropUtils.getScopedPropertyPrefix(prefix);
        
        geometryColumnName = properties.getProperty(prefix + geometryNamePropertyString, "g");
        drawColumnName = properties.getProperty(prefix + drawColumnNamePropertyString, "defaultcolumn");
        drawGraphicNameMode = new Integer(properties.getProperty(prefix + drawGraphicNameModePropertyString, "0"));
        drawColumnNameColor = ColorFactory.parseColor(properties.getProperty(prefix + drawColumnNameColorPropertyString, "FFFF0000"));
        
        colors.setLineColor(properties.getProperty(prefix + lineColorPropertyString, "FF003333"));
        colors.setFillColor(properties.getProperty(prefix + fillColorPropertyString, "FF3366FF"));
        colors.setSelectedLineColor(properties.getProperty(prefix + selectedLineColorPropertyString, "ffffffff"));
        colors.setSelectedFillColor(properties.getProperty(prefix + selectedFillColorPropertyString, "ffffffff"));
        colors.setFillMode(Util.isTrue(properties.getProperty(prefix + fillPropertyString, "true")));
        
    }
    
    //--------------------------------------------------------------------------------------------------------------------NOT FINISHED
    /**
     * The properties and prefix are managed and decoded here, for the standard uses of GeometryDataDisplayLayer.
     * Notes this method allows to define what is present in file "save Map as properties file".
     * @param properties the properties set in the properties file
     * @return Properties.
     */
    public Properties getProperties(Properties properties) {
        properties = super.getProperties(properties); //SimpleDataSorterLayer properties
        Debug.message(Util.className(this), this + " GeometryDataDisplay getProperties(...)");        
        String prefix = PropUtils.getScopedPropertyPrefix(this);
        
        if (colors.getLineColor()!=null) properties.put(prefix + lineColorPropertyString, ColorFactory.getHexColorString(colors.getLineColor()));
        if (colors.getFillColor()!=null) properties.put(prefix + fillColorPropertyString, ColorFactory.getHexColorString(colors.getFillColor()));
        if (colors.getSelectedLineColor()!=null) properties.put(prefix + selectedLineColorPropertyString, ColorFactory.getHexColorString(colors.getSelectedLineColor()));
        if (colors.getSelectedFillColor()!=null) properties.put(prefix + selectedFillColorPropertyString, ColorFactory.getHexColorString(colors.getSelectedFillColor()));
        properties.put(prefix + fillPropertyString, new Boolean(colors.getFillMode()).toString());
        
        properties.put(prefix + geometryNamePropertyString, geometryColumnName);
        properties.put(prefix + drawColumnNamePropertyString, drawColumnName);
        properties.put(prefix + drawGraphicNameModePropertyString, drawGraphicNameMode.toString());
        properties.put(prefix + drawColumnNameColorPropertyString, ColorFactory.getHexColorString(drawColumnNameColor));
        
        return properties;
    }
    //--------------------------------------------------------------------------------------------------------------------
    
    /**
     * Method that we use to generate the list of drawable
     * objects.
     * This method is called by prepare to get OMGraphicList
     * @see #projectionChanged(ProjectionEvent e)
     * @see #renderDataForProjection(Projection proj, java.awt.Graphics g)
     */
    public synchronized OMGraphicList generateGraphics() {
        OMGraphicList omgraphics = new OMGraphicList();
        // No iterator == no data => return.
        if (dataStorageIterator == null) return omgraphics;
        
        // get projection in order to compute the bounding box.
        Projection proj = getProjection();
        if (proj == null) return omgraphics;
        
        // get the bounding Box
        LatLonPoint upperLeft = proj.getUpperLeft();
        LatLonPoint lowerRight = proj.getLowerRight();
        BoundingBox area = new BoundingBox(upperLeft.getLongitude(), upperLeft.getLatitude(),
                lowerRight.getLongitude(), lowerRight.getLatitude());
        
        Debug.message(Util.className(this),this+"GeometryDataDisplayLayer : generateGraphics() using "+ geometryColumnName + " and " + area);
        
        // clear the list.
        omgraphics.clear();
        
        // reset the iterator to the first record.
        dataStorageIterator.reset();
        
        // Find valid object to draw (i.e. their bounding box is inside
        // the area). We will try to draw an object if there is no bounding box for it.
        while (dataStorageIterator.hasNext()) {
            OMDataLocation omdl = dataStorageIterator.next().getOMDataLocation(geometryColumnName);
            if (omdl != null && omdl.getBoundingBox() != null && omdl.getBoundingBox().intersects(area)) {
                OMGraphic omg = omdl.getOMGraphic();
                omg.setVisible(true);
                omg.setLinePaint(colors.getLineColor());
                omg.setFillPaint(colors.getFillColor());
                if (omg instanceof OMGraphicList){
                    ((OMGraphicList)omg).setVague(true);
                }
                omgraphics.add(omg);
            }
        }
        Debug.message(classNameForDebug, this + " using " + omgraphics.size() + " objects.");
        omgraphics.generate(proj);
        return omgraphics;
    }
    
    public Component getGUI() {
        Debug.message(Util.className(this), this + " GeometryDataDisplayLayer.getGUI()");
        if (!createPalette) return null;
        if (box == null) {
            ColorsPanel colorPanel = new ColorsPanel(colors);
            colorPanel.addUnicaenListener(this);
            // try to find if super has a gui
            Component fatherGui = super.getGUI();
            // if no gui, we create one, otherwise we add information to itit.
            
            if (fatherGui==null) {
                fatherGui = new JPanel(new BorderLayout());
            }
            radioPanel = new JPanel(new GridLayout(1, 0));
            radioPanel.setBorder(BorderFactory.createCompoundBorder(BorderFactory.createTitledBorder("Affichage des informations"), BorderFactory.createEmptyBorder(5,5,5,5)));
            
            String noDrawing = new String("Afficher aucune information");
            noDrawingButton = new JRadioButton(noDrawing);
            noDrawingButton.setActionCommand(noDrawing);
            
            String drawingAll = new String("Afficher toutes les informations");
            drawingAllButton = new JRadioButton(drawingAll);
            drawingAllButton.setActionCommand(drawingAll);
            // add group
            drawingGroup = new ButtonGroup();
            drawingGroup.add(noDrawingButton);
            drawingGroup.add(drawingAllButton);
            // add listener
            noDrawingButton.addActionListener(this);
            drawingAllButton.addActionListener(this);
            // add to panel
            radioPanel.add(noDrawingButton);
            radioPanel.add(drawingAllButton);
            ((JComponent)fatherGui).add(radioPanel);
            
            switch (drawGraphicNameMode) {
                case LayerConstants.DRAWING_NAME_MODE_OFF :
                    noDrawingButton.setSelected(true);
                    break;
                case LayerConstants.DRAWING_NAME_MODE_ON :
                    drawingAllButton.setSelected(true);
                    break;
                default :
                    noDrawingButton.setSelected(true);
                    break;
            }
            box=new SimplePalette(fatherGui, colorPanel,dataStorageIterator.getDataGUI());
            box.addComponentListener(this);
        }
        return box;
    }
    
    /** Listens to the radio buttons. */
    public void actionPerformed(ActionEvent e) {
        super.actionPerformed(e);
        if (e.getSource() == noDrawingButton) {
            drawGraphicNameMode = LayerConstants.DRAWING_NAME_MODE_OFF;
            repaint();
        }
        if (e.getSource() == drawingAllButton) {
            drawGraphicNameMode = LayerConstants.DRAWING_NAME_MODE_ON;
            repaint();
        }
    }
    
    /** ComponentListener method. */
    public void componentResized(ComponentEvent e) {}
    
    /** ComponentListener method. */
    public void componentShown(ComponentEvent e) {}
    
    /** ComponentListener method */
    public void componentMoved(ComponentEvent e) {}
    
    /** ComponentListener method. */
    public void componentHidden(ComponentEvent e) {}
    
    /**
     * paint name only if we have to do it, depends on drawGraphicNameMode
     *
     * @param g the Graphics context for painting
     */
    public void paint(Graphics g) {
        super.paint(g);
        Projection projection = getProjection();
        if (projection==null) return;
        // Render data name on graphics ?
        if (drawGraphicNameMode>LayerConstants.DRAWING_NAME_MODE_OFF) {
            // get the text ...
            OMGraphicList omglt = getTextFromOMGraphicList();
            // draw it
            if (omglt!=null) omglt.render(g);
        }
        
    }
    
    public OMGraphicList getTextFromOMGraphicList(){
        // create the list
        OMGraphicList liste = new OMGraphicList();
        // get omgraphic list
        OMGraphicList omgl = getList();
        if (omgl==null) return null;
        // set the size, and method to draw text.
        FontSizer fontsizer = new FontSizer(300000f,1, 8, 20);
        // for all omGraphics get the name and ad it to he list
        for (int i = 0 ; i < omgl.size() ; i++){
            // get omgraphic
            OMGraphic omg = omgl.getOMGraphicAt(i);
            // get index in datastorage
            int index = (Integer)omg.getAttribute(Util.OMGraphicIndex);
            // get DataObjet that contain informations.
            DataObject dataObject = dataStorageIterator.getDataStorage().getElementAt(index);
            // get the centroid
            Centroid centre = dataObject.getOMDataLocation(geometryColumnName).getCentroid();
            // crate the text
            OMText omt = new OMText((float)centre.getLat(),(float)centre.getLon(),(String)dataObject.getData(drawColumnName),OMText.JUSTIFY_CENTER);
            // set the color
            omt.setLinePaint(drawColumnNameColor);
            // set the font sizer
            omt.setFontSizer(fontsizer);
            // add it to the list.
            liste.add(omt);
        }
        
        Projection proj = getProjection();
        liste.generate(proj);
        return liste;
    }

    public void animationModeChanged(MonoValuedFilterDataStorage ads) {
        //throw new UnsupportedOperationException("Not supported yet.");
    }

    
    
}
