// <header>
/**
 *  ThemaMap : The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
package fr.unicaen.openmap.layer.thematicDataDisplayLayer.omGraphics;

import com.bbn.openmap.omGraphics.*;
import fr.unicaen.openmap.layer.thematicDataDisplayLayer.ThematicDataDisplayLayer;
import fr.unicaen.openmap.layer.thematicDataDisplayLayer.util.*;
import fr.unicaen.openmap.layer.thematicDataDisplayLayer.visu.MultiColumnsRep.Orientation;
import fr.unicaen.openmap.layer.thematicDataDisplayLayer.visu.Offsetable;
import java.awt.Paint;
import java.awt.Point;
import java.awt.geom.Point2D;

/**
 * A BoxPlot (box-and-whisker diagram) given 
 * centered on Point(latPoint, lonPoint).
 * <pre>
 *   Q0      Q1        Q2                    Q3            Q4
 *            _______________________________
 *   |-------|_________X_____________________|--------------|
 * </pre>
 * If Q0 Q2 or Q4 is equals to NaN, we don't draw the corresponding line.
 * TBD: Plot the strings VQi if not null. 
 */
public class OMBoxPlot extends OMOffsetable implements Comparable<OMBoxPlot> {

    private double height;
    private double width;

    public OMBoxPlot(Offsetable off, String id, double latPoint, double lonPoint, int xOffset, int yOffset,  double witdh, double[] Q, Paint lineColor, Paint fillColor, Orientation orientation, boolean v, boolean l) {
	this(off, id, latPoint, lonPoint, xOffset, yOffset, witdh, Q, null, lineColor, fillColor, orientation, v, l);
    }
    public OMBoxPlot(Offsetable off, String id, double latPoint, double lonPoint, int xOffset, int yOffset,  double witdh, double[] Q, String[] labels, Paint lineColor, Paint fillColor, Orientation orientation, boolean v, boolean l) {
	// compute maxHeight of this histogram, 
	// max is the max height of all the histograms build by this layer
        super(off, id, latPoint, lonPoint, xOffset, yOffset, v, l);
	height = Q[4]-Q[0];
        this.width = width;
	addGraphics(xOffset, yOffset, latPoint, lonPoint, witdh, lineColor, fillColor, Q, labels, orientation);
        init(xOffset, yOffset);
    }

    protected void addGraphics(int xOffset, int yOffset, double latPoint, double lonPoint, double witdh, Paint lineColor, Paint fillColor, double[] Q, String[] labels, Orientation orientation) {
	//    Q0      Q1        Q2                    Q3            Q4 
	//             _______________________________
	//    |-------|_________X_____________________|--------------|
	//
	//   Made of 5 lines and a rectangle (in this order, in order to make the Q2 line visible) 
	//    |
	//     -------
	//                      X|
	//                                             --------------
	//             _______________________________               |
	//            |_______________________________|
	int halfWidth = (int)(witdh/2);  
	int quarterWidth = (int)(witdh/4);  
	OMRect omRect = null;
	switch (orientation) {
	case UP:
	    if (! Double.isNaN(Q[0])) {
		add(new OMLine(latPoint, lonPoint,
			       -halfWidth,(int)Q[0],
			       halfWidth, (int)Q[0]));
		add(new OMLine(latPoint, lonPoint,
			       0, (int)Q[0],
			       0, (int)Q[1]));
	    }
	    add(new OMLine(latPoint, lonPoint,
			   -quarterWidth, (int)Q[2]-quarterWidth,
			   quarterWidth, (int)Q[2]+quarterWidth));
	    add(new OMLine(latPoint, lonPoint,
			   -quarterWidth, (int)Q[2]+quarterWidth,
			   quarterWidth, (int)Q[2]-quarterWidth));
	    add(new OMLine(latPoint, lonPoint,
			   -halfWidth, (int)Q[2],
			   halfWidth, (int)Q[2]));
	    if (! Double.isNaN(Q[4])) {
		add(new OMLine(latPoint, lonPoint,
			       0, (int)Q[3],
			       0, (int)Q[4]));
		add(new OMLine(latPoint, lonPoint,
			       -halfWidth, (int)Q[4],
			       halfWidth, (int)Q[4]));
	    }
	    add(omRect = new OMRect(latPoint, lonPoint,
				    -halfWidth, (int)Q[1],
				    halfWidth, (int)Q[3]));
	    break;
	case RIGHT:
	    if (! Double.isNaN(Q[0])) {
		add(new OMLine(latPoint, lonPoint,
			       (int)Q[0], halfWidth,
			       (int)Q[0], -halfWidth));
		add(new OMLine(latPoint, lonPoint,
			       (int)Q[0], 0,
			       (int)Q[1], 0));
	    }
	    add(new OMLine(latPoint, lonPoint,
			   (int)Q[2]+quarterWidth, quarterWidth,
			   (int)Q[2]-quarterWidth, -quarterWidth));
	    add(new OMLine(latPoint, lonPoint,
			   (int)Q[2]-quarterWidth, quarterWidth,
			   (int)Q[2]+quarterWidth, -quarterWidth));
	    add(new OMLine(latPoint, lonPoint,
			   (int)Q[2], halfWidth,
			   (int)Q[2], -halfWidth));
	    if (! Double.isNaN(Q[4])) {
		add(new OMLine(latPoint, lonPoint,
			       (int)Q[3], 0,
			       (int)Q[4], 0));
		add(new OMLine(latPoint, lonPoint,
			       (int)Q[4], halfWidth,
			       (int)Q[4], -halfWidth));
	    }
	    add(omRect = new OMRect(latPoint, lonPoint,
				    (int)Q[1], halfWidth,
				    (int)Q[3], -halfWidth));
	    break;
	case LEFT:
	    if (! Double.isNaN(Q[0])) {
		add(new OMLine(latPoint, lonPoint,
			       -(int)Q[0], halfWidth,
			       -(int)Q[0], -halfWidth));
		add(new OMLine(latPoint, lonPoint,
			       -(int)Q[0], 0,
			       -(int)Q[1], 0));
	    }
	    add(new OMLine(latPoint, lonPoint,
			   -(int)Q[2]+quarterWidth, quarterWidth,
			   -(int)Q[2]-quarterWidth, -quarterWidth));
	    add(new OMLine(latPoint, lonPoint,
			   -(int)Q[2]-quarterWidth, quarterWidth,
			   -(int)Q[2]+quarterWidth, -quarterWidth));
	    add(new OMLine(latPoint, lonPoint,
			   -(int)Q[2], halfWidth,
			   -(int)Q[2], -halfWidth));
	    if (! Double.isNaN(Q[4])) {
		add(new OMLine(latPoint, lonPoint,
			       -(int)Q[3], 0,
			       -(int)Q[4], 0));
		add(new OMLine(latPoint, lonPoint,
			       -(int)Q[4], halfWidth,
			       -(int)Q[4], -halfWidth));
	    }
	    add(omRect = new OMRect(latPoint, lonPoint,
				    -(int)Q[1], halfWidth,
				    -(int)Q[3], -halfWidth));
	    break;
	case DOWN:
	    if (! Double.isNaN(Q[0])) {
		add(new OMLine(latPoint, lonPoint,
			       -halfWidth,-(int)Q[0],
			       halfWidth, -(int)Q[0]));
		add(new OMLine(latPoint, lonPoint,
			       0, -(int)Q[0],
			       0, -(int)Q[1]));
	    }
	    add(new OMLine(latPoint, lonPoint,
			   -quarterWidth, -(int)Q[2]-quarterWidth,
			   quarterWidth, -(int)Q[2]+quarterWidth));
	    add(new OMLine(latPoint, lonPoint,
			   -quarterWidth, -(int)Q[2]+quarterWidth,
			   quarterWidth, -(int)Q[2]-quarterWidth));
	    add(new OMLine(latPoint, lonPoint,
			   -halfWidth, -(int)Q[2],
			   halfWidth, -(int)Q[2]));
	    if (! Double.isNaN(Q[4])) {
		add(new OMLine(latPoint, lonPoint,
			       0, -(int)Q[3],
			       0, -(int)Q[4]));
		add(new OMLine(latPoint, lonPoint,
			       -halfWidth, -(int)Q[4],
			       halfWidth, -(int)Q[4]));
	    }
	    add(omRect = new OMRect(latPoint, lonPoint,
				    -halfWidth, -(int)Q[1],
				    halfWidth, -(int)Q[3]));
	    break;
	}

	if (fillColor != null)
	    omRect.setFillPaint(fillColor);
	setLinePaint(lineColor);
	// for (int i = 0; i < 5; i++) System.out.print("\t" + Q[i]); 
	// System.out.println(" :\t" + this.size());
	String toolTip = labels[0]+" [ "+labels[1]+" "+labels[2]+" "+labels[3]+" ] "+labels[4];
	putAttribute(ThematicDataDisplayLayer.VALUE, toolTip);
    }
    
    public int compareTo(OMBoxPlot omb) {
	return (int)Math.signum(this.height - omb.height);
    }

    public int getHeight() {
	return (int) height;
    }

    public int getWidth() {
	return (int) width;
    }

    @Override
    public void translate(int x, int y) {
        super.translate(x, y);
        for(OMGraphic omg : this) {
	    if (omg instanceof OMRect) {
		OMRect omr =(OMRect)omg;
		omr.setLocation(getLat(),
				getLon(),
				x + omr.getLeft(),
				y + omr.getTop(),
				x + omr.getRight(),
				y + omr.getBottom());  
	    } else {
		OMLine oml = (OMLine) omg;
		int[] xys = oml.getPts();
		for (int i = 0; i < 4; i++)
		    xys[0] += i%2 == 0 ? x : y;
		oml.setPts(xys);
	    }
        }
    }
    
}
