// <header>
/**
 *  ThemaMap : The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
package fr.unicaen.openmap.layer.thematicDataDisplayLayer.omGraphics;

import com.bbn.openmap.omGraphics.OMPoly;
import com.bbn.openmap.proj.coords.LatLonPoint;
import fr.unicaen.openmap.layer.thematicDataDisplayLayer.visu.Offsetable;
import java.awt.Point;
import java.awt.geom.Point2D;

/** 
 * @author Jacques.Madelaine@info.unicaen.fr
 */
public class OMRegularPolygon extends OMOffsetableWithData implements Comparable<OMRegularPolygon> {

    private double radius;
    private int nbEdges;
    private double startAngle = 0.0F;
    private int starCoef = 0;
    private OMPoly omPoly;

    public OMRegularPolygon(Offsetable off, String id, double latPoint, double lonPoint, int nbEdges, double radius, boolean v, boolean l) {
        this(off, id, latPoint, lonPoint, nbEdges, radius, 0, 0, 0F, 0, v, l, "");
    }

    public OMRegularPolygon(Offsetable off, String id, double latPoint, double lonPoint, int nbEdges, double radius, int xOffset, int yOffset, double startAngle, int starCoef, Object data, boolean v, boolean l) {
        this(off, id, latPoint, lonPoint, nbEdges, radius, xOffset, yOffset, startAngle, starCoef, v, l, data);
        //	System.out.println("Regular poly "+nbEdges+" nbEdges "+ radius + " radius ["+xOffset+","+yOffset);
    }

    /**
     * Construct a regular polygon defined in XY relatively to the point defined in lat/lon.
     * It has <code>nbEdges</code> edges and the specified <code>radius</code>. 
     * The <code>starCoef</code> is used to build a star as a nbEdges*2 polygon;
     * 0 leed to the regular polygon, 100 leeds to a very slim star (a + for 4 edges), 
     * in between values leed to a proportionnal result.
     * The center of the polygon may be at (xOffset, yOffset) from the lat/lon point.
     * A rotation of startAngle expressed in radians is applied (default to 0).
     */
    public OMRegularPolygon(Offsetable off, String id, double latPoint, double lonPoint, int nbEdges, double radius, int xOffset, int yOffset, double startAngle, int starCoef, boolean cpVisibility, boolean l) {
	this(off, id, latPoint, lonPoint, nbEdges, radius, xOffset, yOffset, startAngle, starCoef, cpVisibility, l, "");
    }
    public OMRegularPolygon(Offsetable off, String id, double latPoint, double lonPoint, int nbEdges, double radius, int xOffset, int yOffset, double startAngle, int starCoef, boolean cpVisibility, boolean l, Object data) {
        super(off, id, latPoint, lonPoint, xOffset, yOffset, cpVisibility, l, data);
        omPoly = new OMPoly(latPoint, lonPoint, new int[]{}, OMPoly.COORDMODE_ORIGIN);
        this.radius = radius;
        this.nbEdges = nbEdges;
        this.setStarCoef(starCoef);
        this.startAngle = startAngle;
        omPoly.setIsPolygon(true);
        setXYs();
        add(omPoly);
        init(xOffset, yOffset);
    }



    /** starCoef must be between 0 and 100 */
    public void setStarCoef(int starCoef) {
        if (starCoef < 0) {
            starCoef = 0;
        } else if (starCoef > 100) {
            starCoef = 100;
        }
        this.starCoef = starCoef;
    }

    // a polygon centered on the point (xOffset, yOffset)
    public void setXYs() {
        double angle = startAngle;
        int[] Xs, Ys;

        if (starCoef == 0) {
            // we build a nbEdges edges regular polygon
            Xs = new int[nbEdges + 1];
            Ys = new int[nbEdges + 1];
            double angleIncr = 2 * Math.PI / nbEdges;
            for (int i = 0; i < nbEdges; i++, angle += angleIncr) {
                Xs[i] = getX() + (int) (radius * Math.cos(angle));
                Ys[i] = getY() + (int) (radius * Math.sin(angle));
            }
            //the last point is equals to the first one in order to have a closed polygon
            Xs[nbEdges] = Xs[0];
            Ys[nbEdges] = Ys[0];
        } else { // a star
            // Note that this code leeds to the same shape when starCoef = 0, but with 2*nbEdges edges.
            Xs = new int[2*nbEdges+1];
            Ys = new int[2*nbEdges+1];
            double angleIncr = Math.PI / nbEdges;
            double radiusForStar = Math.cos(Math.PI / nbEdges) * (1 - starCoef / 100.0) * radius;
            for (int i = 0; i < 2 * nbEdges; i++, angle += angleIncr) {
                Xs[i] = getX() + (int) (radius * Math.cos(angle));
                Ys[i] = getY() + (int) (radius * Math.sin(angle));
                i++;
                angle += angleIncr;
                Xs[i] = getX() + (int) (radiusForStar * Math.cos(angle));
                Ys[i] = getY() + (int) (radiusForStar * Math.sin(angle));
            }
            //the last point is equals to the first one in order to have a closed polygon
            Xs[2*nbEdges] = Xs[0];
            Ys[2*nbEdges] = Ys[0];
        }
        omPoly.setXs(Xs);
        omPoly.setYs(Ys);
    }

    public int[] getXs() {
        return omPoly.getXs();
    }

    public int[] getYs() {
        return omPoly.getYs();
    }

    public double getRadius() {
        return radius;
    }

    public int getWidth() {
        return (int) (2 * radius);
    }

    public int getHeight() {
        return (int) (2 * radius);
    }

    public double getStarCoef() {
        return starCoef;
    }

    public int compareTo(OMRegularPolygon rp) {
        return (int) Math.signum(this.getRadius() - rp.getRadius());
    }

    @Override
    public void translate(int x, int y){
        super.translate(x, y);
        setXYs();
    }

}
