// <header>
/**
 *  ThemaMap : The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
package fr.unicaen.openmap.layer.thematicDataDisplayLayer.omGraphics.legend;

import com.bbn.openmap.omGraphics.*;
import com.bbn.openmap.util.PropUtils;
import com.bbn.openmap.util.Debug;
import fr.unicaen.openmap.layer.legendLayer.LegendLayer;
import fr.unicaen.openmap.layer.thematicDataDisplayLayer.gui.*;
import fr.unicaen.openmap.layer.thematicDataDisplayLayer.omGraphics.*;
import fr.unicaen.openmap.util.Util;

import java.util.Properties;
import javax.swing.*;
import java.awt.Color;
import java.awt.event.*;

/** These graphics are solely in RENDERXY mode 
 * <p>
 * These graphics are resizeable thanks to OMControlPoints that know the OMResizeableGraphic 
 * as there appObjects. These OMControlPoints are controled with a MouseControlPointsListener. 
 * Every mousePressed, mouseReleased and mouseDragged events are forwarded to the controled OMResizeableGraphic 
 * as calls to the methods translate or extend.
 * <p> 
 * In case te MouseControlPointsListener is null, no control points are added,
 * but they are still translatable and extendable.
 * <p>
 * The Refreshable is used to refresh the view after each event.
 * <p>
 * The method add(OMGraphic) is protected. If you want a list @see OMResizeableGraphicList
 * <p>
 * The hierarchy of OMResizeableGraphic is as follows :
 * <ul>
 *   <li> OMResizeableCompass
 *   <li> OMResizeableDisc
 *   <li> OMResizeableFlow
 *   <li> OMResizeableLine
 *   <li> OMResizeablePie
 *   <li> OMResizeableRaster
 *   <li> OMResizeableRect
 *   <li> OMResizeableRegularPolygon
 *   <li> OMResizeableText 
 *   <ul>
 *       <li> OMResizeableTextArea
 *   </ul>
 *   <li> OMResizeableGraphicList // a simple list of independant graphics
 *   <ul>
 *       <li> OMResizeableGroup // a list of graphics that translate all together
 *       <ul>
 *           <li> OMResizeableScale
 *           <li> OMResizeableFrame // a group with a border 
 *           <ul>
 *               <li> OMResizeableColorScaleLegend
 *               <li> OMResizeableDiscLegend
 *               <li> OMResizeableExtensibleFrame
 *               <li> OMResizeableFlowLegend
 *               <li> OMResizeableMultiRegularPolygonLegend
 *               <li> OMResizeableRegularPolygonLegend 
 *           </ul>
 *       </ul>
 *   </ul>
 */
public abstract class OMResizeableGraphic {

    public static final int OMControlPointRadius = 4;
    protected OMGraphicList omControlPoints = new OMGraphicList();
    protected Refreshable refreshable;
    protected JCheckBox visibleCB;
    /** name used in the GUI */
    private String name = "";
    private boolean removableAnchors;

    /** name used in the GUI */
    public void setName(String name) {
        this.name = name;
    }

    /** name used in the GUI */
    public String getName() {
        return name;
    }
    protected MouseControlPointsListener ml = null; // just used to know if controlPoints should be visible or not

    public OMResizeableGraphic(Refreshable r, MouseControlPointsListener ml, String nameForGUI){
        this(r, ml, nameForGUI, true);
    }

    public OMResizeableGraphic(Refreshable r, MouseControlPointsListener ml, String nameForGUI, boolean ra) {
        removableAnchors = ra;
        refreshable = r;
        if (ml != null) {
            add(omControlPoints);
            this.ml = ml;
        }
        setName(nameForGUI);
        addOMControlPointsToMouseListener();

        // Debug.message("Legend", "### new "+Util.className(this)+"@"+hashCode()+" : "+ml);
    }

    public void addOMControlPointsToMouseListener() {
        if (ml != null) {
            ml.addOMControlPoints(omControlPoints);
        }
    }

    protected OMControlPoint addOMControlPoint(OMControlPoint omc) {
        omc.setAppObject(this);
        omControlPoints.add(omc);
        // System.out.println(this+" rf@"+(refreshable==null?"0#":refreshable.hashCode())
        // 		   +" ml@"+(ml==null?"0#":ml.hashCode())+" omControlPoints@"+omControlPoints.hashCode()+" -> "+omControlPoints.size());
        return omc;
    }
    protected OMGraphicList graphics = new OMGraphicList();

    protected void add(OMGraphic g) {
        graphics.add(g);
    }

    public OMGraphicList getGraphics() {
        updateFromGUI();
        return graphics;
    }
    private boolean visible = true;

    protected void setVisible(boolean v) {
        setGraphicsVisible(v);
    }

    public boolean getVisible() {
        return visible;
    }

    protected void setGraphicsVisible(boolean v) {
        // System.out.println("=>setGraphicsVisible to "+v+" for "+ this);
        Debug.message("Legend3", "Legend =>setGraphicsVisible to " + v + " for " + this);
        visible = v;
        graphics.setVisible(v);
        omControlPoints.setVisible(v && ml != null && ml.getOMControlPointVisible());
        refreshVisibleCB();
    }

    public abstract void translate(int dx, int dy);
    //// We could have a default implementation. tbd...
    //public void translate(int dx, int dy) {
    //	for(int i = 0; i < omControlPoints.size(); i++) 
    //	    ((OMControlPoint)omControlPoints.getOMGraphicAt(i)).translate(dx, dy);
    //    }

    public abstract void extend(int dw, int dh);

    /** absolute positionning */
    public void moveTo(int x, int y) {
	translate(x-getXAnchor(), y-getYAnchor());
    }

    public abstract int getXAnchor();

    public abstract int getYAnchor();

    public abstract int getWidth();

    public abstract int getHeight();

    public OMGraphicList getOMControlPoints() {
        return omControlPoints;
    }

    public void refreshGraphics() {
        refreshable.refreshGraphicsAndGUI();
    }

    public void refreshInteractiveGraphics() {
        refreshable.refreshInteractiveGraphics();
    }

    public void released(){
        ((LegendLayer)refreshable).clearInteractiveGraphics();
    }

    public void addToInteractiveGraphic(){
        ((LegendLayer)refreshable).addInteractiveGraphics(this);
    }

    protected void refreshVisibleCB() {
        if (visibleCB != null) {
            visibleCB.setIcon(Util.loadIcon(!getVisible() ? "disabled.png" : "enabled.png"));
            visibleCB.setToolTipText(getVisible() ? "disable" : "enable");
        }
    }

    /** default GUI is empty */
    public JComponent getGUI() {
        return null;
    }
    protected JComponent toolGUI = null;

    /** provides a visible/non visible toggle button */
    public JComponent getToolGUI() {
        if (toolGUI != null) {
            return toolGUI;
        }
        visibleCB = new JCheckBox(Util.loadIcon(!getVisible() ? "disabled.png" : "enabled.png"), getVisible());
        // visibleCB.setText(name);
        visibleCB.setToolTipText(!getVisible() ? "disable" : "enable");
        //visibleCB.setBorder(null);
        visibleCB.setBackground(Color.WHITE);
        //visibleCB.setContentAreaFilled(false);
        visibleCB.setOpaque(false);
        visibleCB.addActionListener(new ActionListener() {

            public void actionPerformed(ActionEvent e) {
                setVisible(visibleCB.isSelected());
                // System.out.println(OMResizeableGraphic.this+" setVisible "+visibleCB.isSelected()+ " = " + getVisible());
                refreshGraphics();
            }
        });

        toolGUI = new JPanel();
        toolGUI.add(visibleCB);
        //ComponentSizer.fixComponentSize(visibleCB, new Dimension(25,25));
        return toolGUI;
    }

    /** used essentially for Texts (avoid typing return in text fields) */
    protected void updateFromGUI() {
    }
    //================================================================
    // PROPERTIES
    //================================================================
    public static final String X_ANCHOR_PROPERTY_NAME = "anchor.x";
    public static final String Y_ANCHOR_PROPERTY_NAME = "anchor.y";
    public static final String WIDTH_PROPERTY_NAME = "width";
    public static final String HEIGHT_PROPERTY_NAME = "height";
    public static final String VISIBILITY_PROPERTY_NAME = "visible";

    /**
     * If you want to set default valu, you have to overload this method
     * (see OMResizeableGlobalPieFrame for default visibility set to false)
     */
    public void setProperties(String prefix, Properties props) {
        if (props == null) {
            return;
        }
        prefix = PropUtils.getScopedPropertyPrefix(prefix);
        //Position
        translate(
                PropUtils.intFromProperties(props, prefix + X_ANCHOR_PROPERTY_NAME, getXAnchor()) - getXAnchor(),
                PropUtils.intFromProperties(props, prefix + Y_ANCHOR_PROPERTY_NAME, getYAnchor()) - getYAnchor());


        //Size
        //Pb with OMResizeableFrame & GUI, see extend() !!

        extend(
                PropUtils.intFromProperties(props, prefix + WIDTH_PROPERTY_NAME, getWidth()) - getWidth(),
                PropUtils.intFromProperties(props, prefix + HEIGHT_PROPERTY_NAME, getHeight()) - getHeight());

        Boolean v = PropUtils.booleanFromProperties(props, prefix + VISIBILITY_PROPERTY_NAME, getVisible());
	// System.out.println("(SET) " + this + prefix + VISIBILITY_PROPERTY_NAME+" "+v + " ---- "+props.get(prefix + VISIBILITY_PROPERTY_NAME));
        setVisible(v);
        // System.out.println("ResizeableGraphic setProps " + this);
    }

    public Properties getProperties(String prefix, Properties props) {
        if (props == null) {
            props = new Properties();
        }
        prefix = PropUtils.getScopedPropertyPrefix(prefix);

        //Position
        props.put(prefix + X_ANCHOR_PROPERTY_NAME, Integer.toString(getXAnchor()));
        props.put(prefix + Y_ANCHOR_PROPERTY_NAME, Integer.toString(getYAnchor()));

        //Size
        props.put(prefix + WIDTH_PROPERTY_NAME, Integer.toString(getWidth()));
        props.put(prefix + HEIGHT_PROPERTY_NAME, Integer.toString(getHeight()));

        //visibility
        // System.out.println("(GET) " + this + prefix + VISIBILITY_PROPERTY_NAME+" "+getVisible());
        props.put(prefix + VISIBILITY_PROPERTY_NAME, String.valueOf(getVisible()));

        // System.out.println(getClass()+ " " +prefix + WIDTH_PROPERTY_NAME+"="+Integer.toString(getWidth()));
        return props;
    }

    public Properties removeAnchorFromProperties(String prefix, Properties props) {
        if (removableAnchors) {
            prefix = PropUtils.getScopedPropertyPrefix(prefix);
            props.remove(prefix + X_ANCHOR_PROPERTY_NAME);
            props.remove(prefix + Y_ANCHOR_PROPERTY_NAME);
        }
        return props;
    }

    public Properties removeSizeFromProperties(String prefix, Properties props) {
        prefix = PropUtils.getScopedPropertyPrefix(prefix);
        props.remove(prefix + WIDTH_PROPERTY_NAME);
        props.remove(prefix + HEIGHT_PROPERTY_NAME);
        return props;
    }

    public void resetPositionProperties(String prefix, Properties props) {
        removeAnchorFromProperties(prefix, props);
    }

    public String toString() {
        return Util.className(this) + " (" + getXAnchor() + ", " + getYAnchor() + ")" + (getVisible() ? " v" : "");
    }

}
