// <header>
/**
 *  ThemaMap : The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
package fr.unicaen.openmap.layer.thematicDataDisplayLayer.omGraphics.legend;

import com.bbn.openmap.omGraphics.*;
import com.bbn.openmap.util.PropUtils;
import fr.unicaen.openmap.layer.thematicDataDisplayLayer.gui.*;
import fr.unicaen.openmap.layer.thematicDataDisplayLayer.omGraphics.*;
import fr.unicaen.openmap.layer.thematicDataDisplayLayer.util.*;
import fr.unicaen.openmap.util.Util;
import fr.unicaen.openmap.util.gui.ComponentSizer;
import java.awt.Color;
import java.awt.Paint;
import java.awt.event.*;
import javax.swing.*;
import java.awt.Dimension;
import java.util.Properties;
import java.util.ArrayList;

/** 
 * OMResizeableRegularPolygonLegend construct a legend for colored proportionnal polygons.
 * <pre>
 *     ..........--------value
 *     ..........
 *     ..........
 *     ..........
 *     ..........
 *     ..........
 * </pre>
 * The asciiart is copyrighted Yannick Lange ;-)</br>
 * In fact, there is no horizontal line between text ant disc in this version. 
 *
 * @author Jacques.Madelaine@info.unicaen.fr
 */
public class OMResizeableRegularPolygonLegend extends OMResizeableFrame {

    protected MinMeanMaxValuesAndSizes minMeanMaxValuesAndSizes;

    public static final String multiRectLegendPropertyString = "multiRectLegend";
    
    /** if columnName is null, draw only 1 polygon (because they are of fixed size. */
    public OMResizeableRegularPolygonLegend(Refreshable r, MouseControlPointsListener ml, int xAnchor, int yAnchor, Resizer resizer, MinMaxValues minMaxValues, int nbEdges, double startAngle, int starCoef, String columnName, Paint  lineColor, Paint fillColor, String name) {
	super(r, ml, xAnchor, yAnchor, 0, 0, columnName, true, name);
	//this.resizer = resizer;
        minMeanMaxValuesAndSizes =
		    new MinMeanMaxValuesAndSizes(resizer, minMaxValues);
	init(nbEdges, startAngle, starCoef, columnName, lineColor, fillColor);
    }

    public void init(int nbEdges, double startAngle, int starCoef, String columnName, Paint  lineColor, Paint fillColor) {

	int deltaY0 = clearAndAddTitle();

	final int offset =  getOMTitleHeight();

	int x0 = getXAnchor() + offset;
	int y0 = getYAnchor() + deltaY0 + offset;

        //0 index is the biggest polygon
	int biggestRadius = minMeanMaxValuesAndSizes.getSize(0);
	y0 += 2 * offset;

	int polygonIndex = minMeanMaxValuesAndSizes.getSize() - 1;
	if (columnName == null && polygonIndex > 1)
	    polygonIndex = 0; 

	double trueRadiusFactor = getTrueRadiusFactor(nbEdges, startAngle);
	int maxTextWidth = 0;
	// in reverse order so that the smallest polygon is on top.
	for(int i = polygonIndex; i >= 0; i--) {
	    int radius = minMeanMaxValuesAndSizes.getSize(i);
	    // Debug.message("Legend2", "LegendPolygon ["+(x0 + biggestRadius)+", "+ (y0 + 2*biggestRadius - radius) + "] " + radius + " " + biggestRadius+ " ["+(x0+biggestRadius)+", "+ (y0+2*biggestRadius-2*radius)+"] "+(minMeanMaxValuesAndSizes.getValue(i)));
	    OMResizeableGroup group = 
		new OMResizeableGroup(refreshable, ml, 
				      x0, (int)Math.round(y0 + 2 * biggestRadius - radius * trueRadiusFactor), 0, 0, "Elem "+i);

	    OMResizeableGraphic feature =
		newOMResizeableFeature(x0, y0, offset, radius, biggestRadius, trueRadiusFactor, nbEdges, startAngle, starCoef, lineColor, fillColor);

	    OMResizeableText text = 
		new OMResizeableText(refreshable, ml, 
				     x0 + 2 * biggestRadius + offset, 
				     (int)Math.round(y0 + 2 * biggestRadius - 2 * radius * trueRadiusFactor),
				     minMeanMaxValuesAndSizes.getValue(i), 
				     OMText.JUSTIFY_LEFT, "symbol value  " + i, true, true);

	    if (maxTextWidth < text.getWidth())
		maxTextWidth = text.getWidth();
	    group.add(feature);
	    group.add(text);
            this.add(group);
	}

	int frameWidth = (int)Math.max(getOMTitleWidth(), 2 * offset + 2 * biggestRadius + maxTextWidth);
	this.extendBorder(x0 + frameWidth, y0 + 2 * biggestRadius + offset);
	recenterTitle(getWidth());
    }

    /** Factor to apply to radius in order to bottom align the polygons 
     * By extension for discs, if nbEdges equals 0, returns 1 
     */
    private double getTrueRadiusFactor(int nbEdges, double startAngle) {
	if (nbEdges == 0)
	    return 1;
	double r = 0;
	double angle = startAngle;
	double angleIncr = 2*Math.PI/nbEdges;
	for(int i = 0; i < nbEdges; i++, angle += angleIncr) { 
	    double y = Math.sin(angle);
	    // System.out.println(y);
	    if (r < y)
		r = y;
	}
	// System.out.println(r);
	return r;
    }

    protected OMResizeableGraphic newOMResizeableFeature(int x0, int y0, int offset, int radius, int biggestRadius, double trueRadiusFactor, int nbEdges, double startAngle, int starCoef, Paint lineColor, Paint fillColor) {
	// mouseListener is null as you can't move or extend it individually, but with its group
	return new OMResizeableRegularPolygon(refreshable, null, 
					      x0 + biggestRadius, 
					      (int)Math.round(y0 + 2 * biggestRadius - radius * trueRadiusFactor), 
					      nbEdges, radius, startAngle, starCoef, 0 /* no line * biggestRadius + offset*/,
					      lineColor, fillColor);
    }

    protected boolean hasAutoAlign() { return true; }

}
