// <header>
/**
 *  ThemaMap : The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
package fr.unicaen.openmap.layer.thematicDataDisplayLayer.util;

import java.awt.Color;
import java.util.ArrayList;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.Properties;

import com.bbn.openmap.util.ColorFactory;

import fr.unicaen.openmap.layer.thematicDataDisplayLayer.gui.ColorScalePanel;

/**
 * ColorScale is a ColorScale that can be modified by the user
 * It is based on a model color scale. 
 */
public class ColorScale extends SimpleColorScale {
    private SortedSet<Integer> colorIsSetByUser;
    private SimpleColorScale shadedColorScale;
    private StandardColorScale model;
    public ColorScale() {
        this(-1, defaultColorScaleInHex, null);
    }

    public ColorScale(int colorNumber, String colorScaleInHex, StandardColorScale model) {
        super(colorNumber, colorScaleInHex);
	resetColorSetByUser();
	if (model == null)
	    model = ColorScalePanel.getStandardShadedColorScale();
	this.model = model;
    }

    public void setProperties(String prefix, Properties properties) {
	super.setProperties(prefix, properties);
	model = null;
	shadedColorScale = null;
	setShaded(false);
    }

    public StandardColorScale getModel() {
	return model;
    }

    public void setModel(StandardColorScale model) {
	this.model = model;
	model.setColorNumber(getColorNumber());
	// copy the colors from the model
	for(int i = 0; i < getColorNumber(); i++)
	    super.setColor(i, model.getColor(i));
	shaded = false;
    }

    private Color getShade(Color min, Color max, double ratio) {
	return new Color(
			 min.getRed() + (int) (ratio * (max.getRed() - min.getRed())),	 
			 min.getGreen() + (int) (ratio * (max.getGreen() - min.getGreen())),	 
			 min.getBlue() + (int) (ratio * (max.getBlue() - min.getBlue())),	 
			 min.getAlpha() + (int) (ratio * (max.getAlpha() - min.getAlpha())));	 
    }

    /**
     * init a shade color map given the colors customized by the user plus the two extremities 
     */
    private void initShadedColorScale() {
	boolean _shaded = shaded;
	shaded = false;
	shadedColorScale = new SimpleColorScale(getColorNumber());
	if (colorIsSetByUser == null) {
	    doShade(0, getColorNumber()-1);
	} else {	    
	    int imin = colorIsSetByUser.first();
	    for(Integer index : colorIsSetByUser) {
		if (imin  == index) 
		    continue;
		doShade(imin, index);
		imin = index;
	    }
	}
	shaded = _shaded;
    }

    // shade between colors at indexes imin and imax 
    private void doShade(int imin, int imax) {
	// System.out.println(colorScaleToString());
	Color min = super.getColor(imin);
	Color max = super.getColor(imax);
	// System.out.println(imin + " " + min+" - "+imax + " "+ max );
	for (int i = imin; i <= imax; i++) {
		shadedColorScale.setColor(i, getShade(min, max, ((double)(i-imin))/(imax-imin)));
		// System.out.println(i + " > " +   getShade(min, max, ((double)(i-imin))/(imax-imin)));
	}
    }

    public void setColorNumber(int nb) {
	int i = getColorNumber();
	// com.bbn.openmap.util.Debug.message("Colorizer", "ColorScale setColorNumber to " + nb + " was "+i + "  "+ this);
	// when shaded just use the last color
	if (shaded) { 
	    for (; i < nb; i++)
		super.setColor(i, Color.WHITE);
	    super.setColor(nb-1, super.getColor(colorNumber-1)); 
	    colorNumber = nb;
	    initShadedColorScale();
	    return;
	}

	// use the model for the first colors
	if (model != null) {
	    model.setColorNumber(nb);
	    // if not modified by the user
	    if (colorIsSetByUser == null) // reset the entire colorscale
		i= 0;
	    for (; i < nb && i < model.getColorNumber(); i++)
		super.setColor(i, model.getColor(i));
	}
	// add random colors if necessary
        for (; i < nb; i++) 
            super.setColor(i, new Color((int) (Math.random() * 255), (int) (Math.random() * 255), (int) (Math.random() * 255)));

        colorNumber = nb;
    }

    private int getIndex(int index) {
	return getReverse() ? getColorNumber() - index -1 : index;
    }

    public Color getColor(int i) {
	int index = getIndex(i);
	if (shaded)
	    return shadedColorScale.getColor(index);
	return super.getColor(index);
    }

    public void setColor(int index, Color color) {
	if (colorIsSetByUser == null) {
	    colorIsSetByUser = new TreeSet<Integer>();
	    colorIsSetByUser.add(0); 
	    colorIsSetByUser.add(getColorNumber()-1); 
	}
	colorIsSetByUser.add(index);
	super.setColor(index, color);
	if (getShaded()) 
	    initShadedColorScale();
    }

    public void resetColorSetByUser() {
	colorIsSetByUser = null;
    }

    private boolean reverse = false;

    public void setReverse(boolean reverse) {
	this.reverse = reverse;	
    }

    public boolean getReverse() {
        return reverse;
    }

    private boolean shaded = false;

    public void setShaded(boolean b) {
	if (b) 
	    initShadedColorScale();
	else
	    resetColorSetByUser();
        this.shaded = b;
    }

    public boolean getShaded() {
        return shaded;
    }

    
    public String getName() {
	return "custom";
    }

    public static Color negative(Color c) {
        return new Color(255 - c.getRed(), 255 - c.getGreen(), 255 - c.getBlue(), c.getAlpha());
    }

}
