// <header>
/**
 *  ThemaMap : The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
package fr.unicaen.openmap.layer.thematicDataDisplayLayer.util;

import java.awt.Color;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Properties;
import java.lang.reflect.Constructor;
import javax.swing.*;
import javax.swing.event.*;

import com.bbn.openmap.util.PropUtils;
import com.bbn.openmap.util.Debug;
import com.bbn.openmap.util.ColorFactory;
import com.bbn.openmap.omGraphics.*;

import fr.unicaen.openmap.util.classification.*;
import fr.unicaen.openmap.dataStorage.*;
import fr.unicaen.openmap.layer.thematicDataDisplayLayer.gui.*;

/**
 * A Colorizer is made of a color scale (i.e. a fixed number of colors) and a
 * classification.
 * <p>
 * May be used to assign a color to a graphical object, given a value.
 * <p>
 * Default colors are:
 * 
 * <pre>
 * c9ffb300 c9ec93c1 c974c476 c980a6ce c9ffff4d c9f4c6de c9a6dba0 c9bfd1e5
 * </pre>
 * 
 * <p>
 * May be set by properties. Ex.:<br>
 * 
 * <pre>
 * prefix.colorScale = c9ffb300 c9ec93c1 c974c476 c980a6ce c9ffff4d c9f4c6de c9a6dba0 c9bfd1e5
 * prefix.classificationClassName = fr.unicaen.openmap.util.classification.NestedMeans
 * prefix.nbClasses = 4
 * </pre>
 */
public class Colorizer implements ColorScaleListener {

    private ColorScale colorScale = new ColorScale();
    private Classification classif;
    public static final String classificationClassNamePropertyString = "classificationClassName";
    public static final String nbClassesPropertyString = "nbClasses";
    /**
     * For CustomizedClassification.
     */
    public static final String BREAKS_PROPERTY_NAME = "breaks";

    private HashSet<ClassificationListener> listenerList = new HashSet<ClassificationListener>();

    public void addClassificationListener(ClassificationListener l) {
        listenerList.add(l);
    }

    public void removeClassificationListener(ClassificationListener l) {
        listenerList.remove(l);
    }

    protected void fireClassificationChanged() {
	// fire only if classification has data ???
	if (classif == null || classif.hasNoData())
	    return;
        int i = 0;
        for (ClassificationListener l : listenerList) {
            Debug.message("Colorizer", "##Colorizer.fireClassificationChanged " + (i++) + " => " + l + " @" + l.hashCode());
            l.classificationChanged(this);
        }
    }

    private HashSet<ColorScaleListener> csListenerList = new HashSet<ColorScaleListener>();

    public void addColorScaleListener(ColorScaleListener l) {
        csListenerList.add(l);
    }

    public void removeColorScaleListener(ColorScaleListener l) {
        csListenerList.remove(l);
    }

    protected void fireColorScaleChanged() {   
        int i = 0;
        for (ColorScaleListener l : csListenerList) {
            Debug.message("Colorizer", "##Colorizer.fireColorScaleChanged " + (i++) + " => " + l);
            l.colorScaleChanged(null); // we can't use colorScalePanel as it always changes!
        }
    }

    public void setProperties(String prefix, Properties properties) {
        colorScale.setProperties(prefix, properties);
        String classificationClassName =
                properties.getProperty(prefix + classificationClassNamePropertyString, null);
        int nbClasses =
                PropUtils.intFromProperties(properties, prefix + nbClassesPropertyString, 4);
        if (classificationClassName != null)
            setClassification(classificationClassName, nbClasses);
       
	// useful in custom mode
	String[] breaksStringArray = PropUtils.stringArrayFromProperties(properties, 
									 prefix + BREAKS_PROPERTY_NAME, "\t");
	if (breaksStringArray != null)
	    classif.setBreaks(breaksStringArray);
    }
 
    private Constructor getClassifConstructor(String className)
            throws java.lang.ClassNotFoundException {
        Class cls = Class.forName(className);
        for (Constructor c : cls.getConstructors()) {
            java.lang.reflect.Type[] params = c.getGenericParameterTypes();
            if (params.length == 2 
		&& params[0] instanceof java.lang.reflect.ParameterizedType
		&& ((java.lang.reflect.ParameterizedType)params[0]).getRawType() == java.util.ArrayList.class
		&& params[1] == int.class ) {
                return c;
            }
        }
	Debug.error("** No Constructor for: " + className);
        return null;
    }

    private void setClassification(String classificationClassName, int nbClasses) {
        Debug.message("Colorizer", "#Colorizer.setClassification("+classificationClassName+","+nbClasses+")");
        Classification oldClassif = classif;
	try {
	    if (classif != null && classif.getClass() == Class.forName(classificationClassName)
		&& nbClasses == classif.getNumberOfClasses()) {
		Debug.message("Colorizer", "#Colorizer.setClassification discarded for " + classif);
		return;
	    }
	} catch(ClassNotFoundException e) {
	    Debug.error("** Error classification class not found: " + classificationClassName);
	    return;
	}

        try {
            classif = (Classification) getClassifConstructor(classificationClassName).newInstance(new Object[]{null, nbClasses});

            if (oldClassif != null) {     
		// System.out.println("#Colorizer.setClassification reinit "+ classif + " with " + oldClassif);
                classif.reInit(oldClassif);
            }

        } catch (Exception e) {
            Debug.error("[Colorizer] *** construction failure for " + classificationClassName);
            e.printStackTrace();
	    if (e.getCause() != null) {
		System.err.print("Caused by ");
		e.getCause().printStackTrace();
	    }
            classif = oldClassif;
        }
	// System.out.println("#Colorizer.setClassification =>"+classif);
    }

    public void resetClassification(String classificationClassName) {
        setClassification(classificationClassName, classif != null ? classif.getNumberOfClasses() 
			  : (colorScale != null && colorScale.getColorNumber() != 0) ? colorScale.getColorNumber() : 6);
        // fireClassificationChanged(); already done by setClassification
    }

    public Properties getProperties(String prefix, Properties properties) {
        colorScale.getProperties(prefix, properties);

        if (classif != null) {
            properties.put(prefix + classificationClassNamePropertyString, classif.getClassName());
            properties.put(prefix + nbClassesPropertyString, String.valueOf(classif.getNumberOfClasses()));
	    if (classif instanceof CustomizedClassification)
		properties.put(prefix + BREAKS_PROPERTY_NAME, classif.getBreaksAsString());
        }

        return properties;
    }

    public void setData(DataStorageIterator dsi, String dataColumnName) {
        if (classif == null) {
            return;
        }
        Debug.message("Colorizer", "Colorizer -> CLASSIF ReInit " + classif + " * dataColName " + dataColumnName);
	
        classif.reInit(dsi, dataColumnName);
        fireClassificationChanged();
    }

    public Color getColor(Object value) {
        if (classif != null) {
	    if (classif instanceof ContinuousClassification) {
		ContinuousClassification.ClassRankAndRatio crar = 
		    ((ContinuousClassification)classif).getClassRankAndRatio(value);
		if (crar == null) 
		    return Color.WHITE;
		int j = crar.getRank();
		double ratio = crar.getRatio();
		Color cj = colorScale.getColor(j);
		Color cj_1 = colorScale.getColor(j+1);
		// System.out.println("ContinuousClassification of "+value + " class " + j + ", ratio " + ratio);
		return new Color(
				 cj.getRed() + (int)((cj_1.getRed()-cj.getRed())*ratio),
				 cj.getGreen() + (int)((cj_1.getGreen()-cj.getGreen())*ratio),
				 cj.getBlue() + (int)((cj_1.getBlue()-cj.getBlue())*ratio),
				 cj.getAlpha() + (int)((cj_1.getAlpha()-cj.getAlpha())*ratio));
	    }
	    // System.out.println(classif + " " + value + " ==> " + classif.getClass(value));
	    return colorScale.getColor(classif.getClass(value));
        }
        return Color.BLACK;
    }

    private ColorizerPanel colorizerPanel; 
    public ColorizerPanel getGUI() {
        colorizerPanel = new ColorizerPanel(this);
        // p.refreshPanel(); no refresh we always rebuild it
        return colorizerPanel;
    }

    public boolean hasForColorScalePanel(ColorScalePanel colorScalePanel) {
	return colorizerPanel != null && colorizerPanel.hasForColorScalePanel(colorScalePanel);
    }

    public ColorScale getColorScale() {
        return colorScale;
    }

    public void setColorScale(ColorScale c) {
        colorScale = c;
        // Classification didn't really change, but we need to propagate colorScale changes.
        fireClassificationChanged();
    }

    public Classification getClassification() {
        return classif;
    }

    public String toString() {
        return "ColorScale: " + colorScale + " Classif:" + (classif == null ? "null" : classif.toString());
    }

    /** get colormap button text from the breaks get from the classification */
    public String getColorText(int i) {
        if (getClassification() != null) {
            return getClassification().getClassText(i);
        }
        return " ";
    }

    public String[] getColorsTexts() {
        if (getClassification() == null) {
            return null;
        }
        return getClassification().getClassesTexts();
    }

    public void setClassification(Classification classif) {
	// System.out.println("=> #Colorizer: set classif " + classif);
        if (this.classif != null) {
	    if (this.classif.getClass() == classif.getClass())
		return;
	    if (this.classif instanceof CustomizedClassification && classif instanceof CustomizedClassification)
		return;
        }
	
        this.classif = classif;

	fireClassificationChanged();
        // System.out.println("<= #Colorizer: set classif " + this.classif);
    }

    public void colorScaleChanged(ColorScalePanel colorScalePanel) {
	if (hasForColorScalePanel(colorScalePanel))
	    fireColorScaleChanged();
    }

}
