// <header>
/**
 *  ThemaMap : The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
package fr.unicaen.openmap.layer.thematicDataDisplayLayer.util;

import java.awt.Font;
import java.util.*;

import com.bbn.openmap.omGraphics.OMText;
import com.bbn.openmap.util.PropUtils;
import com.bbn.openmap.proj.Projection;

/**
 * FontSizer that uses a list of size/scales (intead as in OpenMap that uses a base scale and a multiplier) 
 * @see FontSizerPanel
 * @see fr.unicaen.openmap.layer.thematicDataDisplayLayer.visu.Label
 */
public class FontSizer extends com.bbn.openmap.omGraphics.FontSizer {

    /**
     * Property name for font name.
     */
    public final static String FONT_PROPERTY_NAME = "font";
    /**
     * Property name for font size.
     */
    public final static String FONT_SIZE_PROPERTY_NAME = "font.size";
    /**
     * Property name for font style.
     */
    public final static String FONT_STYLE_PROPERTY_NAME = "font.style";
     /**
     * Property name for total number of ScaleAndSize to keep in properties.
     */
    public final static String TOTAL_PROPERTY_NAME = "fsp.total";
    /**
     * Property name for scale values to keep in properties.
     */
    public final static String SCALE_PROPERTY_NAME = "fsp.scale";
    /**
     * Property name for font size values to keep in properties.
     */
    public final static String SIZE_PROPERTY_NAME = "fsp.size";

    protected transient Font font = OMText.DEFAULT_FONT;
    protected transient Font lastFont = null;

    public FontSizer() {
	super(OMText.DEFAULT_FONT, 1f, 1, 12, 12);
    }

    /**
     * Set the base font.
     */
    public void setFont(Font font) {
        this.font = font;
	lastFont = null;
    }
  /**
     * Get the base font.
     */
    public Font getFont() {
	return font;
    }

    public Font getFont(float scale) {
	return font;
    }

    public Font getScaledFont() {
	return font;
    }

    private int currentFontSize = 0;
    /**
     * Get a font sized for the scale given by the projection.
     */
    public Font getFont(Projection proj) {
        if (proj == null || scales == null ||  scales.size() == 0)
	    return getFont();
	 
	float scale = proj.getScale();
	// compute font size for that projection scale

	//Default value, out of range : no display, usefull
        int size = 0;

	ScaleAndSize lowerSaS = null;
	for (ScaleAndSize current : scales) {
	    // System.out.println(lowerSaS+ " --- " + current + " scale " + scale + " -> " + size);
	    if (scale < current.getScale()) {
		size = lowerSaS == null 
		    ? current.getSize()
		    : current.getSize()
		    - (int)((current.getScale()-scale)/(current.getScale()-lowerSaS.getScale()) * (current.getSize()-lowerSaS.getSize())  );
		// System.out.println(lowerSaS+ " --- " + current + " scale " + scale + " -> " + size);
		break;
	    }
	    lowerSaS = current;
	}
	if (lastFont == null || size != lastFont.getSize())
	    lastFont = new Font(font.getName(), font.getStyle(), size);
	// System.out.println(font+" "+lastFont);
	return lastFont;
    }

    private ArrayList<ScaleAndSize> scales = new ArrayList<ScaleAndSize>();

    private int getMaxFontSize() {
	return scales.size() > 0 ? scales.get(0).getSize() : font.getSize();
    }


    public void addScale() {
	addScale(new ScaleAndSize(1f, getMaxFontSize()));
    }

    public int getScalesSize() {
	return scales.size();
    }

    public ScaleAndSize getScale(int i) {
	return scales.get(i);
    }

    public void removeScale(int i) {
	scales.remove(i);
    }

    /**
     * Store ScaleAndSize behind scale, in ascendant order.
     * @param scaleAndSize contains font size to use with max scale 
     */
    protected void addScale(ScaleAndSize scaleAndSize) {
        for (int i = 0; i < scales.size(); i++)
            if (scaleAndSize.getScale() < scales.get(i).getScale()) {
		// insert at the current position
                scales.add(i, scaleAndSize);
                return;
            }
	// otherwise add to the end
	scales.add(scaleAndSize);
    }

    /**
     * Used to set FontSizerShadow with properties resources.
     * @param prefix string prefix used in the properties
     * @param properties used to retrieve properties
     */
    public void setProperties(String prefix, Properties properties) {
	setFont(new Font(properties.getProperty(prefix + FONT_PROPERTY_NAME, OMText.DEFAULT_FONT.getName()),
			 PropUtils.intFromProperties(properties, prefix + FONT_STYLE_PROPERTY_NAME, OMText.DEFAULT_FONT.getStyle()),
			 PropUtils.intFromProperties(properties, prefix + FONT_SIZE_PROPERTY_NAME, OMText.DEFAULT_FONT.getSize())));

	int total = PropUtils.intFromProperties(properties, prefix + TOTAL_PROPERTY_NAME, 1);
        if (total <= 0) 
	    total = 1;
	scales.clear();
	for (int i = 0; i < total; i++) 
	    addScale(
		     new ScaleAndSize(
				      PropUtils.floatFromProperties(properties, prefix + SCALE_PROPERTY_NAME + String.valueOf(i), 200000000f),
				      PropUtils.intFromProperties(properties, prefix + SIZE_PROPERTY_NAME + String.valueOf(i), getFont().getSize())));
	
    }

    /**
     * Used to keep properties from FontSizerShadow.
     * @param prefix string prefix used in the properties
     * @param properties used to load properties
     * @return Properties.
     */
    public Properties getProperties(String prefix, Properties properties) {
	//Font
	properties.put(prefix + FONT_PROPERTY_NAME, getFont().getName());
	properties.put(prefix + FONT_STYLE_PROPERTY_NAME, String.valueOf(getFont().getStyle()));
	properties.put(prefix + FONT_SIZE_PROPERTY_NAME, String.valueOf(getFont().getSize()));
	properties.put(prefix + TOTAL_PROPERTY_NAME, String.valueOf(scales.size()));
        for (int i = 0; i < scales.size(); i++) {
            properties.put(prefix + SCALE_PROPERTY_NAME + String.valueOf(i), String.valueOf(scales.get(i).getScale()));
            properties.put(prefix + SIZE_PROPERTY_NAME + String.valueOf(i), String.valueOf(scales.get(i).getSize()));
        }
	
	return properties;
    }

    /**
     * Association of scale and font size to use. Just provide accessors.
     */
    static public class ScaleAndSize {

        private float scale;
        private int size;

        public ScaleAndSize(float scale, int size) {
            this.scale = scale;
            this.size = size;
        }

        public float getScale() {
            return scale;
        }

        public int getSize() {
            return size;
        }

        public void setScale(float scale) {
            if (scale < 1)
                return;
            this.scale = scale;
        }

        public void setSize(int size) {
            if (size < 0)
                return;
            this.size = size;
        }

        public String toString() {
            return String.valueOf(scale) + " -- " + String.valueOf(size);
        }
    }

    private List<FontSizerListener> listenerList = new ArrayList<FontSizerListener>();

    public void addFontSizerListener(FontSizerListener l) {
        listenerList.add(l);
    }

    public void removeFontSizerListener(FontSizerListener l) {
        listenerList.remove(l);
    }

    public void fireFontSizerChanged() {
        for (FontSizerListener l : listenerList)
	    l.fontSizerChanged(this);
    }

}
