// <header>
/**
 *  ThemaMap : The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
package fr.unicaen.openmap.layer.thematicDataDisplayLayer.util;


import com.bbn.openmap.event.MapMouseListener;
import com.bbn.openmap.omGraphics.OMText;
import com.bbn.openmap.omGraphics.OMGraphic;
import com.bbn.openmap.omGraphics.OMGraphicList;
import com.bbn.openmap.proj.Projection;
import com.bbn.openmap.util.PropUtils;
import com.bbn.openmap.util.ColorFactory;

import java.awt.Color;
import java.awt.Font;
import java.util.Properties;
import java.util.ArrayList;
import java.util.List;
import javax.swing.JComponent;

import fr.unicaen.openmap.dataStorage.Centroid;
import fr.unicaen.openmap.dataStorage.DataObject;
import fr.unicaen.openmap.dataStorage.DataStorageIterator;
import fr.unicaen.openmap.layer.thematicDataDisplayLayer.gui.LabelizerGUI;
import fr.unicaen.openmap.layer.thematicDataDisplayLayer.gui.ThematicMouseControlPointsListener;
import fr.unicaen.openmap.layer.thematicDataDisplayLayer.omGraphics.OMLabel;
import fr.unicaen.openmap.layer.thematicDataDisplayLayer.visu.Offsetable;
import fr.unicaen.openmap.layer.thematicDataDisplayLayer.visu.Thematic;
import fr.unicaen.openmap.layer.thematicDataDisplayLayer.visu.VisuWithMouseListener;
import fr.unicaen.openmap.layer.thematicDataDisplayLayer.ThematicDataDisplayLayer;

/**
 * Utility class to manage text labels : font, color, offsets
 */
public class Labelizer implements Offsetable, FontSizerListener, OffseterListener {
    public static final String dataColumnNamePropertyString = "dataColumnName";
    public static final String onOffPropertyString = "onOff";
    public static final String fontColorPropertyString = "fontColor";

    private int alignment = OMText.JUSTIFY_CENTER;
    /** internally keep angle in radians ans in degrees to avoid conversion anomalies */
    private double angleInRadians = 0;
    private double angleInDegrees = 0;
    protected FontSizer fontSizer;
    protected Offseter offseter;
    protected MutableString dataColumnName = new MutableString("data");
    private Color fontColor = null;
    /** It is the Thematic thats uses a Labelizer that decides if it is always on or not */
    public final boolean alwaysOn;
    protected boolean onOff = false; // a property

    public Labelizer(VisuWithMouseListener visu) {
	this(visu, true);
    }

    public Labelizer(VisuWithMouseListener visu, boolean alwaysOn) {
	this.visu = visu;
	this.alwaysOn = alwaysOn;
	fontSizer = new FontSizer();
	fontSizer.addFontSizerListener(this);
	offseter = new Offseter();
	offseter.addOffseterListener(this);
    }

    /** 
     * create an OMLabel with the text adhorned with the current Labelizer state.
     * The DataObject is used by the Offseter
     */
    public OMLabel createOMLabel(Centroid centroid, DataObject data, Projection proj) {
	Offseter.OffsetLocationAndKey offset = offseter.getOffsetLocation(data);
	Object dataValue = data.getData(dataColumnName.get());
	if (dataValue == null) {
	    // System.out.println(i + " null data value for " + dataColumnName);
	    return null;
	}

	OMLabel oml =  new OMLabel(this, offset.getKey(), 
				   centroid.getLat(), centroid.getLon(),
				   offset.getX(), offset.getY(),
				   dataValue.toString(),
				   fontSizer.getFont(proj),
				   alignment, angleInRadians,
				   offseter.isVisible(), offseter.isLineVisible());

	if (fontColor != null)
	    oml.setLinePaint(fontColor);

	return oml;
    }

    private VisuWithMouseListener visu;
    public VisuWithMouseListener getVisu() { return visu; }

    public boolean isAlwaysColorized() { return false; }

    public boolean isAlwaysOn() {
	return alwaysOn;
    }

    public boolean isOnOff() {
	return alwaysOn || onOff;
    }

    public void setOnOff(boolean b) { 
	if (alwaysOn)
	    return;
	onOff = b;

	enableGUI();
	refresh();
    }

    protected void enableGUI() {
	if (labelizerGUI != null)
	    labelizerGUI.setEnabled(isOnOff());
    }

    public Color getFontColor() {
	return fontColor;
    }

    public void setFontColor(Color color) {
	// System.out.println("Labelizer setFontColor "+ color + " for " + visu);
	fontColor = color;
    }

    public void setAlignment(int alignment) {
	this.alignment = alignment;
    }

    public int getAlignment() { 
	return alignment; 
    }

    /**
     * In degrees
     */
    public double getRotationAngle() {
        return angleInDegrees;
    }

    public void setRotationAngle(double a) {
	angleInDegrees = a;
	angleInRadians = Math.toRadians(a);
    }

    public Font getFont() { 
	return  fontSizer.getFont(); 
    }
    
    public void setFont(Font font) { 
	fontSizer.setFont(font); 
    }

    public FontSizer getFontSizer() { 
	return  fontSizer; 
    }

    public MutableString getDataColumnName() {
	return dataColumnName;
    }

    private LabelizerGUI labelizerGUI;

    /** provides GUI with several Panels */
    public LabelizerGUI getLabelizerGUI(DataStorageIterator dsi) {
	if (labelizerGUI == null)
	    labelizerGUI = new LabelizerGUI(this, dsi);
	enableGUI();
	return labelizerGUI;
    }

    /** use getLabelizerGUI() instead */
    public JComponent getGUI(DataStorageIterator dsi) {
	return null;
    }
    

    public void refreshGUI(DataStorageIterator dsi) {     
	offseter.refreshGUI(dsi);
	refreshLabelizerGUI();
    }

    public void refreshLabelizerGUI() {
	if (labelizerGUI != null)
	    labelizerGUI.refreshGUI();
    }

    public static final String rotationAnglePropertyString = "rotationAngle";
    public static final String alignmentPropertyString = "alignment";

    public void setProperties(String prefix, Properties properties) {
        getFontSizer().setProperties(prefix, properties);
	setFontColor(ColorFactory.parseColor(properties.getProperty(prefix + fontColorPropertyString, "#FF000000")));
        setRotationAngle(PropUtils.doubleFromProperties(properties, prefix + rotationAnglePropertyString, 0));
        alignment = PropUtils.intFromProperties(properties, prefix + alignmentPropertyString, OMText.JUSTIFY_CENTER);
	dataColumnName.set(properties.getProperty(prefix + dataColumnNamePropertyString, "data"));
	offseter.setProperties(prefix, properties);
	onOff = Boolean.parseBoolean(properties.getProperty(prefix + onOffPropertyString, "false"));
  }

    public Properties getProperties(String prefix, Properties properties) {
	properties = getFontSizer().getProperties(prefix, properties);
	properties.put(prefix + fontColorPropertyString, ColorFactory.getHexColorString(getFontColor()));
        properties.put(prefix + rotationAnglePropertyString, Double.toString(getRotationAngle()));
        properties.put(prefix + alignmentPropertyString, Integer.toString(alignment));
        properties.put(prefix + dataColumnNamePropertyString, dataColumnName.get());
	properties = offseter.getProperties(prefix, properties);
	properties.put(prefix + onOffPropertyString, String.valueOf(onOff));
       return properties;
    }

    public void fireLabelizerChanged() {
	refresh();
    }

    public ThematicMouseControlPointsListener getMouseControlPointsListener() {
        return visu.getMouseControlPointsListener();
    }

    public MapMouseListener getMapMouseListener() {
        return visu.getMapMouseListener();
    }

    public Offseter getOffseter() {
        return offseter;
    }

    public void offseterChanged(Offseter offseter) {
	// System.out.println("Labelizer offseterChanged " + offseter + " == " + getOffseter());
        if (this.offseter == offseter) {
            refresh();
        }
    }

    // Using visu as a proxy for Offsetable

    public void clearInteractiveGraphics() {
	visu.clearInteractiveGraphics();
    }

    public void refresh() {
	visu.refresh();
    }

    public OMGraphicList getInteractiveGraphics() {
	return visu.getInteractiveGraphics();
    }

    public void addInteractiveGraphic(OMGraphic omg) {
	visu.addInteractiveGraphic(omg);
    }

    public void removeInteractiveGraphic(OMGraphic omg) {
	visu.removeInteractiveGraphic(omg);
    }

    public void removeGraphic(OMGraphic g) {
	visu.removeGraphic(g);
    }

    public ThematicDataDisplayLayer getLayer() {
	return visu.getLayer();
    }

    public void fontSizerChanged(FontSizer fs) {
	// System.out.println("Labelizer fontSizerChanged " + fs + " == " + getFontSizer());
	if (fs == getFontSizer())
	    refresh();
    }

}
