// <header>
/**
 *  ThemaMap : The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
package fr.unicaen.openmap.layer.thematicDataDisplayLayer.util;

import com.bbn.openmap.util.Debug;
import java.util.Properties;
import java.text.NumberFormat;

/** grab max, min and mean values and size to use in Legend 
 *  for OMDisc, OMFlow, OMPie, OMHistogram or OMGraph.
 */
public class MinMeanMaxValuesAndSizes {

    public enum typeValue {

        MIN(2),
        MEAN(1),
        MAX(0);
        private int index;

        typeValue(int index) {
            this.index = index;
        }

        public int getIndex(){
            return index;
        }
    }
    /** minimum size of a symbol in the legend */
    public static int MINSIZE = 2;
    /** maximum size of a symbol in the legend */
    public static int MAXSIZE = 60;
    public static final String valuePropertyString = "value";

    static private class ValueAndSize {

        public ValueAndSize(double numericalValue, int size) {
            this.size = size;
            this.value = NumberFormat.getInstance().format(numericalValue);
            this.numericalValue = numericalValue;
        }

        public String toString() {
            return "(" + value + " = " + numericalValue + " -> " + size + ")";
        }
        public String value;
        public int size;
        public double numericalValue;
    }

    private ValueAndSize[] data;
    private Resizer resizer;

    public MinMeanMaxValuesAndSizes(Resizer resizer, MinMaxValues minMaxValues) {
        /*  Ideas for rounding values :
         *    take the power of ten immediately less to the max-min difference
         *    round the max and min to a multiple of this power
         *    for the median value take (max-min)/2 or max/2 if min is less than 30% of max
         *    ...
         */
        this.resizer = resizer;
        double max = minMaxValues.getPMax();
        double min = minMaxValues.getPMin();
        double unit = roundToPowerOf10(max - min, 0);
        Debug.message("Legend", "MinMeanMaxValuesAndSizes " + max + " - " + min + " = " + unit + " [ " + minMaxValues);
        if ((max - min) < 1) {
            data = new ValueAndSize[1];
            data[0] = new ValueAndSize(max, (int) Math.round(resizer.getSize(max)));
            return;
        }
        // round the max
        max = (Math.round(max / unit)) * unit;
        // round down the min
        min = Math.floor(min / unit) * unit;

        int maxSize = (int) Math.round(resizer.getSize(max));
        int minSize = (int) Math.round(resizer.getSize(min));

        Debug.message("Legend2", "MinMeanMaxValuesAndSizes " + unit + " " + min + " < " + max + " --> " + minSize + " < " + maxSize);

        if ((maxSize - minSize) < 1) {
            data = new ValueAndSize[1];
            data[0] = new ValueAndSize(max, maxSize);
            return;
        }

        while (maxSize > MAXSIZE) {
            max -= unit;
            maxSize = (int) Math.round(resizer.getSize(max));
        }
        // What if max == 0 ?
        if (maxSize < MINSIZE) {
            unit /= 10.0;
            max = unit;
            while (maxSize <= MAXSIZE) {
                max += unit;
                maxSize = (int) Math.round(resizer.getSize(max));
            }
        }
        // What if max equals 1 unit ?
        if (max < 3 * unit) {
            unit /= 10.0;
        }

        // store the actual minimum
        ValueAndSize minValueAndSize = new ValueAndSize(min, minSize);

        // try to adjust a minimum with a size larger than MINSIZE
        //           and not too large compare to maxSize
        while (true) {
            while (minSize < MINSIZE) {
                min += unit;
                minSize = (int) Math.round(resizer.getSize(min));
            }

            if (minValueAndSize.size > maxSize / 3 || minSize <= maxSize / 3) {
                break; // minSize ok regarding maxSize
            }
            // what if minSize too large ?
            unit /= 10;
            min = minValueAndSize.numericalValue;
            minSize = minValueAndSize.size;

            // System.out.println(unit + " " + min + " < " + max + " --> " + minSize + " < " + maxSize);
        }

        // System.out.println(unit + " " + min + " " + max);

        // keep only two values if size are too close
        int nbIndex = ((maxSize - minSize) < 5) ? 2 : 3;

        data = new ValueAndSize[nbIndex];
        data[0] = new ValueAndSize(max, maxSize);
        data[nbIndex - 1] = new ValueAndSize(min, minSize);
        if (nbIndex == 3) {
            data[1] = newMeanValueAndSize(data[nbIndex - 1], data[0], resizer);
        }

        Debug.message("Legend", this.toString());
    }

    private ValueAndSize newMeanValueAndSize(ValueAndSize max, ValueAndSize min, Resizer resizer) {
        // if min is to small don't take it into account (in order to get a roud mean)
        double value = min.size <= 2
                ? max.numericalValue / 2
                : min.numericalValue + (max.numericalValue - min.numericalValue) / 2;
	Debug.message("Legend2", "newMeanValueAndSize (min.size " + min.size + ") " + min + " < " + max + " --> " + value);
        return new ValueAndSize(value, (int) Math.round(resizer.getSize(value)));
    }

    private double roundToPowerOf10(double d, int up) {
        return Math.pow(10, up + (int) Math.floor(Math.log10(d)));
    }

    public int getSize() {
        return data == null ? 0 : data.length;
    }

    public int getSize(int i) {
        return data == null || i >= data.length || data[i] == null ? 0 : data[i].size;
    }

    public String getValue(int i) {
        return data[i] == null ? "" : data[i].value;
    }

    public double getNumericValue(int i) {
        return data[i] == null ? 0 : data[i].numericalValue;
    }

    public typeValue getTypeValue(int index){
        for(typeValue v : typeValue.values()){
            if(v.getIndex() == index)
                return v;
        }
        return null;
    }

    public String toString() {
        String ret = "";
        for (int i = 0; i < getSize(); i++) {
            ret += "" + i + " " + data[i] + "\n";
        }
        return ret;
    }

//    public Component getGUI(final int index){
//        JPanel p = new JPanel();
//        SpinnerNumberModel spModel = new SpinnerNumberModel();
//        final JSpinner spinValue = new JSpinner(spModel);
//        spinValue.addChangeListener(new ChangeListener() {
//            public void stateChanged(ChangeEvent arg0) {
//                data[index] = new ValueAndSize((Integer)spinValue.getValue(), data[index].size);
//            }
//        });
//        p.add(new JLabel("Value: "));
//        p.add(spinValue);
//        return p;
//
//    }
    public void setProperties(String prefix, Properties properties) {
        for (int i = 0; i < data.length; i++) {
            ValueAndSize vas = data[i];
	    String valueAsString = properties.getProperty(prefix + valuePropertyString + "." + i);
            double value = valueAsString == null ? vas.numericalValue : Double.valueOf(valueAsString);
            data[i] = new ValueAndSize(value, (int)resizer.getSize(value));
        }

    }

    public Properties getProperties(String prefix, Properties properties) {
        for (int i = 0; i < data.length; i++) {
            ValueAndSize vas = data[i];
            properties.put(prefix + valuePropertyString + "." + i, Double.toString(vas.numericalValue));
        }
        return properties;
    }

    public void setResizer(Resizer resizer) {
        for (ValueAndSize vas : data) {
            vas.size = (int) resizer.getSize(vas.numericalValue);
        }
    }

    public Resizer getResizer(){
        return resizer;
    }

    public void setValue(typeValue t, double value) {
        data[t.getIndex()] = new ValueAndSize(value, (int) resizer.getSize(value));
    }

}
