// <header>
/**
 *  ThemaMap : The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
package fr.unicaen.openmap.layer.thematicDataDisplayLayer.visu;

import com.bbn.openmap.Environment;
import java.util.*;
import java.awt.event.*;

import javax.swing.*;

import java.awt.BorderLayout;
import java.awt.GridLayout;
import java.awt.Color;
import java.awt.FlowLayout;

import fr.unicaen.openmap.dataStorage.*;
import fr.unicaen.openmap.util.Util;
import fr.unicaen.openmap.layer.thematicDataDisplayLayer.*;
import fr.unicaen.openmap.layer.thematicDataDisplayLayer.util.*;
import fr.unicaen.openmap.layer.thematicDataDisplayLayer.gui.*;
import fr.unicaen.openmap.layer.thematicDataDisplayLayer.omGraphics.*;
import fr.unicaen.openmap.layer.thematicDataDisplayLayer.omGraphics.legend.*;

import com.bbn.openmap.omGraphics.*;
import com.bbn.openmap.util.*;
import java.awt.Dimension;
import org.jdesktop.swingx.JXCollapsiblePane;

/**
 * Pie visualization mode.
 * <p>Draw pies in place of discs. The property radiusIsSum fixes the disc dimension. 
 */
public class Pie extends Disc implements AttributesNames, ColorScaleListener {

    public static final String colorDataColumnNamesPropertyString = "colorDataColumnNames";
    public static final String radiusIsSumPropertyString = "radiusIsSum";
    /** names of the column used to draw the pie portions */
    protected String[] colorDataColumnNames = new String[0];
    protected double[] generalAngles;
    protected ColorScale colorScale = new ColorScale();
    protected boolean radiusIsSum = false;
    protected boolean badLegendCache = false;

    public Pie(ThematicDataDisplayLayer layer, String name) {
        this(layer, name, i18n.get(Pie.class, "pie", "Pie"));
    }

    public Pie(ThematicDataDisplayLayer layer, String name, String typeName) {
        super(layer, name, typeName);
        resizer.setWithRadius(true);
        interactive = true;
    }

    protected void buildDrawingAttributes() {
        drawingAttributes = new ThematicDrawingAttributes(this, true, false, false, true, false);
    }

    public void setProperties(String prefix, Properties properties) {
        super.setProperties(prefix, properties);
        colorScale.setProperties(prefix, properties);
        colorDataColumnNames = properties.getProperty(prefix + colorDataColumnNamesPropertyString, "").split("[ \t]+");
        radiusIsSum = Util.isTrue(properties.getProperty(prefix + radiusIsSumPropertyString, new Boolean(radiusIsSum).toString()));
        drawingAttributes.setFillPaint(new Color(0, 0, 0, 0));
        //colors.setFillColor(new Color(0, 0, 0, 0));
    }

    public Properties getProperties(String prefix, Properties properties) {
        properties = super.getProperties(prefix, properties);
        properties = colorScale.getProperties(prefix, properties);
        StringBuffer colorDataColumnNamesString = new StringBuffer();
        for (int i = 0; i < colorDataColumnNames.length; i++) {
            colorDataColumnNamesString.append(colorDataColumnNames[i]);
            colorDataColumnNamesString.append(" ");
        }
        properties.put(prefix + colorDataColumnNamesPropertyString, colorDataColumnNamesString.toString());
        properties.put(prefix + radiusIsSumPropertyString, new Boolean(radiusIsSum).toString());

        return properties;
    }

    public boolean getRadiusIsSum() {
        return radiusIsSum;
    }

    protected void setRadiusIsSum(boolean b) {
        radiusIsSum = b; //note that resizerPanel is enabled or disabled, via the radio buttons group
        refresh();
    }

    @Override
	public OMGraphicList generateGraphics(DataStorageIterator dsi) {
        mouseControlPointsListener.clearControlPoints();

	// if DS is empty return empty list
	if (dsi.getDataColumnNames().length == 0)
	    return new OMGraphicList();

	// length = max(colorScale length, bars number)
	int length = colorDataColumnNames.length;
	if (length > colorScale.getColorNumber())
	    length = colorScale.getColorNumber();	

        if (radiusIsSum) {
	    resizer.setDataColumnName("Sum");
            resizer.setDataValueComputer(new DataSquareRootSumColumnsValueComputer(colorDataColumnNames, length));
            resizer.setIsFixedSize(false); // in order for it to take the DataValueComputer into account
        } else {
	    if (getDataColumnName() == null) 
		resizer.setDataColumnName(dsi.getDataColumnNames()[0]);
            resizer.setDataValueComputer(new DataSquareRootColumnValueComputer(getDataColumnName()));
        }

        resizer.setMinMaxValue(dsi);
        DataSumColumnsValueComputer colordc = new DataSumColumnsValueComputer(colorDataColumnNames,length);
        generalAngles = new double[length];
        int nbPie = 0;
        OMSizeableGraphicList omglDiscs = new OMSizeableGraphicList<OMDisc>();
        OMGraphicList ompl = new OMGraphicList(); // for the OMPoint (Discs of null size)
        boolean offLineVis = offseter.isLineVisible();
        boolean offVis = offseter.isVisible();
        while (dsi.hasNext()) {
            DataObject data = dsi.next();
            OMDataLocation omdl = data.getOMDataLocation(geometryColumnName);
            if (omdl == null) {
                continue;
            }

            Centroid centroid = omdl.getCentroid();
            if (centroid == null) {
                continue;
            }

            double[] values = colordc.getValues(data);
            if (values == null) {
                Debug.message("ThematicDataDisplayLayer", "no Pie because of a null value ");
                continue;
            }
            double sum = values[length];
            boolean zeroValues = true; //true if all values are equal to 0 or NaN
            for (int i = 0; i < length; i++) {
                values[i] *= 360 / sum;
                if(values[i] != 0.0 && !Double.valueOf(values[i]).isNaN()){                  
                    zeroValues = false;
                }
            }
            nbPie++;
            for (int i = 0; i < length; i++) {
                generalAngles[i] += values[i];
            }
            if(zeroValues){
                continue; // pie won't be drawn
            }
            double radius = radiusIsSum
		? resizer.getSize(sum)
		: resizer.getSize(data);
            OMGraphic disc;
            // 0.001 is to be adapted with LatLon Graphics
            if (Math.abs(radius) < 0.001 && getShowNullAsPoint()) { // null sized pie => OMPoint
                disc = new OMPoint(centroid.getLat(), centroid.getLon(), 1);
                ompl.add(disc);
            } else {
		int offX = 0;
		int offY = 0;
		String key = "";

		if (offseter.isOn()) {
		    Offseter.OffsetLocationAndKey loc = offseter.getOffsetLocation(data);
		    offX = loc.getX();
		    offY = loc.getY();
		    key = loc.getKey();
		}
		disc = getZoomableSymbol()
		    ? new OMPie(this, key, centroid.getLat(), centroid.getLon(), 
				Math.abs(radius),
				length, values, colorScale,
				(radiusIsSum ? String.valueOf(sum) : resizer.getString(data)))
		    : new OMPie(this, key, centroid.getLat(), centroid.getLon(), 
				offX, offY, (int) (Math.abs(radius) * 2),
				length, values, colorScale,
				(radiusIsSum ? String.valueOf(sum) : resizer.getString(data)), offVis, offLineVis);
		omglDiscs.add(disc);
            }
	    disc.setLinePaint(radius >= 0 ? drawingAttributes.getLinePaint() 
			      : ColorScale.negative((Color) drawingAttributes.getLinePaint()));
            disc.setStroke(drawingAttributes.getStroke());
	    disc.setMatted(getDrawingAttributes().isMatted());
	    disc.setMattingPaint(getDrawingAttributes().getMattingPaint());
        }

        for (int i = 0; i < generalAngles.length; i++) {
            generalAngles[i] = generalAngles[i] / nbPie;
        }

        if (nbPie > 0) {
            getLayer().thematicChanged(this);
        }

	if (ompl.size() > 0) // add null sized discs
	    omglDiscs.add(ompl);     
	omglDiscs.sort();
        
	return omglDiscs;
    }

    @Override
	public JComponent getGUI(final DataStorageIterator dsi) {

        if (thematicGUI != null) {
            return thematicGUI;
        }
        thematicGUI = (ThematicGUI) super.getGUI(dsi);
        graphicsPanelToggleControlsDisplay(false);


        resizer.setShowNullAsPoint(showNullAsPointCB);
        final JComboBox combo = resizer.getComboSize();
        combo.addItem(i18n.get(this, "radiusIsSumButton", "Proportional to the sum"));
	if (radiusIsSum)
	    combo.setSelectedIndex(combo.getItemCount() - 1);
        combo.addItemListener(new ItemListener() {
		public void itemStateChanged(ItemEvent arg0) {
		    setRadiusIsSum(combo.getSelectedIndex() == combo.getItemCount() - 1);
		    resizer.setDataColumnName("Sum");		
		}
	    });

        colorScalePanel = new AttributesColorScalePanel(colorScale, (AttributesNames) this);
        colorScalePanel.setAvailableNames(getNoColorDataColumnNames(dsi));

	Debug.message("ThematicDataDisplayLayer", "**Pie: getGUI() getNoColorDataColumnNames(dsi) " + getNoColorDataColumnNames(dsi));

        colorScalePanel.addColorScaleListener(this);

        ThematicConfPanel pieConfPan =
	    ThematicConfPanel.createThematicConfPanel(i18n.get(this, "classification", "Classification"), false);

        JPanel wrapClassif = new JPanel();
        wrapClassif.add(pieConfPan);
        pieConfPan.setContent(colorScalePanel);
        thematicGUI.addThematicConfigurationPanel(pieConfPan);
        thematicGUI.add(wrapClassif);

        return thematicGUI;
    }

    private boolean isAColorDataColumnNames(String columnName) {
        for (int i = 0; i < colorDataColumnNames.length && i < colorScale.getColorNumber(); i++) {
            if (colorDataColumnNames[i].equals(columnName)) {
                return true;
            }
        }
        return false;
    }

    public void setColorDataColumnNames(String[] colorDataColumnNames) {
        
        this.colorDataColumnNames = colorDataColumnNames;
        badLegendCache = true;
    }

    /** gives colorDataColumnName[i]. @see AttributesNames interface */
    public String getName(
			  int i) {
        return i >= 0 && i < colorDataColumnNames.length ? colorDataColumnNames[i] : "";
    }
    public final static int GAP = 10;

    private OMResizeableFrame frame = null;
    private OMResizeableColorScaleLegend pieLegend = null;

    @Override
	public OMResizeableGraphic getLegend(Refreshable r, MouseControlPointsListener ml) {

        OMResizeableDiscLegend discLegend = (OMResizeableDiscLegend) super.getLegend(r, ml);

	frame = new OMResizeablePieLegend(r, ml, 0, 0, resizer, generalAngles, 
					  drawingAttributes.getLinePaint(), new Color(0, 0, 0, 0), 
					  colorScale, colorDataColumnNames, 
					  discLegend, getFullName());

        frame.setVisuFrame(true);
        return frame;

    }
    protected AttributesColorScalePanel colorScalePanel;

    //when we change dsi, data change so... refresh list *and* colorscale
    public void refreshGUI(DataStorageIterator dsi) {
        super.refreshGUI(dsi);
        if (colorScalePanel != null) {
	    colorScalePanel.setAvailableNames(getNoColorDataColumnNames(dsi));
	    Debug.message("ThematicDataDisplayLayer", "**Pie: getNoColorDataColumnNames(dsi) " + getNoColorDataColumnNames(dsi));

	    actualiseAttributesNames(dsi);
	    colorScalePanel.refreshPanel(); // redraws all
        }
    }

    /** @see AttributesNames interface */
    public void setAttributesNames(String[] names) {
        setColorDataColumnNames(names);
    }

    /**
     * Remove names from AttributesNames that are no more attributes
     * unless DS is not empty
     */
    private void actualiseAttributesNames(DataStorageIterator dsi) {

	// when installing a filter, it happens that the DS is empty. In that case, just do nothing...
	if (dsi.getDataColumnNames().length == 0)
	    return;

        ArrayList<String> names = new ArrayList<String>();

        for (String name : colorDataColumnNames)
            for (String attribute : dsi.getDataColumnNames())
                if (name.equals(attribute)) {
                    names.add(name);
                    break;
                }

        setAttributesNames(names.toArray(new String[names.size()]));
    }

    protected ArrayList<String> getNoColorDataColumnNames(DataStorageIterator dsi) {
        ArrayList<String> names = new ArrayList<String>();
        for (String name : dsi.getDataColumnNames())
            if (!isAColorDataColumnNames(name))
                names.add(name);
            
        return names;
    }

    public String[] getLegendColorDataColumnNames() {
        try {
            if (badLegendCache) {
                badLegendCache = false;
                return colorDataColumnNames;
            }

            ArrayList<String> colorDataNames = new ArrayList<String>();

            for (OMResizeableGraphic omrg : pieLegend.getResizeableGraphics()) {
                if (omrg instanceof OMResizeableText) {
                    OMResizeableText omrtext = (OMResizeableText) omrg;
                    if (omrtext.getText() != null && !omrtext.getText().equals("")) {
                        colorDataNames.add(omrtext.getText());
                    }
                }
            }
            return colorDataNames.toArray(new String[colorDataNames.size()]);
        } catch (NullPointerException exc) {
            return colorDataColumnNames;
        }

    }

    public void colorScaleChanged(ColorScalePanel colorScalePanel) {
        if (this.colorScalePanel == colorScalePanel) {
            refresh();
        }
    }
}
