// <header>
/**
 *  ThemaMap : The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
package fr.unicaen.openmap.layer.thematicDataDisplayLayer.visu;

import fr.unicaen.openmap.event.RefreshableGUI;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Paint;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemListener;
import java.awt.event.ItemEvent;
import java.util.*;
import javax.swing.*;

import fr.unicaen.openmap.dataStorage.*;
import fr.unicaen.openmap.layer.thematicDataDisplayLayer.*;
import fr.unicaen.openmap.layer.thematicDataDisplayLayer.util.*;
import fr.unicaen.openmap.layer.thematicDataDisplayLayer.omGraphics.*;
import fr.unicaen.openmap.layer.thematicDataDisplayLayer.omGraphics.legend.*;
import fr.unicaen.openmap.layer.thematicDataDisplayLayer.gui.*;
import fr.unicaen.openmap.util.Util;

import com.bbn.openmap.omGraphics.*;
import com.bbn.openmap.util.*;
import java.awt.Dimension;

/**
 * Proportional Symbol (Polygon, Disc, Pictogram) visualization mode.
 *
 * <p>Beware : as the setters call a refresh, never use then directly in setproperties
 * or you will generate many refresh ...
 */
public class ProportionalSymbol extends OffsetableRaw implements ResizerListener {

    public static final String showNullAsPointPropertyString = "showNullAsPoint";
    public static final String modePropertyString = "mode";
    public static final String nbEdgesPropertyString = "nbEdges";
    public static final String starCoefPropertyString = "starCoef";
    public static final String startAnglePropertyString = "startAngle";
    public static final String pictogramFilePropertyString = "pictogramFile";
    public static final String pictogramColumnPropertyString = "pictogramColumn";

    public static enum type {
        DISC,  
	    POLY, 
	    PICTO, // same pictogram for all
	    PICTOS // different pictograms defined in a column
    };
    private type mode;
    private String pictogramFile = null;
    private String pictogramColumnName = null;
    protected SymbolResizer resizer =
            new SymbolResizer(new DataSquareRootColumnValueComputer(""));

    public String getDataColumnName() {
        return resizer.getDataColumnName();
    }

    public void setDataColumnName(String name) {
        resizer.setDataColumnName(name);
        refresh();
    }
    public static int nbDiscEdges = 21;
    public static int nbPolyEdges = 3;
    protected int nbEdges = 21;
    protected int starCoef = 0;
    protected int startAngle = 45; // in degrees
    private boolean showNullAsPoint;

    protected boolean getShowNullAsPoint() {
        return showNullAsPoint;
    }

    protected void setShowNullAsPoint(boolean b) {
        if (showNullAsPoint != b) {
	    Debug.message("ThematicDataDisplayLayer", "##ProportionalSymbolPanel setShowNullAsPoint Changed " + b);
            showNullAsPoint = b;
            refresh();
        }
    }
    protected MinMaxValues minMaxValues;

    public ProportionalSymbol(ThematicDataDisplayLayer layer, String name) {
        this(layer, name, i18n.get(ProportionalSymbol.class,
                "proportionalSymbol", " Proportional symbol"));
    }

    public ProportionalSymbol(ThematicDataDisplayLayer layer, String name,
            String typeName) {
        super(layer, name, typeName);
        resizer.addResizerListener(this);
        mode = type.DISC;
        interactive = true;
    }

    @Override
    public void setProperties(String prefix, Properties properties) {
        super.setProperties(prefix, properties);
        this.showNullAsPoint = Util.isTrue(properties.getProperty(prefix + showNullAsPointPropertyString, "false"));
        resizer.setProperties(prefix, properties);
        this.mode = Enum.valueOf(type.class, properties.getProperty(prefix + modePropertyString, "DISC"));
        this.nbEdges = PropUtils.intFromProperties(properties, prefix + nbEdgesPropertyString, getNbEdges());
        this.startAngle = PropUtils.intFromProperties(properties, prefix + startAnglePropertyString, getStartAngle());
        this.starCoef = PropUtils.intFromProperties(properties, prefix + starCoefPropertyString, getStarCoef());
        this.pictogramFile = properties.getProperty(prefix + pictogramFilePropertyString, "");
	this.pictogramColumnName = properties.getProperty(prefix + pictogramColumnPropertyString, "picto");
    }

    public Properties getProperties(String prefix, Properties properties) {
        properties = super.getProperties(prefix, properties);
        properties.put(prefix + showNullAsPointPropertyString, new Boolean(getShowNullAsPoint()).toString());
        properties = resizer.getProperties(prefix, properties);
        properties.put(prefix + modePropertyString, mode.name());
        properties.put(prefix + nbEdgesPropertyString, Integer.toString(getNbEdges()));
        properties.put(prefix + startAnglePropertyString, Integer.toString(getStartAngle()));
        properties.put(prefix + starCoefPropertyString, Integer.toString(getStarCoef()));
        if (mode == type.PICTO) {
	    if (pictogramFile != null)
		properties.put(prefix + pictogramFilePropertyString, pictogramFile);
        } else if (mode == type.PICTOS) {
	    properties.put(prefix + pictogramColumnPropertyString, pictogramColumnName);
	}

        return properties;
    }

    @Override
    public OMGraphicList generateGraphics(DataStorageIterator dsi) {
        try {
            resizer.setMinMaxValue(dsi);
            OMSizeableGraphicList omglPoly = new OMSizeableGraphicList<OMRegularPolygon>();
	    OMGraphicList ompl = new OMGraphicList(); // for the OMPoint (Polys of null sizes)
            mouseControlPointsListener.clearControlPoints();
            double startAngleInRadians = getStartAngleInRadians();
            boolean offVis = offseter.isVisible();
            boolean offLineVis = offseter.isLineVisible();
            for (int i = 0; dsi.hasNext(); i++) {
                DataObject data = dsi.next();
                OMDataLocation omdl = data.getOMDataLocation(geometryColumnName);
                if (omdl == null) {
                    continue;
                }

                Centroid centroid = omdl.getCentroid();
                if (centroid == null) {
                    continue;
                }

                double radius = resizer.getSize(data);
                double lat = centroid.getLat();
                double lon =centroid.getLon();
                int offX = 0;
                int offY = 0;
                String key = "";
		if (offseter.isOn()) {
		    Offseter.OffsetLocationAndKey loc = offseter.getOffsetLocation(data);
		    offX = loc.getX();
		    offY = loc.getY();
		    key = loc.getKey();
		}
                OMGraphic p = null;
                if (getShowNullAsPoint() && ((int) (2 * Math.abs(radius))) < 2) {
                    p = new OMPoint(lat, lon, 1);
		    ompl.add(p);
                } else {
		    switch(getMode()) {
		    case PICTOS:
			if (pictogramColumnName != null)
			    p = new OMPictogram(this, key, lat, lon,
						offX, offY,
						data.getData(pictogramColumnName).toString(), (int)Math.abs(radius) * 2,
						resizer.getString(data), offVis, offLineVis);
			break;
		    case PICTO:
			if (pictogramFile != null)
			    p = new OMPictogram(this, key, lat, lon,
						offX, offY,
						pictogramFile, (int)Math.abs(radius) * 2,
						resizer.getString(data), offVis, offLineVis);
			break;
		    case POLY:
                        if (nbEdges < 21 || starCoef > 0) { //draw poly
                            p = new OMRegularPolygon(this, key, lat, lon, 
						     nbEdges, Math.abs(radius),
						     offX, offY, startAngleInRadians, starCoef, 
						     resizer.getString(data), offVis, offLineVis);
			    break;
			}
			// fall back to DISC in case nbEdges > 21 and starCoef == 0
		    case DISC:
			//draw circle. Beware use the diameter!
			p = new OMDisc(this, key, lat, lon, 
				       offX, offY, (int) (Math.abs(radius) * 2), 
				       resizer.getString(data), offVis, offLineVis);
		    }
		    if (p != null)
			omglPoly.add(p);
                }               
		if (p != null) {
		    p.setLinePaint(radius >= 0 ? drawingAttributes.getLinePaint() 
				   : ColorScale.negative((Color) drawingAttributes.getLinePaint()));
		    p.setFillPaint(radius >= 0 ? drawingAttributes.getFillPaint() 
				   : ColorScale.negative((Color) drawingAttributes.getFillPaint()));
		    p.setStroke(drawingAttributes.getStroke());
		    p.setMatted(getDrawingAttributes().isMatted());
		    p.setMattingPaint(getDrawingAttributes().getMattingPaint());
		}
            }

	    if (ompl.size() > 0) // add null sized discs
		omglPoly.add(ompl);
            omglPoly.sort();
            initMinMaxValues(omglPoly); // used for the legend

            return omglPoly;
        } catch (Exception e) {
            e.printStackTrace();
        }
        return new OMGraphicList();
    }

    public boolean isReady() {
        if (resizer.getIsFixedSize()) {
            return super.isReady();
        } else {
            return super.isReady() &&
                    !"".equals(resizer.getDataColumnName()) &&
                    resizer.getDataColumnName() != null;
        }
    }

    public String getMissingInfos() {
        String mess = getName() + ":\n";
        if (geometryColumnName.equals("")) {
            mess += i18n.get(this, "geometryColumnName", "Geometry column name") + "\n";
        }
        if (resizer.getDataColumnName().equals("")) {
            mess += i18n.get(ResizerPanel.class, "dataColumnNameLabel", "Data Column Name") + "\n";
        }
        return mess;

    }

    protected JPanel dataPanel; // protected attribute in order for derived class to keep this one!

    public JComponent getGUI(final DataStorageIterator dsi) {
        thematicGUI = (ThematicGUI) super.getGUI(dsi);
        JCheckBox showNullAsPointCB = new JCheckBox(i18n.get(this, "showNullAsPoint", "Show Null As Point"));
        showNullAsPointCB.setSelected(getShowNullAsPoint());
        showNullAsPointCB.addItemListener(new ItemListener() {

            public void itemStateChanged(ItemEvent e) {
                setShowNullAsPoint(((AbstractButton) (e.getItemSelectable())).isSelected());
            }
        });


        ThematicConfPanel resizerConfPanel = (ThematicConfPanel) resizer.getGUI(dsi);

        JPanel wrapResize = new JPanel();
        wrapResize.add(resizerConfPanel);
        thematicGUI.add(wrapResize);
        thematicGUI.addThematicConfigurationPanel(resizerConfPanel);
        JPanel pShowNull = new JPanel(new BorderLayout());
        pShowNull.add(showNullAsPointCB, BorderLayout.WEST);
        pShowNull.setBorder(BorderFactory.createEmptyBorder(0, 50, 0, 0));
        resizer.getProportionalGUI().add(pShowNull);


        ThematicConfPanel proportionalConf = (ThematicConfPanel) getProportionalSymbolGUI(dsi);

        JPanel wrapProp = new JPanel();
        wrapProp.add(proportionalConf);
        thematicGUI.add(wrapProp);
        thematicGUI.addThematicConfigurationPanel(proportionalConf);


        return thematicGUI;
    }
    protected ProportionalSymbolPanel proportionalSymbolPanel = null;

    public ThematicConfPanel getProportionalSymbolGUI(final DataStorageIterator dsi) {
        if (proportionalSymbolPanel == null) {
            proportionalSymbolPanel = new ProportionalSymbolPanel(this, dsi);
        }
        ThematicConfPanel cp = ThematicConfPanel.createThematicConfPanel(
                i18n.get(ProportionalSymbolPanel.class, "title", "drawing"));
        cp.setContent(proportionalSymbolPanel);
        return cp;
    }

    public void setStartAngle(int a) {
        startAngle = a;
        refresh();
    }

    public void setStarCoef(int c) {
        starCoef = c;
        refresh();
    }

    public void setNbEdges(int n) {
        nbEdges = n;
        refresh();
    }

    /** in order for MultiRegularPolygons to overload this method */
    public void setElemRadius(int r) {
        refresh();
    }

    public int getStartAngle() {
        return startAngle;
    }

    public double getStartAngleInRadians() {
        return startAngle * Math.PI / 180.0;
    }

    public int getStarCoef() {
        return starCoef;
    }

    public int getNbEdges() {
        return nbEdges;
    }

    /** in order for MultiRegularPolygons to overload this method */
    public int getElemRadius() {
        return 20;
    }

    protected void initMinMaxValues(OMGraphicList omgl) {
	if (omgl == null || omgl.size() == 0) {
	    minMaxValues = new MinMaxValues(0, 0);
	    return;
	}

	double min = 0, max = 0;
	OMGraphic omg = omgl.getOMGraphicAt(0);
	try {
	    min = Util.parseDouble(((HaveWidthAndData) omgl.getOMGraphicAt(0)).getData());
	} catch (java.lang.ClassCastException e) {
	}
	try {
	    max = Util.parseDouble(((HaveWidthAndData) omgl.getOMGraphicAt(omgl.size() - 1)).getData());
	} catch (java.lang.ClassCastException e) {
	}
	minMaxValues = new MinMaxValues(min, max);
    }

    @Override
    public OMResizeableGraphic getLegend(Refreshable r, MouseControlPointsListener ml) {
        if (minMaxValues == null) {
            return null;
        }
        OMResizeableFrame frame = null;
	switch (getMode()) {
	case POLY:
	    if (nbEdges < 21 || starCoef > 0) { //draw poly
		frame = new OMResizeableRegularPolygonLegend(r, ml, 0, 0, resizer, minMaxValues,
							     nbEdges, getStartAngleInRadians(),
							     starCoef, getDataColumnName(),
							     drawingAttributes.getLinePaint(), drawingAttributes.getFillPaint(), getFullName());
		break;
	    }	
	    // fall back to DISC in case nbEdges > 21 and starCoef == 0
	case DISC:
	    frame = new OMResizeableDiscLegend(r, ml, 0, 0, resizer, minMaxValues, getDataColumnName(),
					       drawingAttributes.getLinePaint(), drawingAttributes.getFillPaint(), getFullName());
	    break;
	case PICTO:
	case PICTOS:
	    return null;
	}
        frame.setVisuFrame(true);
        return frame;
    }

    public void refreshGUI(DataStorageIterator dsi) {
        super.refreshGUI(dsi);
        if (resizer != null)
	    resizer.refreshGUI(dsi);
	if (proportionalSymbolPanel != null)
	    proportionalSymbolPanel.refreshGUI(dsi, this);
    }

    public void refresh() {
	super.refresh();
	if (proportionalSymbolPanel != null)
	    ;// proportionalSymbolPanel.refreshGUI(dsi, this);
    }

    public void resizerChanged(Resizer resizer) {
        if (this.resizer == resizer) {
	    Debug.message("ThematicDataDisplayLayer", "##ProportionalSymbolPanel resizerChanged");
	    refresh();
        }
    }

    public void setMode(type mode) {
        this.mode = mode;
	refresh();
    }

    public type getMode() {
        return this.mode;
    }

    public void setPictogramFile(String pictogram) {
        this.pictogramFile = pictogram;
	refresh();
    }

    public String getPictogramFile() {
        return pictogramFile;
    }

    public void setPictogramColumnName(String pictogram) {
        this.pictogramColumnName = pictogram;
	refresh();
    }

    public String getPictogramColumnName() {
        return pictogramColumnName;
    }

    public SymbolResizer getResizer() {
        return resizer;
    }
}
