// <header>
/**
 *  ThemaMap : The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
package fr.unicaen.openmap.layer.wmsLayer;

import org.geotools.data.wms.WebMapServer;
import org.geotools.data.ows.WMSCapabilities;
import org.geotools.data.wms.response.GetMapResponse;
import org.geotools.data.wms.request.GetMapRequest;
import org.geotools.data.ows.Layer;
import org.geotools.data.ows.OperationType;
import org.geotools.data.ows.CRSEnvelope;

import java.util.ArrayList;
import java.util.Properties;
import java.util.List;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Enumeration;

import java.awt.GridLayout;
import java.awt.Dimension;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.Component;
import java.awt.image.BufferedImage;
import java.awt.geom.Point2D;
import java.awt.image.WritableRaster;

import java.net.URL;

import javax.imageio.ImageIO;
import javax.swing.JPanel;
import javax.swing.JList;
import javax.swing.DefaultListModel;
import javax.swing.JButton;
import javax.swing.ListSelectionModel;
import javax.swing.JScrollPane;
import javax.swing.BoxLayout;
import javax.swing.JComboBox;
import javax.swing.Box;

import com.bbn.openmap.proj.coords.LatLonPoint;
import com.bbn.openmap.layer.OMGraphicHandlerLayer;
import com.bbn.openmap.proj.Projection;
import com.bbn.openmap.omGraphics.OMGraphicList;
import com.bbn.openmap.omGraphics.OMRaster;
import com.bbn.openmap.util.*;

public class WMSLayer
extends OMGraphicHandlerLayer
implements ActionListener
{
	// Properties
	private static final String urlPropertyString = "wmsURL";
	private static final String layersPropertyString = "layersState";
	private static final String projPropertyString = "proj";
	private static final String imgPropertyString = "img";
	private String wmsURL = "";
	private String proj = "";
	private String img = "";

	// BBox
	private double  lonMin;
	private double  lonMax;
	private double  latMin;
	private double  latMax;

	private int    xShift;

	// MWS
	protected WebMapServer wms;	
	protected WMSCapabilities capa;

	// GUI and storage
	private JPanel gui;
	private JList input;
	private JList output;
	private DefaultListModel inputLM;
	private DefaultListModel outputLM;
	private JButton add;
	private JButton remove;
	private JComboBox projCB;
	private JComboBox imgCB;
	private JButton refresh;

	private ArrayList<String> layers;
	private HashMap<String, Boolean> state;

	private WMSListener listener;


	/* ------- Methods ------- */

	public WMSLayer()
	{
		Debug.message("WMSLayer", this + " new WMSLayer()");
		wms = null;
		capa = null;
		listener = null;

		// GUI
		gui = new JPanel();
		gui.setLayout(new BoxLayout(gui, BoxLayout.X_AXIS));

		inputLM = new DefaultListModel();
		outputLM = new DefaultListModel();

		input = new JList(inputLM);
		output = new JList(outputLM);

		JScrollPane inputSP = new JScrollPane(input);
		JScrollPane outputSP = new JScrollPane(output);

		inputSP.setPreferredSize(new Dimension(100,220));
		outputSP.setPreferredSize(new Dimension(100,220));

		input.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		output.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);

		add = new JButton("->");
		add.addActionListener(this);
		remove = new JButton("<-");
		remove.addActionListener(this);

		JPanel inner = new JPanel( new GridLayout(2,1) );
		JPanel right = new JPanel( new GridLayout(1,3) );

		JPanel left = new JPanel();
		left.setLayout( new BoxLayout( left, BoxLayout.Y_AXIS ) );

		refresh = new JButton("Refresh");
		refresh.addActionListener(this);
		projCB = new JComboBox();
		projCB.addActionListener(this);
		projCB.setEditable(true);
		imgCB = new JComboBox();
		imgCB.addActionListener(this);
		imgCB.setEditable(true);

		left.add( Box.createVerticalGlue() );
		left.add(projCB);
		left.add( Box.createRigidArea(new Dimension(0,10)) );
		left.add(imgCB);
		left.add( Box.createRigidArea(new Dimension(0,10)) );
		left.add( refresh );
		left.add( Box.createVerticalGlue() );

		inner.add( add );
		inner.add( remove );

		right.add( inputSP );
		right.add( inner );
		right.add( outputSP );

		gui.add( Box.createHorizontalGlue() );
		gui.add( left );
		gui.add( Box.createRigidArea(new Dimension(50,0)) );
		gui.add( right );
		gui.add( Box.createHorizontalGlue() );
		gui.setVisible(true);
	}

	private void initialize()
	{
		try
		{
			wms = new WebMapServer(new URL(wmsURL));
		}
		catch (java.net.MalformedURLException ex)
		{
			Debug.error("WMS : URL \"" + wmsURL + "\" is malformed.");
			return;
		}
		catch (Exception ex)
		{
			ex.printStackTrace();
			return;
		}

		capa = wms.getCapabilities();

		String projOld = ""+proj;
		for(String p : capa.getLayer().getSrs())
		{
			projCB.addItem(p);
		}

		if( ! proj.equals( projOld ) )
		{
			for( int idx = 0 ; idx < projCB.getItemCount() ; ++idx )
			{
				if( projCB.getItemAt(idx).equals(projOld) )
				{
					projCB.setSelectedIndex(idx);
					break;
				}
			}
		}

		String imgOld = ""+img;

		imgCB.addItem("image/gif");
		imgCB.addItem("image/png");
		imgCB.addItem("image/jpeg");

		if( ! img.equals( imgOld ) )
		{
			for( int idx = 0 ; idx < imgCB.getItemCount() ; ++idx )
			{
				if( imgCB.getItemAt(idx).equals(imgOld) )
				{
					imgCB.setSelectedIndex(idx);
					break;
				}
			}
		}

		if( layers == null && state == null )
		{
			// New Layer
			layers = new ArrayList<String>(capa.getLayerList().size());
			state = new HashMap<String, Boolean>();

			for( Layer l : capa.getLayerList() )
			{
				String name = l.getName();
				if( name != null )
				{
					inputLM.addElement(name);
				}
			}
		}
		else
		{
			// Layer loaded from properties
			for( Layer l : capa.getLayerList() )
			{
				String name = l.getName();
				if( name != null )
				{
					if( layers.contains(name) )
					{
						outputLM.addElement(name);
					}
					else
					{
						inputLM.addElement(name);
					}
				}
			}
		}
		gui.updateUI();
	}

	public void	actionPerformed(ActionEvent e)
	{
		Object source = e.getSource();

		if( source == refresh )
		{
			// Refresh layer
			doPrepare();
		}
		else
		if( source == add )
		{
			// Add WMSLayer to Themamap visu list
			int idx;
			if( ( idx = input.getSelectedIndex() ) != -1 )
			{
				outputLM.addElement( inputLM.remove( idx ) );
				input.setSelectedIndex( (idx < inputLM.size()) ? idx : 0 );
				output.setSelectedIndex( outputLM.size() -1 );
				listener.addWMSLayer( new DummyWMSLayer( ""+outputLM.lastElement(), this, layers, state ) );
			}
		}
		else
		if( source == remove )
		{
			// Remove WMSLayer from Themamap visu list
			int idx;
			if( ( idx = output.getSelectedIndex() ) != -1 )
			{
				inputLM.addElement( outputLM.remove( idx ) );
				output.setSelectedIndex( (idx < outputLM.size()) ? idx : 0 );
				input.setSelectedIndex( inputLM.size() -1 );
				listener.delWMSLayer( this, ""+inputLM.lastElement() );
			}
		}
		else
		if( source == imgCB )
		{
			// Img format changed
			img = (String) imgCB.getSelectedItem();
		}
		else
		if( source == projCB )
		{
			// Projection changed
			proj = (String) projCB.getSelectedItem();
		}

		gui.updateUI();
	}

	public Component getGUI()
	{
		gui.updateUI();
		return gui;
	}  

	protected int getPixel(int x, int y, Projection p, BufferedImage image)
	{
		LatLonPoint pt = p.inverse(x, y);
		double lon = pt.getX();
		double lat = pt.getY();

		try
		{
			return image.getRGB((int)Math.round(image.getMinX() + image.getWidth()  * (lon - lonMin) / (lonMax - lonMin)),
							    (int)Math.round(image.getMinY() + image.getHeight() * (lat - latMin) / (latMax - latMin)));
		}
		catch (ArrayIndexOutOfBoundsException e)
		{
			return 0; // black
		}
	}
    
	public OMGraphicList prepare()
	{
		Projection p = getProjection();
		BufferedImage images[] = getImage(p);
		int w = p.getWidth();
		int h = p.getHeight();

		OMGraphicList omgl = new OMGraphicList();

		if( images.length == 1 )
		{
			BufferedImage image = images[0];

			if (image == null)
			{
				return omgl;
			}

			if (w != image.getWidth() || h != image.getHeight())
			{
				omgl.add( new OMRaster( 0, 0, image) );
				omgl.generate(p);
				return omgl;	
			}

			// Projection de l'image
			BufferedImage pImage = new BufferedImage (w, h, BufferedImage.TYPE_INT_ARGB);
			for(int x = 0; x < w; x++)
				for (int y = 0; y < h; y++)
					pImage.setRGB(x, y, getPixel(x, y, p, image));

			omgl.add( new OMRaster( 0, 0, pImage ) );
		}
		else
		{
			BufferedImage left = images[0];
			BufferedImage right = images[1];

			if (left == null || right == null)
			{
				return omgl;
			}

			LatLonPoint ul =  p.getUpperLeft();
			LatLonPoint lr =  p.getLowerRight();
			int splitX = ((int) p.forward( 0.0 , 180.0).getX());

			lonMin = ul.getLongitude();
			lonMax = 180.0;

			// Projection de l'image de gauche
			BufferedImage pLeft = new BufferedImage (w, h, BufferedImage.TYPE_INT_ARGB);
			for(int x = 0; x < splitX ; x++)
				for (int y = 0; y < h; y++)
					pLeft.setRGB(x, y, getPixel(x, y, p, left));

			lonMin = -180.0;
			lonMax = lr.getLongitude();


			// Projection de l'image de droite
			BufferedImage pRight = new BufferedImage (w, h, BufferedImage.TYPE_INT_ARGB);
			for(int x = splitX ; x < w; x++)
				for (int y = 0; y < h; y++)
					pRight.setRGB(x-splitX, y, getPixel(x, y, p, right));

			omgl.add( new OMRaster( 0, 0, pLeft ) );
			omgl.add( new OMRaster( splitX, 0, pRight ) );
		}

		omgl.generate(p);
		return omgl;	
	}
    
	public BufferedImage[] getImage(Projection p)
	{
		if( capa == null || wms == null || p == null )
		{
			return null;
		}

		GetMapRequest getMap =  wms.createGetMapRequest();
		LatLonPoint.Double ul = p.getUpperLeft();
		LatLonPoint.Double lr = p.getLowerRight();

		Point2D longitude180 = p.forward( 0.0 , 180.0); // Coordonnées de (0.0, 180.0) par rapport à la fenêtre
		if( longitude180.getX() >= 0 && longitude180.getX() <= p.getWidth() )
		{
			BufferedImage left  = getImagePart( p, ul, new LatLonPoint.Double( lr.getLatitude(), 180 ), ((int) p.forward( 0.0 , 180).getX()) );
			BufferedImage right = getImagePart( p, new LatLonPoint.Double( ul.getLatitude(), -180 ), lr , p.getWidth() - ((int) p.forward( 0.0 , -180).getX()) );

			BufferedImage ret[] = {left, right};
			return ret;
		}
		else
		{
			BufferedImage ret[] = { getImagePart( p, ul, lr, p.getWidth() ) };
			return ret;
		}
	}

	public BufferedImage getImagePart( Projection p, LatLonPoint ul, LatLonPoint lr, int imgWidth )
	{
		GetMapRequest getMap =  wms.createGetMapRequest();

		lonMin = ul.getLongitude();
		lonMax = lr.getLongitude();
		latMin = ul.getLatitude();
		latMax = lr.getLatitude();

		String bbox = lonMin + "," +
					  latMax + "," +
			   		  lonMax + "," +
					  latMin;

		String height = Integer.toString(p.getHeight());
		String width = Integer.toString( imgWidth );
		getMap.setBBox(bbox);
		getMap.setVendorSpecificParameter("CRS", proj);
		getMap.setVendorSpecificParameter("SRS", proj);
		getMap.setDimensions(width, height);
		getMap.setFormat(img);

		for( String name : layers )
		{
			if( state.containsKey(name) )
			{
				if( state.get(name) )
				{
					getMap.addLayer( name, (String) null );
				}
			}
		}

		GetMapResponse getMapResponse = null;
		try
		{
			getMapResponse = wms.issueRequest(getMap);
		}
		catch(java.lang.NullPointerException ex)
		{
			return null;
		}
		catch (Exception ex)
		{
			ex.printStackTrace();
			return null;
		}

		if (getMapResponse.getContentType().startsWith("image"))
		{
			try
			{
				return ImageIO.read(getMapResponse.getInputStream());
			}
			catch (Exception ex)
			{
				ex.printStackTrace();
				return null;
			}
		}
		return null;
	}

	public void setProperties(String prefix, java.util.Properties properties)
	// file -> class
	{
		super.setProperties(prefix, properties);
	    prefix = PropUtils.getScopedPropertyPrefix(prefix);
		
	    wmsURL = properties.getProperty(prefix + urlPropertyString);
	    Debug.message("WMSLayer", "[WMSLayer] wmsURL: " + wmsURL + ".");

		String layersStateProp = properties.getProperty(prefix + layersPropertyString);
	    Debug.message("WMSLayer", "[WMSLayer] layers: " + layersStateProp + ".");
		if( layersStateProp != null )
		{
			String layersState[] = layersStateProp.split(";");
			layers = new ArrayList<String>();
			state = new HashMap<String, Boolean>();

			for( int i = 0 ; i< layersState.length ; i+=2 )
			{
				String name = layersState[i];
				if( ! ( (i+1) < layersState.length ) )
				{
					break;
				}

				String active = layersState[i+1];
				layers.add(name);
				state.put(name, Boolean.valueOf(active));
			}
		}

	    proj = properties.getProperty(prefix + projPropertyString);
	    Debug.message("WMSLayer", "[WMSLayer] proj: " + proj + ".");

	    img = properties.getProperty(prefix + imgPropertyString);
	    Debug.message("WMSLayer", "[WMSLayer] img: " + img + ".");

	    initialize();
	}
	
	public Properties getProperties(Properties properties)
	// class -> file
	{
		properties = super.getProperties(properties);
	    String prefix = PropUtils.getScopedPropertyPrefix(this);
	
	    properties.put(prefix + urlPropertyString, wmsURL);

		if( layers != null && state != null )
		{
			String layersState = new String();
			Iterator<String> i = layers.iterator();
			while( i.hasNext() )
			{
				String name = i.next();
				if( i.hasNext() )
				{
					layersState = layersState + name + ";" + state.get(name) + ";" ;
					continue;
				}

				layersState = layersState + name + ";" + state.get(name);
			}

			properties.put(prefix + layersPropertyString, layersState);
		}

	    properties.put(prefix + projPropertyString, proj);
	    properties.put(prefix + imgPropertyString, img);
	    return properties;
	}

	public Properties getPropertyInfo(Properties properties)
	// file -> class
	{
		properties = super.getPropertyInfo(properties);

		properties.put(initPropertiesProperty, urlPropertyString);
	
		PropUtils.setI18NPropertyInfo(i18n,
									  properties,
									  WMSLayer.class,
									  urlPropertyString,
									  "URL", //label
									  "WMS server", //tooltip
									  "fr.unicaen.openmap.util.propertyEditor.WMSServerChooserPropertyEditor");
	    return properties;
	}

	public void associer( WMSListener l )
	{
		listener = l;

		if( ! outputLM.isEmpty() && layers != null && state != null )
		{
			ArrayList<String> layersCopy = new ArrayList<String>(layers);
			HashMap<String, Boolean> stateCopy = new HashMap<String, Boolean>(state);

			for( String name : layersCopy )
			{
				layers.remove(name);
				DummyWMSLayer newLayer = new DummyWMSLayer( name, this, layers, state );
				listener.addWMSLayer( newLayer );
				newLayer.setVisible(stateCopy.get(name));
			}
		}
    }

	public void moveLayer( String name, int offset )
	{
		int layerIdx = layers.indexOf( name );
		if( layerIdx != -1 )
		{
			String tmp = layers.get(layerIdx+offset);
			layers.set(layerIdx+offset, name );
			layers.set(layerIdx, tmp );
		}
	}

}
