// <header>
/**
 *  ThemaMap : The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
package fr.unicaen.openmap.util;

import java.util.*;

import com.bbn.openmap.util.ColorFactory;

import fr.unicaen.openmap.util.Util;
import fr.unicaen.openmap.util.Colors;


/**
 * A class to associate colors with a hierarchy.
 * <br>=> to be deprecated use ColorScale and Classification instead.
 */
public class ColorTable
{
    /**
     * Colors associated with bounds.
     */
    protected TreeMap<String,Colors> colors; //key = inf sup ; object = Colors

    /**
     * Current state : empty or not.
     */
    protected boolean empty = true;


    /**
     * Creates a new and empty ColorTable.
     */    
    public ColorTable()
    {
	this("");
    }

    /**
     * Creates a new ColorTable from String, if possible.
     * This String must have the same structure as given with this toString() method
     *
     * The props parameter <strong>must</strong> be well form;
     * it's better use the method addColors in a program.
     * 
     * @param props well formed String (@see toString())
     */ 
    public ColorTable(String props)
    {
	colors = new TreeMap<String,Colors>();
	parseColors(props);
    }

    /**
     * Returns current state.
     * @return true if empty, false if not.
     */    
    public boolean isEmpty()
    {
	return empty;
    }
    
    /**
     * Used to add Colors associated with given limits.
     * @param inf inferior limit
     * @param sup superior limit
     * @param colors Colors
     * todo: should control if inf < sup
     */
    public void addColors(double inf, double sup, Colors colors)
    {
	String key = inf+" "+sup;
	this.colors.put(key, colors);
	empty = false;
	
	//	System.out.println("ColorTable: New Colors: "+colors+" added with key :"+key);
    }
    
    /**
     * Returns Colors for a given value, if exists.
     * @param value value to test
     * @return Colors if exists (warning : null if not !)
     */
    public Colors getColors(double value)
    {
	Iterator iter = colors.keySet().iterator();
	String key ="";
	double inf;
	double sup;
	
	while(iter.hasNext()) {
	    String next_key = (String)iter.next();

	    String[] inf_sup =  next_key.split(" ");    
	    inf = Double.parseDouble(inf_sup[0]);
	    sup = Double.parseDouble(inf_sup[1]);
	    
	    if ((value >= inf) && (value < sup)) // [inf, sup[
		key = next_key;
	}
	
	return (Colors)(colors.get(key));
    }

    /**
     *
     */
    public TreeMap getColors()
    {
	return colors;
    }

    /**
     *
     */
    public int getColorNumber()
    {
	return colors.size();
    }
    
    /**
     * Used to erase all datas.
     */
    public void reset()
    {
	colors = new TreeMap<String,Colors>();
	empty = true;
    }
    

    //public Properties getProperties()

    /**
     * Returns a String representation for all colors.
     * This String contains a succession of :
     * inferior limit,
     * superior limit,
     * line color,
     * fill color,
     * selected line color,
     * selected fill color,
     * fill mode (boolean).
     *
     * Colors are given in Hex.
     * @return String.
     */
    public String toString()
    {
	Iterator iter = colors.keySet().iterator();
	Colors c;
	String s = "";
	
	while(iter.hasNext())
	    {
		String key = (String)iter.next();
		String[] inf_sup =  key.split(" ");    
		
		c = (Colors)colors.get(key);
		
		s += " "+Double.parseDouble(inf_sup[0]); //inf
		s += " "+Double.parseDouble(inf_sup[1]); //sup
		s += " "+ColorFactory.getHexColorString(c.getLineColor()); //
		s += " "+ColorFactory.getHexColorString(c.getFillColor()); //
		s += " "+ColorFactory.getHexColorString(c.getSelectedLineColor()); //useless
		s += " "+ColorFactory.getHexColorString(c.getSelectedFillColor()); //useless
		s += " "+new Boolean(c.getFillMode()); //
		
	    }
	
	return s;
    }
    
    /**
     * Used to convert a property String in data for ColorTable.
     * This method is used on a result of this.toString()...
     */
    private void parseColors(String propertyString)
    {
	// update colors
	// foreach color: inf / sup / fill color / line color / fill mode
	String[] properties = propertyString.split(" ");
	double inf;
	double sup;
	Colors c;
	
	for (int i = 0; i < properties.length; ) {
		c = new Colors();
		
		inf = Double.parseDouble(properties[i++]);
		sup = Double.parseDouble(properties[i++]);
		c.setLineColor(properties[i++]);
		c.setFillColor(properties[i++]);
		c.setSelectedLineColor(properties[i++]);
		c.setSelectedFillColor(properties[i++]);
		c.setFillMode(Util.isTrue(properties[i++]));
		
		addColors(inf, sup, c);
	}
    }
   
}
