// <header>
/**
 *  ThemaMap : The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
package fr.unicaen.openmap.util;

import java.awt.*;
import java.awt.font.*;
import java.awt.geom.*;
import java.awt.event.*;
import java.util.Vector;
import javax.swing.*;
import javax.swing.border.EmptyBorder;
import javax.swing.border.TitledBorder;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.plaf.basic.BasicButtonUI;

import com.bbn.openmap.I18n;
import com.bbn.openmap.BasicI18n;
import fr.unicaen.openmap.gui.ThemaMapColorChooser;

/**
 * Creates a dialog to set font, size, justification and rotation angle
 */

public class FontDialog extends JDialog implements ListSelectionListener, ActionListener,
						   WindowListener {

    /**
     * Use I18N mecanism to allow internationalization.
     */    
    public final static I18n i18n = new BasicI18n();

    /**
     * Property name for plain font style.
     */
    public final static String PLAIN_STYLE_PROPERTY_STRING = "plain";

    /**
     * Property name for bold font style.
     */
    public final static String BOLD_STYLE_PROPERTY_STRING = "bold";

    /**
     * Property name for italic font style.
     */
    public final static String ITALIC_STYLE_PROPERTY_STRING = "italic";

    /**
     * Property name for bold and italic font style.
     */
    public final static String BOLD_AND_ITALIC_STYLE_PROPERTY_STRING = "bold_italic";

    /**
     * Property name for title.
     */
    public final static String TITLE_PROPERTY_STRING = "title";

    /**
     * Property name for font.
     */
    public static final String FONT_PROPERTY_STRING = "font";

    /**
     * Property name for font style.
     */
    public static final String FONT_STYLE_PROPERTY_STRING = "font.style";

    /**
     * Property name for font size.
     */
    public static final String FONT_SIZE_PROPERTY_STRING = "font.size";

    /**
     * Property name for preview.
     */
    public final static String PREVIEW_PROPERTY_STRING = "preview";

    /**
     * Property name for previewed text.
     */
    public static final String PREVIEW_SAMPLE_PROPERTY_STRING = "preview.text";

    /**
     * Property name for "ok" button.
     */
    public static final String OK_PROPERTY_STRING = "ok";

    /**
     * Property name for "cancel" button.
     */
    public static final String CANCEL_PROPERTY_STRING = "cancel";

    /**
     * Property name for "color" label.
     */
    public static final String COLOR_PROPERTY_STRING = "color";


    private JTextField previewTF, fontTF, styleTF, sizeTF;
    private JButton okB, cancelB;
    private Font font;
    private String fontName = "Arial";
    private int fontStyle = 0;
    private String fontSize = "10";
    private Color fontColor = Color.BLACK;
    private int justify; 
    private double angle = 0.0;
    private JButton sampleColor; 

    private JList fontsJL = new JList(GraphicsEnvironment.getLocalGraphicsEnvironment().getAvailableFontFamilyNames());
    private JList stylesJL = new JList(new Object [] { 
	    i18n.get(FontDialog.class, PLAIN_STYLE_PROPERTY_STRING, "Plain"), 
	    i18n.get(FontDialog.class, BOLD_STYLE_PROPERTY_STRING, "Bold"), 
	    i18n.get(FontDialog.class, ITALIC_STYLE_PROPERTY_STRING, "Italic"),
	    i18n.get(FontDialog.class, BOLD_AND_ITALIC_STYLE_PROPERTY_STRING, "Bold & Italic") }
	);
    private JList sizesJL = new JList(new Object [] { "6", "8", "10", "12", "14", "16", "18",
						      "20", "24", "36", "48", "56", "64", "72"});

    private boolean canceled = false;

    private JPanel createList(String label, JList list, JTextField field) { //, boolean editable) {
	list.addListSelectionListener(this);
	list.setBorder(new EmptyBorder(0, 2, 0, 2));
	field.setEditable(false);
	JPanel top, middle;
	(middle = new JPanel()).setLayout(new BorderLayout());
	middle.add(field, BorderLayout.NORTH);
	middle.add(new JScrollPane(list), BorderLayout.CENTER);
	
	(top = new JPanel()).setLayout(new BorderLayout());
	top.add(middle, BorderLayout.CENTER);
	top.add(new JLabel (label), BorderLayout.NORTH);
	return top;
    }
    
    private int getFontIndex(String name) {
	for (int i = 0; i < fontsJL.getModel().getSize() ; i++)
	    if (((String) fontsJL.getModel().getElementAt(i)).equalsIgnoreCase(name)) return i;
	return -1;
    }
    
    public void selectFont(Font font) {
	int index = getFontIndex(font.getFamily());
	if (index != -1) {
	    fontsJL.setSelectedIndex(index);
	    fontsJL.ensureIndexIsVisible(index);
	}
	fontTF.setText(font.getFamily());
	
	stylesJL.setSelectedIndex(font.getStyle());
	styleTF.setText((String) stylesJL.getSelectedValue());
	sizesJL.setSelectedValue("" + font.getSize(), true);
	sizeTF.setText("" + font.getSize());
    }

    /**
     * If fontColor is null, don't show color button
     */	
    public FontDialog (Font font, Color fontColor, int justify, double angle){
        this((JFrame) null, font, fontColor, justify, angle, null);
    }

    public FontDialog (Font font, Color fontColor, int justify, double angle, Component additionalComponent) {
	this((JFrame) null, font, fontColor, justify, angle, additionalComponent);
    }

    public FontDialog (JFrame jFrame, Font font, Color fontColor, int justify, double angle, Component additionalComponent) {
	super(jFrame, 
	      i18n.get(FontDialog.class, TITLE_PROPERTY_STRING, "Select font ..."), 
	      true);
	this.font = font;
	this.fontColor = fontColor;
	this.angle = angle;

	JPanel message = new JPanel();
	getContentPane().setLayout(new BorderLayout());
	
	previewTF = new JTextField(i18n.get(FontDialog.class,
					    PREVIEW_SAMPLE_PROPERTY_STRING,
					    "0123456789 abcdefijk ABCDEFIJK"
					    )
				   );
	previewTF.setFont(font);
	previewTF.setHorizontalAlignment(
					 FontDialog.getJTextFieldAlignmentFromOMTextAlignment(justify));
	setJustify(justify);
	JPanel top, up, right, preview, bottom;
	(top = new JPanel()).setLayout(new BoxLayout(top, BoxLayout.Y_AXIS));
	(up = new JPanel()).setLayout(new BorderLayout());
	(right = new JPanel()).setLayout(new BorderLayout());
        (preview = new JPanel()).setLayout(new BorderLayout());
	bottom = new JPanel();
	
	preview.setBorder(new TitledBorder(i18n.get(FontDialog.class,
						    PREVIEW_PROPERTY_STRING,
						    "Preview")
					   )
			  );
	preview.add(previewTF, BorderLayout.CENTER);
	
        up.add(BorderLayout.CENTER, createList(i18n.get(FontDialog.class,
							FONT_PROPERTY_STRING,
							"Font"), 
					       fontsJL, 
					       fontTF = new JTextField()
					       )
	       );
        right.add(BorderLayout.CENTER, createList(i18n.get(FontDialog.class,
							   FONT_STYLE_PROPERTY_STRING,
							   "Style"),
						  stylesJL,
						  styleTF = new JTextField()
						  )
		  );
	right.add(BorderLayout.EAST, createList(i18n.get(FontDialog.class,
							 FONT_SIZE_PROPERTY_STRING,
							 "Size"),
						sizesJL,
						sizeTF = new JTextField()
						)
		  );
	sizeTF.setEditable(true);
        up.add(BorderLayout.EAST, right);
        
	top.add(up);

	if(fontColor != null) {
	            
	    JPanel colorPan = new JPanel();
	    sampleColor = new JButton(" ");
	    sampleColor.setBorderPainted(false);
	    sampleColor.setFocusable(false);
	    sampleColor.setBackground(fontColor);
	    sampleColor.setPreferredSize(new Dimension(50,20));
	    sampleColor.addActionListener(new ActionListener(){
		    public void actionPerformed(ActionEvent arg0) {
			setFontColor(ThemaMapColorChooser.showDialog(sampleColor, "", getFontColor()));
			sampleColor.setBackground(getFontColor());
		    }});
	    JLabel colorLabel = new JLabel(i18n.get(FontDialog.class, COLOR_PROPERTY_STRING, "Color"));
	    colorPan.add(colorLabel);
	    colorPan.add(sampleColor);

	    top.add(colorPan);
	}
	
	top.add(createJustifyGUI());
	top.add(createRotationAngleGUI());

        if(additionalComponent != null)
            top.add(additionalComponent);

	top.add(preview);
	top.setBorder(new EmptyBorder(2, 6, 0, 6));

	
	(okB = new JButton(i18n.get(FontDialog.class, OK_PROPERTY_STRING, "Ok"))).addActionListener(this);
	(cancelB = new JButton(i18n.get(FontDialog.class, CANCEL_PROPERTY_STRING, "Cancel"))).addActionListener(this);
	bottom.add(okB);
	bottom.add(cancelB);
	
        getContentPane().add(BorderLayout.CENTER, top);
	getContentPane().add(BorderLayout.SOUTH, bottom);
	
	// last, to ensure all have components have been initialized
	selectFont(font);

	addWindowListener(this);
	setSize(new Dimension(500,350));
    }
    
    protected JComponent createJustifyGUI() {
        Box box = Box.createHorizontalBox();
        ButtonGroup group = new ButtonGroup();
        for (int i = 0; i < 3; i++) {
            JRadioButton rb = new JustifyRadioButton(i);
            if (i == getJustify()) {
                rb.setSelected(true);
            }
            group.add(rb);
            box.add(rb);
        }
        return box;
    }    

    private JSpinner rotationSpinner;

    protected JComponent createRotationAngleGUI() {
	JPanel rotPanel = new JPanel();
        rotPanel.add(new JLabel(i18n.get(this, "angle", "Rotation Angle (degrees) : ")));    
	rotationSpinner = new JSpinner(new SpinnerNumberModel(getRotationAngle(), -360d, 360d, 1d));
        rotationSpinner.addChangeListener(new ChangeListener() {
		public void stateChanged(ChangeEvent ce) {
		    setRotationAngle((Double)(((JSpinner)ce.getSource()).getValue()));
		}
	    });
        rotPanel.add(rotationSpinner);
	return rotPanel;
    }

    public void valueChanged(ListSelectionEvent e) {
	Object source = e.getSource();

        if (source == fontsJL) {
            fontName = (String) fontsJL.getSelectedValue();
	    fontTF.setText((String) fontsJL.getSelectedValue());
        } else if ( source == stylesJL ) {
            fontStyle = stylesJL.getSelectedIndex();
	    styleTF.setText((String) stylesJL.getSelectedValue());
        } else if (source == sizesJL) {
            fontSize = (String) sizesJL.getSelectedValue();
	    sizeTF.setText((String)sizesJL.getSelectedValue());
        }
	font = new Font (fontName, fontStyle, Integer.parseInt(fontSize));
        previewTF.setFont(font);
    }
    
    public boolean isCanceled() { return canceled; }
    
    public void actionPerformed(ActionEvent e) {
	canceled = !(e.getSource() == okB);
	dispose();
    }
    
    public Font getSelectedFont() { return font; }

    public void windowActivated(WindowEvent e) {}
    public void windowClosed(WindowEvent e) {}
    public void windowClosing(WindowEvent e) { canceled = true; dispose(); }
    public void windowDeactivated(WindowEvent e) {}
    public void windowDeiconified(WindowEvent e) {}
    public void windowIconified(WindowEvent e) {}
    public void windowOpened(WindowEvent e) {}
  
    public Color getFontColor() {
	return fontColor;
    }

    public void setFontColor(Color color) {
	fontColor = color;
	if (sampleColor != null)
	    sampleColor.setBackground(fontColor);
    }

    public int getJustify() {
	return justify;
    }

    static public int getJTextFieldAlignmentFromOMTextAlignment(int j) {
	switch(j) {
	case 0:
	    return JTextField.LEFT;
	case 1:
	    return JTextField.CENTER;
	case 2:
	    return JTextField.RIGHT;
	}
	return JTextField.LEFT;
    }

    public void setJustify(int j) {
	justify = j;
	previewTF.setHorizontalAlignment(getJTextFieldAlignmentFromOMTextAlignment(j));
    }

    public double getRotationAngle() {
	return angle;
    }

    public void setRotationAngle(double angle) {
	this.angle = angle;
	if (rotationSpinner != null)
	    rotationSpinner.setValue(angle);
    }

    private static final String[] justifyIconNames = {"left", "center", "right"};
    private static final ImageIcon[] justifyIcons = {
	Util.loadIcon("visu/stock_text_left.png"),
        Util.loadIcon("visu/stock_text_center.png"),
        Util.loadIcon("visu/stock_text_right.png")
    };

    private class JustifyRadioButton extends JRadioButton implements ActionListener {

        private MouseListener buttonMouseListener = new MouseAdapter() {

            public void mouseEntered(MouseEvent e) {
                setBorderPainted(true);
            }

            public void mouseExited(MouseEvent e) {
                setBorderPainted(false);
            }
        };

        private int just;

        public JustifyRadioButton(int just) {
            super();
            this.just = just;
            // use setIcon to ensure size is ok. PreferredSize is just bullshit.
            setIcon(Util.loadIcon("visu/stock_text_" + justifyIconNames[just] + ".png"));
            setToolTipText(i18n.get(FontDialog.class, "justify_" + justifyIconNames[just], "Justify " + justifyIconNames[just]));
            addActionListener(this);
            setOpaque(false);
            setPreferredSize(new Dimension(20, 20));
            setBorder(BorderFactory.createEtchedBorder());
            setBorderPainted(false);
            addMouseListener(buttonMouseListener);
            setUI(new BasicButtonUI());
        }

        public void actionPerformed(ActionEvent e) {
            setJustify(just);
        }

        public void paintComponent(Graphics g) {
            try {
                if (isSelected()) {
                    g.setColor(new Color(100, 200, 150));
		    g.fillRect(2, 2, 16, 16);
                }
                justifyIcons[just].paintIcon(this, g, 2, 2);
            } catch (Exception e) {
                super.paintComponent(g);
            }
        }
    }
  
    /**
     * For testing purposes ... Some fonts will crash your
     * Java VM and you may test them directly with the
     * Chooser ...
     */
    public static void main(String s[]) {
	FontDialog fd = new FontDialog(new Font("SansSerif", 1, 14), null, 0, 0.0);
	fd.setVisible(true);
	System.exit(0);
    }
}
