// <header>
/**
 *  ThemaMap : The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
package fr.unicaen.openmap.util;

import java.util.*;
import fr.unicaen.openmap.dataStorage.Centroid;
import fr.unicaen.openmap.dataStorage.DataObject;
import fr.unicaen.openmap.dataStorage.OMDataLocation;

/**
 * This class can collect column names and extract lat/lon column pairs. 
 * <p>Names are prefixes of "longitude" or "latitude" and must be at least 3 characters long.
 * In case of multiple lat/lon pairs, the prefix must be followed by an underscore ('_') and a name.
 *
 * @author Jacques.Madelaine@info.unicaen.fr
 */
public class LonLatNames {

    private static String longitude = "longitude";
    private static String latitude = "latitude";
	private static String geometry = "geometry_";
	private static String centroid = "centroid_";
	private static String bounding = "bounding_";

    public static String LON = "lon";
    public static String LAT = "lat";
	public static String GEOM = "geom";
	public static String CENTR = "centr";
	public static String BBOX = "bbox";

    /** suffixes of column names starting with longitude or latitude */
    private ArrayList<String> lonLatNames = new ArrayList<String>();
    /** Gives for each suffix a map that gives for each type (LON or LAT) a column name */
    private Map<String, Map<String, String>> lonLatNamesMap = new HashMap<String, Map<String, String>>();

    static private int commonPrefix(String name1, String name2) {
	int n = 0;
	for(; n < name1.length() && n < name2.length(); ++n)
	    if (name1.charAt(n) != name2.charAt(n))
		return n;
	return n;
    }


    static private String extractSuffix(String prefix, String name) {
	int n = commonPrefix(prefix, name);
	if (n < 3)
	    return null;
	
	if (name.length() == n) // name2 is a strict prefix
	    return "";

	if (name.charAt(n) != '_') // the underscore is mandatory
	    return null;

	return name.substring(n+1);
    }

    private void addSuffix(String name, String suffix, String ll) {
	 lonLatNames.add(suffix);
	 Map<String, String> llNames = lonLatNamesMap.get(suffix);
	 if (llNames == null)
	     llNames = new HashMap<String, String>();
	 llNames.put(ll, name);
	 lonLatNamesMap.put(suffix, llNames);
    }

    public String getLonLatType(String colName) {
	String name = colName.toLowerCase();
	String suffix;
	suffix = extractSuffix(longitude, name);
	if (suffix != null) { 
	    addSuffix(colName, suffix, LON);
	    return LON;
	}
	suffix = extractSuffix(latitude, name);
	if (suffix != null) {
	    addSuffix(colName, suffix, LAT);
	    return LAT;
	}
	suffix = extractSuffix(geometry, name);
	if (suffix != null) { 
	    addSuffix(colName, suffix, GEOM);
	    return GEOM;
	}
	suffix = extractSuffix(centroid, name);
	if (suffix != null) { 
	    addSuffix(colName, suffix, CENTR);
	    return CENTR;
	}
	suffix = extractSuffix(bounding, name);
	if (suffix != null) { 
	    addSuffix(colName, suffix, BBOX);
	    return BBOX;
	}

	lonLatNames.add("-");
	return null;
    }


    /** give the suffixes of column names starting with longitude or latitude */
    public String getLonLatName(int i) {
	// if (i >= lonLatNames.size())
	//     return "";
	return lonLatNames.get(i);
    }

    /** give the suffixes of column names starting with longitude or latitude */
    public String getLonLatName(String colName) {
	for(Map.Entry<String, Map<String, String>> entry : lonLatNamesMap.entrySet()) {
	    Map<String, String> map = entry.getValue();
	    if (map.containsValue(colName))
		return entry.getKey();
	}
	return null;
    }

    private HashMap<String, Double> lonMap = new HashMap<String, Double>();
    private HashMap<String, Double> latMap = new HashMap<String, Double>();
    
    private void initCentroidMap() { 
	if (lonLatNames.size() > 0) { // allocated only if needed
	    lonMap = new HashMap<String, Double>();
	    latMap = new HashMap<String, Double>();
	}
    }

    public void putLon(int i, Object token) throws java.lang.NumberFormatException {	    
	lonMap.put(getLonLatName(i), 
		   token instanceof Double ? (Double)token : new Double(token.toString()));
    }

    public void putLat(int i, Object token) throws java.lang.NumberFormatException {	    
	latMap.put(getLonLatName(i), 
		   token instanceof Double ? (Double)token : new Double(token.toString()));
    }

    public void putLon(String suffix, Object token) throws java.lang.NumberFormatException {	    
	lonMap.put(suffix, 
		   token instanceof Double ? (Double)token : new Double(token.toString()));
    }

    public void putLat(String suffix, Object token) throws java.lang.NumberFormatException {	    
	latMap.put(suffix, 
		   token instanceof Double ? (Double)token : new Double(token.toString()));
    }

    private Centroid getCentroid(String name) {
	Double lon = lonMap.get(name);
	Double lat = latMap.get(name);
	if (lon == null || lat == null)
	    return null;
	return new Centroid(lon, lat);
    }

    public void putCentroids(DataObject dataObject) {

	for (String lonLatName : lonLatNames) {
	    Centroid centroid = getCentroid(lonLatName); 
	    if (centroid != null){
		dataObject.addOMDataLocation(mkLonLatName(lonLatName), 
					     new OMDataLocation(null, centroid, null));
            }
	}
	initCentroidMap();
    }

    private String mkLonLatName(String n) {
	// use lonLatNamesMap ...
	Map<String, String> lonLatMap = lonLatNamesMap.get(n);
	return lonLatMap.get(LON)+"/"+lonLatMap.get(LAT);
    }

    public ArrayList<String> getOMGraphicNames() {
	ArrayList<String> omGraphicNames = new ArrayList<String>();

	for (String lonLatName : lonLatNamesMap.keySet())
	    if (! "-".equals(lonLatName))
		omGraphicNames.add(mkLonLatName(lonLatName));
	return omGraphicNames;
    }

    public Set<String> getLonLatNamesSuffixes() {
	return lonLatNamesMap.keySet();
    }

    public String getLonName(String suffix) {
	return lonLatNamesMap.get(suffix).get(LON);
    }

    public String getLatName(String suffix) {
	return lonLatNamesMap.get(suffix).get(LAT);
    }

    public String toString() {
	return lonLatNames.toString()+" "+lonLatNamesMap.toString();
    }

    /** test if name is a geometry column name built from lon and lat data columns */
    public static boolean isAGeometryLonLatName(String name) {
	if (name == null)
	    return false;
	name = name.toLowerCase();
	String[] dataColumnNames = name.split("/");
	if (dataColumnNames.length != 2)
	    return false;
	String lonSuffix = extractSuffix(longitude, dataColumnNames[0]);
	if (lonSuffix == null)
	    return false;
	String latSuffix = extractSuffix(latitude, dataColumnNames[1]);
	if (latSuffix == null)
	    return false;
	return lonSuffix.equals(latSuffix);
    }

    /** returns the name of the lon/lat geometry which colName is a corresponding longitude or latitude */
    public static String getLonLatName(String colName, String[] dataColumnNames, String[] omColumnNames) {
	String colname = colName.toLowerCase();
	{ 
	    // try lonSuffix
	    String lonSuffix = extractSuffix(longitude, colname);
	    if (lonSuffix != null) {
		String prefix = colName+"/";
		for (String g : omColumnNames) {
		    if (g.startsWith(prefix)) {	    
			// search for a lat name with the same suffix
			for (String lat : dataColumnNames) {
			    String latSuffix = extractSuffix(latitude, lat.toLowerCase());
			    if (lonSuffix.equals(latSuffix) && g.equals(prefix+lat))
				return g;
			}
		    }
		}
		return null;
	    }
	}

	{
	    // try latSuffix
	    String latSuffix = extractSuffix(latitude, colname);
	    if (latSuffix == null)
		return null;
	    String suffix = "/"+colName;
	    for (String g : omColumnNames) {
		if (g.endsWith(suffix)) {	    
		    // search for a lat name with the same suffix
		    for (String lon : dataColumnNames) {
			String lonSuffix = extractSuffix(longitude, lon.toLowerCase());
			if (latSuffix.equals(lonSuffix) && g.equals(lon+suffix))
			    return g;
		    }
		}
	    }
	}

	// fail
	return null;
    }

}
