// <header>
/**
 *  ThemaMap : The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
package fr.unicaen.openmap.util;

import java.util.Properties;

import com.bbn.openmap.util.Debug;
import com.bbn.openmap.util.PropUtils;
import com.bbn.openmap.image.WMTConstants;
import com.bbn.openmap.omGraphics.*;
import com.bbn.openmap.layer.OMGraphicHandlerLayer;
import com.bbn.openmap.Layer;

import java.awt.Color;
import java.awt.Dimension;
import java.io.*;
import java.util.Iterator;
import java.util.ListIterator;

import org.apache.batik.svggen.SVGGraphics2D;
import org.apache.batik.dom.GenericDOMImplementation;
import org.apache.batik.dom.util.DOMUtilities;

import org.w3c.dom.Document;
import org.w3c.dom.DOMImplementation;
import org.w3c.dom.Element;

/**
 * This class uses Batik to render real svg data from openmap layers.
 * Tested with version 1.7.beta1 from <a href="http://xmlgraphics.apache.org/batik/">here</a>.
 * This version will do better than a simple render() call on an SVGGraphics2D object :
 *  - layers are represented by SVG groups.
 *  - SVG elements are tagged by openmap classes names.
 */
public class SVGFormatter extends AbstractTextFormatter {
    
    private final String svgNS = "http://www.w3.org/2000/svg";
    private final String xlinkNS = "http://www.w3.org/1999/xlink";
    private final String evtNS = "http://www.w3.org/2001/xml-events";
    private Document document = null;
    private Element svgRoot = null;
    
    public String getFormatLabel() { return WMTConstants.IMAGEFORMAT_SVG; }

    /**
     * Format an Id String from an object. Id must be unique, thus the use
     * of memory address appended to className.
     */
    private String getIdFromObject(Object o) {
	String c = o.getClass().toString();
	String s = o.toString();
	return c.substring(c.lastIndexOf('.') + 1) + "_" + s.substring(s.lastIndexOf('@') + 1);
    }
    
    /**
     * To be used when no Id can be provided to <code>&lt;g&gt;</code> element.
     */
    private void addList(Element parent, OMGraphicList omgl) {
	addList(parent, omgl, null);
    }
    /**
     * Recursive call to this method causes correct hierarchy in SVG formatted document.
     * This methods takes traverseMode to know the order list rendering.
     */
    private void addList(Element parent, OMGraphicList omgl, String id) {
	Object o = null;
	Element g = document.createElementNS(svgNS, "g");
	g.setAttributeNS(null, "id", (id != null) ? id : getIdFromObject(omgl));
	// set the default stroke color to black
	g.setAttributeNS(null, "stroke", "black");

        if (omgl.getTraverseMode() == OMGraphicList.FIRST_ADDED_ON_TOP) {
            ListIterator it = omgl.getTargets().listIterator(omgl.getTargets().size());
            while (it.hasPrevious()) {
		o = it.previous();
		if (o instanceof OMGraphicList) {
		    Object appObj = ((OMGraphicList) o).getAppObject();
		    if (appObj != null && appObj instanceof XMLAppObject)
			addList(g, (OMGraphicList) o, ((XMLAppObject) appObj).getId());
		    else
			addList(g, (OMGraphicList) o);
		} else if (o instanceof OMGraphic)
		    addGraphic(g, (OMGraphic) o);
            }
        } else {
	    Iterator it = omgl.iterator();
	    while (it.hasNext()) {
		o = it.next();
		if (o instanceof OMGraphicList) {
		    addList(g, (OMGraphicList) o);
		} else if (o instanceof OMGraphic)
		    addGraphic(g, (OMGraphic) o);
	    }
        }
	parent.appendChild(g);
    }

    private void addBackgroundColor(/*Element parent, */int width, int height, Color color) {
        Element r = document.createElementNS(svgNS, "rect");
	r.setAttributeNS(null, "width", "" + width);
	r.setAttributeNS(null, "height", "" + height);
        String fillRGB = "rgb(" + color.getRed() + "," +color.getGreen() + "," + color.getBlue() +")" ;
	r.setAttributeNS(null, "fill", fillRGB);
        //System.out.
        String opacity =""+(double)(color.getAlpha()) / 255d + "";
        r.setAttributeNS(null, "fill-opacity", opacity);
	svgRoot.appendChild(r);
    }
    
    /**
     * Append the SVG code corresponding to OMGraphic to the parent Element.
     * This method uses SVGGraphics2D to do the rendering and append the top
     * level group Element generated to parent Element.
     */
    private void addGraphic(Element parent, OMGraphic omg) {
	SVGGraphics2D tmp = new SVGGraphics2D(document);
	omg.render(tmp);
	Element e = tmp.getTopLevelGroup();
	e.setAttributeNS(null, "id", getIdFromObject(omg));
	parent.appendChild(e);
    }
    
    public byte[] getTextFromLayers(Layer[] layers, int width, int height, Color color) {
	String encoding = "UTF-8"; // SVGGraphics2D.DEFAULT_XML_ENCODING est ISO-8859-1
	Debug.message("SVGFormatter", "rendering" + layers.length + "layers");
	
	try {
	    // Get a DOMImplementation.
	    DOMImplementation domImpl = GenericDOMImplementation.getDOMImplementation();
	    
	    // Create a document
	    document = domImpl.createDocument(svgNS, "svg", null);
	    svgRoot = document.getDocumentElement();

	    // some basic informations
	    svgRoot.setAttributeNS(null, "xmlns", svgNS);
	    svgRoot.setAttributeNS(null, "xmlns:xlink", xlinkNS);
	    svgRoot.setAttributeNS(null, "xmlns:ev", evtNS);

	    svgRoot.setAttributeNS(null, "width", "" + width);
	    svgRoot.setAttributeNS(null, "height", "" + height);

	    addBackgroundColor(width, height, color);

	    int ignored = 0;
	    // reverse order to get bottom layer first.
	    for (int i = layers.length -1 ; i >= 0 ; i--) {
		Layer layer = layers[i];
		if (layer instanceof OMGraphicHandlerLayer) {
		    addList(svgRoot, ((OMGraphicHandlerLayer) layer).getList(), ((OMGraphicHandlerLayer) layer).getName());
		} else {
		    ignored++;
		}
	    }
	    if (ignored != 0)
		Debug.message("SVGFormatter", ignored + " layers ignored");
	    
	    Debug.message("SVGFormatter", "Layers [] => SVG document done.");
	    
	    ByteArrayOutputStream stream = new ByteArrayOutputStream();
	    OutputStreamWriter writer = new OutputStreamWriter(stream, encoding);
	    // to avoid problem with char encoding
	    writer.write("<?xml version='1.0' encoding='" + encoding + "'?>");


	    // Basic write (no indent).
	    DOMUtilities.writeDocument(document,writer);
	    writer.flush();
	    
	    return stream.toByteArray();
	} catch (Exception e) {
	    e.printStackTrace();
	}
	return new byte[0];
    }
}
