// <header>
/**
 *  ThemaMap : The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
package fr.unicaen.openmap.util.classification;

import com.bbn.openmap.util.Debug;

import java.util.ArrayList;
import java.util.Collections;
import javax.swing.SpinnerModel;
import javax.swing.SpinnerNumberModel;

/**
 * Produce the same classes used in box plots :
 *   Q1 the lower quartile, M or Q2 the median and Q3 the upper quartile.
 *   D1, D9 the first and nineth deciles
 *   C1, C99, the 1st and 99th centiles.
 * We consider the breaks for the following number of classes
 * <table>
 * <tr><td>  <th>C1<th> D1 <th> Q1 <th> M <th>Q3<th> D9 <th> C99 </tr>
 * <tr><td>3<td>   <td>       <td>  x <td>    <td> x  <td>           <td>                </tr>
 * <tr><td>4<td>   <td>       <td>  x <td> x <td> x  <td>           <td>                </tr>
 * <tr><td>5<td>   <td>  x   <td>  x <td>    <td> x  <td>    x     <td>                </tr>
 * <tr><td>6<td>   <td>  x   <td>  x <td> x <td> x  <td>    x     <td>                </tr>
 * <tr><td>7<td> x<td>  x   <td>  x <td>    <td> x  <td>    x     <td>        x      </tr>
 * <tr><td>8<td> x<td>  x   <td>  x <td> x <td> x  <td>    x     <td>        x      </tr>
 * </table>
 */
public class BoxPlot extends ClassificationWithContrainedNbOfClasses {

    private static Integer _admissibleNbClasses[] = {3, 4, 5, 6, 7, 8};
    protected Integer[]  getAdmissibleNbClasses() {
	return _admissibleNbClasses;
    }

    public BoxPlot(ArrayList<Double> data, int nbClasses) {
	super(data, nbClasses);
    }

    protected void init() {
	if (data == null) return;

	initBreaks();

	initCountsFromBreaks();
    }

    protected void initBreaks() {	
	int N = data.size();

	Quantile quartiles = new Quantile(data, 4);
	// now data is sorted
	breaks = new double[nbClasses+1];
	breaks[0] = data.get(0); // first break is the min value;
	breaks[nbClasses] = data.get(data.size()-1); // last break is the max value

	double Q1 = quartiles.getNthBreak(1);
	double M  = quartiles.getNthBreak(2);
	double Q3 = quartiles.getNthBreak(3);
	double IQR = Q3 - Q1;
	double Q1_15 = Q1-1.5*IQR;
	double Q1_3 = Q1-3*IQR;
	double Q3_15 = Q3+1.5*IQR;
	double Q3_3 = Q3+3*IQR;
	if (Q1_15 < breaks[0]) Q1_15 = breaks[0];
	if (Q1_3 < breaks[0]) Q1_3 = breaks[0];
	if (Q3_15 > breaks[nbClasses]) Q3_15 = breaks[nbClasses];
	if (Q3_3 > breaks[nbClasses]) Q3_3 = breaks[nbClasses];

	switch (nbClasses) {
	case 3:
	    breaks[1] = Q1; breaks[2] = Q3;
	    break;
	case 4:
	    breaks[1] = Q1; breaks[2] = M; breaks[3] = Q3;
	    break;
	case 5:
	    breaks[1] = Q1_15; breaks[2] = Q1; breaks[3] = Q3; breaks[4] = Q3_15;
	    break;
	case 6:
	    breaks[1] = Q1_15; breaks[2] = Q1; breaks[3] = M; breaks[4] = Q3; breaks[5] = Q3_15;
	    break;
	case 7:
	    breaks[1] = Q1_3; breaks[2] = Q1_15; breaks[3] = Q1; breaks[4] = Q3; breaks[5] = Q3_15; breaks[6] = Q3_3;
	    break;
	case 8:
	    breaks[1] = Q1_3; breaks[2] = Q1_15; breaks[3] = Q1; breaks[4] = M; breaks[5] = Q3; breaks[6] = Q3_15; breaks[7] = Q3_3;
	    break;
	default:
	    Debug.error("[Classification] *** Error *** number of classes for Boxplot " + nbClasses);
	}
    }
}
