// <header>
/**
 *  ThemaMap : The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
package fr.unicaen.openmap.util.classification;

import java.util.ArrayList;
import java.util.Collections;
import java.lang.reflect.Constructor;
import javax.swing.*;

import fr.unicaen.openmap.util.Util;
/**
 * Perform a builtin statistical classification of data.
 * <p>You may get breaks and counts of objects per class.
 * <p>breaks[0] and breaks[breaks.length-1] are respectively equals to min and max values
 * <p>The Classification hierarchy is:
 * <pre>
 * Classification 
 *      ClassificationOfDouble 
 *              ClassificationWithContrainedNbOfClasses
 *                      BoxPlot 
 *                      NestedMeans 
 *                      Standardization                         
 *              Classif2Iplus1divK2 
 *              CustomizedClassificationOfDouble implements CustomizedClassification 
 *                      ContinuousClassification 
 *              DiscreteOfDouble implements DiscreteClassification 
 *              EqualWeight 
 *              HierarchicalClassification 
 *              Quantile 
 *              StepClassification 
 *                      ArithmeticStepClassification 
 *      ClassificationOfString  
 *              DiscreteOfString implements DiscreteClassification 
 *                      CustomizedClassificationOfString implements CustomizedClassification 
 *      
 *      Discrete implements DiscreteClassification
 *      Customized implements CustomizedClassification 
 * </pre>
 */
public abstract class Classification {

    protected int nbClasses;
    
    /**
     * @param nbClasses the number of wished classes; some methods may decide to change its value.
     */
    public Classification(int nbClasses) {
	this.nbClasses = nbClasses;
    }
    
    abstract public void reInit(fr.unicaen.openmap.dataStorage.DataStorageIterator dsi,java.lang.String nbClasses);

    public void reInit(Classification c) {
	reInit(c, c.nbClasses);
    }

    abstract public void reInit(Classification c, int nbClasses); 
    
    abstract public boolean hasNoData();

    /** used to inform the classif that its data is out of date */
    abstract public void unsetData(); 
    
    public void setNumberOfClasses(int nbClasses) {
	this.nbClasses = nbClasses;
	// System.out.println("Classification.setNumberOfClasses("+nbClasses+") "+this);
	init();
    }

    public int getNumberOfClasses() { return nbClasses; }
    
    public static final int NbClassesMax = 32;
    /**
     * Return a SpinnerModel compatible with the classification type and the current
     * number of classes.
     */
    public SpinnerModel getClassificationNumberModel() {
	if (nbClasses > NbClassesMax)
	    nbClasses = NbClassesMax;
	if (nbClasses < 1)
	    nbClasses = 1;
	SpinnerModel m =  new SpinnerNumberModel (nbClasses, 1, NbClassesMax, 1);
	m.setValue(nbClasses);
	return m;
    }

    protected abstract void init();

    
    
    protected int[] counts;
    public int[] getCounts() { return counts; }
    abstract protected void initCountsFromBreaks();    

    /**
     * return the class rank of the value d between 1 and nbClasses
     */    
    abstract public int getClass(Object d);

    public String toString() {
	return "@"+hashCode()+" - ("+Util.className(this)+") ";
    }

    abstract public String dump();

    /** returns a text describing the ith class */
    abstract public String getClassText(int i);
    /** returns texts describing the classes */
    abstract public String[] getClassesTexts();
    /** returns the breaks, useful for getProperties() */
    abstract public String getBreaksAsString();

    /*
     * Returns the real class name. Useful for Customized.
     */
    public String getClassName() {
	return getClass().getName();
    }

    /** only overloaded for CustomizedClassification */
    public void setBreaks(String[] s) {
    }

    public Classification getRealClassif() { return this; }

}
