// <header>
/**
 *  ThemaMap : The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
package fr.unicaen.openmap.util.classification;

import com.bbn.openmap.util.Debug;

import java.util.ArrayList;
import java.util.Collections;
import java.lang.reflect.Constructor;
import java.text.NumberFormat;
import javax.swing.*;
import fr.unicaen.openmap.dataStorage.DataObject;
//import fr.unicaen.openmap.util.classification.gui.ClassificationPanel;
import fr.unicaen.openmap.util.Util;

/**
 * Perform a builtin statistical classification of data that must be doubles.
 * <p>You may get breaks and counts of objects per class.
 * <p>breaks[0] and breaks[breaks.length-1] are respectively equals to min and max values
 */
public abstract class ClassificationOfDouble extends Classification {
    protected ArrayList<Double> data;
    
    /**
     * @param data 
     * @param nbClasses the number of wished classes; some methods may decide to change its value.
     */
    public ClassificationOfDouble(ArrayList<Double> data, int nbClasses) {
	super(nbClasses);
	reInit(data); 
    }

    public ClassificationOfDouble(Classification c, int nbClasses) {
	super(nbClasses);
	reInit(c, nbClasses);
    }
    
    public void reInit(fr.unicaen.openmap.dataStorage.DataStorageIterator dsi,java.lang.String dataColumnName) {
	data = new ArrayList<Double>();
        dsi.reset();
	Util.resetMaxNumberFormatExceptionOutput();
	while (dsi.hasNext()) {
	    DataObject dataObject = dsi.next();
	    Object d = dataObject.getData(dataColumnName);
	    if (d == null)
		continue;
	    if ("".equals(d.toString()))
		continue;
	    double value = Util.parseDouble(d);
	    if (Double.isNaN(value) || value == Double.NEGATIVE_INFINITY || value == Double.POSITIVE_INFINITY)
		continue;
	    data.add(value);
	}
        reInit(data);
	dsi.reset();
	Util.resetMaxNumberFormatExceptionOutput();
	// System.out.println("### ClassificationOfString.reInit( "+ dataColumnName + " ) nbClasses = "+ nbClasses);
    }

    public void reInit(ArrayList<Double> data, int nbClasses) {
	this.nbClasses = nbClasses;
	reInit(data);
    }

    public void reInit(ArrayList<Double> data) {
	this.data = data;
	if (hasNoData()) {
	    // Debug.error("[Classification] *** Error *** empty data");
	    return;
	}
	if (nbClasses <= 0) {
	    Debug.error("[Classification] *** Error *** "+nbClasses+" classes wanted");
	    return;
	}	
	init();
    }

    protected void init() {
	if (data == null) return;
	Collections.sort(data);

	// set breaks
	breaks = new double[nbClasses+1];
	breaks[0] = data.get(0); // first break is the min value;
	breaks[nbClasses] = data.get(data.size()-1); // last break is the max value

	initBreaks();

	initCountsFromBreaks();
    }

    abstract protected void initBreaks();

    public void reInit(Classification c, int nbClasses) {
	ArrayList<Double> data;
	c = c.getRealClassif();
	
	if (c instanceof ClassificationOfString) {
	    ClassificationOfString cos = (ClassificationOfString)c;
	    if (cos.data == null)
		data = null;
	    else {
		data = new ArrayList<Double>(cos.data.size());
		for (String s : cos.data)
		    data.add(Util.parseDouble(s));
	    }
	} else if (c instanceof ClassificationOfDouble) {
	    data = ((ClassificationOfDouble)c).data;
	} else {
	    Debug.error("### ClassificationOfDouble "+this.getClass()+" reInit with an unknown Classification! ### "+c.getClass());
	    return;
	}
	if (nbClasses > NbClassesMax)
	    nbClasses = NbClassesMax;
	reInit(data, nbClasses);
    }

    public boolean hasNoData() { return data == null || data.size() == 0; }

    public void unsetData() { data = null; } 

    protected double[] breaks = null;
    /**
     * breaks.size() == nbClasses + 1
     * <br>min == breaks[0] and max == breaks[nbClasses]
     */ 
    public double[] getBreaks() { return breaks; }
    public double getNthBreak(int n) {
	return (breaks != null && n <= breaks.length-1) ? breaks[n]  : 0; 
    }

    protected void initCountsFromBreaks() {
	counts = new int[breaks.length-1];		
	for(int i = 0; i < counts.length; i++)
	    counts[i] = 0;
		
	for(double d : data)
	    for(int j = 0; j < breaks.length-1; j++)
		if (d <= breaks[j+1]) {
		    counts[j]++;
		    break;
		}
    }


    /**
     * return the class rank of the value between 1 and nbClasses
     */
    public int getClass(Object o) {
	if (breaks == null)
	    return 0;
	double d = Util.parseDouble(o);
	if (Double.isNaN(d))
	    return -1;

	return getClass(d);
    }

    public int getClass(double d) {
	// may be we should use a dichotomy ?
	int i = 1;
	for (; i <= nbClasses && i < breaks.length; i++)
	    if (d <= breaks[i])
		return i-1;
	return i-1;
    }

    public String toString() {
	StringBuilder ret = new StringBuilder(super.toString());
	int totalcounts = 0;
	ret.append(nbClasses + " [");
	if (breaks != null) {
	    for (int i = 0; i < breaks.length-1; i++) { 
		ret.append(String.format("%10.4f", breaks[i]));
		if (counts != null) {
		    ret.append("("+ counts[i] + ")");
		    totalcounts += counts[i];
		}
		ret.append(", ");
	    }  
	    ret.append(breaks[breaks.length-1]);
	} 
	ret.append(" ] ");
	if (counts != null)
	    ret.append("(" + totalcounts + ") ");
	if (breaks != null)
	    ret.append(breaks.length);
	return ret.toString();
    }

    public String dump() {
	Collections.sort(data);
	String ret = "Data ("+data.size()+") [";
	for(Double d : data) ret += String.format("%10.4f; ", d);
	return ret + "]";
    }

    /** returns a text describing the ith class */
    public String getClassText(int i) {
	if (breaks == null || i+1 >= breaks.length)
	    return "";
	return ((i == 0 ? "[ " : "] ") 
		+ NumberFormat.getInstance().format(breaks[i])
		+ " - " 
		+ NumberFormat.getInstance().format(breaks[i+1]) 
		+"]");
    }

    /** returns texts describing the classes */
    public String[] getClassesTexts() {
	int nbTexts = 0;
	if (breaks != null) //avoid null pointer
	    nbTexts = breaks.length - 1;
	String[] texts = new String[nbTexts];
	for (int i = 0; i < nbTexts; i++)
	    texts[i] = getClassText(i);
	return texts;
    }

    public String getBreaksAsString() {
	if (breaks == null) 
	    return "";
	StringBuilder s = new StringBuilder();
	for (double b : breaks) {
	    s.append("\t");
	    s.append(b);
	}
	return s.toString();
    }
        
    /** standalone tests */
    public static void main(String[] args) {
	if (args.length < 3) {
	    System.err.println("usage: java fr.unicaen.openmap.util.classification.Classification <className> <nbClasses> <nbData> [<maxValue>]");
	    return;
	}

	ArrayList<Double> data = new ArrayList<Double>();
	String className = args[0];
	int n = Integer.parseInt(args[1]);
	int N = Integer.parseInt(args[2]);
	if (args.length > 4) {
	    N = args.length - 2;
	    for (int i = 2; i < args.length; i++)
		data.add(Double.parseDouble(args[i]));
	} else {
	    double maxValue = 70;
	    if (args.length == 4)
		maxValue = Double.parseDouble(args[3]);
	    
	    for (int i = 0; i < N; i++)
		data.add((Math.random()>0.25 ? +1 : -1)*(double)maxValue*Math.random());
	}
	ClassificationOfDouble c;
        try {
            Class cls = Class.forName("fr.unicaen.openmap.util.classification."+className);
	    Constructor[] constructors = cls.getConstructors();
	    // we should check for the signature of the constructors 
	    // see ThematicDataDisplayLayer.getConstructorWithStringParameter
	    c = (ClassificationOfDouble) constructors[0] .newInstance(new Object [] { data, n });;
        } catch (Exception e) {
            Debug.error("construction failure for "+ className);
	    e.printStackTrace();
	    return;
        }
        
	//System.out.println(c.dump());
	//System.out.println(c);
	int nb = 0, total = 0;
	int previouscc = 1;
	    //System.out.println("Value\tClasse\tCardinal");
	for (double d : data) {
	    int cc = c.getClass(d);
	    //System.out.format("%10.4f\t%d", d, cc);
	    if (cc == previouscc)
		nb++;
	    else {
		//System.out.print("\t" + nb);
		total += nb;
		previouscc = cc;
		nb = 1;
	    }
	    //System.out.println();
	}
	total += nb;
	//System.out.println("\t\t"+ nb);
	//System.out.println("Total\t\t"+ total);
	if (total != N)
	    	//System.out.println("### "+ total + " != " + N + "!!");
	return;
    }
}
