// <header>
/**
 *  ThemaMap : The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008-2012 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
package fr.unicaen.openmap.util.propertyEditor;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.beans.PropertyEditorSupport;
import javax.swing.BorderFactory;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JPanel;
import java.util.Vector;
import java.util.Iterator;
import java.util.Properties;
import java.util.Set;

import com.bbn.openmap.BasicI18n;
import com.bbn.openmap.Environment;
import com.bbn.openmap.I18n;
import com.bbn.openmap.PropertyConsumer;
import com.bbn.openmap.util.Debug;
import com.bbn.openmap.util.PropUtils;

import fr.unicaen.openmap.dataStorage.*;
import fr.unicaen.openmap.dataStorage.postgis.*;
import fr.unicaen.openmap.dataStorage.shapefile.*;
import fr.unicaen.openmap.dataStorage.xmltable.*;
import fr.unicaen.openmap.layer.dataDisplayLayer.SimpleDataDisplayLayer;
import fr.unicaen.openmap.util.Util;

/**
 * Determinate the data storage to use and how to configurate it.
 */
public class DataStoragePropertyEditor extends PropertyEditorSupport implements DataStorageListener {

    public enum DataStorageClass {
	CSVFILE("CSVFile", "fr.unicaen.openmap.dataStorage.CSVFileDataStorage"),
	    SHAPE("Shape", "fr.unicaen.openmap.dataStorage.shapefile.ShapeFileDataStorage"), 
	    DBF("Dbf", "fr.unicaen.openmap.dataStorage.shapefile.DbfDataStorage"),
	    XMLTABLE("XmlTable", "fr.unicaen.openmap.dataStorage.xmltable.XmlTableDataStorage"),
	    POSTGIS("Postgis", "fr.unicaen.openmap.dataStorage.postgis.PostgisDataStorage"), 
	    GML("GML", "fr.unicaen.openmap.dataStorage.gml.GMLDataStorage"),
	    MIF("Mif", "fr.unicaen.openmap.dataStorage.mif.MifDataStorage"),
	    SHARED("Shared", "fr.unicaen.openmap.dataStorage.SharedDataStorage"),
	    NETCDF("NetCDF", "fr.unicaen.openmap.dataStorage.NetCDFDataStorage"),
	    JDBC("JDBC", "fr.unicaen.openmap.dataStorage.JDBCDataStorage"),
	    JSON("JSON", "fr.unicaen.openmap.dataStorage.JSONDataStorage"),
	    ;

	private String index;
	private String className;
	DataStorageClass(String index, String className) {
	    this.index = index;
	    this.className = className;
	}
	public String toString() { return index; }
	public String getIndex() { return index; }
	public String getClassName() { return className; }
	// public static Vector<DataStorageClass> getActiveValues() {
	//     Vector<DataStorageClass> ret = new Vector<DataStorageClass>();
	//     for (DataStorageClass dsc : DataStorageClass.values())
	// 	ret.add(dsc);
	//     return ret;
	// }

	public static String getClassName(String index) {
	    for (DataStorageClass dsc : DataStorageClass.values())
		if (dsc.getIndex().equals(index))
		    return dsc.getClassName();
	    return null;
	}
    }
    
    /**
     * Separator used in getAsText() between property and value.
     */
    public final static String keyValueSeparator = "=";

    /**
     * Separator used in getAsText() between properties.
     */
    public final static String propertySeparator = ";";

    /**
     * Property name for configuration button.
     */
    public final static String CONFIGURE_PROPERTY_STRING = "configure";
    
    /**
     * Use I18N mecanism to allow internationalization.
     * Use I18N.properties file or I18N_fr_FR.properties for french version.
     */
    protected I18n i18n = new BasicI18n();

    /**
     * Contains list of data storage given by dataStoragesNameList.
     */
    protected JComboBox listJCB;

    /**
     * Allows to configure the wanted data storage.
     */
    protected Inspector inspector = new Inspector();

    /**
     * Used to prefix for the current DataStorage class name.
     */
    protected String prefix = " ";



    //useless to have an effective constructor. See description for PropertyEditor
    //    public DataStoragePropertyEditor() { }    



    /**
     * The editor provides a custom editor.
     * @see java.beans.PropertyEditor 
     * @see PropertyEditorSupport
     * @return true
     */
    public boolean supportsCustomEditor() {
        return true;
    }

    /**
     * First step : choose a dataStorage.
     * Second step : Configurate it .
     * @see java.beans.PropertyEditor 
     * @see PropertyEditorSupport
     * @return panel
     */
    public Component getCustomEditor() {
	final JPanel panel = new JPanel();
	final JPanel dsPanel = new JPanel();

	panel.setLayout(new BoxLayout(panel, BoxLayout.Y_AXIS));
	dsPanel.setLayout(new BoxLayout(dsPanel, BoxLayout.Y_AXIS));

	listJCB = new JComboBox(DataStorageClass.values());
	listJCB.setToolTipText(i18n.get(DataStoragePropertyEditor.class, 
					CONFIGURE_PROPERTY_STRING, 
					I18n.TOOLTIP,
					"Selected data storage to configure."));
	listJCB.setSelectedIndex(-1);
	//more dynamism :°
	listJCB.addActionListener(new ActionListener() {

		public void actionPerformed(ActionEvent ae) {
		    JComponent component = null;
		    try {
			//fetch the custom editor
			DataStorage ds = Util.createDataStorage(getDSClassName(), DataStoragePropertyEditor.this);
			prefix = PropUtils.getScopedPropertyPrefix(ds.getPropertyPrefix());
		    
			// inspector.inspectPropertyConsumer(ds); //not exactly what we want : pop-up again, useless buttons...
			component = (JComponent)inspector.createPropertyGUI(ds);
			//remove buttons Ok/Cancel
			component.remove(((BorderLayout)(component.getLayout())).getLayoutComponent(component, BorderLayout.SOUTH));
		    } catch(Exception e) {
			e.printStackTrace();
			component = new JPanel();
			component.add(new JLabel(getDSIndex()+" "+i18n.get(DataStoragePropertyEditor.class, "notAvailable", "not available")));
		    }
		    
		    panel.setVisible(false);
		    dsPanel.removeAll();
		    dsPanel.add(component);
		    panel.setVisible(true);
		}
	    }
	    );

	JPanel listPanel = new JPanel();
	listPanel.add(new JLabel(i18n.get(DataStoragePropertyEditor.class,"dataSourceType","Data source type:")));
	listPanel.add(listJCB);

	panel.add(listPanel);
	panel.add(dsPanel);
	panel.setPreferredSize(new Dimension(500, 300));
	panel.setBorder(BorderFactory.createTitledBorder(""));
	return panel;
    }

    /*
     * Provides the configuration for the wanted data storage.
     * Note : property name for DataStorage is provided by SimpleDataDisplayLayer.dataStorageClassPropertyString
     * Warning: In order to be parsed correctly, spaces semi colon are escaped. Use unEscapePropertyString before reuse.  
     * @see java.beans.PropertyEditor 
     * @see PropertyEditorSupport
     * @return a String that contains the class name and properties for a DataStorage. Each property is separated by a propertySeparator. Property and value are separated by a keyValueSeparator.
     */
    public String getAsText() {
	Properties props = inspector.collectProperties();
	String dsConfig = "";

	for(String key : props.stringPropertyNames())
	    dsConfig += key + keyValueSeparator + Util.escapePropertyString( props.getProperty(key) ) + propertySeparator;	

	//don't forget the DataStorage class name
	String res =  prefix + SimpleDataDisplayLayer.dataStorageClassPropertyString + keyValueSeparator + getDSClassName() + propertySeparator + dsConfig ;
	Debug.message("inspectordetail" , "DataStoragePropertyEditor DSCONFIG: "+ res);

	return res;
    }
    
    /**
     * Just in order to be able to create a 'dummy' DataStorage.
     * @see DataStorageListener
     */
    public void dataStorageEvent (DataStorage source) {
	//Nothing to do	
    }

    /**
     * Provides the currently selected data storage class name.
     * @return the currently selected data storage class name
     */
    protected String getDSClassName() {
	return ((DataStorageClass)listJCB.getSelectedItem()).getClassName();
    }

    /**
     * @return the currently selected data storage index name
     */
    protected String getDSIndex() {
	return ((DataStorageClass)listJCB.getSelectedItem()).getIndex();
    }



    /** 
     * Since we use any propertyEditor as part of GUI, it's sometimes useful to remove default buttons. 
     * @param propertyConsumer The panel to change is provided by this PropertyConsumer.
     * @param inspector The Inspector to use.
     * @see fr.unicaen.openmap.dataStorage.shapefile.ShapeFileDataStorage#getGUI() 
     * @see fr.unicaen.openmap.layer.thematicDataDisplayLayer.ThematicDataDisplayLayer#getGUI()
     */
    public static Component createPropertyGuiWithoutDefaultButtons(final PropertyConsumer propertyConsumer, final Inspector inspector) {

	//cancel button is obsolete here, so...
	Component component = inspector.createPropertyGUI(propertyConsumer);

	if (component instanceof Container) {//JComponent
	    //unreadable code just to remove default buttons Ok / Cancel
	    Container container = (Container) component;
	    container.remove(((BorderLayout) (container.getLayout())).getLayoutComponent(container, BorderLayout.SOUTH));
	}
	return component;
    }

    /** 
     * Remove default buttons. Add a button for validation.
     * @param propertyConsumer The panel to change is provided by this PropertyConsumer.
     */
    public static Component provideCustomPanel(final PropertyConsumer propertyConsumer) {

	final Inspector inspector = new Inspector();

	JPanel pan = new JPanel();
	JButton ok = new JButton(Environment.getI18n().get(DataStoragePropertyEditor.class, "ok", "Ok"));
	ok.addActionListener(new ActionListener() {

		public void actionPerformed(ActionEvent event) {
		    propertyConsumer.setProperties(inspector.collectProperties());
		}
	    });
	pan.add(ok);

	JPanel panel = new JPanel();
	panel.setLayout(new BoxLayout(panel, BoxLayout.Y_AXIS));
	panel.add(createPropertyGuiWithoutDefaultButtons(propertyConsumer, inspector));
	panel.add(pan);

	return panel;
    }
    
    /** 
     * Remove default buttons. Add a button for validation.
     * @param propertyConsumer The panel to change is provided by this PropertyConsumer.
     */
    public static Component provideSimpleCustomPanel(final PropertyConsumer propertyConsumer, Inspector inspector) {

	//final Inspector inspector = new Inspector();

	JPanel pan = new JPanel();
	JButton ok = new JButton(Environment.getI18n().get(DataStoragePropertyEditor.class, "ok", "Ok"));
	/*ok.addActionListener(new ActionListener() {

	  public void actionPerformed(ActionEvent event) {
	  propertyConsumer.setProperties(inspector.collectProperties());
	  }
	  });
	  pan.add(ok);*/

	JPanel panel = new JPanel();
	panel.setLayout(new BoxLayout(panel, BoxLayout.Y_AXIS));
	panel.add(createPropertyGuiWithoutDefaultButtons(propertyConsumer, inspector));
	panel.add(pan);

	return panel;
    }


}

